/*
** elvisdbg.c - debugging window for Gnome GUI for Elvis.
**
** Written by C W Rose.
**
** $Log$
**
** NOTE:
** This app was written to help with the debugging of a Gnome GUI
** for the Elvis editor.  It can be run from the command line for
** testing, but is normally started from Elvis' debugging menu (and
** therefore must be on the command search path), taking data from
** Elvis via a pipe.
**
** It has only one option, --fontname, which sets the font. Since
** the debug data is presented in tabular form it's best to use
** monospace fonts only.  The default is Monospace 8.
**
** The main menu, toolbar, and mouse menu items are placeholders,
** since the program will probably be extended in due course.
**
** When any of the structures are updated, the field counts below
** (eg: ED_FIELDS) must also be changed.
*/

/*! \file elvisdbg.c
**  \brief A debugging window for Gnome GUI for Elvis.
**
** The elvisdbg program is used by some of the entries in the GnomeGui
** debug menu, which is enabled by \#define DEBUG_MENU in guignome.c.
** It displays the current values of various data structures in the Gnome
** GUI front end for Elvis.
*/

/* Description and version number */
#define DESCR "debugging window for Gnome GUI for Elvis."
#define VERSION "Version 0.93 of 1 Jan 08."

/* Assert macro control, ANSI C and Glib */
#undef NDEBUG
#undef G_DISABLE_ASSERT

/* Include files */
#include <linux/limits.h>	/* for PATH_MAX */
#include <assert.h>
#include <errno.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <unistd.h>
#include <fcntl.h>

#define PANGO_ENABLE_BACKEND	/*!< Undocumented, required for font metrics. */
#include <gdk/gdkx.h>
#include <gdk-pixbuf/gdk-pixdata.h>

#include <gtk/gtk.h>
#include <gconf/gconf-client.h>
#include <pango/pango.h>
#include <libbonobo-2.0/libbonobo.h>
#include <libbonoboui-2.0/bonobo.h>
#include <libgnome/libgnome.h>
#include <libgnomeui/libgnomeui.h>

#define EXTERN
#include "guignome.h"

/* Code control defines */

/* Optional features */
#undef DEBUG
#undef TEST

/* Debugging message defines */
#ifdef DBG_PRINT
  #define DPRINT(x) g_print(x)	/*!< A debugging print command. */
#else
  #define DPRINT(x)	/*!< A suppressed debugging print command. */
#endif

/* Standard defines */

/* Local defines */
#define DEFAULT_FONT_NAME "Monospace 8"
#define DEFAULT_PANGO_LANG "en-us"

#define F_SIZ 64	/*!< The font name buffer size. */
#define F_MAX (F_SIZ-1)	/*!< The font name buffer maximum string length. */

/* Used for drawing nodes in the "expose" signal handler */
#define ED_FIELDS 44	/*!< The number of fields in a struct Edwin. */
#define EN_FIELDS 8	/*!< The number of fields in a struct Enode. */
#define GU_FIELDS 124	/*!< The number of fields in a struct GnomeGui. */
#define MGNS 6		/*!< The total node rectangle margin width. */
#define LHS 10		/*!< The child window left hand margin width. */

/* Macros */

/* Structures, typedefs and enums */
typedef unsigned char UCHAR;
typedef unsigned int UINT;
typedef unsigned long ULONG;

/*! Data types used in drawing routines. */
enum Arg {
	X,	/*!< unsigned long type. */
	U,	/*!< unsigned int type. */
	B,	/*!< bitmap type. */
	L,	/*!< long int type. */
	D,	/*!< int type. */
	S,	/*!< string type. */
	C,	/*!< char type. */
	O,	/*!< GdkColor type. */
	G	/*!< GdkGeometry type. */
	};

/* Global data */

#ifndef lint
/* If these strings are static GCC optimises them out of existence */
char id_elvisdbg_c[] = "$Id: elvisdbg.c,v 1.93 2008/01/01 12:00:00 cwr Exp $";
static volatile char Version[] = "@(#) elvisdbg.c: " VERSION;
#endif
static char progname[PATH_MAX + 1]; /*!< The program name. */

/* popt library option handling */
#if 0
/* Copied for reference from popt.h */
struct poptOption {
	const char * longName;	/* may be NULL */
	char shortName;		/* may be '\0' */
	int argInfo;		/* POPT_ARG_NONE, POPT_ARG_STRING,
				** POPT_ARG_INT, POPT_ARG_LONG,
				** POPT_ARG_INCLUDE_TABLE, POPT_ARG_CALLBACK,
				** POPT_ARG_INTL_DOMAIN
				*/
	void * arg;		/* depends on argInfo */
	int val;		/* used in callback -- may be 0 */
	const char * descrip;	/* description for autohelp -- may be NULL */
	const char * argDescrip;/* argument description for autohelp */
  };
#endif

/* popt structures and data */
static char *Fontopt = NULL;	/*!< The popt font option. */
static char Fontname[F_SIZ];	/*!< The current font name. */

/*! struct poptOption is defined in popt.h with fields:
** longname, shortname, argtype, &arg, return, description, help.
*/
static struct poptOption gnome_popt[] =
  {
  {"fontname", 'f', POPT_ARG_STRING, &Fontopt, 0, "font name", "FONT" },
  { NULL, 0, 0, NULL, 0, NULL, NULL }
  };

/* Macros to get values of global options */
#define o_fontname Fontname	/*!< A macro for the current font name. */

/* Input data list */
static struct Node * Data;	/*!< A pointer to the input data structure. */

/* Externals */

/* Function prototypes */
static gint delete_event(GtkWidget *widget, GdkEvent *event, gpointer data);
static void hello(GtkWidget *widget, gpointer data);
static void accelerator(GtkWidget *widget, gpointer data);
static void destroy(GtkWidget *widget, gpointer data);
static int parse_font(char *str);
static char *show_bits(unsigned int value, unsigned int len);
static gboolean gio_handler(GIOChannel *src, GIOCondition cond, gpointer data);
static gint popup_handler(GtkWidget *widget, GdkEvent *event, gpointer *data);
static gboolean expose(GtkWidget *widget, GdkEventExpose *event, gpointer data);
static int draw_link(GtkWidget *widget, GdkGC *gc, int *px, int *py, int w);
static int draw_enode(GtkWidget *widget, GdkGC *gc, struct Enode *pn,
							int *px , int *py);
static int draw_gui(GtkWidget *widget, GdkGC *gc, struct GnomeGui *pg,
							int *px, int *py);
static int draw_edwin(GtkWidget *widget, GdkGC *gc, struct Edwin *pe,
							int *px, int *py);

/* Functions */

/*! delete_event -- Handler for a "delete-event" signal on the main window.
** \param widget -- The main window.
** \param event -- The trigger for the signal.
** \param data -- NULL.
** \returns FALSE: Force a destroy signal.
** \returns TRUE: Otherwise.
*/

/*
** d e l e t e _ e v e n t
**
** delete_event callback.
**
** Returns:	FALSE		Force a destroy signal
**		TRUE		Otherwise
*/
static gint delete_event(GtkWidget *widget, GdkEvent *event, gpointer data)
{
  DPRINT(("delete event occurred\n"));

  return FALSE;
}

/*! hello -- A general-purpose callback.
** \param widget -- The calling widget.
** \param data -- The caller's ID.
** \returns Nothing: Always.
*/

/*
** h e l l o
**
** A general-purpose callback.
**
** Returns:	Nothing		Always
*/
static void hello(GtkWidget *widget, gpointer data)
{
  /* write a message describing the source */
  DPRINT(("Hello from %s\n", (char *) data));
  return;
}

/*! accelerator -- An accelerator callback.
** \param widget -- The calling widget.
** \param data -- NULL.
** \returns Nothing: Always.
*/

/*
** a c c e l e r a t or
**
** Accelerator callback.
**
** Returns:	Nothing		Always
*/
static void accelerator(GtkWidget *widget, gpointer data)
{
  /* write a message describing the source */
  DPRINT(("Hello from the accelerator\n"));
  return;
}

/*! destroy -- Handler for a "destroy" signal on the main window.
** \param widget -- The main window.
** \param data -- NULL.
** \returns Nothing: Always.
*/

/*
** d e s t r o y
**
** Shutdown callback.
**
** Returns:	Nothing		Always
*/
static void destroy(GtkWidget *widget, gpointer data)
{
  gtk_main_quit();
  return;
}

/*! parse_font -- Parse a font (string) option.
** \param str -- The name to be parsed.
** \returns OK: Success.
** \returns NK: Value out of range.
** \returns FAILED: Otherwise.
*/

/*
** p a r s e _ f o n t
**
** Parse a font (string) option.
**
** Returns:	OK		Success
**		NK		Value out of range
**		FAILED		Otherwise
**
** Arguments are those of the standard options parsing functions.
*/
static int parse_font(char *str)
{
  GdkDisplay *display;
  GdkScreen *screen;
  PangoContext *pango_context;
  PangoFontDescription *pd;
  PangoFontFamily **pango_families;
  gint familycount;

  PangoFontMask mask;
  const char *pf;
  int j, valid;
  const char *pn;

  /* Get the default pango context */
  pango_context = NULL;
  display = gdk_display_get_default();
  screen = gdk_display_get_default_screen(display);
  pango_context = gdk_pango_context_get_for_screen(screen);

  /* Bale out early on obvious failure */
  if (pango_context == NULL) return FAILED;

  /* The input string is of the form --option=name, where name can be
  ** any valid font string.
  **
  ** This string has the form "[FAMILY-LIST] [STYLE-OPTIONS] [SIZE]",
  ** where FAMILY-LIST is a comma separated list of families optionally
  ** terminated by a comma, STYLE_OPTIONS is a whitespace separated list
  ** of words where each WORD describes one of style, variant, weight,
  ** or stretch, and SIZE is a decimal number (size in points). Any one
  ** of the options may be absent. If FAMILY-LIST is absent, then the
  ** family_name field of the resulting font description will be
  ** initialized to NULL. If STYLE-OPTIONS is missing, then all style
  ** options will be set to the default values. If SIZE is missing,
  ** the size in the resulting font description will be set to 0.
  */
  pd = pango_font_description_from_string(str); /* eg: Monospace italic 12 */
  pn = pango_font_description_get_family(pd); /* returns family _name_ */
  mask = pango_font_description_get_set_fields(pd);

  DPRINT(("parse_font(): parse check family |%s|\n", pn));

  valid = TRUE;
  for (j = 0 ; j < 6 ; j++) {
	switch (mask & (1 << j)) {
	case 0:
			DPRINT(("Pango attribute %d invalid\n", j));
			valid = FALSE;
			break;
	case (1 << 0):
			DPRINT(("PANGO_FONT_MASK_FAMILY valid\n"));
			break;
	case (1 << 1):
			DPRINT(("PANGO_FONT_MASK_STYLE valid\n"));
			break;
	case (1 << 2):
			DPRINT(("PANGO_FONT_MASK_VARIANT valid\n"));
			break;
	case (1 << 3):
			DPRINT(("PANGO_FONT_MASK_WEIGHT valid\n"));
			break;
	case (1 << 4):
			DPRINT(("PANGO_FONT_MASK_STRETCH valid\n"));
			break;
	case (1 << 5):
			DPRINT(("PANGO_FONT_MASK_SIZE valid\n"));
			break;
	default:
			DPRINT(("Pango attribute %d unknown\n", j));
			valid = FALSE;
			break;
	}
  }

  /* Find out if this family is monospace, and available */
  if (valid) {
	valid = FALSE;
	/* Use only monospace fonts */
	pango_context_list_families(pango_context, &pango_families,
							&familycount);
	for (j = 0 ; j < familycount ; j++) {
		if (pango_font_family_is_monospace(pango_families[j])) {
			pf = pango_font_family_get_name(pango_families[j]);
			DPRINT(("parse_font(): matching |%s| to |%s|\n",
								pf, str));
			if (strncmp(pf, str, strlen(pf) - 1) == SAME) {
				valid = TRUE;
				break;
			}
		}
	}
	if (pango_families != NULL) g_free(pango_families);
  }

  DPRINT(("parse_font():%smatch for |%s|\n", valid ? " " : " no ", pn));

  return valid ? OK : FAILED;
}

/*! show_bits -- Convert an unsigned int to a bitmap string.
** \note The results are stored in a static buffer which is overwritten
** on each call.
** \param value -- Value to be converted.
** \param len -- Maximum output length required.
** \returns Pointer to bitmap: Success
** \returns NULL: Otherwise.
*/

/*
** s h o w _ b i t s
**
** Convert an unsigned int to a bitmap string.
**
** Returns:	Pointer to bitmap	Success
**		NULL			Otherwise
**
** The results are stored in a static buffer which is overwritten
** on each call.
*/
static char *show_bits(unsigned int value, unsigned int len)
{
  size_t t, count;
  unsigned int u;
  static char buff[sizeof(int) * CHAR_BIT + 1];
  char *pc;

  count = sizeof(int) * CHAR_BIT;

  /* Convert the whole integer to a string */
  pc = buff + count;
  *pc-- = '\0';
  for (t = 0 ; t < count ; t++)
	*pc-- = (value & (1 << t)) ? '1' : '0';

  /* Truncate the string to len bits */
  for (pc = buff, t = 0 ; t < (len + 1) ; t++, pc++)
	*pc = *(pc + count - len);
 
  return buff;
}

/*! gio_handler -- Handler for G_IO_IN in g_io_add_watch().
** \param src -- The channel to be watched.
** \param cond -- The condition to be watched for.
** \param data -- A GtkLayout to be drawn on.
** \returns FALSE: Source should be removed.
** \returns TRUE: Otherwise.
*/

/*
** g i o _ h a n d l e r
**
** Handler for G_IO_IN in g_io_add_watch().
**
** Returns:	FALSE	Source should be removed
**		TRUE	Otherwise
*/
static gboolean gio_handler(GIOChannel *src, GIOCondition cond, gpointer data)
{
  GIOStatus gs = G_IO_STATUS_ERROR;
  GError *err = NULL;
  struct Node *pn, *pt;
  int status = TRUE;
  char *pb, *pc;

  if (cond & G_IO_IN) {
	pc = NULL;
	gs = g_io_channel_read_line(
				src,	/* GIOChannel */
				&pc,	/* Line read, NULL on zero length */
				NULL,	/* Length of read data */
				NULL,	/* Position of line terminator */
				&err	/* Error returned */
				);
#if 0
	{ /* Debugging block begin */
	if (gs == G_IO_STATUS_ERROR)
		g_print("gs is ERROR\n");
	else if (gs == G_IO_STATUS_NORMAL)
		g_print("gs is NORMAL\n");
	else if (gs == G_IO_STATUS_EOF)
		g_print("gs is EOF\n");
	else if (gs == G_IO_STATUS_AGAIN)
		g_print("gs is AGAIN\n");
	else
		g_print("gs is unknown\n");

	if (gs == G_IO_STATUS_ERROR)
		g_print("Error |%s|\n", err->message);
	} /* Debugging block end */
#endif

	if (pc != NULL) {
		for (pb = pc ; *pb != '\0' ; pb++)
			/* MT */ ;
		/* Deal with unterminated lines ### */
		if (Data == NULL) {
			pt = malloc(sizeof(struct Node) + (pb - pc) + 1);
			pt->name = (char *) pt + sizeof(struct Node);
			strcpy(pt->name, pc);
			pt->next = NULL;
			Data = pt;
		}
		else {
			for (pn = Data ; pn->next != NULL ; pn = pn->next)
				/* MT */ ;
			if (pn != NULL) {
				pt = malloc(sizeof(struct Node) +
							(pb - pc) + 1);
				pt->name = (char *) pt + sizeof(struct Node);
				strcpy(pt->name, pc);
				pt->next = NULL;
				pn->next = pt;
			}
		}
		free(pc);
		pc = NULL;
	}
  }
  else if (cond & G_IO_ERR) {
	status = FALSE;
  }
  else if (cond & G_IO_HUP) {
	status = FALSE;
  }
  else if (cond & G_IO_NVAL) {
	status = FALSE;
  }

  if (gs == G_IO_STATUS_EOF) status = FALSE;

  /* Update the screen
  gdk_window_process_updates(GTK_LAYOUT(Layout)->bin_window, TRUE);
  gdk_window_process_all_updates();
  */
  gtk_widget_queue_draw(GTK_WIDGET(data));

  return status;
}

/*! popup_handler -- Handler for an "event" signal on the scrolling window.
** \param widget -- The scrolling window.
** \param event -- The trigger for the signal.
** \param data -- The mouse menu.
** \returns TRUE: Event handled.
** \returns FALSE: Otherwise.
*/

/*
** p o p u p _ h a n d l e r
**
** Popup menu handler.
**
** Returns:	TRUE	Event handled
**		FALSE	Otherwise
*/
static gint popup_handler(GtkWidget *widget, GdkEvent *event, gpointer *data)
{
  GdkEventButton *event_button;
  gint status = FALSE;	/* default to event not handled */

  /* Assert macros */
  g_return_val_if_fail(widget != NULL, FALSE);
  g_return_val_if_fail(GTK_IS_MENU(widget), FALSE);
  g_return_val_if_fail(event != NULL, FALSE);

  if (event->type == GDK_BUTTON_PRESS) {
	event_button = (GdkEventButton *) event;
	/* Act on RMB click */
	if (event_button->button == 3) {
		gtk_menu_popup(
			GTK_MENU(widget),
			NULL,
			NULL,
			NULL,
			NULL,
			event_button->button,
			event_button->time
			);
		status = TRUE;
	}
  }

  return status;
}

/*! expose -- Handler for an "expose-event" signal on the GtkLayout.
** \param widget -- A GtkLayout.
** \param event -- The trigger for the signal.
** \param data -- NULL.
** \returns FALSE: Force a destroy signal.
** \returns TRUE: Otherwise.
*/

/*
** e x p o s e _ e v e n t
**
** expose-event callback.
**
** Returns:	TRUE	Event handled
**		FALSE	Otherwise
*/
static gboolean expose(GtkWidget *widget, GdkEventExpose *event, gpointer data)
{
  static PangoLayout *pango_layout = NULL;
  static int w, h;
  static GdkGC *gc = NULL;
  static char buffer[sizeof(struct Enode) + 40] = { '\0' };
  static struct Enode *pe = NULL;
  static struct GnomeGui gui;
  static struct Edwin edwin;
  struct Node *pn;
  size_t t;
  ULONG u;
  int j, x, y, x2, y2, x_sep, y_next;
  UCHAR *pu;
  char *end, *pc;

  if (gc == NULL)
	gc = gdk_gc_new(widget->window);

  if (pango_layout == NULL) {
	pango_layout = gtk_widget_create_pango_layout(widget,
				"01234567890123456789");
	pango_layout_get_pixel_size(pango_layout, &w, &h);
  }

  if (pe == NULL) {
	pe = (struct Enode *) buffer;
	pe->id = 0;
	pe->next = NULL;
	pe->prev = NULL;
	pe->data = NULL;
	pe->name = (char *) pe + sizeof(struct Enode);
	pe->label = (char *) pe + sizeof(struct Enode) + 20;
	pe->signal = 0;
	pe->uiid = 0;
  }

  x = LHS;
  y = 10;
  y_next = y;
  x_sep = 50;

  /* Read the collected data strings, stepping down to the pre-set
  ** y_next position on each return to the LHS.
  */
  for (pn = Data ; pn != NULL ; pn = pn->next) {

	/* We are starting a new row of data */
	if (x == LHS) y = y_next;

	if (strncmp(pn->name, "EDWIN", 5) == SAME) {
		/* Load the local structure from a hex string */
		pc = pn->name + 5;
		for (t = 0, pu = (UCHAR *) &edwin, end = pc ; *end != '\0' &&
			t < sizeof(struct Edwin) ; t++, pu++, pc = end) {
			u = strtoul(pc, &end, 16);
			*pu = (UCHAR) u;
		}

		/* Display the Edwin structure */
		x2 = x;
		y2 = y;
		(void) draw_edwin(widget, gc, &edwin, &x2, &y2);
		x = x2;
		y_next = y2 + 10;
	}
	else if (strncmp(pn->name, "ENODE", 5) == SAME) {
		/* Initialise a local copy */
		j = sscanf(pn->name,
				"%*s %u %p %p %p %18s %18s %u %u",
				&pe->id, (void **) &pe->next,
				(void **) &pe->prev, (void **) &pe->data,
				pe->name, pe->label,
				&pe->signal, &pe->uiid);
		if (j == 8) {
			/* Display a node */
			x2 = x;
			y2 = y;
			(void) draw_enode(widget, gc, pe, &x2, &y2);
			x = x2;
			y_next = y2 + 10;
		}
	}
	else if (strncmp(pn->name, "GUI", 3) == SAME) {
		/* Load the local structure from a hex string */
		pc = pn->name + 3;
		for (t = 0, pu = (UCHAR *) &gui, end = pc ; *end != '\0' &&
			t < sizeof(struct GnomeGui) ; t++, pu++, pc = end) {
			u = strtoul(pc, &end, 16);
			*pu = (UCHAR) u;
		}

		/* Display the GUI structure */
		x2 = x;
		y2 = y;
		(void) draw_gui(widget, gc, &gui, &x2, &y2);
		x = x2;
		y_next = y2 + 10;
	}
	else if (strncmp(pn->name, "LINK", 4) == SAME) {
		/* Display a link */
		x2 = x;
		y2 = y + 10;
		(void) draw_link(widget, gc, &x2, &y2, x_sep);

		/* Writing position y height is unchanged */
		x = x2;
	}
	else if (strncmp(pn->name, "TITLE", 5) == SAME) {
		/* Display a title */
		x = LHS;
		y = y_next;
		for (pc = pn->name ; *pc != '\0' && *pc != ' ' ; pc++)
			/* MT */ ;
		pc++;
		if (*pc != '\0') {
			/* Apparently these don't have to be free'd ??? */
			pango_layout = gtk_widget_create_pango_layout(
								widget, pc);
			gdk_draw_layout(
					GTK_LAYOUT(widget)->bin_window,
					gc,
					x, y,
					pango_layout
					);
		}
		x = LHS;
		y += 20;
		y_next = y;
	}
  }

  return FALSE;
}

/*! draw_link -- Draw a link between output blocks.
** \param widget -- A GtkLayout.
** \param gc -- The GDK graphics context.
** \param px -- The current x-coordinate.
** \param py -- The current y-coordinate.
** \param w -- The width of the link.
** \returns OK: Success.
** \returns FAILED: Otherwise.
*/

/*
** d r a w _ l i n k
**
** Draw a link between output blocks.
**
** Returns:	OK	Success
**		FAILED	Otherwise
*/
static int draw_link(GtkWidget *widget, GdkGC *gc, int *px, int *py, int w)
{
  int x = *px;
  int y = *py;

  gdk_gc_set_line_attributes(
		gc,
		5,			/* Line width */
		GDK_LINE_SOLID,
		GDK_CAP_BUTT,
		GDK_JOIN_BEVEL
		);
  gdk_draw_line(
		GTK_LAYOUT(widget)->bin_window,
		gc,
		/* Why the x offset ??? */
		x - 2, y + 10,		/* From x, y */
		x + w - 2, y + 10	/* To x, y */
		);
  gdk_gc_set_line_attributes(
		gc,
		1,
		GDK_LINE_SOLID,
		GDK_CAP_BUTT,
		GDK_JOIN_BEVEL
		);
  *px = x + w;
  *py = y + 10;

  return OK;
}

/*! draw_enode -- Draw a struct Enode.
** \param widget -- A GtkLayout.
** \param gc -- The GDK graphics context.
** \param pn -- The structure to be drawn.
** \param px -- The current x-coordinate.
** \param py -- The current y-coordinate.
** \returns OK: Success.
** \returns FAILED: Otherwise.
*/

/*
** d r a w _ e n o d e
**
** Draw a struct Enode.
**
** Returns:	OK	Success
**		FAILED	Otherwise
*/
static int draw_enode(GtkWidget *widget, GdkGC *gc, struct Enode *pn,
							int *px, int *py)
{
  static PangoLayout *pango_layout = NULL;
  static int w, h;
  static char buff[EN_FIELDS][64] = { { 0 } };
  static char *formats[EN_FIELDS] = {
			"Id     %u",
			"Next   0x%08lx",
			"Prev   0x%08lx",
			"Data   0x%08lx",
			"Name   ",
			"Label  ",
			"Signal %d",
			"UIid   %d"
			};
  int j, x, y, x2, y2;

  x = *px;
  y = *py;

  for (j = 0 ; j < EN_FIELDS ; j++) {
	switch (j) {
	case 0:
		sprintf(buff[j], formats[j], pn->id);
		break;
	case 1:
		sprintf(buff[j], formats[j], pn->next);
		break;
	case 2:
		sprintf(buff[j], formats[j], pn->prev);
		break;
	case 3:
		sprintf(buff[j], formats[j], pn->data);
		break;
		break;
	case 4:
		strcpy(buff[j], formats[j]);
		strncat(buff[j], pn->name, 12);
		if (buff[j][18] != '\0') buff[j][18] = '#';
		break;
	case 5:
		strcpy(buff[j], formats[j]);
		strncat(buff[j], pn->label, 12);
		if (buff[j][18] != '\0') buff[j][18] = '#';
		break;
	case 6:
		sprintf(buff[j], formats[j], pn->signal);
		break;
	case 7:
		sprintf(buff[j], formats[j], pn->uiid);
		break;
	default:
		break;
	}
  }

  if (pango_layout == NULL) {
	pango_layout = gtk_widget_create_pango_layout(widget,
					"01234567890123456789");
	pango_layout_get_pixel_size(pango_layout, &w, &h);
  }

  /* Calculate the next drawing position */
  x2 = x + w + MGNS;
  y2 = y + EN_FIELDS * (h + MGNS);

  /* Draw the bounding rectangle */
  gdk_draw_rectangle(
		GTK_LAYOUT(widget)->bin_window,
		widget->style->fg_gc[GTK_WIDGET_STATE(widget)],
		FALSE,
		x - MGNS/2, y,				/* x, y */
		w + MGNS, EN_FIELDS * (h + MGNS)	/* w, h */
		);

  /* Draw the data and the separating lines */
  for (j = 0 ; j < EN_FIELDS ; j++, y += h + MGNS) {
	pango_layout_set_text(pango_layout, buff[j], -1);
	gdk_draw_layout(
			GTK_LAYOUT(widget)->bin_window,
			gc,
			x, y + MGNS/2,
			pango_layout
			);
	if (j != (EN_FIELDS - 1)) gdk_draw_line(
			GTK_LAYOUT(widget)->bin_window,
			gc,
			x - MGNS/2, y + h + MGNS,	/* x1, y1 */
			x + w + MGNS/2, y + h + MGNS	/* x2, y2 */
			);
  }

  /* Update the final position */
  *px = x2;
  *py = y2;

  return OK;
}


/*! draw_gui -- Draw a struct GnomeGui.
** \param widget -- A GtkLayout.
** \param gc -- The GDK graphics context.
** \param pg -- The structure to be drawn.
** \param px -- The current x-coordinate.
** \param py -- The current y-coordinate.
** \returns OK: Success.
** \returns FAILED: Otherwise.
*/

/*
** d r a w _ g u i
**
** Draw a struct GnomeGui.
**
** Returns:	OK	Success
**		FAILED	Otherwise
*/
static int draw_gui(GtkWidget *widget, GdkGC *gc, struct GnomeGui *pg,
							int *px, int *py)
{
  static PangoLayout *pango_layout = NULL;
  static int w, h;
  static char buff[GU_FIELDS][128] = { { 0 } };
  struct Entry {
	char *format;
	enum Arg type;
	size_t offset;
	};
  static struct Entry entries[] = {
  { "next               0x%08lx", X, offsetof(struct GnomeGui, next) },
  { "prev               0x%08lx", X, offsetof(struct GnomeGui, prev) },
  { "widget             0x%08lx", X, offsetof(struct GnomeGui, widget) },
  { "toplevel           0x%08lx", X, offsetof(struct GnomeGui, toplevel) },
  { "gwindow            0x%08lx", X, offsetof(struct GnomeGui, gwindow) },
  { "statusbar          0x%08lx", X, offsetof(struct GnomeGui, statusbar) },
  { "msgbar             0x%08lx", X, offsetof(struct GnomeGui, msgbar) },
  { "msgbarbox          0x%08lx", X, offsetof(struct GnomeGui, msgbarbox) },
  { "entry              0x%08lx", X, offsetof(struct GnomeGui, entry) },
  { "entrybox           0x%08lx", X, offsetof(struct GnomeGui, entrybox) },
  { "hpaned             0x%08lx", X, offsetof(struct GnomeGui, hpaned) },
  { "barbox             0x%08lx", X, offsetof(struct GnomeGui, barbox) },
  { "indicators         0x%08lx", X, offsetof(struct GnomeGui, indicators) },
  { "holder             0x%08lx", X, offsetof(struct GnomeGui, holder) },
  { "scrollwin          0x%08lx", X, offsetof(struct GnomeGui, scrollwin) },
  { "menubar            0x%08lx", X, offsetof(struct GnomeGui, menubar) },
  { "notesmenu          0x%08lx", X, offsetof(struct GnomeGui, notesmenu) },
  { "mousemenu          0x%08lx", X, offsetof(struct GnomeGui, mousemenu) },
  { "toolbar            0x%08lx", X, offsetof(struct GnomeGui, toolbar) },
  { "tooltips           0x%08lx", X, offsetof(struct GnomeGui, tooltips) },
  { "menutips           0x%08lx", X, offsetof(struct GnomeGui, menutips) },
  { "acceltab           0x%08lx", X, offsetof(struct GnomeGui, acceltab) },
  { "tooldock           0x%08lx", X, offsetof(struct GnomeGui, tooldock) },
  { "mainactiongroup    0x%08lx", X, offsetof(struct GnomeGui, mainactiongroup) },
  { "mouseactiongroup   0x%08lx", X, offsetof(struct GnomeGui, mouseactiongroup) },
  { "mainuimanager      0x%08lx", X, offsetof(struct GnomeGui, mainuimanager) },
  { "mouseuimanager     0x%08lx", X, offsetof(struct GnomeGui, mouseuimanager) },
  { "gconf_client       0x%08lx", X, offsetof(struct GnomeGui, gconf_client) },
  { "gnome_client       0x%08lx", X, offsetof(struct GnomeGui, gnome_client) },
  { "gnome_program      0x%08lx", X, offsetof(struct GnomeGui, gnome_program) },
  { "window_state       0x%02x", X, offsetof(struct GnomeGui, window_state) },
  { "uid                %u", U, offsetof(struct GnomeGui, uid) },
  { "gconf_id           %u", U, offsetof(struct GnomeGui, gconf_id) },
  { "glog_id            %u", U, offsetof(struct GnomeGui, glog_id) },
  { "snoop_id           %d", U, offsetof(struct GnomeGui, snoop_id) },
  { "mouse_ui_id        %u", U, offsetof(struct GnomeGui, mouse_ui_id) },
  { "entry_id           %u", U, offsetof(struct GnomeGui, entry_id) },
  { "edwin_id           %u", U, offsetof(struct GnomeGui, edwin_id) },
  { "cfgdir_flg         %d", D, offsetof(struct GnomeGui, cfgdir_flg) },
  { "scheme_flg         %d", D, offsetof(struct GnomeGui, scheme_flg) },
  { "block_sel          %d", D, offsetof(struct GnomeGui, block_sel) },
  { "caps_set           %d", D, offsetof(struct GnomeGui, caps_set) },
  { "cur_cmd            %d", D, offsetof(struct GnomeGui, cur_cmd) },
  { "move_cmd           %d", D, offsetof(struct GnomeGui, move_cmd) },
  { "history_name       %.32s", S, offsetof(struct GnomeGui, history_name /* [F_SIZ] */) },
  { "gio_out            0x%08lx", X, offsetof(struct GnomeGui, gio_out) },
  { "drop_type          %d", D, offsetof(struct GnomeGui, drop_type) },
  { "drop_text          %u", U, offsetof(struct GnomeGui, drop_text) },
  { "drop_len           %u", U, offsetof(struct GnomeGui, drop_len) },
  { "drop_fd            %d", D, offsetof(struct GnomeGui, drop_fd) },
  { "main_start_x       %d", D, offsetof(struct GnomeGui, main_start_x) },
  { "main_start_y       %d", D, offsetof(struct GnomeGui, main_start_y) },
  { "main_height        %d", D, offsetof(struct GnomeGui, main_height) },
  { "main_width         %d", D, offsetof(struct GnomeGui, main_width) },
  { "stepper_height     %d", D, offsetof(struct GnomeGui, stepper_height) },
  { "stepper_width      %d", D, offsetof(struct GnomeGui, stepper_width) },
  { "display            0x%08lx", X, offsetof(struct GnomeGui, display) },
  { "screen             0x%08lx", X, offsetof(struct GnomeGui, screen) },
  { "cursor             0x%08lx", X, offsetof(struct GnomeGui, cursor) },
  { "style              0x%08lx", X, offsetof(struct GnomeGui, style) },
  { "theme              0x%08lx", X, offsetof(struct GnomeGui, theme) },
  { "theme_name         %.32s", S, offsetof(struct GnomeGui, theme_name /* [F_SIZ] */) },
  { "pango_context      0x%08lx", X, offsetof(struct GnomeGui, pango_context) },
  { "pango_lang         0x%08lx", X, offsetof(struct GnomeGui, pango_lang) },
  { "def_fontname       %.32s", S, offsetof(struct GnomeGui, def_fontname /* [F_SIZ] */) },
  { "cur_fontname       %.32s", S, offsetof(struct GnomeGui, cur_fontname /* [F_SIZ] */) },
  { "def_print_font     %.32s", S, offsetof(struct GnomeGui, def_print_font /* [F_SIZ] */) },
  { "cur_print_font     %.32s", S, offsetof(struct GnomeGui, cur_print_font /* [F_SIZ] */) },
  { "def_hdr_print_font %.32s", S, offsetof(struct GnomeGui, def_hdr_print_font /* [F_SIZ] */) },
  { "cur_hdr_print_font %.32s", S, offsetof(struct GnomeGui, cur_hdr_print_font /* [F_SIZ] */) },
  { "def_dbg_fontname   %.32s", S, offsetof(struct GnomeGui, def_dbg_fontname /* [F_SIZ] */) },
  { "cur_dbg_fontname   %.32s", S, offsetof(struct GnomeGui, cur_dbg_fontname /* [F_SIZ] */) },
  { "pango_desc         0x%08lx", X, offsetof(struct GnomeGui, pango_desc) },
  { "pango_metrics      0x%08lx", X, offsetof(struct GnomeGui, pango_metrics) },
  { "pango_width        %d", D, offsetof(struct GnomeGui, pango_width) },
  { "pango_height       %d", D, offsetof(struct GnomeGui, pango_height) },
  { "use_def_font       %d", D, offsetof(struct GnomeGui, use_def_font) },
  { "use_def_color      %d", D, offsetof(struct GnomeGui, use_def_color) },
  { "use_def_cursor     %d", D, offsetof(struct GnomeGui, use_def_cursor) },
  { "use_def_print      %d", D, offsetof(struct GnomeGui, use_def_print) },
  { "font_codes         %d", D, offsetof(struct GnomeGui, font_codes /*[32]*/) },
  { "color_types        %04x %04x %04x %08x [..]", O, offsetof(struct GnomeGui, color_types /* [32] */) },
  { "vte_colors         %04x %04x %04x %08x [..]", O, offsetof(struct GnomeGui, vte_colors /* [32] */) },
  { "def_forecolor      %04x %04x %04x %08x", O, offsetof(struct GnomeGui, def_forecolor) },
  { "def_backcolor      %04x %04x %04x %08x", O, offsetof(struct GnomeGui, def_backcolor) },
  { "def_cursorcolor    %04x %04x %04x %08x", O, offsetof(struct GnomeGui, def_cursorcolor) },
  { "def_forename       %.32s", S, offsetof(struct GnomeGui, def_forename /* [F_SIZ] */) },
  { "def_backname       %.32s", S, offsetof(struct GnomeGui, def_backname /* [F_SIZ] */) },
  { "def_cursorname     %.32s", S, offsetof(struct GnomeGui, def_cursorname /* [F_SIZ] */) },
  { "def_cursorblink    %d", D, offsetof(struct GnomeGui, def_cursorblink) },
  { "def_hdrprint       %d", D, offsetof(struct GnomeGui, def_hdrprint) },
  { "cur_forecolor      %04x %04x %04x %08x", O, offsetof(struct GnomeGui, cur_forecolor) },
  { "cur_backcolor      %04x %04x %04x %08x", O, offsetof(struct GnomeGui, cur_backcolor) },
  { "cur_cursorcolor    %04x %04x %04x %08x", O, offsetof(struct GnomeGui, cur_cursorcolor) },
  { "cur_forename       %.32s", S, offsetof(struct GnomeGui, cur_forename /* [F_SIZ] */) },
  { "cur_backname       %.32s", S, offsetof(struct GnomeGui, cur_backname /* [F_SIZ] */) },
  { "cur_cursorname     %.32s", S, offsetof(struct GnomeGui, cur_cursorname /* [F_SIZ] */) },
  { "cur_cursorblink    %d", D, offsetof(struct GnomeGui, cur_cursorblink) },
  { "cur_hdrprint       %d", D, offsetof(struct GnomeGui, cur_hdrprint) },
  { "horiz_bar          0x%08lx", X, offsetof(struct GnomeGui, horiz_bar) },
  { "vert_bar           0x%08lx", X, offsetof(struct GnomeGui, vert_bar) },
  { "horiz_range        0x%08lx", X, offsetof(struct GnomeGui, horiz_range) },
  { "vert_range         0x%08lx", X, offsetof(struct GnomeGui, vert_range) },
  { "horiz_adj          0x%08lx", X, offsetof(struct GnomeGui, horiz_adj) },
  { "vert_adj           0x%08lx", X, offsetof(struct GnomeGui, vert_adj) },
  { "print_dialog       0x%08lx", X, offsetof(struct GnomeGui, print_dialog) },
  { "print_settings     0x%08lx", X, offsetof(struct GnomeGui, print_settings) },
  { "print_operation    0x%08lx", X, offsetof(struct GnomeGui, print_operation) },
  { "prog_name          0x%08lx", X, offsetof(struct GnomeGui, prog_name) },
  { "lastprint          %.32s", S,   offsetof(struct GnomeGui, lastprint /* [PATH_MAX] */) },
  { "dictionary         %.32s", S, offsetof(struct GnomeGui, dictionary /* [F_SIZ] */) },
  { "spellcheck         %.32s", S, offsetof(struct GnomeGui, spellcheck /* [F_SIZ] */) },
  { "vterm              0x%08lx", X, offsetof(struct GnomeGui, vterm) },
  { "def_hints          0x%08lx", X, offsetof(struct GnomeGui, def_hints) },
  { "def_rows           %d", D, offsetof(struct GnomeGui, def_rows) },
  { "def_cols           %d", D, offsetof(struct GnomeGui, def_cols) },
  { "bar_cols           %d", D, offsetof(struct GnomeGui, bar_cols) },
  { "menu_state         %s", B, offsetof(struct GnomeGui, menu_state) },
  { "edhead             0x%08lx", X, offsetof(struct GnomeGui, edhead) },
  { "edcount            %d", D, offsetof(struct GnomeGui, edcount) },
#if 0
  { "ringhead           0x%08lx", X, offsetof(struct GnomeGui, ringhead) },
  { "ringcount          %d", D, offsetof(struct GnomeGui, ringcount) },
#endif
  { "docshead           0x%08lx", X, offsetof(struct GnomeGui, docshead) },
  { "docscount          %d", D, offsetof(struct GnomeGui, docscount) },
  { "noteshead          0x%08lx", X, offsetof(struct GnomeGui, noteshead) },
  { "notescount         %d", D, offsetof(struct GnomeGui, notescount) }
  };
  size_t t;
  ULONG value;
  int j, x, y, x2, y2;
  GdkColor color = { 0, 0, 0, 0 };

#if 0
  g_print("sizes %u\n", sizeof(entries) / sizeof(entries[0]));
#endif
  assert((sizeof(entries) / sizeof(entries[0])) == GU_FIELDS);

  x = *px;
  y = *py;

  for (t = 0 ; t < (sizeof(entries) / sizeof(entries[0])) ; t++) {
	switch (entries[t].type) {
	case X:
		value = *(ULONG *) ((char *) pg + entries[t].offset);
		sprintf(buff[t], entries[t].format, value);
		/* Lose excessive NULLs */
		if (value == 0)
			*(buff[t] + 20) = '\0';
		break;
	case U:
		sprintf(buff[t], entries[t].format,
				*(UINT *) ((char *) pg + entries[t].offset));
		break;
	case B:
		sprintf(buff[t], entries[t].format,
		show_bits(*(UINT *) ((char *) pg + entries[t].offset), 8));
		break;
	case D:
		sprintf(buff[t], entries[t].format,
				*(int *) ((char *) pg + entries[t].offset));
		break;
	case S:
		/* Mark the end of truncated strings */
		buff[t][50] = '\0';
		sprintf(buff[t], entries[t].format,
				(char *) ((char *) pg + entries[t].offset));
		if (buff[t][50] != '\0') { 
			buff[t][50] = '#';
			buff[t][51] = '\0';
		}
		break;
	case O:
		color = *(GdkColor *) ((char *) pg + entries[t].offset);
		sprintf(buff[t], entries[t].format,
			color.red, color.green, color.blue, color.pixel);
		break;
	default:
		sprintf(buff[t], "Unknown entry type %d", entries[t].type);
		break;
	}
  }

  if (pango_layout == NULL) {
	pango_layout = gtk_widget_create_pango_layout(widget,
		"012345678901234567890123456789012345678901234567890123456789"
		);
	pango_layout_get_pixel_size(pango_layout, &w, &h);
  }

  /* Calculate the next drawing position */
  x2 = x + w + MGNS;
  y2 = y + GU_FIELDS * (h + MGNS);

  /* Draw the bounding rectangle */
  gdk_draw_rectangle(
		GTK_LAYOUT(widget)->bin_window,
		widget->style->fg_gc[GTK_WIDGET_STATE(widget)],
		FALSE,
		x - MGNS/2, y,				/* x, y */
		w + MGNS, GU_FIELDS * (h + MGNS)	/* w, h */
		);

  /* Draw the data and the separating lines */
  for (j = 0 ; j < GU_FIELDS ; j++, y += h + MGNS) {
	pango_layout_set_text(pango_layout, buff[j], -1);
	gdk_draw_layout(
			GTK_LAYOUT(widget)->bin_window,
			gc,
			x, y + MGNS/2,
			pango_layout
			);
	if (j != (GU_FIELDS - 1)) gdk_draw_line(
			GTK_LAYOUT(widget)->bin_window,
			gc,
			x - MGNS/2, y + h + MGNS,	/* x1, y1 */
			x + w + MGNS/2, y + h + MGNS	/* x2, y2 */
			);
  }

  /* Update the final position */
  *px = x2;
  *py = y2;

  return OK;
}

/*! draw_edwin -- Draw a struct Edwin.
** \param widget -- A GtkLayout.
** \param gc -- The GDK graphics context.
** \param pe -- The structure to be drawn.
** \param px -- The current x-coordinate.
** \param py -- The current y-coordinate.
** \returns OK: Success.
** \returns FAILED: Otherwise.
*/

/*
** d r a w _ e d w i n
**
** Draw an EDWIN structure.
**
** Returns:	OK	Success
**		FAILED	Otherwise
*/
static int draw_edwin(GtkWidget *widget, GdkGC *gc, struct Edwin *pe,
							int *px, int *py)
{
  static PangoLayout *pango_layout = NULL;
  static int w, h;
  static char buff[ED_FIELDS][128] = { { 0 } };
  struct Entry {
	char *format;
	enum Arg type;
	size_t offset;
	};
  static struct Entry entries[] = {
  { "id              %u", U,      offsetof(struct Edwin, id) },
  { "prev            0x%08lx", X, offsetof(struct Edwin, prev) },
  { "next            0x%08lx", X, offsetof(struct Edwin, next) },
  { "gn_gui          0x%08lx", X, offsetof(struct Edwin, gn_gui) },
  { "argshead        0x%08lx", X, offsetof(struct Edwin, argshead) },
  { "argscount       %d", D, offsetof(struct Edwin, argscount) },
  { "st_id           %u", U,      offsetof(struct Edwin, st_id) },
  { "ms_id           %u", U,      offsetof(struct Edwin, ms_id) },
  { "edname          0x%08lx", X, offsetof(struct Edwin, edname) },
  { "edtitle         0x%08lx", X, offsetof(struct Edwin, edtitle) },
  { "ro_flg          %02x", C,    offsetof(struct Edwin, ro_flg) },
  { "keybuff         %.20s", S,   offsetof(struct Edwin, keybuff /* [BUF_SIZ] */) },
  { "keycount        %d", D,      offsetof(struct Edwin, keycount) },
  { "vterm           0x%08lx", X, offsetof(struct Edwin, vterm) },
  { "eventbox        0x%08lx", X, offsetof(struct Edwin, eventbox) },
  { "key_id          %u", U,      offsetof(struct Edwin, key_id) },
  { "mouse_dn_id     %u", U,      offsetof(struct Edwin, mouse_dn_id) },
  { "mouse_up_id     %u", U,      offsetof(struct Edwin, mouse_up_id) },
  { "mouse_mv_id     %u", U,      offsetof(struct Edwin, mouse_mv_id) },
  { "scroll_id       %u", U,      offsetof(struct Edwin, scroll_id) },
  { "attr_mask       %u", U,      offsetof(struct Edwin, attr_mask) },
  { "clr_mask        %u", U,      offsetof(struct Edwin, clr_mask) },
  { "fg_clr          %ld", L,     offsetof(struct Edwin, fg_clr) },
  { "bg_clr          %ld", L,     offsetof(struct Edwin, bg_clr) },
  { "use_def_font    %d", D,      offsetof(struct Edwin, use_def_font) },
  { "use_def_color   %d", D,      offsetof(struct Edwin, use_def_color) },
  { "use_def_cursor  %d", D,      offsetof(struct Edwin, use_def_cursor) },
  { "colormap        0x%08lx", X, offsetof(struct Edwin, colormap) },
  { "forecolor       %04x %04x %04x %08x", O, offsetof(struct Edwin, forecolor) },
  { "backcolor       %04x %04x %04x %08x", O, offsetof(struct Edwin, backcolor) },
  { "cursorcolor     %04x %04x %04x %08x", O, offsetof(struct Edwin, cursorcolor) },
  { "cursorblink     %d", D,      offsetof(struct Edwin, cursorblink) },
  { "fontname        %.20s", S,   offsetof(struct Edwin, fontname /* [F_SIZ] */) },
  { "pango_lang      0x%08lx", X, offsetof(struct Edwin, pango_lang) },
  { "pango_context   0x%08lx", X, offsetof(struct Edwin, pango_context) },
  { "pango_desc      0x%08lx", X, offsetof(struct Edwin, pango_desc) },
  { "pango_metrics   0x%08lx", X, offsetof(struct Edwin, pango_metrics) },
  { "pango_width     %d", D,      offsetof(struct Edwin, pango_width) },
  { "pango_height    %d", D,      offsetof(struct Edwin, pango_height) },
  { "hints (b,i)     %d %d %d %d [..]", G, offsetof(struct Edwin, hints) },
  { "xpad            %d", D,      offsetof(struct Edwin, xpad) },
  { "ypad            %d", D,      offsetof(struct Edwin, ypad) },
  { "cols            %ld", L,     offsetof(struct Edwin, cols) },
  { "rows            %ld", L,     offsetof(struct Edwin, rows) }
  };
  size_t t;
  ULONG value;
  int j, x, y, x2, y2;
  GdkColor color = { 0, 0, 0, 0 };
  GdkGeometry hint = { 0, 0, 0, 0, 0, 0, 0, 0, 0.0, 0.0, 0 };

  assert((sizeof(entries) / sizeof(entries[0])) == ED_FIELDS);

  x = *px;
  y = *py;

  for (t = 0 ; t < (sizeof(entries) / sizeof(entries[0])) ; t++) {
	switch (entries[t].type) {
	case X:
		value = *(ULONG *) ((char *) pe + entries[t].offset);
		sprintf(buff[t], entries[t].format, value);
		/* Lose the excessive NULLs */
		if (value == 0)
			*(buff[t] + 17) = '\0';
		break;
	case U:
		sprintf(buff[t], entries[t].format,
				*(UINT *) ((char *) pe + entries[t].offset));
		break;
	case L:
		sprintf(buff[t], entries[t].format,
				*(long int *) ((char *) pe + entries[t].offset));
		break;
	case D:
		sprintf(buff[t], entries[t].format,
				*(int *) ((char *) pe + entries[t].offset));
		break;
	case S:
		/* Mark the end of truncated strings */
		buff[t][35] = '\0';
		sprintf(buff[t], entries[t].format,
				(char *) ((char *) pe + entries[t].offset));
		if (buff[t][35] != '\0') { 
			buff[t][35] = '#';
			buff[t][36] = '\0';
		}
		break;
	case C:
		/* Simply using %c doesn't work; assume CHAR_WIDTH 8 */
		sprintf(buff[t], entries[t].format,
			(int) (*((char *) pe + entries[t].offset) & 0xff));
		break;
	case O:
		color = *(GdkColor *) ((char *) pe + entries[t].offset);
		sprintf(buff[t], entries[t].format,
			color.red, color.green, color.blue, color.pixel);
		break;
	case G:
		hint = *(GdkGeometry *) ((char *) pe + entries[t].offset);
		sprintf(buff[t], entries[t].format,
#if 0
			hint.min_width, hint.min_height,
			hint.max_width, hint.max_height,
			hint.base_width, hint.base_height,
			hint.width_inc, hint.height_inc
#else
			hint.base_width, hint.base_height,
			hint.width_inc, hint.height_inc
#endif
			);
		break;
	default:
		sprintf(buff[t], "Unknown entry type %d", entries[t].type);
		break;
	}
  }

  if (pango_layout == NULL) {
	pango_layout = gtk_widget_create_pango_layout(widget,
		"0123456789012345678901234567890123456789"
		);
	pango_layout_get_pixel_size(pango_layout, &w, &h);
  }

  /* Calculate the next drawing position */
  x2 = x + w + MGNS;
  y2 = y + ED_FIELDS * (h + MGNS);

  /* Draw the bounding rectangle */
  gdk_draw_rectangle(
		GTK_LAYOUT(widget)->bin_window,
		widget->style->fg_gc[GTK_WIDGET_STATE(widget)],
		FALSE,
		x - MGNS/2, y,				/* x, y */
		w + MGNS, ED_FIELDS * (h + MGNS)	/* w, h */
		);

  /* Draw the data and the separating lines */
  for (j = 0 ; j < ED_FIELDS ; j++, y += h + MGNS) {
	pango_layout_set_text(pango_layout, buff[j], -1);
	gdk_draw_layout(
			GTK_LAYOUT(widget)->bin_window,
			gc,
			x, y + MGNS/2,
			pango_layout
			);
	if (j != (ED_FIELDS - 1)) gdk_draw_line(
			GTK_LAYOUT(widget)->bin_window,
			gc,
			x - MGNS/2, y + h + MGNS,	/* x1, y1 */
			x + w + MGNS/2, y + h + MGNS	/* x2, y2 */
			);
  }

  /* Update the final position */
  *px = x2;
  *py = y2;

  return OK;
}

/*! main -- main function of elvisdbg.c.
** \param argc -- Number of elements in argv.
** \param argv -- Array of pointers to argument strings.
** \returns 0: Always.
*/

/*
** m a i n
**
** Main function of elvisdbg.c
*/
int main(int argc, char *argv[])
{
  GtkWidget *window;
  GtkWidget *statusbar;
  GtkWidget *vbox;
  GtkWidget *layout;
  GtkWidget *scrolledwin;
  GtkWidget *menubar;
  GtkWidget *submenu;
  GtkWidget *subsubmenu;
  GtkWidget *mainmenu_1;
  GtkWidget *mainmenu_2;
  GtkWidget *menuitem_1;
  GtkWidget *menuitem_2;
  GtkWidget *submenuitem_1;
  GtkWidget *submenuitem_2;
  GtkWidget *mousesubmenu;
  GtkWidget *mousemenuitem_1;
  GtkWidget *mousemenuitem_2;
  GtkWidget *toolbar;
  GtkToolItem *tb_open;
  GtkToolItem *tb_close;
  GtkTooltips *tooltips;
  GtkAccelGroup *acceltab;
  GtkAdjustment *hadj, *vadj;

  PangoFontDescription *pango_desc;
  PangoLayout *pango_layout;
  int w, h;

  GnomeProgram *app;
  const char **args;
  poptContext gnome_ctx;

  GIOChannel *gio;
  guint ev_id;
  char *pc, *pf;

  guint st_id;
  volatile size_t t;
  char buff[128];

  /* Park the program name, not the path, somewhere accessible */
  for (pf = pc = argv[0] ; *pc != '\0' ; pc++)
	if (*pc == SEPCHR) pf = pc;
  if (*pf == SEPCHR) pf++;
  progname[0] = '\0';
  strncat(progname, pf, PATH_MAX);

  /* Initialise GTK+ */
  gtk_init(&argc, &argv);

  /* We assume PACKAGE and VERSION are set to the program name and version
  ** number respectively. Also, assume that 'options' is a global array of
  ** poptOptions.
  */
  app = gnome_program_init(
			"elvisdbg",
			VERSION,
			LIBGNOMEUI_MODULE,
			argc,
			argv,
			GNOME_PROGRAM_STANDARD_PROPERTIES, /* Set lib paths */
			GNOME_PARAM_HUMAN_READABLE_NAME, "elvisdbg",
			GNOME_PARAM_ENABLE_SOUND, FALSE,
			GNOME_PARAM_POPT_TABLE, gnome_popt,
			/* GNOME_PARAM_POPT_CONTEXT, &gnome_ctx, FAILS */
			GNOME_PARAM_NONE
			);
  /* Apparently a work-around for a bug in GNOME_PARAM_POPT_CONTEXT */
  g_object_get(G_OBJECT(app), GNOME_PARAM_POPT_CONTEXT, &gnome_ctx, NULL);
  args = poptGetArgs(gnome_ctx);

  /* Check that --fontname=FONT gives a valid font */
  buff[0] = '\0';
  if (Fontopt != NULL) {
	if (parse_font(Fontopt) == OK) {
		Fontname[0] = '\0';
		strncat(Fontname, Fontopt, F_MAX);
	}
	else {
		sprintf(buff, "Font %.60s not found", Fontopt);
		strcpy(Fontname, DEFAULT_FONT_NAME);
	}
  }
  else
	strcpy(Fontname, DEFAULT_FONT_NAME);

#if 0
  { /* Debugging block begin */
  int j;
  for (j = 0 ; j < argc ; j++)
	g_print("argv %d is |%s|\n", j, argv[j]);
  g_print("new_args is %s\n", args == NULL ? "NULL" : "valid");
  for (j = 0 ; args != NULL && args[j] != NULL ; j++) {
	g_print("args[%d] is |%s|\n", j, args[j] == NULL ? "NULL" : args[j]);
  g_print("fontname |%s|\n", o_fontname);
  }
  } /* Debugging block end */
#endif

  /* Set up the input channel */
  if (args == NULL) {
	/* We're reading from stdin */
	gio = g_io_channel_unix_new(0);
  }
  else {
	/* Any remaining argument must be the single input file permitted */
	if ((gio = g_io_channel_new_file(args[0], "r", NULL)) == NULL) {
		DPRINT(("In %s: cannot open file %s\n", progname, args[0]));
		exit(EXIT_FAILURE);
	}
  }
  g_io_channel_set_encoding(gio, NULL, NULL);
  /* g_io_channel_set_buffered(gio, FALSE);
  */
  g_io_channel_set_flags(gio,
	g_io_channel_get_flags(gio) | G_IO_FLAG_NONBLOCK, NULL);

  /* Create a new window */
  window = gtk_window_new(GTK_WINDOW_TOPLEVEL);
  gtk_window_set_title(GTK_WINDOW(window), "Elvis GUI data structures");
  /* Main window border is _outside_ the window */
  gtk_container_border_width(GTK_CONTAINER(window), 2);
  gtk_widget_set_usize(GTK_WIDGET(window), 500, 500);

  /* The delete_event signal is given by the window manager */
  g_signal_connect(
		G_OBJECT(window),
		"delete-event",
		G_CALLBACK(delete_event),
		NULL
		);

  /* Run if gtk_widget_destroy() is called on the window,
  ** or if the "delete-event" callback returns FALSE.
  */
  g_signal_connect(
		G_OBJECT(window),
		"destroy",
		G_CALLBACK(destroy),
		NULL
		);

  /* Build a statusbar */
  statusbar = gtk_statusbar_new();

  /* Build a toolbar */
  toolbar = gtk_toolbar_new();

  /* Add an open button */
  tb_open = gtk_tool_button_new_from_stock(GTK_STOCK_ADD);
  gtk_toolbar_insert(GTK_TOOLBAR(toolbar), tb_open, -1);
  g_signal_connect(
		G_OBJECT(tb_open),
		"clicked",
		G_CALLBACK(hello),
		"Open"
		);

  /* Add a close button */
  tb_close = gtk_tool_button_new_from_stock(GTK_STOCK_CLOSE);
  gtk_toolbar_insert(GTK_TOOLBAR(toolbar), tb_close, -1);
  g_signal_connect(
		G_OBJECT(tb_close),
		"clicked",
		G_CALLBACK(hello),
		"Close"
		);

  /* Build a tooltips holder */
  tooltips = gtk_tooltips_new();

  /* Add tooltips to the toolbar buttons */
  gtk_tool_item_set_tooltip(tb_open, tooltips, "Open", NULL);
  gtk_tool_item_set_tooltip(tb_close, tooltips, "Close", NULL);

  /* Toolbar tooltips are enabled by default, and this function
  ** doesn't seem to disable them. The enable/disable functions
  ** control _all_ tooltips, not just those on a toolbar.
  */
  gtk_widget_show(toolbar);
  gtk_toolbar_set_tooltips(GTK_TOOLBAR(toolbar), FALSE);

  /* Build a sub sub menu - _don't_ use gtk_show_widget() on this */
  subsubmenu = gtk_menu_new();

  submenuitem_1 = gtk_menu_item_new_with_label("SubItem_1");
  gtk_menu_shell_append(GTK_MENU_SHELL(subsubmenu), submenuitem_1);
  submenuitem_2 = gtk_menu_item_new_with_label("SubItem_2");
  gtk_menu_shell_append(GTK_MENU_SHELL(subsubmenu), submenuitem_2);

  /* Build a sub menu - _don't_ use gtk_show_widget() on this */
  submenu = gtk_menu_new();

  /* Add a couple of menu items to the sub menu */
  menuitem_1 = gtk_menu_item_new_with_label("Item_1");
  gtk_menu_shell_append(GTK_MENU_SHELL(submenu), menuitem_1);
  menuitem_2 = gtk_menu_item_new_with_label("Item_2");
  gtk_menu_shell_append(GTK_MENU_SHELL(submenu), menuitem_2);

  /* Add the cascading menu */
  gtk_menu_item_set_submenu(GTK_MENU_ITEM(menuitem_2), subsubmenu);

  /* Create a menu item to act as the root of the sub menu */
  mainmenu_1 = gtk_menu_item_new_with_label("Main_1");

  /* Attach the sub menu to the root menu item */
  gtk_menu_item_set_submenu(GTK_MENU_ITEM(mainmenu_1), submenu);

  /* Build a menubar */
  menubar = gtk_menu_bar_new();

  /* Attach the root of the menu to the menu bar */
  gtk_menu_bar_append(GTK_MENU_BAR(menubar), mainmenu_1);

  /* Add a second menu item to the menubar */
  mainmenu_2 = gtk_menu_item_new_with_label("Main_2");
  gtk_menu_bar_append(GTK_MENU_BAR(menubar), mainmenu_2);

  /* Connect the main menu items to callbacks */
  g_signal_connect(
		G_OBJECT(menuitem_1),
		"activate",
		G_CALLBACK(hello),
		"Item_1"
		);
  g_signal_connect(
		G_OBJECT(menuitem_2),
		"activate",
		G_CALLBACK(hello),
		"Item_2"
		);

  /* Size the layout to match the default font */
  layout = gtk_layout_new(NULL, NULL);
  pango_layout = gtk_widget_create_pango_layout(layout,
					"01234567890123456789");
  pango_layout_get_pixel_size(pango_layout, &w, &h);
  gtk_layout_set_size(GTK_LAYOUT(layout), w * 20U, h * 200U);
  gtk_widget_add_events(layout, GDK_BUTTON_PRESS_MASK);
  g_signal_connect(
		G_OBJECT(layout),
		"expose-event",
		G_CALLBACK(expose),
		NULL
		);

  /* Initialise scrollbar adjustments, needed for scrollbar arrows */
  hadj = GTK_ADJUSTMENT(gtk_adjustment_new(0.0, 0.0, w * 20.0, w * 2.0,
							w * 100.0, 0.0));
  vadj = GTK_ADJUSTMENT(gtk_adjustment_new(0.0, 0.0, h * 200.0, h * 2.0,
							h * 100.0, 0.0));

  /* Set up a layout in a scrollbox */
  scrolledwin = gtk_scrolled_window_new(hadj, vadj);
  gtk_scrolled_window_set_policy(GTK_SCROLLED_WINDOW(scrolledwin),
			GTK_POLICY_ALWAYS, GTK_POLICY_ALWAYS);
  /* gtk_container_set_border_width(GTK_CONTAINER(scrolledwin), 10);
  */

#if 1
  gtk_container_add(GTK_CONTAINER(scrolledwin), layout);
#else
  gtk_scrolled_window_add_with_viewport(
			GTK_SCROLLED_WINDOW(scrolledwin), layout);
#endif

  /* Build and load a multi-item container */
  vbox = gtk_vbox_new(
		FALSE,		/* homogenous */
		1		/* spacing */
		);
  gtk_box_pack_end(
		GTK_BOX(vbox),	/* box */
		statusbar,	/* child */
		FALSE,		/* expand */
		FALSE,		/* fill */
		0		/* padding */
		);

  gtk_box_pack_end(GTK_BOX(vbox), scrolledwin, TRUE, TRUE, 0);
  gtk_box_pack_end(GTK_BOX(vbox), toolbar, FALSE, FALSE, 0);
  gtk_box_pack_end(GTK_BOX(vbox), menubar, FALSE, FALSE, 0);

  /* Add the container to the main window */
  gtk_container_add(GTK_CONTAINER(window), vbox);

  /* Build an accelerator table */
  acceltab = gtk_accel_group_new();

  /* Attach it to the main window */
  gtk_window_add_accel_group(GTK_WINDOW(window), acceltab);

  /* Attach it to the (first) submenu */
  gtk_menu_set_accel_group(GTK_MENU(submenu), acceltab);

  /* Build a mouse sub menu */
  mousesubmenu = gtk_menu_new();

  /* Add two items to the mouse sub menu */
  mousemenuitem_1 = gtk_menu_item_new_with_label("Mouse_1");
  gtk_menu_shell_append(GTK_MENU_SHELL(mousesubmenu), mousemenuitem_1);
  mousemenuitem_2 = gtk_menu_item_new_with_label("Mouse_2");
  gtk_menu_shell_append(GTK_MENU_SHELL(mousesubmenu), mousemenuitem_2);

  /* Connect the mouse menu items to callbacks */
  g_signal_connect(
		G_OBJECT(mousemenuitem_1),
		"activate",
		G_CALLBACK(hello),
		"Mouse_1"
		);
  g_signal_connect(
		G_OBJECT(mousemenuitem_2),
		"activate",
		G_CALLBACK(hello),
		"Mouse_2"
		);

  /* connect a popup menu handler */
  gtk_signal_connect_object(
			GTK_OBJECT(scrolledwin),
			"event",
			GTK_SIGNAL_FUNC(popup_handler),
			GTK_OBJECT(mousesubmenu)
			);

  /* Sort out fonts
  if ((pango_lang = gtk_get_default_language()) == NULL)
	pango_lang = pango_language_from_string(DEFAULT_PANGO_LANG);
  */
  pango_desc = pango_font_description_from_string(Fontname);
  gtk_widget_modify_font(layout, pango_desc);

  /* Display the newly created widgets */
  gtk_widget_show(GTK_WIDGET(tb_open));
  gtk_widget_show(GTK_WIDGET(tb_close));
  gtk_widget_show(mainmenu_1);
  gtk_widget_show(mainmenu_2);
  gtk_widget_show(menuitem_1);
  gtk_widget_show(menuitem_2);
  gtk_widget_show(submenuitem_1);
  gtk_widget_show(submenuitem_2);
  /* omit submenu */
  /* omit subsubmenu */
  gtk_widget_show(menubar);
  gtk_widget_show(statusbar);
  gtk_widget_show(layout);
  gtk_widget_show(scrolledwin);
  gtk_widget_show(vbox);

  /* tooltips are enabled by default.
  gtk_tooltips_enable(tooltips);
  */

  /* Show the mouse menu sub-items */
  gtk_widget_show(mousemenuitem_1);
  gtk_widget_show(mousemenuitem_2);

  /* Read the input */
  ev_id = g_io_add_watch(
		gio,			/* Channel */
#if 1
		G_IO_IN,		/* Watch condition */
#else
		G_IO_IN | G_IO_ERR |
		G_IO_HUP | G_IO_NVAL,
#endif
		gio_handler,		/* Handler */
		(void *) layout		/* Handler data */
		);

  /* Display any error messages on the statusbar */
  if (buff[0] != '\0') {
	st_id = gtk_statusbar_get_context_id(GTK_STATUSBAR(statusbar), "1");
	gtk_statusbar_push(GTK_STATUSBAR(statusbar), st_id, buff);
  }

  /* Display the window */
  gtk_widget_show(window);

  /* Enter the main event loop */
  gtk_main();

  return 0;
}

/*
** EOF
*/

