/*
** guignome.c - Gnome GUI for the Elvis editor.
**
** Written by C W Rose.
**
** $Log$
**
** NOTE:
** This front-end has been developed using GTK+ 2.8.8 and Gnome 2.12.2,
** and GTK+ 2.10.9 and Gnome 2.16.2.  The main difference between the
** versions seems to be the addition of the print library to Gnome.
**
** The code contains a lot of comments mostly extracted from the GTK+
** documentation describing various library calls and their uses. It
** probably won't run on earlier versions, certainly not unless PRINT_MENU
** is #undefined; printing is then available only through Elvis' standard
** print commands.  However, the code is reasonably future-proof, since
** the use of deprecated GTK+ calls has been avoided.  The one exception
** is the deprecated gnome_entry history widget, for which there seems
** to  be no substitute.
**
** Error checking with GTK+ is a problem, since most library functions
** have no return codes.  However, this program's error checking could
** still be more extensive.
**
** If the startup sizes and positions aren't the defaults, the main window
** bounces around a lot.  Gnome needs the window to be realised before
** making adjustments, but the process should be smoother.  It can be
** done - gnome-terminal is an example.
**
** Color and font handling works, but is still a bit odd.  The combination
** of Elvis' ideas of fonts, of GTK's ideas, and of VTE's ideas, is not a
** happy one.
**
** Scrolling via the scrollbar also works, but the handling of scrollbar
** movement in the main Elvis code doesn't use exact numbers, so scroll
** movements (most noticeably towards the end of a file) aren't precise.
**
** Points of confusion have been marked with ???.
** Points of breakdown have been marked with ###.
**
** To add the debugging menu, set the #define DEBUG_MENU below.
*/

/* Description and version number */
#define DESCR "Gnome GUI for Elvis."
#define GNOMEGUI_VERSION "Version 0.93 of 1 Jan 08."

/* Assert macro control, ANSI C and GLib */
#define NDEBUG
#define G_DISABLE_ASSERT

/* Needed only for access(), fdopen() and kill() */
#define _POSIX_SOURCE

/* Include files */
#include <linux/limits.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <assert.h>
#include <ctype.h>
#include <errno.h>
#include <fcntl.h>
#include <math.h>
#include <pwd.h>
#include <signal.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <unistd.h>
#include <wait.h>
/* Undocumented, but required for font metrics */
#define PANGO_ENABLE_BACKEND
#include <gdk/gdkx.h>
#include <gdk-pixbuf/gdk-pixdata.h>
#include <gtk/gtk.h>
#include <gconf/gconf-client.h>
/* GNU gettext i18n is not used.
#include <glib.h>
#include <glib/gi18n.h>
*/
#include <pango/pango.h>
#include <libbonobo.h>
#include <bonobo.h>
#include <libgnome/libgnome.h>
#include <libgnomeui/libgnomeui.h>
#include <vte/vte.h>

#include "../elvis.h"

#define EXTERN
#include "guignome.h"

/* Icon pixel buffers */
#include "elvis-icon.h"
#include "down_std-icon.h"
#include "down_blk-icon.h"
#include "down_wht-icon.h"
#include "down_lge-icon.h"
#include "gnbl-icon.h"
#include "rdbl-icon.h"
#include "gngn-icon.h"
#include "rdgn-icon.h"

/* Code control defines */

/* Optional features */
#define DEPRECATED	/* Use the deprecated gnome_entry widget */
#define PRINT_MENU	/* Add a Print sub-menu to the main menu */
#undef CANCEL_ACTION	/* Add a toolbar Cancel button */
#undef G_MEM_PROFILE	/* Use GLib memory profiling */
#undef DEBUG_MENU	/* Add a Debugging menu */
#undef DEBUG_MOUSE	/* Add mouse test code */
#undef DEBUG
#undef TEST

/* Experimental code */
#undef ADD_UNUSED	/* Add functions currently unused */
#undef CHG_ARGSLIST	/* Change the current window's arglist */
#undef SAVE_CONFIG	/* Menu item to save and restore configurations */
#undef SESSION_MGR	/* Enable session-management */

/* Debugging message defines.
** Switch first and third entries for effect.
*/
#define DPRINT_A(x) (g_print x)
#undef DPRINT_A			/* :args list, newfiles(), newcmd() */
#undef DPRINT_AA		/* Extra information */
#define DPRINT_A(x)

#define DPRINT_B(x) (g_print x)
#undef DPRINT_B			/* Clipboard */
#define DPRINT_B(x)

#define DPRINT_C(x) (g_print x)
#undef DPRINT_C			/* Color parsing */
#undef DPRINT_CC		/* Extra information */
#define DPRINT_C(x)

#define DPRINT_D(x) (g_print x)
#undef DPRINT_D			/* Drag-drop and selection code */
#undef DPRINT_DD		/* Extra information */
#define DPRINT_D(x)

#define DPRINT_E(x) (g_print x)
#undef DPRINT_E			/* Edwin code */
#define DPRINT_E(x)

#define DPRINT_F(x) (g_print x)
#undef DPRINT_F			/* Function names */
#define DPRINT_F(x)

#define DPRINT_G(x) (g_print x)
#undef DPRINT_G			/* GConf handling code */
#define DPRINT_G(x)

#define DPRINT_H(x) (g_print x)
#undef DPRINT_H			/* Main window code */
#undef DPRINT_HH		/* Extra information */
#define DPRINT_H(x)

#define DPRINT_I(x) (g_print x)
#undef DPRINT_I			/* init() code */
#undef DPRINT_II		/* Extra information */
#define DPRINT_I(x)

#define DPRINT_J(x) (g_print x)
#undef DPRINT_J			/* Menuing code */
#undef DPRINT_JJ		/* Extra information */
#define DPRINT_J(x)

#define DPRINT_K(x) (g_print x)
#undef DPRINT_K			/* Key handler code */
#define DPRINT_K(x)

#define DPRINT_L(x) (g_print x)
#undef DPRINT_L			/* get_icons() code */
#undef DPRINT_LL		/* Extra information */
#define DPRINT_L(x)

#define DPRINT_M(x) (g_print x)
#undef DPRINT_M			/* Mouse handler code */
#undef DPRINT_MM		/* Extra information */
#define DPRINT_M(x)

#define DPRINT_N(x) (g_print x)
#undef DPRINT_N			/* Font parsing */
#undef DPRINT_NN		/* Extra information */
#define DPRINT_N(x)

#define DPRINT_P(x) (g_print x)
#undef DPRINT_P			/* Documents and notepad handling */
#undef DPRINT_PP		/* Extra information */
#define DPRINT_P(x)

#define DPRINT_Q(x) (g_print x)
#undef DPRINT_Q			/* Local argument parsing */
#undef DPRINT_QQ		/* Extra information */
#define DPRINT_Q(x)

#define DPRINT_R(x) (g_print x)
#undef DPRINT_R			/* Resize window */
#define DPRINT_R(x)

#define DPRINT_S(x) (g_print x)
#undef DPRINT_S			/* Scrollbar */
#define DPRINT_S(x)

#define DPRINT_T(x) (g_print x)
#undef DPRINT_T			/* System test code */
#define DPRINT_T(x)

#define DPRINT_V(x) (g_print x)
#undef DPRINT_V			/* Low level function names */
#define DPRINT_V(x)

#define DPRINT_W(x) (g_print x)
#undef DPRINT_W			/* Very low level writes */
#undef DPRINT_WW		/* Extra information */
#define DPRINT_W(x)

#define DPRINT_X(x) (g_print x)
#undef DPRINT_X			/* Linked list code */
#define DPRINT_X(x)

#define DPRINT_Y(x) (g_print x)
#undef DPRINT_Y			/* Shells and OS commands */
#define DPRINT_Y(x)

#define DPRINT_Z(x) (g_print x)
#undef DPRINT_Z			/* Printing */
#undef DPRINT_ZZ		/* Extra information */
#define DPRINT_Z(x)

/* Standard defines */

/* Local defines */

/* Program name */
#define PROGRAM_NAME "elvis"

/* Program logo icon name */
#define LOGO_ELVIS "elvis-logo"

/* Debugging window program name */
#define DEBUG_ELVIS "elvisdbg"

/* Elvis' configuration directory, beneath HOME.
** To use the HOME directory itself, undefine this.
*/
#undef CFGDIR_ELVIS
#define CFGDIR_ELVIS ".elvis"

/* Scratch file name, used for default config */
#define FILE_ELVIS "file.elvis"

/* GTK+ configuration file */
#define GTKRC_ELVIS "gtkrc.elvis"

/* Elvis' GConf namespace header; see also gconfkeys[] */
#define GCONF_PATH "/apps/elvis"

/* Gnome entry widget history name (id) */
#define HISTORY_ELVIS "history-elvis"

/* Fallback shell program */
#define SHELL_PRG "/bin/sh"

/* Dictionary program */
#define DICT_UTIL "dict"

/* Spelling program */
#define SPELL_UTIL "spell"

/* Man program commands, returning first/all pages */
#define MAN_ONE "man "
#define MAN_ALL "man -a "

/* Default color scheme */
#define DEFAULT_SCHEME "blue-on-white"

/* Default font name */
#define DEFAULT_FONT_NAME "Monospace 10"

/* Default print font names */
#define DEFAULT_PRINT_FONT_NAME "Monospace 10"

/* Default debugging font name */
#define DEFAULT_DBG_FONT_NAME "Monospace 8"

/* Default Pango language */
#define DEFAULT_PANGO_LANG "en-us"

/* Gnome Yelp (help) file */
#define YELP_ELVIS "elvis.xml"

/* Docslist file name in Elvis' configuration directory */
#define DOCSLIST_ELVIS "docslist.elvis"

/* Noteslist file name in Elvis' configuration directory */
#define NOTESLIST_ELVIS "noteslist.elvis"

/* Notepad file name, in Elvis' configuration directory */
#define NOTEPAD_ELVIS "notepad.elvis"

/* Default clipboard.  GDK_SELECTION_CLIPBOARD is the default
** GTK+ clipboard.  Should we use GDK_SELECTION_PRIMARY,
** the primary X selection. ???
#define DEFAULT_CLIPBOARD GDK_SELECTION_CLIPBOARD
*/

/* Bonobo server query used to check if Gnome is running.
** An alternative is GNOME_NotificationAreaApplet_Factory.
** Surely there must be a better way of recognising Gnome ??? ###
*/
#define BONOBO_QUERY_1 "((iid=='OAFIID:GNOME_SettingsDaemon')AND(_active))"
#define BONOBO_QUERY_2 "((iid=='OAFIID:GNOME_Wncklet_Factory')AND(_active))"
#define BONOBO_QUERY_3 "((iid=='OAFIID:GNOME_MainMenu_Factory')AND(_active))"

/* Width of bar useable on standard screens */
#define STATUS_BAR_SIZE 150

/* Cook's Constant for window size adjustment */
#define WIN_SHIM 2

/* Spacing between containers' children */
#define CHILD_SEP 3

/* Used in the VTE terminal library but not exposed */
#define VT_CUR_BG 29

/* Maximum path length (now defined in headers)
#define PATH_MAX 2048
*/

/* These are local extensions to the EXCMD enum.
** :sNext doesn't work, so :wi +/- is needed as
** a replacement, but that is a command and its
** argument, not a single enum.  The EXCMD enums
** run to 126, so these #defines start well clear.
** EX_MAIN marks an initialising call from main().
** EX_NAME marks a update of the argslist.
** A kludge, in fact.
*/
#define EX_MAIN 256
#define EX_NAME 257
#define EX_WINUP 258
#define EX_WINDN 259

/* Macros */

/* dec_and_test() returns TRUE if it obtains the lock (sets it to 0) */
typedef gint LOCK;
#define LOCK_INIT 1
#define INITLOCK(x) (x = 1)
#define TRYLOCK(x) \
	(g_atomic_int_dec_and_test(&x) ? 1 : (g_atomic_int_inc(&x), 0))
#define UNLOCK(x) g_atomic_int_inc(&x)

/* A mutex can be used only with threads enabled
typedef GStaticMutex LOCK;
#define LOCK_INIT G_STATIC_MUTEX_INIT
#define INITLOCK(x) g_static_mutex_init(&x)
#define TRYLOCK(x) g_static_mutex_trylock(&x)
#define UNLOCK(x) g_static_mutex_unlock(&x)
*/

/* Structures, typedefs and enums */
typedef unsigned char UCHAR;
typedef unsigned int UINT;
typedef unsigned long ULONG;

/* Key translation table.
**
** For the function keys, the 16-bit GDK keyvalues are converted to a
** hexadecimal string preceded by ^K.  How does this fit in with the
** mapped keys ???
*/
#if 0
struct Keyvalue {
	guint	keyval;		/* Symbols are from gdk/gdkkeysyms.h */
	int	ctrlstate;	/* -1 = ignore, 0 = must be 0, 1 = must be 1 */
	int	altstate;	/* -1 = ignore, 0 = must be 0, 1 = must be 1 */
	int	nchars;		/* Number of valid chars, or 0 for table end */
	char	* keystr;	/* Keystring output */
  };
#endif
static struct Keyvalue keyvalues[] = {
	/* Value		ctl alt len string */
	{ 'a'			,  1, -1,  1, "\001" },
	{ 'a'			,  1, -1,  1, "\001" },
	{ 'b'			,  1, -1,  1, "\002" },
	{ 'c'			,  1, -1,  1, "\003" },
	{ 'd'			,  1, -1,  1, "\004" },
	{ 'e'			,  1, -1,  1, "\005" },
	{ 'f'			,  1, -1,  1, "\006" },
	{ 'g'			,  1, -1,  1, "\007" },
	{ 'h'			,  1, -1,  1, "\010" },
	{ 'i'			,  1, -1,  1, "\011" },
	{ 'j'			,  1, -1,  1, "\012" },
	{ 'k'			,  1, -1,  1, "\013" },
	{ 'l'			,  1, -1,  1, "\014" },
	{ 'm'			,  1, -1,  1, "\015" },
	{ 'n'			,  1, -1,  1, "\016" },
	{ 'o'			,  1, -1,  1, "\017" },
	{ 'p'			,  1, -1,  1, "\020" },
	{ 'q'			,  1, -1,  1, "\021" },
	{ 'r'			,  1, -1,  1, "\022" },
	{ 's'			,  1, -1,  1, "\023" },
	{ 't'			,  1, -1,  1, "\024" },
	{ 'u'			,  1, -1,  1, "\025" },
	{ 'v'			,  1, -1,  1, "\026" },
	{ 'w'			,  1, -1,  1, "\027" },
	{ 'x'			,  1, -1,  1, "\030" },
	{ 'y'			,  1, -1,  1, "\031" },
	{ 'z'			,  1, -1,  1, "\032" },
	{ 'A'			,  1, -1,  1, "\001" },
	{ 'B'			,  1, -1,  1, "\002" },
	{ 'C'			,  1, -1,  1, "\003" },
	{ 'D'			,  1, -1,  1, "\004" },
	{ 'E'			,  1, -1,  1, "\005" },
	{ 'F'			,  1, -1,  1, "\006" },
	{ 'G'			,  1, -1,  1, "\007" },
	{ 'H'			,  1, -1,  1, "\010" },
	{ 'I'			,  1, -1,  1, "\011" },
	{ 'J'			,  1, -1,  1, "\012" },
	{ 'K'			,  1, -1,  1, "\013" },
	{ 'L'			,  1, -1,  1, "\014" },
	{ 'M'			,  1, -1,  1, "\015" },
	{ 'N'			,  1, -1,  1, "\016" },
	{ 'O'			,  1, -1,  1, "\017" },
	{ 'P'			,  1, -1,  1, "\020" },
	{ 'Q'			,  1, -1,  1, "\021" },
	{ 'R'			,  1, -1,  1, "\022" },
	{ 'S'			,  1, -1,  1, "\023" },
	{ 'T'			,  1, -1,  1, "\024" },
	{ 'U'			,  1, -1,  1, "\025" },
	{ 'V'			,  1, -1,  1, "\026" },
	{ 'W'			,  1, -1,  1, "\027" },
	{ 'X'			,  1, -1,  1, "\030" },
	{ 'Y'			,  1, -1,  1, "\031" },
	{ 'Z'			,  1, -1,  1, "\032" },
	{ GDK_F1		, -1, -1,  5, "\013ffde" },
	{ GDK_F2		, -1, -1,  5, "\013ffdf" },
	{ GDK_F3		, -1, -1,  5, "\013ffc0" },
	{ GDK_F4		, -1, -1,  5, "\013ffc1" },
	{ GDK_F5		, -1, -1,  5, "\013ffc2" },
	{ GDK_F6		, -1, -1,  5, "\013ffc3" },
	{ GDK_F7		, -1, -1,  5, "\013ffc4" },
	{ GDK_F8		, -1, -1,  5, "\013ffc5" },
	{ GDK_F9		, -1, -1,  5, "\013ffc6" },
	{ GDK_F10		, -1, -1,  5, "\013ffc7" },
	{ GDK_F11		, -1, -1,  5, "\013ffc8" },
	{ GDK_F12		, -1, -1,  5, "\013ffc9" },
	/* These keys need to be translated from GDK codes to ASCII */
#if 1
	{ GDK_BackSpace		, -1, -1,  1, "\010" },
	{ GDK_Tab		, -1, -1,  1, "\011" },
	{ GDK_Return		, -1, -1,  1, "\015" },
	{ GDK_Escape		, -1, -1,  1, "\033" },
#else
	{ GDK_BackSpace		, -1, -1,  5, "\013ff08" },
	{ GDK_Tab		, -1, -1,  5, "\013ff09" },
	{ GDK_Return		, -1, -1,  5, "\013ff0d" },
	{ GDK_Escape		, -1, -1,  5, "\013ff1b" },
#endif
	{ GDK_Left		, -1, -1,  5, "\013ff51" },
	{ GDK_Right		, -1, -1,  5, "\013ff53" },
	{ GDK_Up		, -1, -1,  5, "\013ff52" },
	{ GDK_Down		, -1, -1,  5, "\013ff54" },
	{ GDK_Page_Down		, -1, -1,  5, "\013ff56" },
	{ GDK_Page_Up		, -1, -1,  5, "\013ff55" },
	{ GDK_Home		, -1, -1,  5, "\013ff50" },
	{ GDK_End		, -1, -1,  5, "\013ff57" },
	{ GDK_Insert		, -1, -1,  5, "\013ff63" },
#if 1
	{ GDK_KP_Enter		, -1, -1,  1, "\015" },
#else
	{ GDK_KP_Enter		, -1, -1,  5, "\013ff8d" },
#endif
	{ GDK_KP_Home		, -1, -1,  5, "\013ff95" },
	{ GDK_KP_Left		, -1, -1,  5, "\013ff96" },
	{ GDK_KP_Up		, -1, -1,  5, "\013ff97" },
	{ GDK_KP_Right		, -1, -1,  5, "\013ff98" },
	{ GDK_KP_Down		, -1, -1,  5, "\013ff99" },
	{ GDK_KP_Prior		, -1, -1,  5, "\013ff9a" },
	{ GDK_KP_Page_Up	, -1, -1,  5, "\013ff9a" },
	{ GDK_KP_Next		, -1, -1,  5, "\013ff9b" },
	{ GDK_KP_Page_Down	, -1, -1,  5, "\013ff9b" },
	{ GDK_KP_End		, -1, -1,  5, "\013ff9c" },
	{ GDK_KP_Begin		, -1, -1,  5, "\013ff9d" },
	{ GDK_KP_Insert		, -1, -1,  5, "\013ff9e" },
#if 1
	{ GDK_KP_Delete		, -1, -1,  1, "\177" },
	{ GDK_Delete		, -1, -1,  1, "\177" },
#else
	{ GDK_KP_Delete		, -1, -1,  5, "\013ff9f" },
	{ GDK_Delete		, -1, -1,  5, "\013ffff" },
	{ GDK_VoidSymbol	, -1, -1,  6, "ffffff"   },
#endif
	{ '\0'			,  0,  0,  0, "" }
  };

/* Default key mapping table.
**
** These keys are mapped automatically on startup.
**
** From X11/X.h: typedef unsigned long XID; typedef XID KeySym;
** KeySym symbols are from gdk/gdkkeysyms.h.
** From elvis/map.h: typedef unsigned int MAPFLAGS;
*/
#if 0
struct Mapkey {
	char *label;
	KeySym sym;
	char *cooked;
	MAPFLAGS flags;
  };
#endif
static struct Mapkey mapkeys[] = {
	{ "<Up>",	GDK_Up,		"k",	MAP_ALL },
	{ "<KP_Up>",	GDK_KP_Up,	"k",	MAP_ALL },
	{ "<Down>",	GDK_Down,	"j",	MAP_ALL },
	{ "<KP_Down>",	GDK_KP_Down,	"j",	MAP_ALL },
	{ "<Left>",	GDK_Left,	"h",	MAP_ALL },
	{ "<KP_Left>",	GDK_KP_Left,	"h",	MAP_ALL },
	{ "<Right>",	GDK_Right,	"l",	MAP_ALL },
	{ "<KP_Right>",	GDK_KP_Right,	"l",	MAP_ALL },
	{ "<Prior>",	GDK_Prior,	"\002",	MAP_ALL },
	{ "<KP_Prior>",	GDK_KP_Prior,	"\002",	MAP_ALL },
	{ "<Next>",	GDK_Next,	"\006",	MAP_ALL },
	{ "<KP_Next>",	GDK_KP_Next,	"\006",	MAP_ALL },
	{ "<Home>",	GDK_Home,	"^",	MAP_ALL },
	{ "<KP_Home>",	GDK_KP_Home,	"^",	MAP_ALL },
	{ "<Begin>",	GDK_Begin,	"^",	MAP_ALL },
	{ "<KP_Begin>",	GDK_KP_Begin,	"^",	MAP_ALL },
	{ "<End>",	GDK_End,	"$",	MAP_ALL },
	{ "<KP_End>",	GDK_KP_End,	"$",	MAP_ALL },
	{ "<Insert>",	GDK_Insert,	"i",	MAP_ALL },
	{ "<KP_Insert>",GDK_KP_Insert,	"i",	MAP_ALL },
	{ "<Delete>",	GDK_Delete,	"x",	MAP_ALL },
	{ "<KP_Delete>",GDK_KP_Delete,	"x",	MAP_ALL },
	{ "<Undo>",	GDK_Undo,	"u",	MAP_ALL },
  };

/* GConf keys and handlers.
**
** Note that this layout must match the separate elvis.schemas file
** used to initialise the GConf database.
*/
static struct Gconfkey {
	const char *key ;	/* GConf key string, including GCONF_PATH */
	const char *option;	/* Equivalent parsers[] option prefix */
	size_t index ;		/* Index into optvals[] */
 } gconfkeys[] = {
	{ GCONF_PATH "/edit/defaultcols", "--defaultcols=", 0 },
	{ GCONF_PATH "/edit/defaultrows", "--defaultrows=", 1 },
	{ GCONF_PATH "/edit/terminalfont", "--terminalfont=", 2 },
	{ GCONF_PATH "/edit/forecolor", "--forecolor=", 3 },
	{ GCONF_PATH "/edit/backcolor", "--backcolor=", 4 },
	{ GCONF_PATH "/edit/cursorcolor", "--cursorcolor=", 5 },
	{ GCONF_PATH "/edit/cursorblink", "--cursorblink=", 6 },
	{ GCONF_PATH "/ui/dictionary", NULL, 7 },
	{ GCONF_PATH "/ui/docscount", NULL, 8 },
	{ GCONF_PATH "/ui/hidemouse", NULL, 9 },
	{ GCONF_PATH "/ui/manmode", NULL, 10 },
	{ GCONF_PATH "/ui/newmode", NULL, 11 },
	{ GCONF_PATH "/ui/scrollwheel", NULL, 12 },
	{ GCONF_PATH "/ui/shellcols", NULL, 13 },
	{ GCONF_PATH "/ui/shellrows", NULL, 14 },
	{ GCONF_PATH "/ui/shellmenu", NULL, 15 },
	{ GCONF_PATH "/ui/showtoolbar", "--showtoolbar=", 16 },
	{ GCONF_PATH "/ui/spellcheck", NULL, 17 },
	{ GCONF_PATH "/ui/startposition", NULL, 18 },
#ifdef PRINT_MENU
	{ GCONF_PATH "/print/headerfont", "--headerfont=", 19 },
	{ GCONF_PATH "/print/printerfont", "--printerfont=", 20 },
	{ GCONF_PATH "/print/printheader", NULL, 21 },
	{ GCONF_PATH "/print/printmenu", NULL, 22 },
#endif
	{ NULL, NULL, 0 }
  };

/* Instances of this structure are used to describe options.
** Copied for reference from elvis/options.h
*/
#if 0
typedef struct optdesc_s
{
   char	*longname
   char *shortname;
   CHAR *(*asstring) P_((struct optdesc_s *opt, OPTVAL *val));
   int	(*isvalid) P_((struct optdesc_s *opt, OPTVAL *val, CHAR *newval));
   char *limit;
   int	(*store) P_((struct optdesc_s *opt, OPTVAL *val, CHAR *newval));
   CHAR	*dflt;
#ifdef FEATURE_AUTOCMD
   ELVBOOL event; /* if ElvTrue, send events when option changes */
#endif
} OPTDESC;
#endif

/* Names of global options for the Gnome GUI, with range check values.
**
** Must be sychronised with optvals[].  Values are stored apart from
** their descriptions because for some options the values will vary
** for each window or buffer, but their descriptions won't.
**
** Elvis has a number of option range check and set functions:
** optisnumber() Check a number's validity.
** optisstring() Check a strings validity (all are valid).
** optispacked() Check a series of strings are valid.
** optisoneof()  Check a string's validity against a space-delimited list.
** optistab()    Check if a string is an acceptable list of tabstop positions.
** optnstring()  Convert a "number" value to a string.
** optsstring()  Convert a "string" value to a string.
** opt1string()  Convert a "one of" value to a string.
** opttstring()  Convert a "tab" value to a string.
**
** Note that the definition of the OPTDESC structure will change if
** FEATURE_AUTOCMD, normally defined, is undefined.
**
** To simplify range checking, only number arguments have defaults.
** String arguments have their defaults checked as they are needed.
** Check functions must accept their arguments as strings.
**
** The functions below are declared early for initialisation.
*/
static int set_blink(struct optdesc_s *, OPTVAL *, CHAR *);
static int set_cursor(struct optdesc_s *, OPTVAL *, CHAR *);
static int set_fore(struct optdesc_s *, OPTVAL *, CHAR *);
static int set_back(struct optdesc_s *, OPTVAL *, CHAR *);
static int set_font(struct optdesc_s *, OPTVAL *, CHAR *);
static int set_cols(struct optdesc_s *, OPTVAL *, CHAR *);
static int set_rows(struct optdesc_s *, OPTVAL *, CHAR *);
static int set_tool(struct optdesc_s *, OPTVAL *, CHAR *);
static int set_print(struct optdesc_s *, OPTVAL *, CHAR *);

/* longname, shortname, asstring(), isvalid(), limit, store(), dflt, event */
static OPTDESC optdescs[] = {
	{"defaultcols", "defcols", optnstring, optisnumber,
			MIN_COLS_S ":" MAX_COLS_S, set_cols, DEF_COLS_S, 0},
	{"defaultrows", "defrows", optnstring, optisnumber,
			MIN_ROWS_S ":" MAX_ROWS_S, set_rows, DEF_ROWS_S, 0},
	{"terminalfont", "termfont", optsstring, optisstring,
			NULL, set_font, NULL, 0},
	{"forecolor", "foreclr", optsstring, optisstring,
			NULL, set_fore, NULL, 0},
	{"backcolor", "backclr", optsstring, optisstring,
			NULL, set_back, NULL, 0},
	{"cursorcolor", "csrclr", optsstring, optisstring,
			NULL, set_cursor, NULL, 0},
	{"cursorblink", "csrblink", NULL, NULL,
			"0:1", set_blink, "1", 0},
	{"dictionary", "dict", optsstring, optisstring,
			NULL, NULL, NULL, 0},
	{"docscount", "docs", optnstring, optisnumber,
			"0:" DOC_MAX_S, NULL, DOC_COUNT_S, 0},
	{"hidemouse", "hide", NULL, NULL,
			"0:1", NULL, "0", 0},
	{"manmode", "man", NULL, NULL,
			"0:1", NULL, "1", 0},
	{"newmode", "new", NULL, NULL,
			"0:1", NULL, "1", 0},
	{"scrollwheel", "scrlwhl", NULL, NULL,
			"0:1", NULL, "1", 0},
	{"shellcols", "shcols", optnstring, optisnumber,
			MIN_COLS_S ":" MAX_COLS_S, NULL, DEF_COLS_S, 0},
	{"shellrows", "shrows", optnstring, optisnumber,
			MIN_ROWS_S ":" MAX_ROWS_S, NULL, DEF_ROWS_S, 0},
	{"shellmenu", "shmenu", NULL, NULL,
			"0:1", NULL, "0", 0},
	{"showtoolbar", "tools", NULL, NULL,
			"0:1", set_tool, "1", 0},
	{"spellcheck", "spell", optsstring, optisstring,
			NULL, NULL, NULL, 0},
	{"startposition", "start", NULL, NULL,
			"0:1", NULL, "0", 0},
	{"headerfont", "hdrfont", optsstring, optisstring,
			NULL, set_print, NULL, 0},
	{"printerfont", "prnfont", optsstring, optisstring,
			NULL, set_print, NULL, 0},
	{"printheader", "prnhdr", NULL, NULL,
			"0:1", NULL, "1", 0},
	{"printmenu", "print", NULL, NULL,
			"0:1", NULL, "1", 0}
#ifdef DEBUG_MENU
	,{"debugfont", "dbgfont", optsstring, optisstring,
			NULL, set_font, NULL, 0}
#endif
  };

/* Instances of this structure are used to store values of options.
** Copied for reference from elvis/options.h
*/
#if 0
#define OPT_SET		0x001	/* value has been changed */
#define OPT_SHOW	0x002	/* value should be shown */
#define OPT_HIDE	0x004	/* value should be hidden even if ":set all" */
#define OPT_LOCK	0x008	/* value can never be changed */
#define OPT_UNSAFE	0x010	/* value can't be changed if "safer" */
#define OPT_FREE	0x020	/* call safefree() on the value before */
				/* freeing/changing it */
#define OPT_REDRAW	0x040	/* changing this option forces a redraw */
#define OPT_SCRATCH	0x080	/* changing forces redraw from scratch */
#define OPT_NODFLT	0x100	/* no default value */
typedef struct
{
  union
  {
	long	number;		/* for numeric options */
	ELVBOOL	boolean;	/* for boolean options */
	char	character;	/* for "one of" options */
	CHAR	*string;	/* for string options */
	short	*tab;		/* tabstop positions */
	void	*pointer;	/* for any other type of option */
  } value;
  short flags;			/* flags for the option */
} OPTVAL;
#endif

/* Values of global options for the Gnome GUI, initialised to defaults.
**
** Must be sychronised with optdescs[].  Some defaults can be reset from
** the command line, some from GConf (see gconfkeys[]), and some from
** Elvis' :set command.
**
** The above values are ORed together in the "flags" field of an option
** to tell elvis how to deal with it.  Initially, all flags should be 0.
** Elvis maintains the "set" and "show" flags itself; the "save" flag is'
** maintained by each option's store() function.  (This is according to
** options.h; however, there appears to _be_ no "save" flag ???).  In fact,
** the OPT_REDRAW and OPT_SCRATCH don't seem to make any difference.
**
** Most of the options can be altered from the GUI and/or from GConf,
** so they are "locked" here to avoid confusion.
**
** Note that before C99 only the first-named element of a union could
** be initialised in a declaration, so the code below usually invokes
** undefined behaviour.
*/
static OPTVAL optvals[] = {
	/* OPTVAL: union (first element long) value, short flags */
	{ {DEF_COLS }, OPT_REDRAW },		/* 0 default cols */
	{ {DEF_ROWS }, OPT_REDRAW },		/* 1 default rows */
	{ { 0 }, OPT_REDRAW },			/* 2 terminal font */
	{ { 0 }, OPT_REDRAW },			/* 3 foreground color */
	{ { 0 }, OPT_REDRAW },			/* 4 background color */
	{ { 0 }, OPT_REDRAW },			/* 5 cursor color */
	{ { TRUE }, OPT_REDRAW },		/* 6 cursor blink */
	{ { 0 }, OPT_LOCK },			/* 7 dictionary utility */
	{ { DOC_COUNT }, OPT_LOCK },		/* 8 document list size */
	{ { FALSE }, OPT_LOCK },		/* 9 mouse autohide */
	{ { TRUE }, OPT_LOCK | OPT_REDRAW },	/* 10 manpage mode */
	{ { TRUE }, OPT_LOCK | OPT_REDRAW },	/* 11 new window mode */
	{ { TRUE }, OPT_LOCK },			/* 12 scroll direction */
	{ { DEF_COLS }, OPT_LOCK },		/* 13 shell cols */
	{ { DEF_ROWS }, OPT_LOCK },		/* 14 shell rows */
	{ { FALSE }, OPT_LOCK },		/* 15 shell menu */
	{ { TRUE }, OPT_REDRAW },		/* 16 show toolbar */
	{ { 0 }, OPT_LOCK },			/* 17 spellcheck utility */
	{ { FALSE }, OPT_LOCK },		/* 18 start position */
	{ { 0 }, OPT_LOCK },			/* 19 print job header font */
	{ { 0 }, 0 },				/* 20 print job printer font */
	{ { TRUE }, OPT_LOCK },			/* 21 print header */
	{ { TRUE }, OPT_LOCK }			/* 22 print menu */
#ifdef DEBUG_MENU
	,{ { 0 }, OPT_LOCK | OPT_REDRAW }	/* 19/23 debugging font */
#endif
  };

/* Macros to get the values of each global option.
** Must be sychronised with optvals[].
** #defines are set for some of the widely-used indices.
*/
#define O_DCOLS 0
#define o_defaultcols optvals[O_DCOLS].value.number
#define O_DROWS 1
#define o_defaultrows optvals[O_DROWS].value.number
#define O_TFONT 2
#define o_terminalfont optvals[O_TFONT].value.string
#define O_FCOLOR 3
#define o_forecolor optvals[O_FCOLOR].value.pointer
#define O_BCOLOR 4
#define o_backcolor optvals[O_BCOLOR].value.pointer
#define O_CCOLOR 5
#define o_cursorcolor optvals[O_CCOLOR].value.pointer
#define O_CBLINK 6
#define o_cursorblink optvals[O_CBLINK].value.boolean
#define O_DICT 7
#define o_dictionary optvals[O_DICT].value.string
#define O_DOCS 8
#define o_docscount optvals[O_DOCS].value.number
#define O_HMOUSE 9
#define o_hidemouse optvals[O_HMOUSE].value.boolean
#define O_MANM 10
#define o_manmode optvals[O_MANM].value.boolean
#define O_NEWM 11
#define o_newmode optvals[O_NEWM].value.boolean
#define O_SCROLL 12
#define o_scrollwheel optvals[O_SCROLL].value.boolean
#define O_SHCOLS 13
#define o_shellcols optvals[O_SHCOLS].value.number
#define O_SHROWS 14
#define o_shellrows optvals[O_SHROWS].value.number
#define O_SHMENU 15
#define o_shellmenu optvals[O_SHMENU].value.number
#define O_TOOL 16
#define o_showtoolbar optvals[O_TOOL].value.boolean
#define O_SPELL 17
#define o_spellcheck optvals[O_SPELL].value.string
#define O_START 18
#define o_startposition optvals[O_START].value.boolean
#define O_HFONT 19
#define o_headerfont optvals[O_HFONT].value.string
#define O_PFONT 20
#define o_printerfont optvals[O_PFONT].value.string
#define O_PHDR 21
#define o_printheader optvals[O_PHDR].value.boolean
#define O_PRINT 22
#define o_printmenu optvals[O_PRINT].value.boolean
#define O_DFONT 23
#ifdef DEBUG_MENU
#define o_debugfont optvals[O_DFONT].value.string
#endif

/* Command line options parser setup; functions declared for initialisation */
static int parse_int(char *str, size_t index, size_t opt_index);
static int parse_two_ints(char *str, size_t index, size_t opt_index);
static int parse_bool(char *str, size_t index, size_t opt_index);
static int parse_color(char *str, size_t index, size_t opt_index);
static int parse_font(char *str, size_t index, size_t opt_index);
static int parse_string(char *str, size_t index, size_t opt_index);

static struct Parse {
	char *name;		/* Option string prefix */
	size_t len;		/* Length of string to match */
	int cmd;		/* Command line option available */
	int (*parse)(char *, size_t, size_t); /* Parser function, called with */
				/* (option name, parsers index, option idx) */
	size_t idx;		/* Index of option(s) in optdescs[] */
	char *usage;		/* Usage string */
  } parsers[] = {
	/* The order of these names must follow the order of the optdescs[]
	** array where they match. Since some command line options set more
	** than one optdescs[] options not all of them will correspond.
	** Most of these options can't be set from the command line, but
	** their entries are used to collect eg. parsing information in
	** one place.
	*/
	{ "--defaultcols", sizeof("--defaultcols") - 1, TRUE,  &parse_int, 0,
			"=M     Set default terminal columns to M" },
	{ "--defaultrows", sizeof("--defaultrows") - 1, TRUE,  &parse_int, 1,
			"=N     Set default terminal rows to N" },
	{ "--terminalfont", sizeof("--terminalfont") - 1, TRUE,  &parse_font, 2,
			"=name Set default terminal font name and size" },
	{ "--forecolor", sizeof("--forecolor") - 1, TRUE,  &parse_color, 3,
			"=name    Set default foreground color" },
	{ "--backcolor", sizeof("--backcolor") - 1, TRUE,  &parse_color, 4,
			"=name    Set default background color" },
	{ "--cursorcolor", sizeof("--cursorcolor") - 1, TRUE,  &parse_color, 5,
			"=name  Set default cursor color" },
	{ "--cursorblink", sizeof("--cursorblink") - 1, TRUE,  &parse_bool, 6,
			"=on|off Set default cursor blink" },
	/* The following options can't be set from the command line */
	{ "--dictionary", sizeof("--dictionary") - 1, FALSE, &parse_string, 7,
			"=name   Set dictionary program name" },
	{ "--docscount", sizeof("--docscount") - 1, FALSE, &parse_int, 8,
			"=N       Set size of document list" },
	{ "--hidemouse", sizeof("--hidemouse") - 1, FALSE, &parse_bool, 9,
			"=yes|no  Set mouse autohide mode" },
	{ "--manmode", sizeof("--manmode") - 1, FALSE, &parse_bool, 10,
			"=one|all   Set man page search mode" },
	{ "--newmode", sizeof("--newmode") - 1, FALSE, &parse_bool, 11,
			"=new|cur   Set default new window open mode" },
	{ "--scrollwheel", sizeof("--scrollwheel") - 1, FALSE, &parse_bool, 12,
			"=up|dn Set scrollwheel scroll direction" },
	{ "--shellcols", sizeof("--shellcols") - 1, FALSE, &parse_int, 13,
			"=M       Set default shell columns to M" },
	{ "--shellrows", sizeof("--shellrows") - 1, FALSE, &parse_int, 14,
			"=N       Set default shell rows to N" },
	{ "--shellmenu", sizeof("--shellmenu") - 1, FALSE, &parse_bool, 15,
			"=yes|no  Set default shell menu visibility" },
	/* A command line argument */
	{ "--showtoolbar", sizeof("--showtoolbar") - 1, TRUE, &parse_bool, 16,
			"=yes|no Set default toolbar visibility" },
	{ "--spellcheck", sizeof("--spellcheck") - 1, FALSE, &parse_string, 17,
			"=name   Set spellcheck program name" },
	{ "--startpos", sizeof("--startpos") - 1, FALSE, &parse_bool, 18,
			"=no|yes   Set fixed window start position" },
	{ "--headerfont", sizeof("--headerfont") - 1, FALSE, &parse_font, 19,
			"=name   Set job header font name and size" },
	/* A command line argument */
	{ "--printerfont", sizeof("--printerfont") - 1, TRUE, &parse_font, 20,
			"=name  Set default printing font name and size" },
	{ "--printheader", sizeof("--printheader") - 1, FALSE, &parse_bool, 21,
			"=yes|no Print job header on each page" },
	{ "--printmenu", sizeof("--printmenu") - 1, FALSE, &parse_bool, 22,
			"=yes|no  Activate main print menu" },
#ifdef DEBUG_MENU
	{ "--debugfont", sizeof("--debugfont") - 1, 0, &parse_font, O_DFONT,
			"=name    Set default debugging font" },
#endif
	/* A command line argument */
	{ "--geometry", sizeof("--geometry") - 1, TRUE, &parse_two_ints, 0,
			"=NxM      Set default terminal rows and columns" }
  };

/* GTKrc file option names and set flags */
static struct RCoption {
	char * name;		/* Name to match */
	size_t len;		/* Length of name */
	int set;		/* Option set flag */
  } rcoptions[] = {
	{ "geometry=", sizeof("geometry=") - 1, FALSE },
	{ "lastprint=", sizeof("lastprint=") - 1, FALSE }
  };

/* Selections in font and color buttons and dialogs */
struct Select {
	GtkWidget *button;	/* Color or font or toggle button */
	GtkWidget *label;	/* Label widget associated with button */
	GtkWidget *box;		/* Container for button and label widgets */
	GdkColor *initial;	/* Initial value of button */
	GdkColor color;		/* Color set on button */
	int setting;		/* Flag set on button */
	char *name;		/* Title of the label widget */
  };

/* Global variables */

#ifndef lint
/* If these strings are static GCC optimises them out of existence */
char id_guignome_c[] = "$Id: guignome.c,v 1.93 2008/01/01 12:00:00 cwr Exp $";
static volatile char Version[] = "@(#) guignome.c: " GNOMEGUI_VERSION;
#endif

/* These may be used before init() is run */
static int Exists = FALSE;		/* GUI availability */
static int StopThat = FALSE;		/* Used for user overrides */
static int InitDone = FALSE;		/* Program initialisation complete */
static char BadFile[64];		/* Bad file name for init() errors */
static GtkClipboard *Clipboard = NULL;	/* Gnome's default clipboard */
static LOCK entry_lock;			/* Entry box lock */
#ifdef DEBUG_MENU
static GPid Child = 0;			/* Child's  PID */
static int Key_flg = 0;			/* Key debugging print flag */
#endif
#ifdef DPRINT_CC
static char color_names[32][32];	/* Color palette names for debugging */
#endif

/* These get used as default initialisers */
static struct Fixedcolors {
	char *name;
	GdkColor color;
  } fixedcolors[] = {
	{ "black", {0, 0x0000, 0x0000, 0x0000 } },
	{ "gray", {0, 0x7fff, 0x7fff, 0x7fff } },
	{ "grey", {0, 0x7fff, 0x7fff, 0x7fff } },
	{ "white", {0, 0xffff, 0xffff, 0xffff } },
	{ "red", {0, 0xffff, 0x0000, 0x0000 } },
	{ "yellow", {0, 0xffff, 0xffff, 0x0000 } },
	{ "green", {0, 0x0000, 0xffff, 0x0000 } },
	{ "cyan", {0, 0x0000, 0xffff, 0xffff } },
	{ "blue", {0, 0x0000, 0x0000, 0xffff } },
	{ "magenta", {0, 0xffff, 0x0000, 0xffff } }
  };
#define color_black   fixedcolors[0].color
#define color_gray    fixedcolors[1].color
#define color_grey    fixedcolors[2].color
#define color_white   fixedcolors[3].color
#define color_red     fixedcolors[4].color
#define color_yellow  fixedcolors[5].color
#define color_green   fixedcolors[6].color
#define color_cyan    fixedcolors[7].color
#define color_blue    fixedcolors[8].color
#define color_magenta fixedcolors[9].color

/* Externals */

/* Function prototypes */

/*
** Main menu actions:
** empty_action -- A placeholder for the main menu.
** open_action -- Open an existing file and make it current.
** save_action -- Save a file, under a new name if necessary.
** saveas_action -- Save a file using the Save File dialog.
** close_action -- Close the current file.
** quit_action -- Quit the Gnome GUI.
** yank_all_action -- Yank all the lines in a file.
** move_action -- Move to a specified line.
** clear_history_action - Clear the search history.
** dictionary_action -- Show a dictionary entry for the current word.
** spellcheck_action -- Spellcheck the current word.
** manpage_action -- Show the man page for the current word.
** note_action -- A placeholder for the notepad submenu.
** notepad_action -- Open a notepad file.
** shell_action -- Run a command shell.
** show_tools_action -- Show or hide the main toolbar.
** startpos_action -- Set or unset the startup postion option.
** default_size_action -- Restore the default terminal size.
** docs_how_action -- Set the window opening mode for documents.
** cursor_action -- Configure the cursor for the current window.
** color_action -- Select colors for the current window.
** font_action -- Select fonts for the current window.
** config_action -- Save and load file list (ring) configurations.
** docs_clear_action -- Clear the Documents list and menu.
** get_print_filename -- Get the name of a file to be printed.
** print_action -- Print all or part of a file.
** print_setup_action -- Set the printing fonts and header.
** print_cancel_action -- Cancel the current print job.
** help_action -- Show the Elvis help (Gnome Yelp) file.
** goto_url -- Helper function for about_action dialog.
** about_action -- Show the "About Elvis" dialog window.
**
** Debug Menu actions:
** debug_action -- Run specific functions for debugging.
** child_exit -- Child exit watch function.
** selection_test -- Dialog to debug clipboard selection.
** target_handler -- Handler for clicked signal, a selection request.
** selection_received -- Handler for selection-received signal.
** selection_acquire -- Handler for toggle signal, to control a selection.
** selection_blocked -- Handler for selection-clear-event signal.
** target_set -- Handler for selection-get signal.
**
** Toolbar actions:
** add_r_action -- Add a new window (ie. edit ring) and make it current.
** rmv_r_action -- Drop the current window (ie. edit ring).
** add_f_action -- Add a file to the current arglist and make it current.
** rmv_f_action -- Remove a file from the current arglist.
** cancel_action -- Cancel the current task.
** ex_action -- Pass an EX command to Elvis.
**
** Popup menu actions:
** re_action -- Reverse or repeat the last edit action(s).
** drag_action -- Drag text to another window.
** cut_action -- Cut text to the clipboard.
** copy_action -- Copy text to the clipboard.
** paste_action -- Paste text from the clipboard.
** delete_action -- Delete text.
**
** Manipulate the menus and the toolbar:
** create_main_menu -- Create the main menu.
** create_mouse_menu -- Create the mouse menu.
** config_menus -- Configure the main and mouse menus.
** switch_toolbar_menu -- Switch the show/hide toolbar menu entries.
** switch_startpos_menu -- Switch the set/unset start position menu entries.
** switch_docs_open_menu -- Switch the current/new document open menu entries.
** switch_manpage_menu -- Switch the one/all man page search menu entries.
** update_undolevel_menu -- Update the undolevel edit menu entry.
** create_toolbar -- Create a Gnome toolbar.
** switch_toolbar -- Show or hide the toolbar.
**
** Signal handlers:
** snoop_handler -- Handler for key snooper.
** popup_handler -- Handler for popup signal.
** motion_handler -- Handler for motion-notify-event (mouse) signals.
** scroll_handler -- Handler for scroll-event (mouse) signals.
** mouse_handler -- Handler for mouse button and movement signals.
** range_handler -- Handler for range change-value signals.
** theme_handler -- Handler for theme changed signal.
** gconf_notify_handler -- Handler for GConfClient change notification.
** client_save_handler -- Handler for session manager save-yourself signal.
** client_die_handler -- Handler for session manager die signal.
** drag_data_handler -- Handler for drag-data-get signal (outgoing data).
** drag_begin_handler -- Handler for drag-begin signal.
** drag_end_handler -- Handler for drag-end signal.
** drop_data_handler -- Handler for drag-data-received signal (incoming data).
** drag_drop_handler -- Handler for drag-drop signal.
** entry_handler -- Handler for activate signal from entry box.
** filename_handler - Handler for activate signal in File menu.
** documents_handler - Handler for activate signal in Documents menu.
** v_val_chg_handler -- Handler for value-changed signal in the vertical slider.
** del_handler -- Handler for delete-event signal.
** glog_handler -- Handler for GLib message logger.
** gio_handler -- Handler for a GIOChannel watcher.
** key_press_handler -- Handler for key-press-event signal.
** destroy_handler -- Handler for a main window destroy signal.
** resize_handler -- Handler for a main window configure-event signal.
** state_handler -- Handler for a main window window-state-event signal.
** default_font_handler -- Handler for font dialog default button.
** default_color_handler -- Handler for color dialog default button.
** default_cursor_handler -- Handler for cursor dialog default button.
** terminal_font_handler -- Handler for terminal font-set signal.
** color_handler -- Handler for color-set signal.
** cursor_handler -- Handler for toggled signal.
** begin_print_handler -- Handler for GtkPrintOperation begin-print signal.
** draw_page_handler -- Handler for GtkPrintOperation draw-page signal.
** end_print_handler -- Handler for GtkPrintOperation end-print signal.
** done_print_handler -- Handler for GtkPrintOperation done signal.
** print_name_handler -- Handler for print file name dialog
**                       file chooser button selection-changed signal.
** default_print_name_handler -- Handler for print file name dialog
**                               default button.
** print_header_handler -- Handler for print setup dialog toggle button.
** print_font_handler -- Handler for print setup dialog
**                       font button font-set signal.
** default_print_setup_handler -- Handler for print setup dialog default button.
** hello -- Placeholder callback.
**
** Initialisation:
** init_dand -- Initialise the drag and drop interface.
** init_defaults -- Initialise application settings and parse arguments.
** init_gnome_win -- Initialise the Gnome application main window.
** init_interface -- Initialise GLib, GTK+ and Gnome.
** show_warning -- Show warning messages from initialisation.
** firstcmd -- Execute Elvis' mainfirstcmd() function.
** load_settings -- Load the startup settings.
** store_settings -- Store the startup settings.
** get_print_settings -- Get the current print settings.
** load_fileslist - Load a linked list of filenames.
** store_docslist - Store the document list.
**
** Miscellaneous:
** check_gconf -- Check the GConf settings database.
** request_clip -- Request a selection from the clipboard.
** url_to_path -- Convert a URL to a local path.
** undot_path -- Remove "." and ".." entries from a path.
** get_id -- Get a unique id number.
** get_exname -- Get the name of an EXCMD enum.
** get_icons -- Get the program logo and any icons not from stock.
** get_tools_icon -- Get the appropriate toolbar Tools menu icon.
** size_main_window -- Show and set the main window size.
** update_geometry -- Update the VTE terminal geometry.
** write_buff -- Write text to a window.
** terminal -- Run a terminal shell.
** trap_handler -- Handle Unix signals requesting a controlled shutdown,
**
** Manipulate edit rings:
** create_edit_ring -- Create an edit ring.
** destroy_edit_ring -- Destroy an edit ring.
** goto_edit_ring -- Move to another edit ring.
** update_edit_files -- Update the edit ring files.
** load_files_menu -- Add filenames to the Files menu.
** unload_files_menu -- Remove filenames from the Files menu.
**
** Argument parsing:
** parse_int -- Parse a single integer option.
** parse_two_ints -- Parse a double integer option.
** parse_bool -- Parse a boolean option.
** parse_color -- Parse a color (string) option.
** parse_font -- Parse a font (string) option.
** parse_string -- Parse a string option.
** parse_opts -- Parse the global argument string for Gnome GUI options.
** get_default_colormap -- Get the default (system) color map.
** set_default_colors -- Set the default colors.
** check_font_family -- Check that a font family is available.
** check_mono_family -- Check that a monospace font family is available.
** get_mono_fonts -- Get a monospace font, or a default.
** list_fonts - List the available monospace fonts for debugging.
** set_current_font -- Set the current font.
**
** Set commands:
** set_blink -- Update the cursor blink using :set.
** set_cursor -- Update the cursor color using :set.
** set_fore -- Update the foreground color using :set.
** set_back -- Update the background color using :set.
** set_font -- Update the terminal fonts using :set.
** set_cols -- Update the terminal columns using :set.
** set_rows -- Update the terminal rows using :set.
** set_tool -- Update the toolbar display using :set.
** set_print -- Update the printer fonts using :set.
**
** Document and Notepad handling:
** add_docsmenu - Add an item to the Documents menu.
** rmv_docsmenu - Remove an item from the Documents menu.
** add_notesmenu - Add an item to the Notebooks submenu.
**
** Editing windows:
** create_edwin -- Create a file editing window.
** destroy_edwin -- Destroy a file editing window.
** add_edwin -- Add an EDWIN structure to a linked list.
** rmv_edwin -- Remove a EDWIN structure from a linked list and destroy it.
** init_enode -- Initialise an ENODE's type-specific fields.
**
** Linked list handling:
** add_node -- Add a node to a linked list.
** check_node -- Check for a node in a linked list.
** del_node -- Delete a node from a linked list.
** find_node -- Find a node in a linked list.
** show_node -- Show a node from a linked list.
** del_list -- Delete a linked list.
** show_list -- Show a linked list.
** octstr -- Convert a string to printable characters and octal values.
**
** Gnome GUI functions.
** test -- See if Gnome is running.
** init -- Initialise the editor's GUI.
** usage -- Display GUI-dependent options.
** loop -- Run the main loop.
** poll -- Check for user interaction.
** term -- Shut down the Gnome GUI.
** creategw -- Create an editing window.
** destroygw -- Destroy an editing window.
** focusgw -- Move the focus to another editing window.
** retitle -- Retitle the main window.
** reset -- Reset the Gnome GUI as far as is possible.
** flush -- Flush an editing window.
** moveto -- Move the cursor to a given character cell.
** draw -- Put some text on the vte_terminal widget.
** shift -- Shift the screen contents right.
** scroll -- Scroll the editing window.
** clrtoeol -- Clear to end of line.
** textline -- Write a line to the screen.
** beep -- Audible (and visual?) bell.
** guimsg -- Update the message bar of the Gnome window.
** scrollbar -- Set the size of the scroll bar.
** statusbar -- Update the status bar of the Gnome window.
** keylabel -- Convert <label> to key code.
** clipopen -- Open the clipboard.
** clipwrite -- Write text to the clipboard.
** clipread -- Read text from the clipboard.
** clipclose -- Close the clipboard.
** addcolor -- Add a color to the main color table.
** freecolor -- Remove a color from the main color table.
** setbg -- Set the terminal background color from the color palette.
** guicmd -- Update the X11 GUI toolbar.
** tabcmd -- Used by the tags command, somehow.
** save -- Save the window position.
** wildcard -- Filename wildcard expansion.
** pgopen -- Run an OS command connected to the Elvis process by pipes.
** pgclose -- Shut down an OS command.
** stop -- Run a command shell.
** newfiles -- Update the GUI's current filenames.
*/

static void empty_action(GtkAction *action, gpointer data);
static void open_action(GtkAction *action, gpointer data);
static void save_action(GtkAction *action, gpointer data);
static void saveas_action(GtkAction *action, gpointer data);
static void close_action(GtkAction *action, gpointer data);
static void quit_action(GtkAction *action, gpointer data);
static void yank_all_action(GtkAction *action, gpointer data);
static void move_action(GtkAction *action, gpointer data);
static void clear_history_action(GtkAction *action, gpointer data);
static void dictionary_action(GtkAction *action, gpointer data);
static void spellcheck_action(GtkAction *action, gpointer data);
static void manpage_action(GtkAction *action, gpointer data);
static void note_action(GtkAction *action, gpointer data);
static void notepad_action(GtkAction *action, gpointer data);
static void shell_action(GtkAction *action, gpointer data);
static void show_tools_action(GtkAction *action, gpointer data);
static void startpos_action(GtkAction *action, gpointer data);
static void default_size_action(GtkAction *action, gpointer data);
static void docs_how_action(GtkAction *action, gpointer data);
static void cursor_action(GtkWidget *widget, gpointer data);
static void color_action(GtkWidget *widget, gpointer data);
static void font_action(GtkWidget *widget, gpointer data);
#ifdef SAVE_CONFIG
static void config_action(GtkWidget *widget, gpointer data);
#endif
static void docs_clear_action(GtkAction *action, gpointer data);
static int get_print_filename(struct Printjob *print_job, int exists_flg);
static void print_action(GtkAction *action, gpointer data);
static void print_setup_action(GtkAction *action, gpointer data);
static void print_cancel_action(GtkAction *action, gpointer data);
static void help_action(GtkAction *action, gpointer data);
static void goto_url(GtkAboutDialog *about, const gchar *url, gpointer data);
static void about_action(GtkAction *action, gpointer data);
#ifdef DEBUG_MENU
static void debug_action(GtkAction *action, gpointer data);
static void child_exit(GPid pid, gint status, gpointer data);
static void selection_test(void);
static void target_handler(GtkWidget *, gpointer);
static void selection_received(GtkWidget *, GtkSelectionData *, guint,
								gpointer);
static void selection_acquire(GtkWidget *, gpointer);
static gint selection_blocked(GtkWidget *, GdkEventSelection *, gpointer);
static void target_set(GtkWidget *, GtkSelectionData *, guint, guint, gpointer);
#endif
static void add_r_action(GtkAction *action, gpointer data);
static void rmv_r_action(GtkAction *action, gpointer data);
#ifdef CHG_ARGSLIST
static void add_f_action(GtkAction *action, gpointer data);
static void rmv_f_action(GtkAction *action, gpointer data);
#endif
#ifdef CANCEL_ACTION
static void cancel_action(GtkToolButton *button, gpointer data);
#endif
static void ex_action(GtkToolButton *button, gpointer data);
static void re_action(GtkAction *action, gpointer data);
static void drag_action(GtkAction *action, gpointer data);
static void cut_action(GtkAction *action, gpointer data);
static void copy_action(GtkAction *action, gpointer data);
static void paste_action(GtkAction *action, gpointer data);
static void delete_action(GtkAction *action, gpointer data);
static int create_main_menu(GNGUI *gg);
static int create_mouse_menu(GNGUI *gg);
static void config_menus(GNGUI *gg);
static void switch_toolbar_menu(void);
static void switch_startpos_menu(void);
static void switch_docs_open_menu(void);
static void switch_manpage_menu(void);
static void update_undolevel_menu(int, int);
static int create_toolbar(GNGUI *gg);
static void switch_toolbar(void);
static gint snoop_handler(GtkWidget *widget, GdkEventKey *event, gpointer data);
#ifdef ADD_UNUSED
static gint popup_handler(GtkWidget *widget, GdkEvent *event, gpointer data);
static gint motion_handler(GtkWidget *widget, GdkEvent *event, gpointer data);
#endif
static gint scroll_handler(GtkWidget *widget, GdkEvent *event, gpointer data);
static gint mouse_handler(GtkWidget *widget, GdkEvent *event, gpointer data);
static gboolean range_handler(GtkRange *, GtkScrollType, gdouble, gpointer);
static void theme_handler(GtkIconTheme *, gpointer);
static void gconf_notify_handler(GConfClient *, guint, GConfEntry *, gpointer);
#ifdef SESSION_MGR
static gint client_save_handler(GnomeClient *, int, GnomeSaveStyle,
				int, GnomeInteractStyle, int, gpointer);
static gint client_die_handler(GnomeClient *, gpointer);
#endif
static void drag_data_handler(GtkWidget *, GdkDragContext *,
				GtkSelectionData *, guint, guint, gpointer);
static void drag_begin_handler(GtkWidget *, GdkDragContext *, gpointer);
static void drag_end_handler(GtkWidget *, GdkDragContext *, gpointer);
static void drop_data_handler(GtkWidget *, GdkDragContext *, gint, gint,
				GtkSelectionData *, guint, guint, gpointer);
static gboolean drag_drop_handler(GtkWidget *, GdkDragContext *,
				gint, gint, guint, gpointer);
static void entry_handler(GtkEditable *, gpointer);
static void filename_handler(GtkWidget *, gpointer);
static void documents_handler(GtkWidget *, gpointer);
static gint v_val_chg_handler(GtkAdjustment *, gpointer);
#ifdef ADD_UNUSED
static gint del_handler(GtkWidget *widget, GdkEvent *event, gpointer data);
#endif
static void glog_handler(const gchar *, GLogLevelFlags, const gchar *,
								gpointer);
static gboolean gio_handler(GIOChannel *, GIOCondition, gpointer);
static gint key_press_handler(GtkWidget *, GdkEventKey *, gpointer);
static void destroy_handler(GtkWidget *, gpointer);
static gint resize_handler(GtkWidget *, GdkEventConfigure *, gpointer);
static gboolean state_handler(GtkWidget *, GdkEventWindowState *, gpointer);
static void default_font_handler(GtkWidget *, gpointer);
static void default_color_handler(GtkWidget *, gpointer);
static void default_cursor_handler(GtkWidget *, gpointer);
static void terminal_font_handler(GtkWidget *, gpointer);
static void color_handler(GtkWidget *, gpointer);
static void cursor_handler(GtkWidget *, gpointer);
static void begin_print_handler(GtkPrintOperation *, GtkPrintContext *,
							gpointer);
static void draw_page_handler(GtkPrintOperation *, GtkPrintContext *, gint,
							gpointer);
static void end_print_handler(GtkPrintOperation *, GtkPrintContext *,
							gpointer);
static void done_print_handler(GtkPrintOperation *, GtkPrintOperationResult *,
							gpointer);
static void print_name_handler(GtkWidget *, gpointer);
static void default_print_name_handler(GtkWidget *, gpointer);
static void print_header_handler(GtkWidget *, gpointer);
static void print_font_handler(GtkWidget *, gpointer);
static void default_print_setup_handler(GtkWidget *, gpointer);
#ifdef ADD_UNUSED
static void hello(GtkWidget *, gpointer);
#endif
static int init_dand(GNGUI *gg);
static int init_defaults(int argc, char *argv[]);
static int init_gnome_win(GNGUI *gg);
static int init_interface(GNGUI *gg);
static int load_settings(void);
static gboolean show_warning(gpointer data);
static gboolean firstcmd(gpointer data);
static int store_settings(int posn);
#ifdef DEBUG_MENU
static void get_print_settings(void);
#endif
static int load_fileslist(int type);
static int store_docslist(void);
static int check_gconf(void);
static size_t request_clip(char **ppt, size_t maxlen);
static int url_to_path(char *path, char *data, size_t len);
static int undot_path(char *oldpath, char *newpath, size_t len);
static UINT get_id(void);
static char * get_exname(int);
static int get_icons(void);
static int get_tools_icon(void);
static void size_main_window(GNGUI *gg, int actual);
static void update_geometry(GNGUI *gg);
static void write_buff(EDWIN *pw, char *buff, size_t num);
static int terminal(char *cmd);
static void trap_handler(int signum);
static int create_edit_ring(ENODE **ppn, char **list, int next);
static int destroy_edit_ring(ENODE **ppn);
static int goto_edit_ring(GNGUI *gg, int count, int dirn);
static int update_edit_files(char **list, int next);
static int load_files_menu(ENODE *pn);
static int unload_files_menu(ENODE *pn);
static int parse_int(char *str, size_t index, size_t opt_index);
static int parse_two_ints(char *str, size_t index, size_t opt_index);
static int parse_bool(char *str, size_t index, size_t opt_index);
static int parse_color(char *str, size_t index, size_t opt_index);
static int parse_font(char *str, size_t index, size_t opt_index);
static int parse_string(char *str, size_t index, size_t opt_index);
static int parse_opts(int argc, char **argv);
#ifdef ADD_UNUSED
static int get_default_colormap(void);
#endif
static int set_default_colors(GNGUI *gg);
static int check_font_family(char *, int);
static int check_mono_family(char *);
static int get_mono_fonts(char *, char *);
#ifdef DPRINT_NN
static void list_fonts(PangoFontFamily *family);
#endif
static int set_current_font(GNGUI *gg, char *fontname, int type);
static int set_blink(struct optdesc_s *, OPTVAL *, CHAR *);
static int set_cursor(struct optdesc_s *, OPTVAL *, CHAR *);
static int set_fore(struct optdesc_s *, OPTVAL *, CHAR *);
static int set_back(struct optdesc_s *, OPTVAL *, CHAR *);
static int set_font(struct optdesc_s *, OPTVAL *, CHAR *);
static int set_cols(struct optdesc_s *, OPTVAL *, CHAR *);
static int set_rows(struct optdesc_s *, OPTVAL *, CHAR *);
static int set_tool(struct optdesc_s *, OPTVAL *, CHAR *);
static int set_print(struct optdesc_s *, OPTVAL *, CHAR *);
static int add_docsmenu(ENODE *pn);
static int rmv_docsmenu(ENODE *pn);
static int add_notesmenu(ENODE *pn);
static int create_edwin(EDWIN **ppw);
static int destroy_edwin(EDWIN **ppw);
static int add_edwin(EDWIN *pw, int posn);
static int rmv_edwin(EDWIN *pw);
static int init_enode(ENODE *pn);
static int add_node(ENODE **ppn, char *name, void *data, int posn);
static int check_node(ENODE *pn, char *name);
static int del_node(ENODE *pn, char *name, void **pdata);
static int find_node(ENODE **ppn, char *name, int full, int update);
static int show_node(ENODE *pn, char *header, int disp);
static int del_list(ENODE **ppn);
static int show_list(ENODE *pn);
#ifdef ADD_UNUSED
static char * octstr(char *pc);
#endif

/* The Gnome GUI calls the following Elvis functions directly:
**
** msgflush() is called directly before calls to eventex(), and after some
** msg(MSG_ERROR) calls; Elvis seems to handle the other cases itself.
** ioabsolute() is used to derive absolute paths for file handling.
** drawextext() and winofgw() are used to force a full screen update.
** morepush() is supplies the "Hit <Enter> ..." message.
**
** msg()
** msgflush()
**
** drawextext()
** eventclick()
** eventcreate()
** eventdestroy()
** eventdraw()
** eventex()
** eventkeys()
** eventresize()
** eventscroll()
** ioabsolute()
** morepush()
** winofgw()
*/

/* The main Elvis program calls the following functions via the
** Elvis GUI when it wants the Gnome GUI to do something.
** The functions newfiles(), newcmd() and newundo() have global scope
** as a temporary measure; they need to be included in the Elvis GUI. ###
**
** int	(*test) P_((void));
** int	(*init) P_((int argc, char **argv));
** void	(*usage) P_((void));
** void	(*loop) P_((void));
** ELVBOOL (*poll) P_((ELVBOOL reset));
** void	(*term) P_((void));
** ELVBOOL (*creategw) P_((char *name, char *attributes));
** void	(*destroygw) P_((GUIWIN *gw, ELVBOOL force));
** ELVBOOL (*focusgw) P_((GUIWIN *gw));
** void	(*retitle) P_((GUIWIN *gw, char *name));
** void	(*reset) P_((void));
** void	(*flush) P_((void));
** void	(*moveto) P_((GUIWIN *gw, int column, int row));
** void	(*draw) P_((GUIWIN *gw, long fg, long bg, int bits, CHAR *text,
**                               int len));
** ELVBOOL (*shift) P_((GUIWIN *gw, int qty, int rows));
** ELVBOOL (*scroll) P_((GUIWIN *gw, int qty, ELVBOOL notlast));
** ELVBOOL (*clrtoeol) P_((GUIWIN *gw));
** void	(*textline) P_((GUIWIN *gw, CHAR *text, int len));
** void	(*beep) P_((GUIWIN *gw));
** This clashes with the prototype in message.h, but conforms to gui.h.
** ELVBOOL (*msg) P_((GUIWIN *gw, MSGIMP imp, CHAR *text, int len));
** void	(*scrollbar) P_((GUIWIN *gw, long top, long bottom, long nlines));
** status() upsets Doxygen 1.4.7, since it's also used as a variable.
** ELVBOOL (*statusbar) P_((GUIWIN *gw, CHAR *cmd, long line,
**                               long column, _CHAR_ learn, char *mode));
** int	(*keylabel) P_((CHAR *given, int givenlen, CHAR **label, CHAR **raw));
** ELVBOOL (*clipopen) P_((ELVBOOL forwrite));
** int (*clipwrite) P_((CHAR *text, int len));
** int (*clipread) P_((CHAR *text, int len));
** void (*clipclose) P_((void));
** ELVBOOL (*color) P_((int fontcode, CHAR *colorname, ELVBOOL isfg,
**                               long *colorptr, unsigned char rgb[3]));
** void	(*freecolor) P_((long color, ELVBOOL isfg));
** void	(*setbg) P_((GUIWIN *gw, long bg));
** ELVBOOL (*guicmd) P_((GUIWIN *gw, char *extra));
** ELVBOOL (*tabcmd) P_((GUIWIN *gw, _CHAR_ key2, long count));
** void (*save) P_((BUFFER buf, GUIWIN *gw));
** int (*wildcard) P_((char *names, char *buf, int bufsize, ELVBOOL single));
** ELVBOOL (*prgopen) P_((char *command, ELVBOOL willwrite, ELVBOOL willread));
** int (*prgclose) P_((void));
** RESULT (*stop) P_((ELVBOOL alwaysfork));
*/

/* Elvis GUI function prototypes */

static int	test(void);
static	int	init(int argc, char **argv);
static void	usage(void);
static void	loop(void);
static ELVBOOL	poll(ELVBOOL reset);
static void	term(void);
static ELVBOOL	creategw(char *name, char *attributes);
static void	destroygw(GUIWIN *gw, ELVBOOL force);
static ELVBOOL	focusgw(GUIWIN *gw);
static void	retitle(GUIWIN *gw, char *name);
static void	reset(void);
static void	flush(void);
static void	moveto(GUIWIN *gw, int column, int row);
static void	draw(GUIWIN *gw, long fg, long bg, int bits, CHAR *text,
								int len);
static ELVBOOL  shift(GUIWIN *gw, int qty, int rows);
static ELVBOOL	scroll(GUIWIN *gw, int qty, ELVBOOL notlast);
static ELVBOOL	clrtoeol(GUIWIN *gw);
static void	textline(GUIWIN *gw, CHAR *text, int len);
static void	beep(GUIWIN *gw);
/* msg() clashes with the prototype in message.h, but conforms to gui.h */
static ELVBOOL	guimsg(GUIWIN *gw, MSGIMP imp, CHAR *text, int len);
static void	scrollbar(GUIWIN *gw, long top, long bottom, long nlines);
/* status() upsets Doxygen 1.4.7, since it's also used as a variable */
static ELVBOOL	statusbar(GUIWIN *gw, CHAR *cmd, long line, long column,
						_CHAR_ learn, char *mode);
static int	keylabel(CHAR *given, int givenlen, CHAR **label, CHAR **raw);
static ELVBOOL	clipopen(ELVBOOL forwrite);
static int	clipread(CHAR *text, int len);
static int	clipwrite(CHAR *text, int len);
static void	clipclose(void);
static ELVBOOL	addcolor(int fontcode, CHAR *colorname, ELVBOOL isfg,
					long *colorptr, unsigned char rgb[3]);
static void	freecolor(long color, ELVBOOL isfg);
static void	setbg(GUIWIN *gw, long bg);
static ELVBOOL	tabcmd(GUIWIN *gw, _CHAR_ key2, long count);
static ELVBOOL	guicmd(GUIWIN *gw, char *extra);
static void	save(BUFFER buf, GUIWIN *gw);
static int	wildcard(char *names, char *buf, int bufsize, ELVBOOL single);
static ELVBOOL	pgopen(char *command, ELVBOOL willwrite, ELVBOOL willread);
static int	pgclose(void);
static RESULT	stop(ELVBOOL alwaysfork);
	/* These three need to be part of the Elvis GUI structure ### */
	int newfiles(char **list, int next);
	int newcmd(int cmd, int flg);
	int newundo(int cmd, long int level, long int maxlevel);
	/* Used only for debugging */
	void memprofile(void);

/* Elvis GUI interface structure initialisation */

GUI guignome = {
  /* Data and flags */
  "gnome",		/* gui name */
  "Gnome user interface", /* gui description */
  FALSE,		/* exonly: no visual editing */
  FALSE,		/* newblank: inserted chars are undefined ??? */
  TRUE,			/* minimizeclr: minimise length in clrtoel() */
  FALSE,		/* scrolllast: scroll the last line */
  FALSE,		/* shiftrows: shift rows <=>. True for X11 GUI ??? */
  0,			/* movecost: time cost of moving the cursor */
  0,			/* nopts: number of GUI options */
  NULL,			/* optdescs: list of GUI options */
  /* Functions */
  test,			/* test interface availability */
  init,			/* initialise interface */
  usage,		/* usage message */
  loop,			/* main loop of editor */
  poll,			/* check for user interaction */
  term,			/* shut down the GUI */
  creategw,		/* create an editing window */
  destroygw,		/* destroy an editing window */
  focusgw,		/* focusgw: switch editing window focus */
  retitle,		/* change the editing window title */
  reset,		/* reset the editing window */
  flush,		/* flush output to the screen */
  moveto,		/* update the cursor position */
  draw,			/* (re-)draw the screen */
  NULL,			/* shift(): shift the screen right */
  scroll,		/* scroll the screen */
  clrtoeol,		/* clear to the eol */
  NULL,			/* textline(): write a line to the screen */
  beep,			/* visual or audible bell */
  guimsg,		/* msg() => guimsg(): write a message to the GUI */
  scrollbar,		/* update the scrollbar */
  statusbar,		/* write to the status line */
  keylabel,		/* convert <label> to key code */
  clipopen,		/* clipoard open */
  clipwrite,		/* clipboard write */
  clipread,		/* clipboard read */
  clipclose,		/* clipboard close */
  addcolor,		/* color() => addcolor(): set the colors for fonts */
  freecolor,		/* free a color */
  NULL,			/* setbg(): set the background color */
  NULL,			/* guicmd(): update the X11 GUI toolbar */
  NULL,			/* tabcmd(): used by tags lookup. (Unused in X11.) */
  NULL,			/* save(): save the window state */
  NULL,			/* wildcard(): file name expansion */
  NULL,			/* pgopen() =>prgopen(): run an OS command */
  NULL,			/* pgclose() =>prgclose(): shut down an OS command */
  stop			/* run a shell */
};

/*
** Menu descriptions for use by the GTK UI Manager.
** Unfortunately menu bar separators can't be used to force
** right-justification as toolbar separators can; apparently
** this is a feature of the Gnome UI.
*/

/* Main menu description.
**
** NB. <separator expand=\"true\"/> doesn't work with menus.
*/
static char menu_string[] = " \
<ui> \
  <menubar> \
    <menu name=\"FileMenu\" action=\"FileMenu_Action\"> \
      <menuitem name=\"New\" action=\"New_Action\"/> \
      <menuitem name=\"Open\" action=\"Open_Action\"/> \
      <separator/> \
      <menuitem name=\"Save\" action=\"Save_Action\"/> \
      <menuitem name=\"Save As\" action=\"Save_as_Action\"/> \
      <separator/> \
      <placeholder name=\"Empty\"/> \
      <separator/> \
      <menuitem name=\"Close\" action=\"Close_Action\"/> \
      <menuitem name=\"Quit\" action=\"Quit_Action\"/> \
    </menu> \
    <menu name=\"EditMenu\" action=\"EditMenu_Action\"> \
      <menuitem name=\"Undo\" action=\"Undo_Action\"/> \
      <menuitem name=\"Redo\" action=\"Redo_Action\"/> \
      <separator/> \
      <menuitem name=\"Cut\" action=\"Cut_Action\"/> \
      <menuitem name=\"Copy\" action=\"Copy_Action\"/> \
      <menuitem name=\"Paste\" action=\"Paste_Action\"/> \
      <menuitem name=\"Delete\" action=\"Delete_Action\"/> \
      <separator/> \
      <menuitem name=\"Yank All\" action=\"Yank_all_Action\"/> \
    </menu> \
    <menu name=\"SearchMenu\" action=\"SearchMenu_Action\"> \
      <menuitem name=\"Find Next\" action=\"Find_next_Action\"/> \
      <menuitem name=\"Find Previous\" action=\"Find_prev_Action\"/> \
      <menuitem name=\"Find Again\" action=\"Find_again_Action\"/> \
      <separator/> \
      <menuitem name=\"Replace\" action=\"Replace_Action\"/> \
      <menuitem name=\"Replace Again\" action=\"Replace_again_Action\"/> \
      <separator/> \
      <menuitem name=\"Goto\" action=\"Goto_Action\"/> \
      <menuitem name=\"Goto Again\" action=\"Goto_again_Action\"/> \
      <separator/> \
      <menuitem name=\"Clear History\" action=\"Clear_history_Action\"/> \
    </menu> \
    <menu name=\"ToolsMenu\" action=\"ToolsMenu_Action\"> \
      <menuitem name=\"Dictionary\" action=\"Dictionary_Action\"/> \
      <menuitem name=\"Spellcheck\" action=\"Spellcheck_Action\"/> \
      <menuitem name=\"Manpage\" action=\"Manpage_one_Action\"/> \
      <menuitem name=\"Manpage All\" action=\"Manpage_all_Action\"/> \
      <menuitem name=\"Notepads\" action=\"Note_Action\"/> \
      <menuitem name=\"Shell\" action=\"Shell_Action\"/> \
    </menu> \
    <menu name=\"OptionsMenu\" action=\"OptionsMenu_Action\"> "
#ifdef SAVE_CONFIG
     "<menuitem name=\"Save config\" action=\"Save_config_Action\"/> \
      <menuitem name=\"Load config\" action=\"Load_config_Action\"/> \
    <separator/> "
#endif
     "<menuitem name=\"Show Toolbar\" action=\"Show_toolbar_Action\"/> \
      <menuitem name=\"Hide Toolbar\" action=\"Hide_toolbar_Action\"/> \
      <menuitem name=\"Fix Position\" action=\"Fix_startpos_Action\"/> \
      <menuitem name=\"Free Position\" action=\"Free_startpos_Action\"/> \
      <menuitem name=\"Default Size\" action=\"Default_size_Action\"/> \
      <menuitem name=\"Open Current\" action=\"Open_cur_Action\"/> \
      <menuitem name=\"Open New\" action=\"Open_new_Action\"/> \
      <menuitem name=\"Cursor\" action=\"Cursor_Action\"/> \
      <menuitem name=\"Color\" action=\"Color_Action\"/> \
      <menuitem name=\"Font\" action=\"Font_Action\"/> \
    </menu> \
    <menu name=\"DocsMenu\" action=\"DocsMenu_Action\"> \
      <menuitem name=\"Clear List\" action=\"Docs_clear_Action\"/> \
      <placeholder name=\"Empty\"/> \
    </menu> "
#ifdef PRINT_MENU
   "<menu name=\"PrintMenu\" action=\"PrintMenu_Action\"> \
      <menuitem name=\"Print Setup\" action=\"Print_setup_Action\"/> \
      <menuitem name=\"Print Cancel\" action=\"Print_cancel_Action\"/> \
      <menuitem name=\"Print File\" action=\"Print_Action\"/> \
    </menu> "
#endif
   "<menu name=\"HelpMenu\" action=\"HelpMenu_Action\"> \
      <menuitem name=\"Contents\" action=\"Help_Action\"/> \
      <menuitem name=\"About\" action=\"About_Action\"/> \
    </menu> "
#ifdef DEBUG_MENU
   "<separator/> \
    <menu name=\"DebugMenu\" action=\"DebugMenu_Action\"> \
      <menuitem name=\"Debug\" action=\"Debug_Debug_Action\"/> \
      <menuitem name=\"Docs\" action=\"Debug_Docs_Action\"/> \
      <menuitem name=\"Edwin\" action=\"Debug_Edwin_Action\"/> \
      <menuitem name=\"Enode\" action=\"Debug_Enode_Action\"/> \
      <menuitem name=\"Gui\" action=\"Debug_GUI_Action\"/> \
      <menuitem name=\"Keys\" action=\"Debug_Keys_Action\"/> \
      <menuitem name=\"Menu\" action=\"Debug_Menu_Action\"/> \
      <menuitem name=\"Notes\" action=\"Debug_Notes_Action\"/> \
      <menuitem name=\"Memory\" action=\"Debug_Memory_Action\"/> \
      <menuitem name=\"Print\" action=\"Debug_Print_Action\"/> \
      <menuitem name=\"Scroll\" action=\"Debug_Scroll_Action\"/> \
      <menuitem name=\"Select\" action=\"Debug_Select_Action\"/> \
    </menu> "
#endif
   "</menubar> \
</ui> \
";

/* Main menu actions.
**
** The main menu is built from the GtkActionEntry menu_entries[],
** and if this array is changed the references to it in menu_strings[]
** must also be changed, to keep tooltips synchronised.
**
** Main menu GtkActionEntry elements.
** Fields: Name, ID, Label, Accelerator, Tooltip, Callback.
** The marked character of the label is the accelerator.
*/
static GtkActionEntry menu_entries[] = {

	/* Main menu headings */
	{ "FileMenu_Action", NULL, "_File", NULL, NULL, NULL },
	{ "EditMenu_Action", NULL, "_Edit", NULL, NULL, NULL },
	{ "SearchMenu_Action", NULL, "_Search", NULL, NULL, NULL },
	{ "ToolsMenu_Action", NULL, "_Tools", NULL, NULL, NULL },
	{ "OptionsMenu_Action", NULL, "_Options", NULL, NULL, NULL },
	{ "DocsMenu_Action", NULL, "_Documents", NULL, NULL, NULL },
#ifdef PRINT_MENU
	{ "PrintMenu_Action", NULL, "_Print", NULL, NULL, NULL },
#endif
	{ "HelpMenu_Action", NULL, "_Help", NULL, NULL, NULL },
	{ "DebugMenu_Action", NULL, "Debug", NULL, NULL, NULL },

	/* File sub-menu */
	{ "New_Action", GTK_STOCK_NEW,
		"_New", "",
		"Open a new file",
		G_CALLBACK(move_action) },

	{ "Open_Action", GTK_STOCK_OPEN,
		"_Open", "",
		"Open an existing file",
		G_CALLBACK(open_action) },

	{ "Save_Action", GTK_STOCK_SAVE,
		"_Save", "",
		"Save the current file",
		G_CALLBACK(save_action) },

	{ "Save_as_Action", GTK_STOCK_SAVE_AS,
		"Sa_ve As", "",
		"Save as ...",
		G_CALLBACK(saveas_action) },

	{ "Close_Action", GTK_STOCK_CLOSE,
		"_Close", "",
		"Close the current file",
		G_CALLBACK(close_action) },

	{ "Quit_Action", GTK_STOCK_QUIT,
		"_Abandon", "",
		"Quit without saving",
		G_CALLBACK(quit_action) },

		/* Edit sub-menu */
	{ "Undo_Action", GTK_STOCK_UNDO,
		"_Undo", "",
		"Undo the last change",
		G_CALLBACK(re_action) },

	{ "Redo_Action", GTK_STOCK_REDO,
		"_Redo", "",
		"Reverse the last Undo",
		G_CALLBACK(re_action) },

	{ "Cut_Action", GTK_STOCK_CUT,
		"Cu_t", "",
		"Cut the current block",
		G_CALLBACK(cut_action) },

	{ "Copy_Action", GTK_STOCK_COPY,
		"_Copy", "",
		"Copy to the clipboard",
		G_CALLBACK(copy_action) },

	{ "Paste_Action", GTK_STOCK_PASTE,
		"_Paste", "",
		"Paste from the clipboard",
		G_CALLBACK(paste_action) },

	{ "Delete_Action", GTK_STOCK_DELETE,
		"_Delete", "",
		"Delete the current block",
		G_CALLBACK(delete_action) },

	{ "Yank_all_Action", GTK_STOCK_SELECT_ALL,
		"_Yank All", "",
		"Yank the entire file",
		G_CALLBACK(yank_all_action) },

		/* Search sub-menu */
	{ "Find_next_Action", GTK_STOCK_FIND,
		"_Find Next", "",
		"Search forwards",
		G_CALLBACK(move_action) },

	{ "Find_prev_Action", GTK_STOCK_FIND,
		"Find _Previous", "",
		"Search backwards",
		G_CALLBACK(move_action) },

	{ "Find_again_Action", GTK_STOCK_FIND,
		"Find _Again", "",
		"Repeat the previous search",
		G_CALLBACK(move_action) },

	{ "Replace_Action", GTK_STOCK_FIND_AND_REPLACE,
		"_Replace", "",
		"Replace a string",
		G_CALLBACK(move_action) },

	{ "Replace_again_Action", GTK_STOCK_FIND_AND_REPLACE,
		"Replace Again", "",
		"Repeat the previous change",
		G_CALLBACK(move_action) },

	{ "Goto_Action", GTK_STOCK_JUMP_TO,
		"_Goto", "",
		"Go to a line number",
		G_CALLBACK(move_action) },

	{ "Goto_again_Action", GTK_STOCK_JUMP_TO,
		"Goto Again", "",
		"Repeat the previous Goto",
		G_CALLBACK(move_action) },

	{ "Clear_history_Action", GTK_STOCK_CLEAR,
		"Clear History", "",
		"Clear the search history",
		G_CALLBACK(clear_history_action) },

		/* Tools sub-menu */
	{ "Dictionary_Action", GTK_STOCK_SPELL_CHECK,
		"_Dictionary", "",
		"Display a dictionary reference",
		G_CALLBACK(dictionary_action) },

	{ "Spellcheck_Action", GTK_STOCK_SPELL_CHECK,
		"Spell_check", "",
		"Check a words spelling",
		G_CALLBACK(spellcheck_action) },

	{ "Manpage_one_Action", GTK_STOCK_INFO,
		"_Manpage", "",
		"Display the first man page found",
		G_CALLBACK(manpage_action) },

	{ "Manpage_all_Action", GTK_STOCK_INFO,
		"_Manpage all", "",
		"Display all the man pages found",
		G_CALLBACK(manpage_action) },

	{ "Note_Action", GTK_STOCK_EDIT,
		"Notepads", "",
		"Notepad files",
		G_CALLBACK(note_action) },

	{ "Notepad_Action", GTK_STOCK_EDIT,
		"_Notepad", "",
		"Edit a note pad",
		G_CALLBACK(notepad_action) },

	{ "Shell_Action", GTK_STOCK_EXECUTE,
		"_Shell", "",
		"Run a command shell",
		G_CALLBACK(shell_action) },

		/* Options sub-menu */
	{ "Show_toolbar_Action", GTK_STOCK_YES,
		"_Show Toolbar", "",
		"Show the toolbar",
		G_CALLBACK(show_tools_action) },

	{ "Hide_toolbar_Action", GTK_STOCK_NO,
		"_Hide Toolbar", "",
		"Hide the toolbar",
		G_CALLBACK(show_tools_action) },

	{ "Fix_startpos_Action", GTK_STOCK_YES,
		"F_ix Position", "",
		"Fix the main window startup position",
		G_CALLBACK(startpos_action) },

	{ "Free_startpos_Action", GTK_STOCK_NO,
		"F_ree Position", "",
		"Free the main window startup position",
		G_CALLBACK(startpos_action) },

	{ "Default_size_Action", GTK_STOCK_ZOOM_100,
		"_Default Size", "",
		"Restore the terminal to its default size",
		G_CALLBACK(default_size_action) },

	{ "Open_cur_Action", GTK_STOCK_GOTO_BOTTOM,
		"Open _Current", "",
		"Open new documents in the current window",
		G_CALLBACK(docs_how_action) },

	{ "Open_new_Action", GTK_STOCK_GOTO_LAST,
		"Open _New", "",
		"Open new documents in a new window",
		G_CALLBACK(docs_how_action) },

	{ "Cursor_Action", GTK_STOCK_MEDIA_PLAY,
		"C_ursor", "",
		"Set the current terminal cursor style",
		G_CALLBACK(cursor_action) },

	{ "Color_Action", GTK_STOCK_SELECT_COLOR,
		"C_olor", "",
		"Set the current terminal colors",
		G_CALLBACK(color_action) },

	{ "Font_Action", GTK_STOCK_SELECT_FONT,
		"_Font", "",
		"Set the current terminal fonts",
		G_CALLBACK(font_action) },

#ifdef SAVE_CONFIG
	{ "Save_config_Action", GTK_STOCK_SAVE,
		"Save config", "",
		"Save the current configuration",
		G_CALLBACK(config_action) },

	{ "Load_config_Action", GTK_STOCK_REVERT_TO_SAVED,
		"Load config", "",
		"Load the current configuration",
		G_CALLBACK(config_action) },
#endif

		/* Documents sub-menu */
	{ "Docs_clear_Action", GTK_STOCK_CLEAR,
		"Clear List", "",
		"Clear the Documents list",
		G_CALLBACK(docs_clear_action) },

#ifdef PRINT_MENU
		/* Print sub-menu */
	{ "Print_setup_Action", GTK_STOCK_PRINT,
		"Print _Setup", "",
		"Set the printing fonts and header",
		G_CALLBACK(print_setup_action) },

	{ "Print_cancel_Action", GTK_STOCK_PRINT,
		"Print _Cancel", "",
		"Cancel the current print job",
		G_CALLBACK(print_cancel_action) },

	{ "Print_Action", GTK_STOCK_PRINT,
		"_Print File", "",
		"Select and Print a file",
		G_CALLBACK(print_action) },
#endif

	/* Help sub-menu */
	{ "Help_Action", GTK_STOCK_HELP,
		"_Contents", "F1",		/* F1 handled by key snooper */
		"Help for Elvis",
		G_CALLBACK(help_action) },

	{ "About_Action", GTK_STOCK_ABOUT,
		"_About", "",
		"Show the About message",
		G_CALLBACK(about_action) },

#ifdef DEBUG_MENU
		/* Debug sub-menu */
	{ "Debug_Debug_Action", GTK_STOCK_ABOUT,
		"Debug", "",
		"Debug",
		G_CALLBACK(debug_action) },

	{ "Debug_Docs_Action", GTK_STOCK_ABOUT,
		"Docslist", "",
		"Show Docslist",
		G_CALLBACK(debug_action) },

	{ "Debug_Edwin_Action", GTK_STOCK_ABOUT,
		"Edwin", "",
		"Show Edwin",
		G_CALLBACK(debug_action) },

	{ "Debug_Enode_Action", GTK_STOCK_ABOUT,
		"Enode", "",
		"Show Enode",
		G_CALLBACK(debug_action) },

	{ "Debug_GUI_Action", GTK_STOCK_ABOUT,
		"GUI", "",
		"Show GUI",
		G_CALLBACK(debug_action) },

	{ "Debug_Keys_Action", GTK_STOCK_ABOUT,
		"Keys", "",
		"Show Keys",
		G_CALLBACK(debug_action) },

	{ "Debug_Menu_Action", GTK_STOCK_ABOUT,
		"Menu", "",
		"Show Menu",
		G_CALLBACK(debug_action) },

	{ "Debug_Notes_Action", GTK_STOCK_ABOUT,
		"Notes", "",
		"Show Notes",
		G_CALLBACK(debug_action) },

	{ "Debug_Memory_Action", GTK_STOCK_ABOUT,
		"Memory", "",
		"Show Memory",
		G_CALLBACK(debug_action) },

	{ "Debug_Print_Action", GTK_STOCK_ABOUT,
		"Print", "",
		"Show Print",
		G_CALLBACK(debug_action) },

	{ "Debug_Scroll_Action", GTK_STOCK_ABOUT,
		"Scroll", "",
		"Show Scroll",
		G_CALLBACK(debug_action) },

	{ "Debug_Select_Action", GTK_STOCK_ABOUT,
		"Select", "",
		"Show Select",
		G_CALLBACK(debug_action) },
#endif

	/* Put last for convenience */
	{ "Empty_Action", NULL,			/* name, stock id */
		"empty", "",			/* label, accelerator */
		"Placeholder",			/* tooltip */
		G_CALLBACK(empty_action) }	/* callback */
  };

/* Pop-up menu description */
static char mouse_str[] = " \
<ui> \
  <popup> \
    <menuitem name=\"Undo\" action=\"Undo_Action\"/> \
    <menuitem name=\"Redo\" action=\"Redo_Action\"/> \
    <separator/> \
    <menuitem name=\"Drag\" action=\"Drag_Action\"/> \
    <separator/> \
    <menuitem name=\"Cut\" action=\"Cut_Action\"/> \
    <menuitem name=\"Copy\" action=\"Copy_Action\"/> \
    <menuitem name=\"Paste\" action=\"Paste_Action\"/> \
    <menuitem name=\"Delete\" action=\"Delete_Action\"/> \
  </popup> \
</ui> \
";

/* Pop-up menu actions - NB. no accelerators */
static GtkActionEntry mouseentries[] = {
	{ "Undo_Action", GTK_STOCK_UNDO,
		"Undo", "",
		"Undo the last action",
		G_CALLBACK(re_action) },

	{ "Redo_Action", GTK_STOCK_REDO,
		"Redo", "",
		"Redo the last action",
		G_CALLBACK(re_action) },

	{ "Drag_Action", GTK_STOCK_DND,
		"Drag", "",
		"Drag text to another window",
		G_CALLBACK(drag_action) },

	{ "Cut_Action", GTK_STOCK_CUT,
		"Cut", "",
		"Cut to the clipboard",
		G_CALLBACK(cut_action) },

	{ "Copy_Action", GTK_STOCK_COPY,
		"Copy", "",
		"Copy to the clipboard",
		G_CALLBACK(copy_action) },

	{ "Paste_Action", GTK_STOCK_PASTE,
		"Paste", "",
		"Paste from the clipboard",
		G_CALLBACK(paste_action) },

	{ "Delete_Action", GTK_STOCK_DELETE,
		"Delete", "",
		"Delete a text block",
		G_CALLBACK(delete_action) }
  };

/* Toolbar toolbutton data */
typedef void (* TOOL_CALL)(GtkToolButton *, gpointer);
static struct Tool {
	GtkToolItem *button;	/* The button in question */
	const gchar *icon;	/* The button's icon (type ???) */
	char *signal;		/* The signal to be connected */
	TOOL_CALL callback;	/* The callback for the signal */
	gpointer data;		/* The data for the callback */
	char *label;		/* The label for the button */
	char *tip;		/* The tooltip for the button */
	char *cmd;		/* The Elvis (ex) command to be run */
  } tools[] = {
	/* Since there's currently no easy way to modify an arglist
	** the file open and file remove functions have been dropped
	** from the toolbar. ###
	*/
	{ NULL, GTK_STOCK_ADD, "clicked",
			(TOOL_CALL) add_r_action, (gpointer) &gn_gui,
			"Add", "Add an edit ring", NULL },
	{ NULL, GTK_STOCK_REMOVE, "clicked",
			(TOOL_CALL) rmv_r_action, (gpointer) &gn_gui,
			"Drop", "Drop the current edit ring", NULL },
	{ NULL, GTK_STOCK_MEDIA_REWIND, "clicked", ex_action, NULL,
			"<-Ring", "Go to the preceeding ring", "wi -" },
	{ NULL, GTK_STOCK_MEDIA_FORWARD, "clicked", ex_action, NULL,
			"Ring->", "Go to the following ring", "wi +" },
#define LIST_RINGS 4
	{ NULL, NULL, "clicked", ex_action, NULL,
			"Rings", "List the edit rings", "wi" },
#ifdef CHG_ARGSLIST
	{ NULL, GTK_STOCK_OPEN, "clicked",
			(TOOL_CALL) add_f_action, (gpointer) &gn_gui,
			"Add", "Add a file to the current ring", NULL },
	{ NULL, GTK_STOCK_CLOSE, "clicked",
			(TOOL_CALL) rmv_f_action, (gpointer) &gn_gui,
			"Drop", "Drop a file from the current ring", NULL },
#endif
	{ NULL, GTK_STOCK_GO_BACK, "clicked", ex_action, NULL,
			"<-File", "Go to the preceeding file", "N" },
	{ NULL, GTK_STOCK_GO_FORWARD, "clicked", ex_action, NULL,
			"File->", "Go to the following file", "n" },
	{ NULL, GTK_STOCK_GO_UP, "clicked", ex_action, NULL,
			"Back", "HTML back button", "pop" },
	{ NULL, GTK_STOCK_MEDIA_STOP, "clicked",
			(TOOL_CALL) default_size_action, (gpointer) &gn_gui,
			"Default size", "Restore the default window size",
			NULL },
#ifdef CANCEL_ACTION
	{ NULL, GTK_STOCK_CANCEL, "clicked", cancel_action, NULL,
			"Cancel", "Cancel the current task", NULL },
#endif
	{ NULL, GTK_STOCK_EXECUTE, "clicked", ex_action, NULL,
			"Shell", "Run a command shell", "sh" }
  };

/* Tooltips apparently can't be added as a group to menus generated
** by gtk_ui_manager; hence the need for this array, which repeats
** the data in the main menu array above.  This is very annoying.
*/
static struct Menu_str {
	int ref;	/* Index into menu_entries[] */
	char *str;	/* Path from menu_string */
  } menu_strings[] = {
#ifdef PRINT_MENU
	{ 30, "/ui/menubar/ToolsMenu/Dictionary" },
	{ 31, "/ui/menubar/ToolsMenu/Spellcheck" },
	{ 32, "/ui/menubar/ToolsMenu/Manpage" },
	{ 33, "/ui/menubar/ToolsMenu/Manpage All" },
	{ 34, "/ui/menubar/ToolsMenu/Notepads" },
	/* Empty */
	{ 36, "/ui/menubar/ToolsMenu/Shell" },
	{ 37, "/ui/menubar/OptionsMenu/Show Toolbar" },
	{ 38, "/ui/menubar/OptionsMenu/Hide Toolbar" },
	{ 39, "/ui/menubar/OptionsMenu/Fix Position" },
	{ 40, "/ui/menubar/OptionsMenu/Free Position" },
	{ 41, "/ui/menubar/OptionsMenu/Default Size" },
	{ 42, "/ui/menubar/OptionsMenu/Open Current" },
	{ 43, "/ui/menubar/OptionsMenu/Open New" },
	{ 44, "/ui/menubar/OptionsMenu/Cursor" },
	{ 45, "/ui/menubar/OptionsMenu/Color" },
	{ 46, "/ui/menubar/OptionsMenu/Font" },
	{ 47, "/ui/menubar/DocsMenu/Clear List" },
	{ 48, "/ui/menubar/PrintMenu/Print Setup" },
	{ 49, "/ui/menubar/PrintMenu/Print Cancel" },
	{ 50, "/ui/menubar/PrintMenu/Print File" },
	{ 51, "/ui/menubar/HelpMenu/Contents" },
	{ 52, "/ui/menubar/HelpMenu/About" }
#else
	{ 29, "/ui/menubar/ToolsMenu/Dictionary" },
	{ 30, "/ui/menubar/ToolsMenu/Spellcheck" },
	{ 31, "/ui/menubar/ToolsMenu/Manpage" },
	{ 32, "/ui/menubar/ToolsMenu/Manpage All" },
	{ 33, "/ui/menubar/ToolsMenu/Notepads" },
	/* Empty */
	{ 35, "/ui/menubar/ToolsMenu/Shell" },
	{ 36, "/ui/menubar/OptionsMenu/Show Toolbar" },
	{ 37, "/ui/menubar/OptionsMenu/Hide Toolbar" },
	{ 38, "/ui/menubar/OptionsMenu/Fix Position" },
	{ 39, "/ui/menubar/OptionsMenu/Free Position" },
	{ 40, "/ui/menubar/OptionsMenu/Default Size" },
	{ 41, "/ui/menubar/OptionsMenu/Open Current" },
	{ 42, "/ui/menubar/OptionsMenu/Open New" },
	{ 43, "/ui/menubar/OptionsMenu/Cursor" },
	{ 44, "/ui/menubar/OptionsMenu/Color" },
	{ 45, "/ui/menubar/OptionsMenu/Font" },
	{ 46, "/ui/menubar/DocsMenu/Clear List" },
	{ 47, "/ui/menubar/HelpMenu/Contents" },
	{ 48, "/ui/menubar/HelpMenu/About" }
#endif
  };

/* Functions */

/* Main help menu actions */

/*
** e m p t y _ a c t i o n
**
** A placeholder for the main menu.
**
** Returns:	Nothing		Always
*/
static void empty_action(GtkAction *action, gpointer data)
{
  /* GNGUI *gg = (GNGUI *) data; */
  g_print("empty_action(): action |%s|\n", gtk_action_get_name(action));

  return;
}

/* File Menu */

/*
** o p e n _ a c t i o n
**
** Open an existing file and make it current.
**
** Returns:	Nothing		Always
*/
static void open_action(GtkAction *action, gpointer data)
{
  GNGUI *gg = (GNGUI *) data;
  GtkWidget *dialog;
  GtkWidget *toggle;
  int len, ro_flg;
  char *pf, buff[PATH_MAX + 8] = { '\0' };

  DPRINT_F(("open_action()\n"));

  /* Get the new filename from the file chooser dialog */
  pf = NULL;
  ro_flg = FALSE;
  strcpy(buff, "next ");
  len = strlen(buff);
  dialog = gtk_file_chooser_dialog_new(
				"Open File",
				NULL,
				GTK_FILE_CHOOSER_ACTION_OPEN,
				GTK_STOCK_CANCEL, GTK_RESPONSE_CANCEL,
				GTK_STOCK_OPEN, GTK_RESPONSE_ACCEPT,
				NULL
				);
  gtk_file_chooser_set_local_only(GTK_FILE_CHOOSER(dialog), TRUE);

  /* This adds a R/O toggle to the file chooser dialog */
  toggle = gtk_check_button_new_with_label ("Open file read-only");
  gtk_widget_show(toggle);
  gtk_file_chooser_set_extra_widget(GTK_FILE_CHOOSER(dialog), toggle);
  if (gtk_dialog_run(GTK_DIALOG(dialog)) == GTK_RESPONSE_ACCEPT) {
	pf = gtk_file_chooser_get_filename(GTK_FILE_CHOOSER(dialog));
	ro_flg = gtk_toggle_button_get_active(GTK_TOGGLE_BUTTON(toggle));
	if (pf != NULL)
		strncat(buff, ioabsolute(pf), sizeof(buff) - 9);
	g_free(pf);
  }
  gtk_widget_destroy(dialog);

  /* Ask Elvis to load the new filename */
  if (buff[len] != '\0') {

	/* Update the arglist */
	msgflush();
	eventex(gg->edhead, buff, ElvFalse);

	/* This will _usually_ find the buffer.  There seems no
	** definitive way to get a buffer from a filename ???
	*/
	if (ro_flg)
		o_readonly(buffind(buff + len)) = 1;

	/* Force a screen update */
	(void) eventdraw(gg->edhead);
  }

  return;
}


/*
** s a v e _ a c t i o n
**
** Save a file, under a new name if necessary.
**
** Returns:	Nothing		Always
*/
static void save_action(GtkAction *action, gpointer data)
{
  GNGUI *gg = (GNGUI *) data;

  DPRINT_F(("save_action()\n"));

  /* Flush all message buffers and call back into Elvis */
  msgflush();
  eventex(gg->edhead, "write", ElvFalse);

  return;
}


/*
** s a v e a s _ a c t i o n
**
** Save a file using the Save File dialog.
**
** Returns:	Nothing		Always
*/
static void saveas_action(GtkAction *action, gpointer data)
{
  GNGUI *gg = (GNGUI *) data;
  GtkWidget *dialog;
  int ut_flg, ex_flg;
  char buff[PATH_MAX];
  char *pf;

  DPRINT_F(("saveas_action()\n"));

  /* Get the current filename */
  strcpy(buff, g_get_home_dir());
  strcat(buff, SEPSTR);
  pf = gg->edhead->edname;

  /* There seems no way to get the un-named buffer flag from Elvis ??? ### */
  ut_flg = (strncmp(pf, "Elvis untitled ", sizeof("Elvis untitled") - 1)
								== SAME);
  if (pf != NULL && *pf == SEPCHR)
	strcpy(buff, pf);
  else if (pf != NULL && !ut_flg)
	strncat(buff, pf, PATH_MAX - 1 - strlen(buff));
  else
	strncat(buff, "Elvis-untitled-file", PATH_MAX - 1 - strlen(buff));

  /* Does the file already exist ? (Check for writeable elsewhere) */
  ex_flg = (!ut_flg && access(buff, R_OK) == 0);

  /* Open the Standard Save File dialog */
  dialog = gtk_file_chooser_dialog_new(
				"Save File",
				NULL,
				GTK_FILE_CHOOSER_ACTION_SAVE,
				GTK_STOCK_CANCEL, GTK_RESPONSE_CANCEL,
				GTK_STOCK_SAVE, GTK_RESPONSE_OK,
				NULL
				);
  gtk_dialog_set_default_response(GTK_DIALOG(dialog), GTK_RESPONSE_OK);

  /* Stick to the local filesystem */
  gtk_file_chooser_set_local_only(GTK_FILE_CHOOSER(dialog), TRUE);
  /* Bring up the standard file overwrite confirmation dialog */
  gtk_file_chooser_set_do_overwrite_confirmation(GTK_FILE_CHOOSER(dialog),
								TRUE);

  /* If we are saving an unnamed buffer */
  if (ut_flg) {
	/* Save to a default file name */
	gtk_file_chooser_set_current_folder(GTK_FILE_CHOOSER(dialog),
							g_get_home_dir());
	gtk_file_chooser_set_current_name(GTK_FILE_CHOOSER(dialog), buff);
  }
  else if (ex_flg) {
	/* Save to an existing file */
	gtk_file_chooser_set_filename(GTK_FILE_CHOOSER(dialog), buff);
  }
  else {
	/* Save to a new file */
	gtk_file_chooser_set_current_name(GTK_FILE_CHOOSER(dialog), buff);
  }

  /* Work out what selections were made in the dialog */
  if (gtk_dialog_run(GTK_DIALOG(dialog)) == GTK_RESPONSE_OK) {
	pf = gtk_file_chooser_get_filename(GTK_FILE_CHOOSER(dialog));
	buff[0] = '\0';
	if (gtk_file_chooser_get_do_overwrite_confirmation(
					GTK_FILE_CHOOSER(dialog))) {
		strcpy(buff, "write! ");
		strncat(buff, pf, PATH_MAX - 1 - 8 - 1);
	}
	else {
		strcpy(buff, "write ");
		strncat(buff, pf, PATH_MAX - 1 - 7 - 1);
	}
	/* Flush all message buffers and call back into Elvis */
	msgflush();
	eventex(gg->edhead, buff, ElvFalse);
	g_free(pf);
  }
  gtk_widget_destroy(dialog);

  return;
}


/*
** c l o s e _ a c t i o n
**
** Close the current file.
**
** Returns:	Nothing		Always
**
** To work properly, this would involve rewriting the arglist. ###
*/
static void close_action(GtkAction *action, gpointer data)
{
  GNGUI *gg = (GNGUI *) data;

  DPRINT_F(("close_action()\n"));

  /* Flush all message buffers and call back into Elvis.
  ** Should this command be "close" or "quit" ???
  */
  msgflush();
  eventex(gg->edhead, "quit", ElvFalse);

  return;
}


/*
** q u i t _ a c t i o n
**
** Quit the Gnome GUI.
**
** Returns:	Nothing		Always
**
** How much cleaning up should this do ???
*/
static void quit_action(GtkAction *action, gpointer data)
{
  GNGUI *gg = (GNGUI *) data;

  DPRINT_F(("quit_action()\n"));

  /* Flush all message buffers and call back into Elvis */
  msgflush();

  /* :qall! doesn't work; the changes to arglist have upset Elvis'
  ** windowing or buffer code, so that when there are multiple
  ** windows, destroygw recurses.
  ** Workaround is to close each window in turn. ###
  eventex(gg->edhead, "quall!", ElvFalse);
  */
  while (gg->edhead != NULL)
	eventex(gg->edhead, "q!", ElvFalse);

  return;
}

/* Edit Menu actions */

/*
** y a n k _ a l l _ a c t i o n
**
** Yank all the lines in a file.
**
** Returns:	Nothing		Always
*/
static void yank_all_action(GtkAction *action, gpointer data)
{
  GNGUI *gg = (GNGUI *) data;

  DPRINT_F(("yank_all_action()\n"));

  msgflush();
  eventex(gg->edhead, "1,$y\"^", ElvFalse);

  return;
}

/* Search Menu actions */

/*
** m o v e _ a c t i o n
**
** Move to a specified line.
**
** Returns:	Nothing		Always
*/
static void move_action(GtkAction *action, gpointer data)
{
  GNGUI *gg = (GNGUI *) data;
  const gchar *name;
  size_t t;
  int cmd, prev_cmd;
  struct Action {
	int cmd;
	char *name;
  } actions[] = {
	{ FIND_NEXT, "Find_next_Action" },
	{ FIND_PREV, "Find_prev_Action" },
	{ FIND_AGAIN, "Find_again_Action" },
	{ REPLACE, "Replace_Action" },
	{ REPLACE_AGAIN, "Replace_again_Action" },
	{ GOTO, "Goto_Action" },
	{ GOTO_AGAIN, "Goto_again_Action" },
	{ FILE_NAME, "New_Action" }
	/* { MOVE_ERR, NULL } */
  };

  /* Get the (unique) action name */
  name = gtk_action_get_name(action);

  DPRINT_F(("move_action()\n"));

  cmd = MOVE_ERR;
  for (t = 0 ; t < sizeof(actions) / sizeof(actions[0]) ; t++) {
	if (strcmp(name, actions[t].name) == SAME) {
		cmd = actions[t].cmd;
		break;
	}
  }

  /* Must be updated for use in handler */
  prev_cmd = gg->move_cmd;
  gg->move_cmd = cmd;

  if (cmd == FIND_NEXT || cmd == FIND_PREV || cmd == REPLACE ||
					cmd == GOTO || cmd == FILE_NAME) {
	if (TRYLOCK(entry_lock)) {
		/* Switch the RHS of hpaned for an entry widget */
		gtk_widget_reparent(gg->msgbar, gg->msgbarbox);
		gtk_widget_hide(gg->msgbar);
		gtk_widget_reparent(gg->entry, gg->hpaned);
		gtk_widget_show(gg->entry);
		g_signal_handler_unblock(G_OBJECT(gg->entry), gg->entry_id);

		/* The entry widget has the focus until it completes */
		gtk_widget_grab_focus(gnome_entry_gtk_entry(
						(GnomeEntry *) gg->entry));

		/* Block the signal and release the lock in entry_handler().
		g_signal_handler_block(G_OBJECT(gg->entry), gg->entry_id);
		UNLOCK(entry_lock);
		*/
	}
	DPRINT_F(("move_action(): entry_lock%slocked\n",
				entry_lock == 0 ? " " : " un"));

	/* Update the main menu */
	if (cmd == FIND_NEXT) {
		gg->menu_state |= MENU_NEXT;
		gg->menu_state &= ~MENU_PREV;
	}
	else if (cmd == FIND_PREV) {
		gg->menu_state |= MENU_PREV;
		gg->menu_state &= ~MENU_NEXT;
	}
	else if (cmd == REPLACE)
		gg->menu_state |= MENU_REPL;
	else if (cmd == GOTO)
		gg->menu_state |= MENU_GOTO;
	config_menus(gg);
  }
  else if (cmd == FIND_AGAIN || cmd == REPLACE_AGAIN || cmd == GOTO_AGAIN) {
	/* Calling with GtkEditable NULL implies non-signal use */
	entry_handler(NULL, (gpointer) gg);
  }
  else {
	msg(MSG_ERROR, "Unknown search command");
	msgflush();
  }

  return;
}


/*
** c l e a r _ h i s t o r y _ a c t i o n
**
** Clear the search history.
**
** Returns:	Nothing		Always
*/
static void clear_history_action(GtkAction *action, gpointer data)
{
  GNGUI *gg = (GNGUI *) data;

  gnome_entry_clear_history(GNOME_ENTRY(gg->entry));

  /* Update the main and mouse menus */
  gg->menu_state &= ~MENU_NEXT;
  gg->menu_state &= ~MENU_PREV;
  gg->menu_state &= ~MENU_REPL;
  gg->menu_state &= ~MENU_GOTO;
  config_menus(gg);

  return;
}

/* Tools Menu actions */

/*
** d i c t i o n a r y _ a c t i o n
**
** Show a dictionary entry for the current word.
**
** Returns:	Nothing		Always
*/
static void dictionary_action(GtkAction *action, gpointer data)
{
  size_t t;
  int rc;
  char cmd[2 * F_MAX];

  DPRINT_F(("dictionary_action()\n"));

  sprintf(cmd, "%s ", o_dictionary);
  t = strlen(cmd);

  /* The program DICT_UTIL's presence is checked at startup,
  ** and the menu option greyed out if DICT_UTIL isn't found.
  */
  if ((rc = getwordatcursor(cmd + t, 2 * F_MAX - 1 - (int) t)) == 0) {
	sprintf(cmd, "%s %s", o_dictionary, "has been given no argument");
	msg(MSG_ERROR, cmd);
	msgflush();
  }
  else
	rc = gnome_execute_terminal_shell(g_get_home_dir(), cmd);

  return;
}


/*
** s p e l l c h e c k _ a c t i o n
**
** Spellcheck the current word.
**
** Returns:	Nothing		Always
**
** Elvis has a built-in spellchecker, but the system version may be
** more powerful or configurable, so it is available here.
*/
static void spellcheck_action(GtkAction *action, gpointer data)
{
  GNGUI *gg = &gn_gui;
  size_t t;
  int rc;
  char cmd[2 * F_MAX];

  DPRINT_F(("spellcheck_action()\n"));

  sprintf(cmd, "!%s ", o_spellcheck);
  t = strlen(cmd);

  /* The program SPELL_UTIL's presence is checked at startup,
  ** and the menu option removed if SPELL_UTIL isn't found.
  */
  if ((rc = getwordatcursor(cmd + t, 2 * F_MAX - 1 - (int) t)) == 0) {
	sprintf(cmd, "%s %s", o_spellcheck, "has been given no argument");
	msg(MSG_ERROR, cmd);
	msgflush();
  }
  else {
	msgflush();
	eventex(gg->edhead, cmd, ElvFalse);
	(void) eventdraw(gg->edhead);
  }

  return;
}


/*
** m a n p a g e _ a c t i o n
**
** Show the man page for the current word.
**
** Returns:	Nothing		Always
*/
static void manpage_action(GtkAction *action, gpointer data)
{
  /* GNGUI *gg = (GNGUI *) data; */
  size_t t;
  int rc;
  char cmd[64];

  DPRINT_F(("manpage_action()\n"));

  if (o_manmode) {
	strcpy(cmd, MAN_ONE);
	t = sizeof(MAN_ONE) - 1;
  }
  else {
	strcpy(cmd, MAN_ALL);
	t = sizeof(MAN_ALL) - 1;
  }

  /* Get the word (if any) at the cursor */
  if ((rc = getwordatcursor(cmd + t, 64 - (int) t)) == 0) {
	msg(MSG_ERROR, "man has been given no argument");
	msgflush();
  }
  else
	rc = gnome_execute_terminal_shell(g_get_home_dir(), cmd);

  return;
}


/*
** n o t e _ a c t i o n
**
** A placeholder for the notepad submenu.
**
** Returns:	Nothing		Always
**
** gtk_ui_manager_add_ui_from_string() is pretty limited; this
** is a workaround to add a sub-menu to a menu entry. The entry
** is called from both the parent menu and the child, which is
** why it's not quite sure what its arguments are.  It finds
** out by checking for the existence of a useable data pointer.
*/
static void note_action(GtkAction *action, gpointer data)
{

  DPRINT_F(("note_action()\n"));

  /* The widget may be either a GtkMenuItem or a GtkAction */
#if 1
  if (data != NULL && *(char *) data != '\0')
	notepad_action(NULL, data);
#else
  if (data != NULL && *(char *) data != '\0') {
	notepad_action(NULL, data);
	g_print("note_action(): menu |%s|\n", (char *) data);
  }
  else
	g_print("note_action(): action |%s|\n",
				gtk_action_get_name(action));
#endif

  return;
}


/*
** n o t e p a d _ a c t i o n
**
** Open a notepad file.
**
** Returns:	Nothing		Always
**
** This may be called from the main menu, or the toolbar, or directly.
*/
static void notepad_action(GtkAction *action, gpointer data)
{
  GNGUI *gg = &gn_gui;
  int refresh, flg, status = OK;
  char *pc = NULL;
  char buff[PATH_MAX];

  DPRINT_F(("notepad_action()\n"));

#ifdef DPRINT_PP
  { /* Debugging block begin */
  if (GTK_IS_ACTION(action))
	g_print("notepad_action(): called from |%s|\n",
			gtk_action_get_name(action));
  else if (GTK_IS_TOOL_BUTTON(action))
	g_print("notepad_action(): called from |%s|\n",
			gtk_tool_button_get_label((GtkToolButton *) action));
  else if (GTK_IS_MENU_ITEM(action))
	g_print("notepad_action(): called directly\n");
  else
	g_print("notepad_action(): called from unknown point\n");
  g_print("notepad_action(): called data |%s|\n", (char *) data);
  } /* Debugging block end */
#endif

  /* The Elvis notepad file uses a relative path */
  flg = TRUE;
  if (GTK_IS_TOOL_BUTTON(action) ||	/* just open NOTEPAD_ELVIS */
		strcmp(data, (char *) NOTEPAD_ELVIS) == SAME) {
#ifdef CFGDIR_ELVIS
	if (gg->cfgdir_flg)
		pc = gnome_util_prepend_user_home(
				CFGDIR_ELVIS "/" NOTEPAD_ELVIS);
	else
		pc = gnome_util_prepend_user_home(NOTEPAD_ELVIS);
#else
	pc = gnome_util_prepend_user_home(NOTEPAD_ELVIS);
#endif
  }
  /* The other notepad strings are always treated as absolute paths */
  else if (*(char *)data == '/') {
	pc = (char *) data;
	flg = FALSE;
  }
  /* Something very odd has happened */
  else {
	status = FAILED;
	flg = FALSE;
  }

  /* If all's well, try to open the file */
  if (status == OK && pc != NULL) {
	strcpy(buff, "snext ");
	strncat(buff, pc, sizeof(buff) - 7);
	msgflush();
	eventex(gg->edhead, buff, ElvFalse);
	/* Give GTK+ time to stabilise */
	while (gtk_events_pending()) gtk_main_iteration();
	refresh = o_exrefresh;
	o_exrefresh = 1;
	(void) eventdraw(gg->edhead);
	o_exrefresh = refresh;
  }

  /* Tidy up */
  if (flg) g_free(pc);
  pc = NULL;

  return;
}


/*
** s h e l l _ a c t i o n
**
** Run a command shell.
**
** Returns:	Nothing		Always
**
** Elvis then calls back and runs the GUI's "stop" action.
*/
static void shell_action(GtkAction *action, gpointer data)
{
  GNGUI *gg = (GNGUI *) data;

  DPRINT_F(("shell_action()\n"));

  msgflush();
  eventex(gg->edhead, "sh", ElvFalse);
  (void) eventdraw(gg->edhead);

  return;
}


/* Options Menu actions */

/*
** s h o w _ t o o l s _ a c t i o n
**
** Show or hide the toolbar.
**
** Returns:	Nothing		Always
*/
static void show_tools_action(GtkAction *action, gpointer data)
{
  GNGUI *gg = (GNGUI *) data;
  const gchar *name;

  DPRINT_F(("show_tools_action()\n"));

  name = gtk_action_get_name(action);

  /* Update the GConf database.  GConf's call back into
  ** gconf_notify_handler() then updates the menu entry.
  */
  if (strcmp(name, "Show_toolbar_Action") == SAME)
	gconf_client_set_bool(gg->gconf_client,
				GCONF_PATH "/ui/showtoolbar", TRUE, NULL);
  else if (strcmp(name, "Hide_toolbar_Action") == SAME)
	gconf_client_set_bool(gg->gconf_client,
				GCONF_PATH "/ui/showtoolbar", FALSE, NULL);

  return;
}


/*
** s t a r t p o s _ a c t i o n
**
** Set or unset the startup postion option.
**
** Returns:	Nothing		Always
**
*/
static void startpos_action(GtkAction *action, gpointer data)
{
  GNGUI *gg = (GNGUI *) data;
  const gchar *name;

  DPRINT_F(("startpos_action()\n"));

  name = gtk_action_get_name(action);

  /* Update the GConf database.  GConf's call back into
  ** gconf_notify_handler() then updates the menu entry.
  */
  if (strcmp(name, "Fix_startpos_Action") == SAME)
	gconf_client_set_bool(gg->gconf_client,
				GCONF_PATH "/ui/startposition", TRUE, NULL);
  else if (strcmp(name, "Free_startpos_Action") == SAME)
	gconf_client_set_bool(gg->gconf_client,
				GCONF_PATH "/ui/startposition", FALSE, NULL);

  return;
}


/*
** d e f a u l t _ s i z e _ a c t i o n
**
** Restore the default terminal size.
**
** Returns:	Nothing		Always
*/
static void default_size_action(GtkAction *action, gpointer data)
{
  GNGUI *gg = (GNGUI *) data;

  /* Nothing to do and no-one to do it with */
  if (gg->edhead == NULL || gg->edhead->vterm == NULL) return;

  /* This function doesn't work if the window is fullscreen */
  if (gg->window_state & GDK_WINDOW_STATE_MAXIMIZED) {
	gtk_window_unmaximize(GTK_WINDOW(gg->gwindow));
	/* Give GTK+ time to stabilise */
	while (gtk_events_pending()) gtk_main_iteration();
	gg->edhead->rows = vte_terminal_get_row_count(
					VTE_TERMINAL(gg->edhead->vterm));
	gg->edhead->cols = vte_terminal_get_column_count(
					VTE_TERMINAL(gg->edhead->vterm));
	/* Adjustment for window scrollbar */
	gg->edhead->cols -= gg->bar_cols;
	gg->main_height = 0;
	gg->main_width = 0;
  }
  else {
	gg->edhead->cols = gg->def_cols;
	gg->edhead->rows = gg->def_rows;
	gg->main_height = 0;
	gg->main_width = 0;
	size_main_window(gg, TRUE);
  }

  return;
}


/*
** d o c s _ h o w _ a c t i o n
**
** Set the window opening mode for documents.
**
** Returns:	Nothing		Always
*/
static void docs_how_action(GtkAction *action, gpointer data)
{
  GNGUI *gg = (GNGUI *) data;
  const gchar *name;

  DPRINT_F(("docs_how_action()\n"));

  name = gtk_action_get_name(action);

  /* Update the GConf database.  GConf's call back into
  ** gconf_notify_handler() then updates the menu entry.
  */
  if (strcmp(name, "Open_cur_Action") == SAME)
	gconf_client_set_bool(gg->gconf_client,
			GCONF_PATH "/ui/newmode", TRUE, NULL);
  else if (strcmp(name, "Open_new_Action") == SAME)
	gconf_client_set_bool(gg->gconf_client,
			GCONF_PATH "/ui/newmode", FALSE, NULL);

  return;
  }


/*
** c u r s o r _ a c t i o n
**
** Configure the cursor for the current window.
**
** Returns:	Nothing		Always
*/
static void cursor_action(GtkWidget *widget, gpointer data)
{
  GNGUI *gg = (GNGUI *) data;
  struct Select sels[] = {
	{ NULL, NULL, NULL, NULL, { 0,0,0,0 }, 0, "Color is" },
	{ NULL, NULL, NULL, NULL, { 0,0,0,0 }, 0, "Blink is" },
	{ NULL, NULL, NULL, NULL, { 0,0,0,0 }, 0, "Defaults" }
  };
  GtkWidget *dialog;
  size_t t;
  int def_button, status;

  DPRINT_F(("cursor_action()\n"));

  /* Nothing to do and no-one to do it with */
  if (gg->edhead == NULL) return;

  /* Start with the edit window's current cursor, not the main windows' */
  sels[0].initial = &gg->edhead->cursorcolor;
  sels[1].setting = gg->edhead->cursorblink;

  /* Create the dialog */
  dialog = gtk_dialog_new_with_buttons(
		"Cursor setup",			/* Title */
		GTK_WINDOW(gg->gwindow),	/* Parent window */
		GTK_DIALOG_DESTROY_WITH_PARENT,	/* Destroy with main window */
		GTK_STOCK_CANCEL,		/* Stock ID for button one */
		GTK_RESPONSE_CANCEL,		/* Response for button one */
		GTK_STOCK_OK,			/* Stock ID for button two */
		GTK_RESPONSE_OK,		/* Response for button two */
		NULL				/* Terminator */
		);

  /* Reposition the dialogs OK and Cancel buttons */
  gtk_button_box_set_layout(GTK_BUTTON_BOX(GTK_DIALOG(dialog)->action_area),
							GTK_BUTTONBOX_SPREAD);

  /* We need pairs of buttons and labels stacked vertically in the dialog */
  for (t = 0 ; t < (sizeof(sels) / sizeof(sels[0]) - 1) ; t++) {
	if (t == 0) {
		sels[t].button = gtk_color_button_new_with_color(
						&gg->edhead->cursorcolor);
		gtk_color_button_set_title(GTK_COLOR_BUTTON(sels[t].button),
						sels[t].name);
	}
	else if (t == 1) {
		sels[t].button = gtk_toggle_button_new_with_label(
				gg->edhead->cursorblink ? "On" : "Off");
		gtk_toggle_button_set_active(GTK_TOGGLE_BUTTON(sels[t].button),
				gg->edhead->cursorblink);
	}
	sels[t].label = gtk_label_new(sels[t].name);
	sels[t].box = gtk_hbox_new(TRUE, 2);
	gtk_container_add(GTK_CONTAINER(sels[t].box), sels[t].label);
	gtk_container_add(GTK_CONTAINER(sels[t].box), sels[t].button);
	gtk_widget_show(sels[t].button);
	gtk_widget_show(sels[t].label);
	gtk_widget_show(sels[t].box);
	gtk_container_add(GTK_CONTAINER(GTK_DIALOG(dialog)->vbox), sels[t].box);
  }

  /* Add the default button */
  def_button = t;
  sels[t].button = gtk_toggle_button_new_with_label(
		gg->edhead->use_def_cursor ? "Yes" : "No");
  gtk_toggle_button_set_active(
		GTK_TOGGLE_BUTTON(sels[t].button), gg->edhead->use_def_cursor);
  sels[t].label = gtk_label_new(sels[t].name);
  sels[t].box = gtk_hbox_new(TRUE, 2);
  gtk_container_add(GTK_CONTAINER(sels[t].box), sels[t].label);
  gtk_container_add(GTK_CONTAINER(sels[t].box), sels[t].button);
  gtk_widget_show(sels[t].label);
  gtk_widget_show(sels[t].button);
  gtk_widget_show(sels[t].box);
  gtk_container_add(GTK_CONTAINER(GTK_DIALOG(dialog)->vbox), sels[t].box);

  /* Connect the default buttons state to its labels */
  g_signal_connect(
		G_OBJECT(sels[t].button),	/* Widget */
		"toggled",			/* Signal */
		G_CALLBACK(default_cursor_handler),/* Callback */
		(gpointer) sels			/* User data */
		);

  /* Connect the color button to its handler */
  g_signal_connect(
		G_OBJECT(sels[0].button),
		"color-set",
		G_CALLBACK(color_handler),
		(gpointer) sels[def_button].button /* Default button */
		);

  /* Connect the blink button to its handler */
  g_signal_connect(
		sels[1].button,
		"toggled",
		G_CALLBACK(cursor_handler),
		(gpointer) sels[def_button].button /* Default button */
		);

  /* gtk_dialog_run() gives what is effectively a modal dialog,
  ** but signals and timers still work.
  */
  if (gtk_dialog_run(GTK_DIALOG(dialog)) == GTK_RESPONSE_OK) {
	/* If necessary, reset current cursor settings to global defaults */
	gg->edhead->use_def_cursor = gtk_toggle_button_get_active(
				GTK_TOGGLE_BUTTON(sels[def_button].button));
	if (gg->edhead->use_def_cursor) {
		gtk_color_button_set_color(GTK_COLOR_BUTTON(sels[0].button),
						&gg->def_cursorcolor);
		gtk_toggle_button_set_active(GTK_TOGGLE_BUTTON(sels[1].button),
						gg->def_cursorblink);
	}
	/* The cursor is set on a per-terminal basis */
	gtk_color_button_get_color(GTK_COLOR_BUTTON(sels[0].button),
							&sels[0].color);
	sels[1].setting = gtk_toggle_button_get_active(
					GTK_TOGGLE_BUTTON(sels[1].button));

	/* Update edit window and main window current cursor settings */
	gg->edhead->cursorcolor = gg->cur_cursorcolor = sels[0].color;
	if (gg->edhead->use_def_cursor) {
		strcpy(gg->cur_cursorname, gg->def_cursorname);
	}
	else {
		sprintf(gg->cur_cursorname, "#%04x%04x%04x",
			sels[0].color.red, sels[0].color.green,
			sels[0].color.blue);
	}
	/* Update the :set options */
	optisstring(&optdescs[O_CCOLOR], &optvals[O_CCOLOR],
						gg->cur_cursorname);
	gg->edhead->cursorblink = gg->cur_cursorblink = sels[1].setting;
	o_cursorblink = sels[1].setting;

	/* Update color and blink here */
	vte_terminal_set_color_cursor(VTE_TERMINAL(gg->edhead->vterm),
						&gg->edhead->cursorcolor);
	vte_terminal_set_cursor_blinks(VTE_TERMINAL(gg->edhead->vterm),
						gg->edhead->cursorblink);
	status = OK;
  }
  else
	status = FAILED;

  gtk_widget_destroy(dialog);

  return;
}


/*
** c o l o r _ a c t i o n
**
** Select colors for the current window.
**
** Returns:	Nothing		Always
*/
static void color_action(GtkWidget *widget, gpointer data)
{
  GNGUI *gg = (GNGUI *) data;
  struct Select sels[] = {
	{ NULL, NULL, NULL, NULL, { 0,0,0,0 }, 0, "Background" },
	{ NULL, NULL, NULL, NULL, { 0,0,0,0 }, 0, "Foreground" },
	{ NULL, NULL, NULL, NULL, { 0,0,0,0 }, 0, "Defaults" }
  };
  GtkWidget *dialog;
  size_t t;
  int def_button, status;
  char buff[16];

  DPRINT_F(("color_action()\n"));

  /* Nothing to do and no-one to do it with */
  if (gg->edhead == NULL) return;

  /* Start with the edit window's current colors, not the main windows' */
  sels[FORE].initial = &gg->edhead->forecolor;
  sels[BACK].initial = &gg->edhead->backcolor;

  /* Create the dialog */
  dialog = gtk_dialog_new_with_buttons(
		"Color selection",		/* Title */
		GTK_WINDOW(gg->gwindow),	/* Parent window */
		GTK_DIALOG_DESTROY_WITH_PARENT,	/* Destroy with main window */
		GTK_STOCK_CANCEL,		/* Stock ID for button one */
		GTK_RESPONSE_CANCEL,		/* Response for button one */
		GTK_STOCK_OK,			/* Stock ID for button two */
		GTK_RESPONSE_OK,		/* Response for button two */
		NULL				/* Terminator */
		);

  /* Reposition the dialogs OK and Cancel buttons */
  gtk_button_box_set_layout(GTK_BUTTON_BOX(GTK_DIALOG(dialog)->action_area),
							GTK_BUTTONBOX_SPREAD);

  /* We need pairs of buttons and labels stacked vertically in the dialog */
  for (t = 0 ; t < (sizeof(sels) / sizeof(sels[0]) - 1) ; t++) {
	sels[t].button = gtk_color_button_new_with_color(sels[t].initial);
	/* Set the title of the color selection button */
	gtk_color_button_set_title(GTK_COLOR_BUTTON(sels[t].button),
							sels[t].name);
	sels[t].label = gtk_label_new(sels[t].name);
	sels[t].box = gtk_hbox_new(TRUE, 2);
	gtk_container_add(GTK_CONTAINER(sels[t].box), sels[t].label);
	gtk_container_add(GTK_CONTAINER(sels[t].box), sels[t].button);
	gtk_widget_show(sels[t].label);
	gtk_widget_show(sels[t].button);
	gtk_widget_show(sels[t].box);
	gtk_container_add(GTK_CONTAINER(GTK_DIALOG(dialog)->vbox),
				sels[t].box);
  }

  /* Add the default button */
  def_button = t;
  sels[t].button = gtk_toggle_button_new_with_label(
				gg->edhead->use_def_color ? "Yes" : "No");
  gtk_toggle_button_set_active(GTK_TOGGLE_BUTTON(sels[t].button),
				gg->edhead->use_def_color);
  sels[t].label = gtk_label_new(sels[t].name);
  sels[t].box = gtk_hbox_new(TRUE, 2);
  gtk_container_add(GTK_CONTAINER(sels[t].box), sels[t].label);
  gtk_container_add(GTK_CONTAINER(sels[t].box), sels[t].button);
  gtk_widget_show(sels[t].label);
  gtk_widget_show(sels[t].button);
  gtk_widget_show(sels[t].box);
  gtk_container_add(GTK_CONTAINER(GTK_DIALOG(dialog)->vbox), sels[t].box);

  /* Connect the default buttons state to its labels */
  g_signal_connect(
		G_OBJECT(sels[t].button),	/* Widget */
		"toggled",			/* Signal */
		G_CALLBACK(default_color_handler),/* Callback */
		(gpointer) sels			/* User data */
		);

  /* Connect the color buttons to the default toggle button label */
  for (t = 0 ; t < (sizeof(sels) / sizeof(sels[0]) - 1) ; t++) {
	g_signal_connect(
			G_OBJECT(sels[t].button),
			"color-set",
			G_CALLBACK(color_handler),
			(gpointer) sels[def_button].button
			);
  }

  /* gtk_dialog_run() gives what is effectively a modal dialog,
  ** but signals and timers still work.
  */
  if (gtk_dialog_run(GTK_DIALOG(dialog)) == GTK_RESPONSE_OK) {
	/* If necessary, reset current terminal colors to global defaults */
	gg->edhead->use_def_color = gtk_toggle_button_get_active(
				GTK_TOGGLE_BUTTON(sels[def_button].button));
	if (gg->edhead->use_def_color) {
		gtk_color_button_set_color(GTK_COLOR_BUTTON(sels[BACK].button),
						&gg->def_backcolor);
		gtk_color_button_set_color(GTK_COLOR_BUTTON(sels[FORE].button),
						&gg->def_forecolor);
	}

	/* Colors are set on a per-terminal basis */
	gtk_color_button_get_color(GTK_COLOR_BUTTON(sels[BACK].button),
							&sels[BACK].color);
	gtk_color_button_get_color(GTK_COLOR_BUTTON(sels[FORE].button),
							&sels[FORE].color);

	/* Update edit window and main window current colors */
	gg->edhead->backcolor = gg->cur_backcolor = sels[BACK].color;
	gg->edhead->forecolor = gg->cur_forecolor = sels[FORE].color;
	if (gg->edhead->use_def_color) {
		strcpy(gg->cur_backname, gg->def_backname);
		strcpy(gg->cur_forename, gg->def_forename);
	}
	else {
		sprintf(gg->cur_backname, "#%04x%04x%04x",
			sels[BACK].color.red, sels[BACK].color.green,
			sels[BACK].color.blue);
		sprintf(gg->cur_forename, "#%04x%04x%04x",
			sels[FORE].color.red, sels[FORE].color.green,
			sels[FORE].color.blue);
	}
	/* Update the :set options */
	optisstring(&optdescs[O_FCOLOR], &optvals[O_FCOLOR], gg->cur_forename);
	optisstring(&optdescs[O_BCOLOR], &optvals[O_BCOLOR], gg->cur_backname);

	vte_terminal_set_color_background(
				VTE_TERMINAL(gg->edhead->vterm),
				&gg->edhead->backcolor
				);
	vte_terminal_set_color_foreground(
				VTE_TERMINAL(gg->edhead->vterm),
				&gg->edhead->forecolor
				);

	/* Update the screen - a full reset seems the only way ??? */
	msgflush();
	strcpy(buff, "\033c");
	write_buff(gg->edhead, buff, strlen(buff));
	eventexpose((GUIWIN *)gg->edhead, 0, 0,
			gg->edhead->rows - 1, gg->edhead->cols - 1);
#if 0
	/*  Update the full color table ??? */
	vte_terminal_set_colors(
			VTE_TERMINAL(gg->edhead->vterm),
			NULL,
			NULL,
			gg->vte_colors,
			16
			);
#endif
	status = OK;
  }
  else
	status = FAILED;

  gtk_widget_destroy(dialog);

  return;
}


/*
** f o n t _ a c t i o n
**
** Select fonts for the current window.
**
** Returns:	Nothing		Always
**
** VTE is used here as a terminal, and thus can't easily handle alternate
** fonts.  It seems that the only way to use multiple fonts in a window
** is via a full Pango implementation, which wouldn't be easy to integrate
** with the standard Elvis code. ???
*/
static void font_action(GtkWidget *widget, gpointer data)
{
  GNGUI *gg = (GNGUI *) data;
  int def_button;
  const gchar *fontname;
  struct Select sels[] = {
	{ NULL, NULL, NULL, NULL, { 0,0,0,0 }, 0, "Current" },
#ifdef DEBUG_MENU
	{ NULL, NULL, NULL, NULL, { 0,0,0,0 }, 0, "Debug" },
#endif
	{ NULL, NULL, NULL, NULL, { 0,0,0,0 }, 0, "Default" }
  };
  GtkWidget *dialog;
  size_t t;
  int status;

  DPRINT_N(("font_action() old fontname |%s|\n", gg->cur_fontname));

  /* Nothing to do and no-one to do it with */
  if (gg->edhead == NULL) return;

  /* Create the dialog */
  dialog = gtk_dialog_new_with_buttons(
		"Terminal font selection",	/* Title */
		GTK_WINDOW(gg->gwindow),	/* Parent window */
		GTK_DIALOG_DESTROY_WITH_PARENT,	/* Destroy with main window */
		GTK_STOCK_CANCEL,		/* Stock ID for button one */
		GTK_RESPONSE_CANCEL,		/* Response for button one */
		GTK_STOCK_OK,			/* Stock ID for button two */
		GTK_RESPONSE_OK,		/* Response for button two */
		NULL				/* Terminator */
		);

  /* Reposition the dialogs OK and Cancel buttons */
  gtk_button_box_set_layout(GTK_BUTTON_BOX(GTK_DIALOG(dialog)->action_area),
							GTK_BUTTONBOX_SPREAD);

  /* We need pairs of buttons and labels stacked vertically in the dialog */
  for (t = 0 ; t < (sizeof(sels) / sizeof(sels[0]) - 1) ; t++) {
	if (t == 0)
		sels[t].button = gtk_font_button_new_with_font(
						gg->edhead->fontname);
	if (t == 1)
		sels[t].button = gtk_font_button_new_with_font(
						gg->cur_dbg_fontname);
	/* Set the title of the font selection button */
	gtk_font_button_set_title(GTK_FONT_BUTTON(sels[t].button),
							sels[t].name);
	sels[t].label = gtk_label_new(sels[t].name);
	sels[t].box = gtk_hbox_new(TRUE, 2);
	/* There seems to be no way to produce columns of unequal width
	gtk_box_set_homogeneous(GTK_BOX(sels[t].box), FALSE);
	*/
	gtk_container_add(GTK_CONTAINER(sels[t].box), sels[t].label);
	gtk_container_add(GTK_CONTAINER(sels[t].box), sels[t].button);
	gtk_widget_show(sels[t].label);
	gtk_widget_show(sels[t].button);
	gtk_widget_show(sels[t].box);
	gtk_container_add(GTK_CONTAINER(GTK_DIALOG(dialog)->vbox), sels[t].box);
  }

  /* Add the default button */
  def_button = t;
  sels[t].button = gtk_toggle_button_new_with_label(
				gg->edhead->use_def_font ? "Yes" : "No");
  gtk_toggle_button_set_active(GTK_TOGGLE_BUTTON(sels[t].button),
				gg->edhead->use_def_font);
  sels[t].label = gtk_label_new(sels[t].name);
  sels[t].box = gtk_hbox_new(TRUE, 2);
  /* See above
  gtk_box_set_homogeneous(GTK_BOX(sels[t].box), FALSE);
  */
  gtk_container_add(GTK_CONTAINER(sels[t].box), sels[t].label);
  gtk_container_add(GTK_CONTAINER(sels[t].box), sels[t].button);
  gtk_widget_show(sels[t].label);
  gtk_widget_show(sels[t].button);
  gtk_widget_show(sels[t].box);
  gtk_container_add(GTK_CONTAINER(GTK_DIALOG(dialog)->vbox), sels[t].box);

  /* Connect the default buttons state to its labels */
  g_signal_connect(
		G_OBJECT(sels[t].button),	/* Widget */
		"toggled",			/* Signal */
		G_CALLBACK(default_font_handler),/* Callback */
		(gpointer) sels			/* User data */
		);

  /* Connect the font buttons to the default toggle button label */
  for (t = 0 ; t < (sizeof(sels) / sizeof(sels[0]) - 1) ; t++) {
	g_signal_connect(
			G_OBJECT(sels[t].button),
			"font-set",
			G_CALLBACK(terminal_font_handler),
			(gpointer) sels[def_button].button
			);
  }

  /* gtk_dialog_run() gives what is effectively a modal dialog,
  ** but signals and timers still work.  There seems no way to
  ** filter the fonts presented by the dialog so that only
  ** Monospace fonts are shown ???
  */
  if (gtk_dialog_run(GTK_DIALOG(dialog)) == GTK_RESPONSE_OK) {
	/* If necessary, reset current terminal fonts to global defaults */
	gg->edhead->use_def_font = gtk_toggle_button_get_active(
				GTK_TOGGLE_BUTTON(sels[def_button].button));
	if (gg->edhead->use_def_font) {
		/* This won't update the GUI unless the font is monospace */
		(void) set_current_font(gg, gg->def_fontname, TRUE);
#ifdef DEBUG_MENU
		strncpy(gg->cur_dbg_fontname, gg->def_dbg_fontname, F_MAX);
#endif
	}
	else {
		fontname = gtk_font_button_get_font_name(
					GTK_FONT_BUTTON(sels[0].button));
		set_current_font(gg, (char *) fontname, TRUE);
#ifdef DEBUG_MENU
		strncpy(gg->cur_dbg_fontname, gtk_font_button_get_font_name(
				GTK_FONT_BUTTON(sels[1].button)), F_MAX);
#endif
	}
#ifdef DEBUG_MENU
	/* Update the :set option */
	optisstring(&optdescs[O_DFONT], &optvals[O_DFONT],
						gg->cur_dbg_fontname);
#endif
	DPRINT_N(("font_action(): new font name |%s|\n",
						gg->edhead->fontname));
#ifdef DPRINT_NN
	g_print("font_action(): terminal rows %ld columns %ld\n",
		vte_terminal_get_row_count(VTE_TERMINAL(gg->edhead->vterm)),
		vte_terminal_get_column_count(VTE_TERMINAL(gg->edhead->vterm))
		);
	g_print("font_action(): setting terminal font and size\n");
#endif
	vte_terminal_set_font_from_string(VTE_TERMINAL(gg->edhead->vterm),
						gg->edhead->fontname);
	/* Refresh the main window size */
	gg->main_height = 0;
	gg->main_width = 0;
	update_geometry(gg);
	size_main_window(gg, TRUE);

#ifdef DPRINT_NN
	g_print("font_action(): final terminal rows %ld columns %ld\n",
		vte_terminal_get_row_count(VTE_TERMINAL(gg->edhead->vterm)),
		vte_terminal_get_column_count(VTE_TERMINAL(gg->edhead->vterm))
		);
#endif
	status = OK;
  }
  else
	status = FAILED;

  gtk_widget_destroy(dialog);

  return;
}

#ifdef SAVE_CONFIG
/*
** c o n f i g _ a c t i o n
**
** Save and load file list (ring) configurations.
**
** Returns:	Nothing		Always
**
** This is still a placeholder.
*/
static void config_action(GtkWidget *widget, gpointer data)
{
  /* GNGUI *gg = (GNGUI *) data; */

  DPRINT_F(("config_action()\n"));

  return;
}
#endif

  /* Documents Menu actions */

/*
** d o c s _ c l e a r _ a c t i o n
**
** Clear the Documents list and menu.
**
** Returns:	Nothing		Always
*/
static void docs_clear_action(GtkAction *action, gpointer data)
{
  GNGUI *gg = (GNGUI *) data;
  struct Enode *pn;

  DPRINT_F(("docs_clear_action()\n"));

  if (gg->docshead == NULL) return;

  for (pn = gg->docshead ; pn != NULL ; pn = pn->next) {
	rmv_docsmenu(pn);
	if (pn->next == gg->docshead) break;
  }
  del_list(&gg->docshead);
  gg->docscount = 0;

  return;
}

/* Print Menu actions */

/*
** g e t _ p r i n t _ f i l e n a m e
**
** Get the name of a file to be printed.
**
** Returns:	OK		Success
**		FAILED		Otherwise
*/
static int get_print_filename(struct Printjob *print_job, int exists_flg)
{
  GNGUI *gg = &gn_gui;
  struct Select sels[] = {
	{ NULL, NULL, NULL, NULL, { 0,0,0,0 }, 0, "Filename" },
	{ NULL, NULL, NULL, NULL, { 0,0,0,0 }, 0, "Default" }
  };
  size_t t;
  int def_button, status;
  gchar *pg;

  DPRINT_F(("get_print_filename()\n"));

  /* See if the last file printed can still be used as a fallback */
  if (access(gg->lastprint, R_OK) != 0)
	gg->lastprint[0] = '\0';

  /* Create the dialog */
  gg->print_dialog = gtk_dialog_new_with_buttons(
		"Print file name",		/* Title */
		GTK_WINDOW(gg->gwindow),	/* Parent window */
		GTK_DIALOG_DESTROY_WITH_PARENT,	/* Destroy with main window */
		GTK_STOCK_CANCEL,		/* Stock ID for button one */
		GTK_RESPONSE_CANCEL,		/* Response for button one */
		GTK_STOCK_OK,			/* Stock ID for button two */
		GTK_RESPONSE_OK,		/* Response for button two */
		NULL				/* Terminator */
		);

  /* Reposition the dialogs OK and Cancel buttons */
  gtk_button_box_set_layout(GTK_BUTTON_BOX(GTK_DIALOG(
			gg->print_dialog)->action_area), GTK_BUTTONBOX_SPREAD);

  /* The OK button only works when we have a valid file */
  gtk_dialog_set_response_sensitive(GTK_DIALOG(gg->print_dialog),
						GTK_RESPONSE_OK, exists_flg);

  /* We need pairs of buttons and labels stacked vertically in the dialog */
  for (t = 0 ; t < (sizeof(sels) / sizeof(sels[0]) - 1) ; t++) {
	if (t == 0) {
		sels[t].button = gtk_file_chooser_button_new(
			"File selection", GTK_FILE_CHOOSER_ACTION_OPEN);
		if (exists_flg) {
			gtk_file_chooser_set_filename(
					GTK_FILE_CHOOSER(sels[t].button),
					gg->edhead->edname);
		}
		else if (gg->lastprint[0] != '\0') {
			gtk_file_chooser_set_filename(
					GTK_FILE_CHOOSER(sels[t].button),
					gg->lastprint);
		}
		else {
			gtk_file_chooser_set_current_folder(
					GTK_FILE_CHOOSER(sels[t].button),
					g_get_home_dir());
		}
	}
	sels[t].label = gtk_label_new(sels[t].name);
	sels[t].box = gtk_hbox_new(TRUE, 2);
	sels[t].setting = exists_flg;
	/* There seems to be no way to produce columns of unequal width
	gtk_box_set_homogeneous(GTK_BOX(sels[t].box), FALSE);
	*/
	gtk_container_add(GTK_CONTAINER(sels[t].box), sels[t].label);
	gtk_container_add(GTK_CONTAINER(sels[t].box), sels[t].button);
	gtk_widget_show(sels[t].label);
	gtk_widget_show(sels[t].button);
	gtk_widget_show(sels[t].box);
	gtk_container_add(GTK_CONTAINER(GTK_DIALOG(gg->print_dialog)->vbox),
							sels[t].box);
  }

  /* Add the default button */
  def_button = t;
  sels[t].button = gtk_toggle_button_new_with_label(exists_flg ? "Yes" : "No");
  gtk_toggle_button_set_active(GTK_TOGGLE_BUTTON(sels[t].button), exists_flg);
  sels[t].label = gtk_label_new(sels[t].name);
  sels[t].box = gtk_hbox_new(TRUE, 2);
  /* See above
  gtk_box_set_homogeneous(GTK_BOX(sels[t].box), FALSE);
  */
  gtk_container_add(GTK_CONTAINER(sels[t].box), sels[t].label);
  gtk_container_add(GTK_CONTAINER(sels[t].box), sels[t].button);
  gtk_widget_show(sels[t].label);
  gtk_widget_show(sels[t].button);
  gtk_widget_show(sels[t].box);
  gtk_container_add(GTK_CONTAINER(GTK_DIALOG(gg->print_dialog)->vbox),
							sels[t].box);

  /* Connect the default button's state to its labels */
  g_signal_connect(
		G_OBJECT(sels[t].button),	/* Widget */
		"toggled",			/* Signal */
		G_CALLBACK(default_print_name_handler),/* Callback */
		(gpointer) sels			/* User data */
		);

  /* Connect the file selection button to its handler */
  g_signal_connect(
		G_OBJECT(sels[0].button),	/* Widget */
		"selection-changed",		/* Signal */
		G_CALLBACK(print_name_handler), /* Callback */
		(gpointer) sels			/* User data */
		);

  /* gtk_dialog_run() gives what is effectively a modal dialog,
  ** but signals and timers still work.
  */
  status = FAILED;
  if (gtk_dialog_run(GTK_DIALOG(gg->print_dialog)) == GTK_RESPONSE_OK) {
	pg = gtk_file_chooser_get_filename((GtkFileChooser *) sels[0].button);
	if (pg != NULL) {
		print_job->filename[0] = '\0';
		strncat(print_job->filename, pg,
					sizeof(print_job->filename) - 1);
		g_free(pg);
		pg = NULL;
		status = OK;
	}
  }

  gtk_widget_destroy(gg->print_dialog);
  gg->print_dialog = NULL;

  return status;
}


/*
** p r i n t _ a c t i o n
**
** Print all or part of a file.
**
** Returns:	Nothing		Always
*/
static void print_action(GtkAction *action, gpointer data)
{
  GNGUI *gg = &gn_gui;
  struct Printjob *print_job = NULL;
  char username[32];
  static char *months[] = { "Jan", "Feb", "Mar", "Apr", "May", "Jun",
				"Jul", "Aug", "Sep", "Oct", "Nov", "Dec" };
  time_t cal;
  struct tm now;
  struct passwd *pw;
  GtkPrintOperationResult rc;
  GError *error;
  int exists_flg;
  char *pb, *pc;

  DPRINT_F(("print_action()\n"));

  /* Print operations can't overlap */
  if (gg->print_operation != NULL) {
	DPRINT_Z(("print_action(): print operation in progress\n"));
	return;
  }

  /* Initialise the Printjob structure */
  if ((print_job = g_try_malloc(sizeof(struct Printjob))) == NULL) {
	msg(MSG_ERROR, "Cannot allocate memory for printing");
	msgflush();
	return;
  }
  print_job->pango_head_desc = NULL;
  print_job->pango_body_desc = NULL;
  print_job->tmpname = NULL;
  print_job->buff = NULL;

  /* Is a printable file already available ? */
  if (access(gg->edhead->edname, R_OK) == 0)
	exists_flg = TRUE;
  else
	exists_flg = FALSE;

  /* Get a filename */
  if (get_print_filename(print_job, exists_flg) != OK) {
	g_free(print_job);
	return;
  }

  /* Sort out the current time */
  cal = time(NULL);
  now = *localtime(&cal);

  /* Sort out the current user name */
  strcpy(username, "Unknown user");
  if ((pw = getpwuid(getuid())) != NULL) {
	if (pw->pw_gecos != NULL && strlen(pw->pw_gecos) != 0)
		strncpy(username, pw->pw_gecos, sizeof(username) - 1);
	else if (pw->pw_name != NULL && strlen(pw->pw_name) != 0)
		strncpy(username, pw->pw_name, sizeof(username) - 1);
  }

  /* Set up the current print job */
  if ((print_job->fpin = fopen(print_job->filename, "r")) == NULL) {
	msg(MSG_ERROR, "[s]Cannot open file $1 for printing",
						print_job->filename);
	msgflush();
	g_free(print_job);
	return;
  }

  /* Store the current filename in case it's needed again */
  gg->lastprint[0] = '\0';
  strncat(gg->lastprint, print_job->filename, sizeof(gg->lastprint) - 1);

  /* Derive the actual filename from the path string */
  for (pb = pc = print_job->filename ; *pc != '\0' ; pc++)
	if (*pc == '/') pb = pc;
  if (*pb == '/') pb++;
  print_job->upper_header[0] = '\0';
  strncat(print_job->upper_header, pb, sizeof(print_job->upper_header) - 1);
  sprintf(print_job->lower_header, "Printed at %02d:%02d on %d %s %d by %s.",
			now.tm_hour, now.tm_min,
			now.tm_mday, months[now.tm_mon],
			now.tm_year + 1900,
			username);

  /* Initialise the print operation */
  gg->print_operation = gtk_print_operation_new();
  if (gg->print_operation == NULL) {
	DPRINT_Z(("print_action(): printing not available.\n"));
	g_free(print_job);
	return;
  }

  /* Why do most of the pango functions deal with pixels ??? */
  gtk_print_operation_set_unit(gg->print_operation, GTK_UNIT_PIXEL);

  /* How can we confirm that this is set ??? */
  gtk_print_operation_set_allow_async(gg->print_operation, TRUE);

  /* Restore the print settings */
  if (gg->print_settings != NULL) {
	gtk_print_operation_set_print_settings(
		gg->print_operation, gg->print_settings);
  }

  /* Connect the page formatting and printing signals */
  g_signal_connect(
	G_OBJECT(gg->print_operation),
	"begin-print",
	G_CALLBACK(begin_print_handler),
	print_job
	);

  g_signal_connect(
	G_OBJECT(gg->print_operation),
	"draw-page",
	G_CALLBACK(draw_page_handler),
	print_job
	);

  g_signal_connect(
	G_OBJECT(gg->print_operation),
	"end-print",
	G_CALLBACK(end_print_handler),
	print_job
	);

  /* Needed for asynchronous operation, used anyway for wrapup */
  g_signal_connect(
	G_OBJECT(gg->print_operation),
	"done",
	G_CALLBACK(done_print_handler),
	print_job
	);

  /* rc = gtk_print_operation_run(
  **                    GtkPrintOperation op    A GtkPrintOperation.
  **                    GtkPrintOperationAction action  The action to start.
  **                    GtkWindow *parent       The dialogs parent, or NULL.
  **                    GError **error          The error return, or NULL.
  **                    );
  **
  ** Returns :  The result of the print operation.
  **
  ** The action parameter to gtk_print_operation_run() determines what action
  ** the print operation should perform.
  **
  ** GTK_PRINT_OPERATION_ACTION_PRINT_DIALOG Show the print dialog.
  ** GTK_PRINT_OPERATION_ACTION_PRINT   Start to print without showing the
  **                                    print dialog, based on the current
  **                                    print settings.
  ** GTK_PRINT_OPERATION_ACTION_PREVIEW Show the print preview.
  ** GTK_PRINT_OPERATION_ACTION_EXPORT  Export to a file. This requires
  **                                    the export-filename property
  **                                    to be set.
  **
  ** A return value of GTK_PRINT_OPERATION_RESULT_APPLY means that the
  ** printing was completed successfully. In this case, it is a good idea to
  ** obtain the current print settings with
  ** gtk_print_operation_get_print_settings()
  ** and store them for reuse with the next print operation.  This is handled
  ** by the "done" signal handler.
  */
  error = NULL;
  rc = gtk_print_operation_run(
			gg->print_operation,
			GTK_PRINT_OPERATION_ACTION_PRINT_DIALOG,
			GTK_WINDOW(gg->gwindow),
			&error
			);

  return;
}


/*
**  p r i n t _ s e t u p _ a c t i o n
**
** Set the printing fonts and header.
**
** Returns:	Nothing		Always
*/
static void print_setup_action(GtkAction *action, gpointer data)
{
  GNGUI *gg = (GNGUI *) data;
  /* #define P_HEAD_INDEX 0
  ** #define P_BODY_INDEX 1
  ** #define P_FLG_INDEX 2
  */
  struct Select sels[] = {
	{ NULL, NULL, NULL, NULL, { 0,0,0,0 }, 0, "Header font" },
	{ NULL, NULL, NULL, NULL, { 0,0,0,0 }, 0, "Body font" },
	{ NULL, NULL, NULL, NULL, { 0,0,0,0 }, 0, "Add header" },
	{ NULL, NULL, NULL, NULL, { 0,0,0,0 }, 0, "Defaults" }
  };
  GtkWidget *dialog;
  size_t t;
  int def_button, status;

  DPRINT_F(("print_setup_action()\n"));

  /* Create the dialog */
  dialog = gtk_dialog_new_with_buttons(
		"Printing setup",		/* Title */
		GTK_WINDOW(gg->gwindow),	/* Parent window */
		GTK_DIALOG_DESTROY_WITH_PARENT,	/* Destroy with main window */
		GTK_STOCK_CANCEL,		/* Stock ID for button one */
		GTK_RESPONSE_CANCEL,		/* Response for button one */
		GTK_STOCK_OK,			/* Stock ID for button two */
		GTK_RESPONSE_OK,		/* Response for button two */
		NULL				/* Terminator */
		);

  /* Reposition the dialogs OK and Cancel buttons */
  gtk_button_box_set_layout(GTK_BUTTON_BOX(GTK_DIALOG(dialog)->action_area),
							GTK_BUTTONBOX_SPREAD);

  /* We need pairs of buttons and labels stacked vertically in the dialog */
  for (t = 0 ; t < (sizeof(sels) / sizeof(sels[0]) - 1) ; t++) {
	if (t == P_HEAD_INDEX) {	/* Set header font */
		sels[t].button = gtk_font_button_new_with_font(
							gg->cur_hdr_print_font);
		gtk_font_button_set_title(GTK_FONT_BUTTON(sels[t].button),
							sels[t].name);
	}
	else if (t == P_BODY_INDEX) {	/* Set body font */
		sels[t].button = gtk_font_button_new_with_font(
							gg->cur_print_font);
		gtk_font_button_set_title(GTK_FONT_BUTTON(sels[t].button),
							sels[t].name);
	}
	else if (t == P_FLG_INDEX) {	/* Show header flg */
		sels[t].button = gtk_toggle_button_new_with_label(
					gg->cur_hdrprint ? "Yes" : "No");
		gtk_toggle_button_set_active(GTK_TOGGLE_BUTTON(sels[t].button),
					gg->cur_hdrprint);
	}
	sels[t].label = gtk_label_new(sels[t].name);
	sels[t].box = gtk_hbox_new(TRUE, 2);
	/* There seems to be no way to produce columns of unequal width
	gtk_box_set_homogeneous(GTK_BOX(sels[t].box), FALSE);
	*/
	gtk_container_add(GTK_CONTAINER(sels[t].box), sels[t].label);
	gtk_container_add(GTK_CONTAINER(sels[t].box), sels[t].button);
	gtk_widget_show(sels[t].label);
	gtk_widget_show(sels[t].button);
	gtk_widget_show(sels[t].box);
	gtk_container_add(GTK_CONTAINER(GTK_DIALOG(dialog)->vbox), sels[t].box);
  }

  /* Add the default button */
  def_button = t;
  sels[t].button = gtk_toggle_button_new_with_label(
				gg->use_def_print ? "Yes" : "No");
  gtk_toggle_button_set_active(GTK_TOGGLE_BUTTON(sels[t].button),
				gg->use_def_print);
  sels[t].label = gtk_label_new(sels[t].name);
  sels[t].box = gtk_hbox_new(TRUE, 2);
  /* See above
  gtk_box_set_homogeneous(GTK_BOX(sels[t].box), FALSE);
  */
  gtk_container_add(GTK_CONTAINER(sels[t].box), sels[t].label);
  gtk_container_add(GTK_CONTAINER(sels[t].box), sels[t].button);
  gtk_widget_show(sels[t].label);
  gtk_widget_show(sels[t].button);
  gtk_widget_show(sels[t].box);
  gtk_container_add(GTK_CONTAINER(GTK_DIALOG(dialog)->vbox), sels[t].box);

  /* Connect the default buttons state to its labels */
  g_signal_connect(
		G_OBJECT(sels[t].button),	/* Widget */
		"toggled",			/* Signal */
		G_CALLBACK(default_print_setup_handler),/* Callback */
		(gpointer) sels			/* User data */
		);

  /* Connect the font buttons to their handlers */
  g_signal_connect(
		G_OBJECT(sels[P_HEAD_INDEX].button),
		"font-set",
		G_CALLBACK(print_font_handler),
		(gpointer) sels[def_button].button /* Default button */
		);
  g_signal_connect(
		G_OBJECT(sels[P_BODY_INDEX].button),
		"font-set",
		G_CALLBACK(print_font_handler),
		(gpointer) sels[def_button].button /* Default button */
		);

  /* Connect the header print flag button to its handler */
  g_signal_connect(
		sels[P_FLG_INDEX].button,
		"toggled",
		G_CALLBACK(print_header_handler),
		(gpointer) sels[def_button].button /* Default button */
		);

  /* gtk_dialog_run() gives what is effectively a modal dialog,
  ** but signals and timers still work.
  */
  if (gtk_dialog_run(GTK_DIALOG(dialog)) == GTK_RESPONSE_OK) {
	/* If necessary, reset current printer fonts to defaults */
	gg->use_def_print = gtk_toggle_button_get_active(
				GTK_TOGGLE_BUTTON(sels[def_button].button));
	if (gg->use_def_print) {
		strncpy(gg->cur_hdr_print_font, gg->def_hdr_print_font, F_MAX);
		strncpy(gg->cur_print_font, gg->def_print_font, F_MAX);
		gg->cur_hdrprint = gg->def_hdrprint;
	}
	else {
		strncpy(gg->cur_hdr_print_font, gtk_font_button_get_font_name(
			GTK_FONT_BUTTON(sels[P_HEAD_INDEX].button)), F_MAX);
		strncpy(gg->cur_print_font, gtk_font_button_get_font_name(
			GTK_FONT_BUTTON(sels[P_BODY_INDEX].button)), F_MAX);
		gg->cur_hdrprint = gtk_toggle_button_get_active(
			GTK_TOGGLE_BUTTON(sels[P_FLG_INDEX].button));
	}
	/* Update the :set options */
	optisstring(&optdescs[O_PFONT], &optvals[O_PFONT], gg->cur_print_font);
	optisstring(&optdescs[O_HFONT], &optvals[O_HFONT],
							gg->cur_hdr_print_font);
	status = OK;
  }
  else
	status = FAILED;

  gtk_widget_destroy(dialog);

  return;
}


/*
**  p r i n t _ c a n c e l _ a c t i o n
**
** Cancel the current print job.
**
** Returns	Nothing		Always
**
** This shuts the print operation down via end_print() and done_print().
*/
static void print_cancel_action(GtkAction *action, gpointer data)
{
  GNGUI *gg = &gn_gui;

  DPRINT_F(("print_cancel()\n"));

  if (gg->print_operation != NULL) {
	gtk_print_operation_cancel(gg->print_operation);
	gg->print_operation = NULL;
  }

  return;
}

/* Help Menu actions */

/*
** h e l p _ a c t i o n
**
** Show the Elvis help (Gnome Yelp) file.
**
** Returns:	Nothing		Always
*/
static void help_action(GtkAction *action, gpointer data)
{
  GError *error = NULL;
  gboolean status;

  DPRINT_F(("help_action()\n"));

  status = gnome_help_display(
		YELP_ELVIS,
		NULL,
		&error
		);

  if (error != NULL) {
	msg(MSG_ERROR, error->message);
	msgflush();
	free(error);
  }

  return;
}


/*
** g o t o _ u r l
**
** Helper function for about_action dialog.
**
** Returns:	Nothing		Always
*/
static void goto_url(GtkAboutDialog *about, const gchar *url, gpointer data)
{
  gnome_url_show(url, NULL);

  return;
}


/*
** a b o u t _ a c t i o n
**
** Show the "About Elvis" dialog window.
**
** Returns:	Nothing		Always
*/
static void about_action(GtkAction *action, gpointer data)
{
  GNGUI *gg = (GNGUI *) data;

  static GtkAboutDialogActivateLinkFunc geturl;
  GError *error = NULL;
  GdkPixbuf *logo;

  const gchar *authors[] = {
	"Steve Kirkendall <kirkenda@cs.pdx.edu>",
	NULL
  };

  const gchar *documenters[] = {
	"Steve Kirkendall <kirkenda@cs.pdx.edu>",
	NULL
  };

  const gchar *copyright = COPY1;

  const gchar *comments =
	"Elvis is an extended VI clone which runs almost everywhere.\n"
	"Gnome GUI " GNOMEGUI_VERSION;
#if 0
  const gchar *translators = "There are no translators.\n";
#endif

  DPRINT_F(("about_action()\n"));

  logo = gtk_icon_theme_load_icon(
				gg->theme,
				LOGO_ELVIS,	/* icon name */
				48,		/* size */
		GTK_ICON_LOOKUP_USE_BUILTIN,	/* flags */
				&error
				);

  /* Attach the support site URL linkup to the About dialog window */
  geturl = gtk_about_dialog_set_url_hook(goto_url, NULL, NULL);

  /* Build and show the dialog window */
  gtk_show_about_dialog(GTK_WINDOW(gg->gwindow),
			"name", "Elvis",
			"version", VERSION,
			"copyright", copyright,
			"comments", comments,
			"website", "http://elvis.vi-editor.org",
			"authors", authors,
			"documenters", documenters,
			/* Useful only when building with gettext()
			"translator_credits", translators,
			*/
			"logo", logo,
			NULL
			);

  return;
}

/* Debug Menu actions */

#ifdef DEBUG_MENU
/*
** d e b u g _ a c t i o n
**
** Run specific functions for debugging.
**
** Returns:	Nothing		Always
*/
static void debug_action(GtkAction *action, gpointer data)
{
  GNGUI *gg = (GNGUI *) data;
  static int child_fdin;
  static char *argv[] = { NULL, NULL, NULL, NULL };
  GPid child_pid;
  struct Edwin *pd, *pr;
  struct Enode *pe, *pn;
  size_t t;
  int flg, status;
  gboolean rc;
  gchar *pc, *pf;
  enum { Other, Docs, Edwin, Enode, Gui, Notes } act;
  UCHAR *pu;
  char *pb, buff[80], font[80];

  /* Get the name of the menu item */
  strncpy(buff, gtk_action_get_name(action), sizeof(buff) - 1);

  g_print("debug_action(): name |%s|\n", buff);

  /* Make a note of what we are doing */
  if (strcmp(buff, "Debug_Docs_Action") == SAME) {
	act = Docs;
  }
  else if (strcmp(buff, "Debug_Edwin_Action") == SAME) {
	act = Edwin;
  }
  else if (strcmp(buff, "Debug_Enode_Action") == SAME) {
	act = Enode;
  }
  else if (strcmp(buff, "Debug_GUI_Action") == SAME) {
	act = Gui;
  }
  else if (strcmp(buff, "Debug_Notes_Action") == SAME) {
	act = Notes;
  }
  else
	act = Other;

  /* Some actions need the child window initialised */
  if (act != Other && Child == 0) {
	/* Sort out a path and file name */
	if ((pf = g_find_program_in_path(DEBUG_ELVIS)) == NULL) {
		vte_terminal_feed(VTE_TERMINAL(gg->edhead->vterm), "\007", 1L);
		g_print("debug_action(): cannot find |%s|\n",
						DEBUG_ELVIS);
		return;
	}
	argv[0] = pf;

	/* Add a debugging font option */
	sprintf(font, "--fontname=%s", gg->cur_dbg_fontname);
	argv[1] = font;

	rc = g_spawn_async_with_pipes(
		NULL,		/* child's CWD, or NULL for parent's */
		argv,		/* child's argument vector */
		NULL,		/* child's env, or NULL for parent's */
		G_SPAWN_DO_NOT_REAP_CHILD,	/* GSpawnFlags */
		NULL,		/* child_setup function */
		NULL,		/* user data for child_setup */
		&child_pid,	/* child's process ID, or NULL */
		&child_fdin,	/* child's stdin fd, or NULL */
		NULL,		/* child's stdout fd, or NULL */
		NULL,		/* child's stderr fd, or NULL */
		NULL		/* error code */
	);
	if (rc) {
		g_child_watch_add(child_pid, child_exit, NULL);
		flg = fcntl(child_fdin, F_GETFL);
		(void) fcntl(child_fdin, F_SETFL, flg | O_NONBLOCK);
		/* Park the child's PID somewhere accessible */
		Child = child_pid;
	}
	else {
		vte_terminal_feed(VTE_TERMINAL(gg->edhead->vterm), "\007", 1L);
		g_print("debug_action(): cannot exec %s\n", pf);
		return;
	}
  }

  /* Run the named function */
  if (strcmp(buff, "Debug_Debug_Action") == SAME) {
	g_print("No action taken.\n  VTE terminal:   rows %ld columns %ld\n",
		vte_terminal_get_row_count(VTE_TERMINAL(gg->edhead->vterm)),
		vte_terminal_get_column_count(VTE_TERMINAL(gg->edhead->vterm))
		);
	g_print("  Current window: rows %ld columns %ld\n",
		gg->edhead->rows,
		gg->edhead->cols
		);
	g_print("  Main window:    height %d width %d\n",
		gg->main_height,
		gg->main_width
		);
	g_print(
	"  adj->lower %f upper %f value %f\n  adj->step %f page %f size %f\n",
	gg->vert_adj->lower, gg->vert_adj->upper, gg->vert_adj->value,
	gg->vert_adj->step_increment, gg->vert_adj->page_increment,
	gg->vert_adj->page_size);
  }
  else if (strcmp(buff, "Debug_Keys_Action") == SAME) {
	Key_flg = !Key_flg;
	g_print("Key_flg %s\n", Key_flg ? "True" : "False");
  }
  else if (strcmp(buff, "Debug_Memory_Action") == SAME) {
#ifdef G_MEM_PROFILE
	g_mem_profile();
#else
	g_print("GLib memory profiler not installed\n");
#endif
  }
  else if (strcmp(buff, "Debug_Menu_Action") == SAME) {
	update_undolevel_menu(15, 5);
  }
  else if (strcmp(buff, "Debug_Print_Action") == SAME) {
	get_print_settings();
  }
  else if (strcmp(buff, "Debug_Scroll_Action") == SAME) {
	g_print("No action taken\n");
  }
  else if (strcmp(buff, "Debug_Select_Action") == SAME) {
	selection_test();
  }
  else if (Child == 0) {
	g_print("Child process not running\n");
  }

  /* The remaining cases handle the output of data to the child */
  if (act == Docs || act == Notes) {
	pc = (act == Docs) ? "doc" : "note";
	pn = (act == Docs) ? gg->docshead : gg->noteshead;
	sprintf(buff, "TITLE gg->%sshead 0x%08lx %sslist\n",
				pc, (ULONG) pn, pc);
	status = write(child_fdin, buff, strlen(buff));
	if (status == -1) {
		child_exit(Child, 0, NULL);
		return;
	}
	for (pe = pn ; pe != NULL ; pe = pe->next) {
		/* Fairly horrible casts, but it seems the simplest
		** way to move the data.
		*/
		sprintf(buff,
		"ENODE %u 0x%08lx 0x%08lx 0x%08lx %.18s %.18s %u %u\n",
			pe->id, (ULONG) pe->next,
			(ULONG) pe->prev, (ULONG) pe->data,
			pe->name, pe->label,
			pe->signal, pe->uiid
			);
		if (write(child_fdin, buff, strlen(buff)) == -1) {
			child_exit(Child, 0, NULL);
			return;
		}
		if (pe->next == pn) break;
		pc = "LINK\n";
		if (write(child_fdin, pc, strlen(pc)) == -1) {
			child_exit(Child, 0, NULL);
			return;
		}
	}
  }
  else if (act == Edwin) {
	/* Write a title */
	sprintf(buff, "TITLE gg->edhead 0x%08lx struct Edwin\n",
						(ULONG) gg->edhead);
	status = write(child_fdin, buff, strlen(buff));
	if (status == -1) {
		child_exit(Child, 0, NULL);
		return;
	}

	/* Convert the EDWIN to ASCII hex */
	if ((pb = malloc(4 * sizeof(struct Edwin))) == NULL) {
		g_print("debug_action(): malloc failed\n");
		return;
	}
	strcpy(pb, "EDWIN ");

	for (pd = gg->edhead ; pd != NULL ; pd = pd->next) {
		for (pc = pb + 6, pu = (UCHAR *) pd, t = 0 ;
				t < sizeof(struct Edwin) ; t++, pc += 3, pu++)
			sprintf(pc, " %2x", *pu);
		strcpy(pc, "\n");

		/* Write the string out to the child */
		if (write(child_fdin, pb, strlen(pb)) == -1) {
			child_exit(Child, 0, NULL);
			free(pb);
			return;
		}
		if (pd->next == gg->edhead) break;
		pc = "LINK\n";
		if (write(child_fdin, pc, strlen(pc)) == -1) {
			child_exit(Child, 0, NULL);
			free(pb);
			return;
		}
	}

	/* Print any members of the file rings */
	for (pr = gg->edhead ; pr != NULL ; pr = pr->next) {
		if (pr->argshead != NULL) {
			/* Write a title */
			sprintf(buff, "TITLE EDWIN id %d struct Enode\n",
								pr->id);
			status = write(child_fdin, buff, strlen(buff));
			if (status == -1) {
				child_exit(Child, 0, NULL);
				return;
			}
			/* Write the linked list */
			for (pe = pr->argshead ; pe != NULL ; pe = pe->next) {
				static char name[PATH_MAX + 1];
				/* Fairly horrible casts, but it seems
				** the simplest way to move the data.
				** However, the child's parser (sscanf)
				** can't cope with spaces in filenames.
				*/
				strcpy(name, pe->name);
				for (pf = pc = name ; *pc != '\0' ; pc++) {
					if (*pc == ' ') *pc = '?';
					if (*pc == SEPCHR) pf = pc;
				}
				if (*pf == SEPCHR) pf++;
				sprintf(buff,
			"ENODE %u 0x%08lx 0x%08lx 0x%08lx %.18s %.18s %u %u\n",
					pe->id, (ULONG) pe->next,
					(ULONG) pe->prev, (ULONG) pe->data,
					name, pf,
					pe->signal, pe->uiid
					);
				if (write(child_fdin, buff, strlen(buff))
								== -1) {
					child_exit(Child, 0, NULL);
					return;
				}
				if (pe->next == pr->argshead) break;
				pc = "LINK\n";
				if (write(child_fdin, pc, strlen(pc)) == -1) {
					child_exit(Child, 0, NULL);
					return;
				}
			}
		}
		if (pr->next == gg->edhead) break;
	}

	if (pb != NULL) free(pb);
	pb = NULL;
  }
  else if (act == Enode) {
	/* Write a title */
	sprintf(buff, "TITLE Enode output not installed\n");
	status = write(child_fdin, buff, strlen(buff));
	if (status == -1) {
		child_exit(Child, 0, NULL);
		return;
	}
  }
  else if (act == Gui) {
	/* Write a title */
	sprintf(buff, "TITLE Gnome GUI\n");
	status = write(child_fdin, buff, strlen(buff));
	if (status == -1) {
		child_exit(Child, 0, NULL);
		return;
	}

	/* Convert the GUI to ASCII hex */
	if ((pb = malloc(4 * sizeof(struct GnomeGui))) == NULL) {
		g_print("debug_action(): malloc failed\n");
		return;
	}
	strcpy(pb, "GUI ");

	for (pc = pb + 4, pu = (UCHAR *) &gn_gui, t = 0 ;
			t < sizeof(struct GnomeGui) ; t++, pc += 3, pu++)
		sprintf(pc, " %2x", *pu);
	strcpy(pc, "\n");

	/* Write the string out to the child */
	if (write(child_fdin, pb, strlen(pb)) == -1) {
		child_exit(Child, 0, NULL);
		free(pb);
		return;
	}

	if (pb != NULL) free(pb);
	pb = NULL;
  }

  return;
}


/*
** c h i l d _ e x i t
**
** Child exit watch function.
**
** Returns:	Nothing		Always
*/
  void child_exit(GPid pid, gint status, gpointer data)
{
  DPRINT_F(("child_exit()\n"));

  if (Child != 0) {
	g_spawn_close_pid(Child);	/* A no-op under Unix */
	Child = 0;
  }

  return;
}


/*
** s e l e c t i o n _ t e s t
**
** Dialog to debug clipboard selection.
**
** Returns:	Nothing		Always
*/
static void selection_test(void)
{
  GNGUI *gg = &gn_gui;
  static GtkWidget *selection;
  static GtkSelectionData selection_data;
  static int have_selection = FALSE;
  GtkWidget *abutton;
  GtkWidget *sbutton;
  GtkWidget *dialog;
  int status;

  DPRINT_F(("selection_test()\n"));

  /* Create a widget with which the selection can be associated */
  selection = gtk_invisible_new();

  /* Create the dialog */
  dialog = gtk_dialog_new_with_buttons(
		"Selection",			/* Title */
		GTK_WINDOW(gg->gwindow),	/* Parent window */
		GTK_DIALOG_DESTROY_WITH_PARENT,	/* Destroy with main window */
		GTK_STOCK_OK,			/* Stock ID for button one */
		GTK_RESPONSE_OK,		/* Response for button one */
		NULL				/* Terminator */
		);

  sbutton = gtk_button_new_with_label("Select");
  g_signal_connect(
		G_OBJECT(sbutton),		/* Widget */
		"clicked",			/* Signal */
		G_CALLBACK(target_handler),	/* Callback */
		(gpointer) selection		/* User data */
		);
  gtk_widget_show(sbutton);
  gtk_container_add(GTK_CONTAINER(GTK_DIALOG(dialog)->vbox), sbutton);

  abutton = gtk_toggle_button_new_with_label("Acquire");
  g_signal_connect(
		G_OBJECT(abutton),
		"toggled",
		G_CALLBACK(selection_acquire),
		(gpointer) &have_selection
		);
  gtk_widget_show(abutton);
  gtk_container_add(GTK_CONTAINER(GTK_DIALOG(dialog)->vbox), abutton);

  /* Handler to be called when the information is requested.
  ** widget, selection, and target identify the requests this
  ** handler will manage.  Info can be used in the handler.
  ** A request will trigger the selection-get signal.
  */
  gtk_selection_add_target(
		selection,			/* Widget */
		GDK_SELECTION_PRIMARY,		/* Selection */
		GDK_SELECTION_TYPE_STRING,	/* Target */
		1				/* Info */
		);

  /* Connect the various selection signals */
  g_signal_connect(
		G_OBJECT(selection),
		"selection-clear-event",
		G_CALLBACK(selection_blocked),
		(gpointer) &have_selection
		);

  /* This is never called ### */
  g_signal_connect(
		G_OBJECT(selection),
		"selection-get",
		G_CALLBACK(target_set),
		(gpointer) &selection_data
		);

  g_signal_connect(
		G_OBJECT(selection),
		"selection-received",
		G_CALLBACK(selection_received),
		NULL
		);

  /* gtk_dialog_run() gives what is effectively a modal dialog,
  ** but signals and timers still work.
  */
  if (gtk_dialog_run(GTK_DIALOG(dialog)) == GTK_RESPONSE_OK)
	status = OK;
  else
	status = FAILED;

  gtk_widget_destroy(dialog);

  return;
}


/*
** t a r g e t _ h a n d l e r
**
** Handler for clicked signal, a selection request.
**
** Returns:	Nothing		Always
**
** A clicked signal handler has arguments:
** GtkWidget *widget
** gpointer user_data
** return none
*/
static void target_handler(GtkWidget *widget, gpointer data)
{
  static GdkAtom targets_atom = GDK_NONE;
  GtkWidget *selection = (GtkWidget *) data;
  int status;

  DPRINT_F(("target_handler()\n"));

  /* Get the atom corresponding to the string "TARGETS" */
  if (targets_atom == GDK_NONE)
	targets_atom = gdk_atom_intern("TARGETS", FALSE);

  /* Request the "TARGETS" target for the primary selection.
  ** Return is TRUE on success, FALSE if the request could
  ** not be processed.
  */
  status = gtk_selection_convert(
		selection,		/* The requesting widget */
		GDK_SELECTION_PRIMARY,	/* Which selection to get */
		targets_atom,		/* Form of information (eg. string) */
		GDK_CURRENT_TIME	/* Time of request */
		);

  return;
}


/*
** s e l e c t i o n _ r e c e i v e d
**
** Handler for selection-received signal.
** This handles selection returns asynchronously.
**
** Returns:	Nothing		Always
**
** A selection-received signal handler has arguments:
** GtkWidget *widget
** GtkSelectionData *data
** guint time
** gpointer user_data
** return none
*/
static void selection_received(GtkWidget *widget,
	GtkSelectionData *selection, guint ts, gpointer data)
{
  GdkAtom *atoms;
  size_t t;
  char *pn;

  DPRINT_F(("selection_received()\n"));

  /* Check to see if the retrieval succeeded */
  if (selection->length < 0) {
	g_print("Selection retrieval failed\n");
	return;
  }
  /* Make sure we got the data in the expected form */
  if (selection->type != GDK_SELECTION_TYPE_ATOM) {
	g_print("Selection \"TARGETS\" was not returned as atoms!\n");
	return;
  }

  /* Print out the atoms we received */
  atoms = (GdkAtom *) selection->data;

  for (t = 0 ; t < (selection->length / sizeof(GdkAtom)); t++) {
	pn = gdk_atom_name(atoms[t]);
	g_print("Name %lu |%s|\n", (ULONG) t, pn == NULL ? "(bad atom)" : pn);
  }

  return;
}


/*
** s e l e c t i o n _ a c q u i r e
**
** Handler for toggle signal, to control a selection.
**
** Returns:	Nothing		Always
**
** The have_selection flag is updated as a side effect.
**
** A toggle signal handler has arguments:
** GtkWidget *widget
** gpointer user_data
** return none
*/
static void selection_acquire(GtkWidget *widget, gpointer data)
{
  int *have_selection, status;

  DPRINT_F(("selection_acquire()\n"));

  have_selection = (int *) data;

  /* When the button is pressed, try to own or release a selection */
  if (GTK_TOGGLE_BUTTON(widget)->active) {

	/* Claim ownership of a selection. Returns TRUE on success. */
	*have_selection = gtk_selection_owner_set(
				widget,			/* Claimant */
				GDK_SELECTION_PRIMARY,	/* Selection to claim */
				GDK_CURRENT_TIME	/* Timestamp */
				);
	g_print("selection_acquire(): selection claim %s\n",
			have_selection ? "succeeded" : "failed");
  }
  else if (*have_selection) {
	/* Check if we own the selection, and then clear it by
	** setting the owner to NULL.
	*/
	if (gdk_selection_owner_get(GDK_SELECTION_PRIMARY) == widget->window) {
		status = gtk_selection_owner_set(
				NULL,
				GDK_SELECTION_PRIMARY,
				GDK_CURRENT_TIME
				);
		g_print("selection_acquire(): selection release %s\n",
					status ? "succeeded" : "failed");
	}
	else
		g_print("selection_acquire(): selection not owned\n");

	*have_selection = FALSE;
  }

  return;
}


/*
** s e l e c t i o n _ b l o c k e d
**
** Handler for selection-clear-event signal.
** Called when another application claims the selection.
**
** Returns:	TRUE		Always
**
** The have_selection flag is updated as a side effect.
**
** A selection-clear-event signal handler has arguments:
** GtkWidget *widget
** GdkEventSelection *event
** gpointer user_data
** return handled
*/
static gint selection_blocked(GtkWidget *widget,
		GdkEventSelection *event, gpointer data)
{
  int *have_selection = (int *) data;

  DPRINT_F(("selection_blocked()\n"));

  *have_selection = FALSE;

  /* Continue propagation ??? */
  return FALSE;
}


/*
** t a r g e t _ s e t
**
** Handler for selection-get signal.
** This just gives the time as the selection.
**
** Returns:	Nothing		Always
**
** A selection-get signal handler has arguments:
** GtkWidget *widget
** GtkSelectionData *data
** guint info
** guint time
** gpointer user_data
** return none
*/
static void target_set(GtkWidget *widget, GtkSelectionData *selection,
			guint info, guint t, gpointer data)
{
  gchar *timestr;
  time_t current_time;

  DPRINT_F(("target_set()\n"));

  current_time = time(NULL);
  timestr = asctime(localtime(&current_time));

  /* Stores data into a GtkSelectionData object. Must be called from
  ** a selection handler callback. Zero-terminates the stored data.
  ** No return value.
  */
  gtk_selection_data_set(
		(GtkSelectionData *) data, /* GtkSelectionData structure * */
		GDK_SELECTION_TYPE_STRING, /* Type of selection data */
		8,			/* Format (bits in a unit) */
		(UCHAR *) timestr,	/* Data (will be copied) */
		(int) strlen(timestr)	/* Length of the data */
		);

  return;
  }
#endif

/*
** a d d _ r _ a c t i o n
**
** Add a new window (ie. edit ring) and make it current.
**
** Returns:	Nothing		Always
*/
static void add_r_action(GtkAction *action, gpointer data)
{
  GNGUI *gg = (GNGUI *) data;

  DPRINT_F(("add_r_action()\n"));

  /* Flush all message buffers and call back into Elvis */
  msgflush();
  eventex(gg->edhead, "snew", ElvFalse);

  return;
}


/*
** r m v _ r _ a c t i o n
**
** Drop the current window (ie. edit ring).
**
** Returns:	Nothing		Always
*/
static void rmv_r_action(GtkAction *action, gpointer data)
{
  GNGUI *gg = (GNGUI *) data;

  DPRINT_F(("rmv_r_action()\n"));

  /* Flush all message buffers and call back into Elvis */
  msgflush();
  eventex(gg->edhead, "quit!", ElvFalse);

  return;
}

#ifdef CHG_ARGSLIST
/*
** a d d _ f _ a c t i o n
**
** Add a file to the current arglist and make it current.
**
** Returns:	Nothing		Always
**
** This doesn't really seem possible without more help from
** the main Elvis code. ###
*/
static void add_f_action(GtkAction *action, gpointer data)
{
  DPRINT_F(("add_f_action()\n"));

  return;
}


/*
** r m v _ f _ a c t i o n
**
** Remove a file from the current arglist.
**
** Returns:	Nothing		Always
**
** This doesn't really seem possible without more help from
** the main Elvis code. ###
*/
static void rmv_f_action(GtkAction *action, gpointer data)
{
  DPRINT_F(("rmv_f_action()\n"));

  return;
}
#endif

#ifdef CANCEL_ACTION
/*
** c a n c e l _ a c t i o n
**
** Cancel the current task.
**
** Returns:	Nothing		Always
**
** Whatever is going on obviously needs to check the global
** variable StopThat regularly.
*/
static void cancel_action(GtkToolButton *button, gpointer data)
{
  StopThat = TRUE;
  return;
}
#endif

/*
** e x _ a c t i o n
**
** Pass an EX command to Elvis.
**
** Returns:	Nothing		Always
*/
static void ex_action(GtkToolButton *button, gpointer data)
{
  GNGUI *gg = (GNGUI *) data;
  char buff[64];
  size_t index;

  DPRINT_F(("ex_action()\n"));

  /* Get the button's name */
  strncpy(buff, gtk_tool_button_get_label(button), sizeof(buff) - 1);

  DPRINT_F(("ex_action(): |%s|\n", buff));

  /* Get the command string from the tools array and pass it to Elvis */
  for (index = 0 ; index < (sizeof(tools) / sizeof(tools[0])) ; index++) {
	if (strcmp(buff, tools[index].label) == SAME) {
		msgflush();
		eventex(gg->edhead, tools[index].cmd, ElvFalse);
		(void) eventdraw(gg->edhead);
		break;
	}
  }

  return;
}


/*
** r e _ a c t i o n
**
** Reverse or repeat the last edit action(s).
**
** Returns:	Nothing		Always
*/
static void re_action(GtkAction *action, gpointer data)
{
  GNGUI *gg = (GNGUI *) data;
  const gchar *name;
  char *cmd;

  /* Get the name of the menu item */
  name = gtk_action_get_name(action);

  DPRINT_F(("re_action()\n"));

  if (strcmp(name, "Undo_Action") == SAME)
	cmd = "undo";
  else if (strcmp(name, "Redo_Action") == SAME)
	cmd = "redo";
  else
	cmd = NULL;

  if (cmd != NULL) {
	msgflush();
	eventex(gg->edhead, cmd, ElvFalse);
	(void) eventdraw(gg->edhead);
  }

  return;
}


/*
** d r a g _ a c t i o n
**
** Drag text to another window.
**
** Returns:	Nothing		Always
*/
static void drag_action(GtkAction *action, gpointer data)
{
  GNGUI *gg = (GNGUI *) data;
  GtkTargetList *targetlist;
  GtkTargetEntry targets[] = {
	{ "text/plain", 0, 1 },
	{ "text/uri-list", 0, 2 }
  };
  GdkEvent *event;

  DPRINT_F(("drag_action()\n"));

  gg->drop_type = PUT_TXT;

  /* We supply only the text/plain type */
  targetlist = gtk_target_list_new(targets, 1);
  event = gdk_event_new(GDK_3BUTTON_PRESS);

  /* gtk_drag_begin initiates a drag on the source side. The function needs
  ** to be used only when the application is starting drags itself, and is
  ** not needed when gtk_drag_source_set() is used.
  ** widget :	the source widget.
  ** targets :	The targets (data formats) which the source can provide.
  ** actions :	A bitmask of the allowed drag actions for this drag.
  ** button :	The button the user clicked to start the drag.
  ** event :	The event that triggered the start of the drag.
  ** Returns :	the context for this drag.
  */
  gtk_drag_begin(gg->holder, targetlist, GDK_ACTION_COPY, 3, event);

  return;
}


/*
** c u t _ a c t i o n
**
** Cut text to the clipboard.
**
** Returns:	Nothing		Always
*/
static void cut_action(GtkAction *action, gpointer data)
{
  GNGUI *gg = (GNGUI *) data;

  DPRINT_F(("cut_action()\n"));

  /* Call back into Elvis to start the process */
  eventclick(gg->edhead, -1, -1, CLICK_CUT);

  /* Force a screen update */
  if (gg->edhead != NULL) (void) eventdraw(gg->edhead);

  /* Update the block select and the main and mouse menus */
  gg->block_sel = NO;
  gg->menu_state &= ~MENU_MARK;
  config_menus(gg);

  return;
}


/*
** c o p y _ a c t i o n
**
** Copy text to the clipboard.
**
** Returns:	Nothing		Always
*/
static void copy_action(GtkAction *action, gpointer data)
{
  GNGUI *gg = (GNGUI *) data;

  DPRINT_F(("copy_action()\n"));

  /* Call back into Elvis to start the process */
  eventclick(gg->edhead,  -1, -1, CLICK_YANK);

  return;
}


/*
** p a s t e _ a c t i o n
**
** Paste text from the clipboard.
**
** Returns:	Nothing		Always
*/
static void paste_action(GtkAction *action, gpointer data)
{
  GNGUI *gg = (GNGUI *) data;

  DPRINT_F(("paste_action()\n"));

  /* Call back into Elvis to start the process */
  eventclick(gg->edhead, -1, -1, CLICK_PASTE);

  /* Force a screen update */
  if (gg->edhead != NULL) (void) eventdraw(gg->edhead);

  return;
}


/*
** d e l e t e _ a c t i o n
**
** Delete text.
**
** Returns:	Nothing		Always
*/
static void delete_action(GtkAction *action, gpointer data)
{
  GNGUI *gg = (GNGUI *) data;

  DPRINT_F(("delete_action()\n"));

  /* Call back into Elvis to start the process */
  eventclick(gg->edhead, -1, -1, CLICK_DELETE);

  /* Force a screen update */
  if (gg->edhead != NULL) (void) eventdraw(gg->edhead);

  /* Update the block select and the main and mouse menus */
  gg->block_sel = NO;
  gg->menu_state &= ~MENU_MARK;
  config_menus(gg);

  return;
}


/*
** c r e a t e _ m a i n _ m e n u
**
** Create the main menu.
**
** Returns:	OK		Success
**		FAILED		Otherwise
*/
static int create_main_menu(GNGUI *gg)
{
  GtkUIManager *mainuimanager;
  GtkActionGroup *mainactiongroup;
  static GtkTooltips *pt;
  GtkWidget *widget;
  static GtkWidget *note;
  GError *error;
  size_t t;
  int uiid, signal, status = OK;

  DPRINT_F(("create_main_menu()\n"));

  /* Build the full default set of menus.  Note that they form a
  ** single action group, which might be better in sections. ???
  ** Displaying tooltips seems to require a separate iteration over
  ** the whole list, getting the widgets and setting the tooltips. ???
  */
  mainactiongroup = gtk_action_group_new("Actions");
  gtk_action_group_add_actions(
		mainactiongroup,
		menu_entries,
		sizeof(menu_entries) / sizeof(menu_entries[0]),
		(gpointer) gg
		);
  gg->mainactiongroup = mainactiongroup;

  /* Add the action group to the UI Manager */
  mainuimanager = gtk_ui_manager_new();
  gg->mainuimanager = mainuimanager;
  gtk_ui_manager_insert_action_group(mainuimanager, mainactiongroup, 0);

  /* Read in the UI from a fixed string */
  error = NULL;
  uiid = gtk_ui_manager_add_ui_from_string(mainuimanager, menu_string,
							-1, &error);
#if 0
  { /* Debugging block begin */
  if (error == NULL)
	g_print("create_main_menu(): main ui id %d\n", uiid);
  else
	g_print("create_main_menu(): main ui error %s\n", error->message);
  } /* Debugging block end */
#endif

  /* Add the permanent menu tooltips */
  pt = gtk_tooltips_new();
  for (t = 0 ; t < (sizeof(menu_strings) / sizeof(menu_strings[0])) ; t++) {
	gtk_tooltips_set_tip(
		pt,
		gtk_ui_manager_get_widget(mainuimanager, menu_strings[t].str),
		menu_entries[menu_strings[t].ref].tooltip,
		NULL
		);
  }
  gtk_tooltips_enable(pt);

  /* Used only by the Tools, Options, Documents and Help menus */
  gg->menutips = gtk_tooltips_new();
  gtk_tooltips_enable(gg->menutips);

  /* Add any changes from command line arguments */
  switch_toolbar_menu();
  switch_startpos_menu();
  switch_docs_open_menu();
  switch_manpage_menu();

  /* Add the head of a cascading notepads menu */
  note = gtk_menu_item_new_with_label(NOTEPAD_ELVIS);

  gg->notesmenu = gtk_menu_new();
  gtk_menu_shell_append(GTK_MENU_SHELL(gg->notesmenu), note);

  widget = gtk_ui_manager_get_widget(mainuimanager,
				"/ui/menubar/ToolsMenu/Notepads");
  gtk_menu_item_set_submenu(GTK_MENU_ITEM(widget), gg->notesmenu);

  /* Connect the note_action() function to the menu item */
  signal = g_signal_connect(
			G_OBJECT(note),
			"activate",
			G_CALLBACK(note_action),
			NOTEPAD_ELVIS
			);

  /* Show the result */
  gtk_widget_show(note);

  /* Get the menu widget */
  gg->menubar = gtk_ui_manager_get_widget(mainuimanager, "/ui/menubar");
  DPRINT_J(("create_main_menu(): menubar is %s\n",
				gg->menubar == NULL ? "NULL" : "valid"));

  /* Add it to the main Gnome window */
  gnome_app_set_menus(GNOME_APP(gg->gwindow), GTK_MENU_BAR(gg->menubar));

  return status;
}


/*
** c r e a t e _ m o u s e _ m e n u
**
** Create the mouse menu.
**
** Returns:	OK		Success
**		FAILED		Otherwise
*/
static int create_mouse_menu(GNGUI *gg)
{
  GtkActionGroup *mouseactiongroup;
  GtkUIManager *mouseuimanager;
  GError *error;
  int status = OK;

  DPRINT_F(("create_mouse_menu()\n"));

  /* Get an action group and put it into a UI manager */
  mouseactiongroup = gtk_action_group_new("Actions");
  gtk_action_group_add_actions(
		mouseactiongroup,
		mouseentries,
		sizeof(mouseentries) / sizeof(mouseentries[0]),
		(gpointer) gg
		);
  mouseuimanager = gtk_ui_manager_new();
  gtk_ui_manager_insert_action_group(mouseuimanager, mouseactiongroup, 0);

  /* Read in the UI from a buffer (or from an XML file) */
  error = NULL;
  gg->mouse_ui_id = gtk_ui_manager_add_ui_from_string(
			mouseuimanager, mouse_str, -1, &error);
  if (error == NULL) {
	DPRINT_J(("create_mouse_menu(): mouse ui id %d\n", gg->mouse_ui_id));
	gg->mousemenu = gtk_ui_manager_get_widget(mouseuimanager, "/ui/popup");
	DPRINT_J(("create_mouse_menu(): mousemenu is %s\n",
				gg->mousemenu == NULL ? "NULL" : "valid"));
	if (gg->mousemenu == NULL)
			status = FAILED;
  }
  else {
	DPRINT_J(("create_mouse_menu(): mouse ui error %s\n",
						error->message));
	status = FAILED;
  }

  /* Store the result */
  if (status == OK)
	gg->mouseuimanager = mouseuimanager;
  else
	gg->mouseuimanager = NULL;

  return status;
}


/*
** c o n f i g _ m e n u s
**
** Configure the main and mouse menus.
**
** Returns:	Nothing		Always
*/
static void config_menus(GNGUI *gg)
{
  static GtkAction *next_act = NULL;
  static GtkAction *prev_act = NULL;
  unsigned int u, state;
  gboolean flg;

  DPRINT_F(("config_menus()\n"));

  /* Menu entry sensitivities
  #define MENU_UNDO 1
  #define MENU_REDO 2
  #define MENU_NEXT 4
  #define MENU_PREV 8
  #define MENU_REPL 16
  #define MENU_GOTO 32
  #define MENU_COPY 64
  #define MENU_CUT 128
  #define MENU_ENTRIES 8
  #define MENU_MARK 256
  */

  /* Initialise the search ActionEntries once, not on each call */
  if (next_act == NULL) {
	next_act = gtk_ui_manager_get_action(gg->mainuimanager,
				"/ui/menubar/SearchMenu/Find Next");
	prev_act = gtk_ui_manager_get_action(gg->mainuimanager,
				"/ui/menubar/SearchMenu/Find Previous");
  }

  /* A marked block enables various actions */
  state = gg->menu_state;
  if (state & MENU_MARK) {
	state |= MENU_COPY;
	state |= MENU_CUT;
  }
  else {
	state &= ~MENU_COPY;
	state &= ~MENU_CUT;
  }
  /* This clash should never occur; resolve in favour of MENU_NEXT */
  if ((state & MENU_NEXT) && (state & MENU_PREV)) {
	DPRINT_J(("config_menus(): bad state, NEXT and PREV both set\n"));
	state &= ~MENU_PREV;
  }
  gg->menu_state = state;

  /* Clear the Find Again and Find Next/Prev menu entries */
  if (!(state & MENU_NEXT) && !(state & MENU_PREV)) {
	gtk_widget_set_sensitive(gtk_ui_manager_get_widget(
				gg->mainuimanager,
				"/ui/menubar/SearchMenu/Find Again"),
				FALSE);
	g_object_set((GObject *) prev_act, "stock-id", GTK_STOCK_FIND, NULL);
	g_object_set((GObject *) next_act, "stock-id", GTK_STOCK_FIND, NULL);
  }

  /* Set the Find Next/Again icons to the current search type */
  if (state & MENU_NEXT) {
	g_object_set((GObject *) prev_act, "stock-id", GTK_STOCK_FIND, NULL);
	g_object_set((GObject *) next_act, "stock-id", GTK_STOCK_APPLY, NULL);
  }
  else if (state & MENU_PREV) {
	g_object_set((GObject *) prev_act, "stock-id", GTK_STOCK_APPLY, NULL);
	g_object_set((GObject *) next_act, "stock-id", GTK_STOCK_FIND, NULL);
  }

  /* Scan the menu state bitmap and update the menu bar entries */
  for (u = 0 ; u < MENU_ENTRIES ; u++) {
	switch (u) {
	case 0:
		flg = (gboolean) (state & MENU_UNDO);
		gtk_widget_set_sensitive(gtk_ui_manager_get_widget(
					gg->mainuimanager,
					"/ui/menubar/EditMenu/Undo"),
					flg);
		gtk_widget_set_sensitive(gtk_ui_manager_get_widget(
					gg->mouseuimanager,
					"/ui/popup/Undo"),
					flg);
		break;
	case 1:
		flg = (gboolean) (state & MENU_REDO);
		gtk_widget_set_sensitive(gtk_ui_manager_get_widget(
					gg->mainuimanager,
					"/ui/menubar/EditMenu/Redo"),
					flg);
		gtk_widget_set_sensitive(gtk_ui_manager_get_widget(
					gg->mouseuimanager,
					"/ui/popup/Redo"),
					flg);
		break;
	case 2:
		flg = (gboolean) (state & MENU_NEXT);
		if (flg) {
			gtk_widget_set_sensitive(gtk_ui_manager_get_widget(
					gg->mainuimanager,
					"/ui/menubar/SearchMenu/Find Again"),
					TRUE);
		}
		break;
	case 3:
		flg = (gboolean) (state & MENU_PREV);
		if (flg) {
			gtk_widget_set_sensitive(gtk_ui_manager_get_widget(
					gg->mainuimanager,
					"/ui/menubar/SearchMenu/Find Again"),
					TRUE);
		}
	case 4:
		flg = (gboolean) (state & MENU_REPL);
		gtk_widget_set_sensitive(gtk_ui_manager_get_widget(
					gg->mainuimanager,
					"/ui/menubar/SearchMenu/Replace Again"),
					flg);
		break;
	case 5:
		flg = (gboolean) (state & MENU_GOTO);
		gtk_widget_set_sensitive(gtk_ui_manager_get_widget(
					gg->mainuimanager,
					"/ui/menubar/SearchMenu/Goto Again"),
					flg);
		break;
	case 6:
		flg = (gboolean) (state & MENU_COPY);
		gtk_widget_set_sensitive(gtk_ui_manager_get_widget(
					gg->mainuimanager,
					"/ui/menubar/EditMenu/Copy"),
					flg);
		gtk_widget_set_sensitive(gtk_ui_manager_get_widget(
					gg->mouseuimanager,
					"/ui/popup/Drag"),
					flg);
		gtk_widget_set_sensitive(gtk_ui_manager_get_widget(
					gg->mouseuimanager,
					"/ui/popup/Copy"),
					flg);
		break;
	case 7:
		flg = (gboolean) (state & MENU_CUT);
		gtk_widget_set_sensitive(gtk_ui_manager_get_widget(
					gg->mainuimanager,
					"/ui/menubar/EditMenu/Cut"),
					flg);
		gtk_widget_set_sensitive(gtk_ui_manager_get_widget(
					gg->mainuimanager,
					"/ui/menubar/EditMenu/Delete"),
					flg);
		gtk_widget_set_sensitive(gtk_ui_manager_get_widget(
					gg->mouseuimanager,
					"/ui/popup/Cut"),
					flg);
		gtk_widget_set_sensitive(gtk_ui_manager_get_widget(
					gg->mouseuimanager,
					"/ui/popup/Delete"),
					flg);
		break;
	default:
		break;
	}
  }

  return;
}


/*
** s w i t c h _ t o o l b a r _ m e n u
**
** Switch the show/hide toolbar menu entries.
**
** Returns:	Nothing		Always
*/
static void switch_toolbar_menu(void)
{
  GNGUI *gg = &gn_gui;
  GtkWidget *widget;

  DPRINT_F(("switch_toolbar_menu()\n"));

  /* Hide whichever toolbar menu entry is irrelevant */
  widget = gtk_ui_manager_get_widget(gg->mainuimanager,
				"/ui/menubar/OptionsMenu/Show Toolbar");
  if (o_showtoolbar)
	gtk_widget_hide(widget);
  else
	gtk_widget_show(widget);

  widget = gtk_ui_manager_get_widget(gg->mainuimanager,
				"/ui/menubar/OptionsMenu/Hide Toolbar");
  if (!o_showtoolbar)
	gtk_widget_hide(widget);
  else
	gtk_widget_show(widget);

  return;
}


/*
** s w i t c h _ s t a r t p o s _ m e n u
**
** Switch the set/unset start position menu entries.
**
** Returns:	Nothing		Always
*/
static void switch_startpos_menu(void)
{
  GNGUI *gg = &gn_gui;
  GtkWidget *widget;

  DPRINT_F(("switch_startpos_menu()\n"));

  /* Hide whichever starting position menu entry is irrelevant */
  widget = gtk_ui_manager_get_widget(gg->mainuimanager,
				"/ui/menubar/OptionsMenu/Fix Position");
  if (o_startposition)
	gtk_widget_hide(widget);
  else
	gtk_widget_show(widget);

  widget = gtk_ui_manager_get_widget(gg->mainuimanager,
				"/ui/menubar/OptionsMenu/Free Position");
  if (!o_startposition)
	gtk_widget_hide(widget);
  else
	gtk_widget_show(widget);

  return;
}


/*
** s w i t c h _ d o c s _ o p e n _ m e n u
**
** Switch the current/new document open menu entries.
**
** Returns:	Nothing		Always
*/
static void switch_docs_open_menu(void)
{
  GNGUI *gg = &gn_gui;
  GtkWidget *widget;

  DPRINT_F(("switch_docs_open_menu()\n"));

  /* Hide whichever docs open position menu entry is irrelevant */
  widget = gtk_ui_manager_get_widget(gg->mainuimanager,
				"/ui/menubar/OptionsMenu/Open Current");
  if (o_newmode)
	gtk_widget_hide(widget);
  else
	gtk_widget_show(widget);

  widget = gtk_ui_manager_get_widget(gg->mainuimanager,
				"/ui/menubar/OptionsMenu/Open New");
  if (!o_newmode)
	gtk_widget_hide(widget);
  else
	gtk_widget_show(widget);

  return;
}


/*
** s w i t c h _ m a n p a g e _ m e n u
**
** Switch the one/all man page return menu entries.
**
** Returns:	Nothing		Always
*/
static void switch_manpage_menu(void)
{
  GNGUI *gg = &gn_gui;
  GtkWidget *widget;

  DPRINT_F(("switch_manpage_menu()\n"));

  /* Hide whichever manpage menu entry is irrelevant */
  widget = gtk_ui_manager_get_widget(gg->mainuimanager,
				"/ui/menubar/ToolsMenu/Manpage");
  if (o_manmode)
	gtk_widget_show(widget);
  else
	gtk_widget_hide(widget);

  widget = gtk_ui_manager_get_widget(gg->mainuimanager,
				"/ui/menubar/ToolsMenu/Manpage All");
  if (!o_manmode)
	gtk_widget_show(widget);
  else
	gtk_widget_hide(widget);

  return;
}


/*
** u p d a t e _ u n d o l e v e l _ m e n u
**
** Update the undolevel edit menu entry.
**
** Returns:	Nothing		Always
*/
static void update_undolevel_menu(int undo, int redo)
{
  GNGUI *gg = &gn_gui;
  GtkUIManager *self = NULL;
  GtkAction *act = NULL;
  static char u_label[64], r_label[64];
  char u_path[] = "/ui/menubar/EditMenu/Undo";
  char r_path[] = "/ui/menubar/EditMenu/Redo";

  DPRINT_F(("update_undolevel_menu()\n"));

  /* Get the UI Manager */
  self = gg->mainuimanager;
  DPRINT_J(("update_undolevel_menu(): self is %s\n",
				self == NULL ? "NULL" : "valid"));

  /* Get the Undo action and reset the label */
  act = gtk_ui_manager_get_action(self, u_path);
  sprintf(u_label, "%s %d", "_Undo", undo);
  g_object_set((GObject *) act, "label", u_label, NULL);

  /* Get the Redo action and reset the label */
  act = gtk_ui_manager_get_action(self, r_path);
  sprintf(r_label, "%s %d", "_Redo", redo);
  g_object_set((GObject *) act, "label", r_label, NULL);

  /* Force a menu update */
  gtk_ui_manager_ensure_update(self);

  return;
}


/*
** c r e a t e _ t o o l b a r
**
** Create a Gnome toolbar.
**
** Returns:	OK		Success
**		FAILED		Otherwise
**
** NB. Returns early on failure.
*/
static int create_toolbar(GNGUI *gg)
{
  size_t index;
  int status, new;

  DPRINT_F(("create_toolbar()\n"));

  status = OK;
  new = FALSE;

  /* Initialisation */
  if (gg->toolbar == NULL) {
	gg->toolbar = gtk_toolbar_new();
	if (gg->tooltips == NULL)
		gg->tooltips = gtk_tooltips_new();
	new = TRUE;
  }

  /* Poor excuse for error checking */
  if (gg->toolbar == NULL || gg->tooltips == NULL) {
	gg->toolbar = NULL;
	gg->tooltips = NULL;
	return FAILED;
  }

  /* Initialise the data elements */
  if (new) {
	/* Load the toolbar and the tooltips from the tools[] array */
	for (index = 0 ; index < (sizeof(tools) / sizeof(tools[0])) ; index++) {
		if (index == LIST_RINGS) {
			tools[index].button =
				gtk_tool_button_new(
					(GtkWidget *) tools[index].icon, NULL);
		}
		else {
			tools[index].button =
				gtk_tool_button_new_from_stock(
					tools[index].icon);
		}
		gtk_tool_button_set_label(GTK_TOOL_BUTTON(tools[index].button),
						tools[index].label);
		gtk_toolbar_insert(GTK_TOOLBAR(gg->toolbar),
						tools[index].button, -1);
		/* Add the callback to the button */
		g_signal_connect(
			G_OBJECT(tools[index].button),
			tools[index].signal,
			G_CALLBACK(tools[index].callback),
			tools[index].data ? tools[index].data :
						(gpointer) gg
			);
		/* Add a tooltip to the button */
		gtk_tool_item_set_tooltip(
			tools[index].button,
			gg->tooltips,
			tools[index].tip,
			NULL
			);
		/* Make the button visible */
		gtk_widget_show(GTK_WIDGET(tools[index].button));
	}

	/* Toolbar tooltips are enabled by default, and this function
	** doesn't seem to disable them. The enable/disable functions
	** control _all_ tooltips, not just those on a toolbar.
	gtk_tooltips_disable(GTK_TOOLTIPS *tooltips);
	*/
	gtk_toolbar_set_tooltips(GTK_TOOLBAR(gg->toolbar), FALSE);
  }


  /* Create a new BonoboDockItem widget containing a toolbar, and add it
  ** to the app's dock with the specified layout information. Note that
  ** if automatic layout configuration is enabled, the layout is overridden
  ** by the saved configuration, if any.
  */
#if 0
  gnome_app_add_toolbar(
	GnomeApp *app,			/* A GnomeApp widget */
	GtkToolbar *toolbar,		/* Toolbar to be added to app's dock */
	const gchar *name,		/* Dock item containing the toolbar */
	BonoboDockItemBehavior behavior,/* Behavior for the new dock item */
	BonoboDockPlacement placement,	/* Placement for the new dock item */
	gint band_num,			/* Dock's band number */
	gint band_position,		/* Dock's position in the band */
	gint offset			/* Offset from previous dock item */
	);
  typedef enum {
  BONOBO_DOCK_ITEM_BEH_NORMAL = 0,
  BONOBO_DOCK_ITEM_BEH_EXCLUSIVE = 1 << 0,
  BONOBO_DOCK_ITEM_BEH_NEVER_FLOATING = 1 << 1,
  BONOBO_DOCK_ITEM_BEH_NEVER_VERTICAL = 1 << 2,
  BONOBO_DOCK_ITEM_BEH_NEVER_HORIZONTAL = 1 << 3,
  BONOBO_DOCK_ITEM_BEH_LOCKED = 1 << 4
  } BonoboDockItemBehavior;
#endif

  /* A floating dock can't be selected for re-attachment;
  ** this is almost certainly a bug in libbonoboui-2.10.1,
  ** so for now keep the dock mobile but attached.
  ** The alternative would be to rewrite the whole thing
  ** with GtkHandleBox(); floating toolbars aren't worth
  ** the bother. ###
  */
  gnome_app_add_toolbar(
		GNOME_APP(gg->gwindow),
		GTK_TOOLBAR(gg->toolbar),
		"Toolbar",
		BONOBO_DOCK_ITEM_BEH_EXCLUSIVE |
		BONOBO_DOCK_ITEM_BEH_NEVER_VERTICAL |
#if 0
		BONOBO_DOCK_ITEM_BEH_NORMAL,
#else
		BONOBO_DOCK_ITEM_BEH_NEVER_FLOATING,
#endif
		BONOBO_DOCK_TOP,
		1,		/* Menu is band 0 */
		0,
		0
		);

  /* Make a note of the dock widget (a BonoboDockItem *) */
  gg->tooldock = gnome_app_get_dock_item_by_name(
			GNOME_APP(gg->toplevel), GNOME_APP_TOOLBAR_NAME);

  /* The toolbar can be hidden, but not the tooldock.  If the default
  ** is to be no toolbar, it's easier to set up the dock here and then
  ** switch it off when the first edit window is shown.  Otherwise the
  ** GTK+ main window resize code gets pretty confused.
  gtk_widget_hide(GTK_WIDGET(gg->tooldock));
  */

  return status;
}


/*
** s w i t c h _ t o o l b a r
**
** Show or hide the toolbar.
**
** Returns:	Nothing		Always
*/
static void switch_toolbar(void)
{
  GNGUI *gg = &gn_gui;
  GtkRequisition req;
  gint height, width;

  DPRINT_F(("switch_toolbar()\n"));

  if (gg->tooldock != NULL) {
	if (o_showtoolbar)
		gtk_widget_show(GTK_WIDGET(gg->tooldock));
	else
		gtk_widget_hide(GTK_WIDGET(gg->tooldock));
  }

  /* Update the main window */
  gtk_widget_size_request(gg->toolbar, &req);
  gtk_window_get_size((GtkWindow *) gg->toplevel, &width, &height);
  if (o_showtoolbar)
	height += req.height;
  else
	height -= req.height;
  /* resize_handler() mustn't change the vterm size */
  gg->main_height = 0;
  gg->main_width = 0;
  /* Resize the main window accordingly */
  gtk_window_resize((GtkWindow *) gg->toplevel, width, height);

  return;
}


/*
** s n o o p _ h a n d l e r
**
** Handler for key snooper.
**
** Returns:	TRUE		Key handled.
**		FALSE		Key still needs handling.
**
** Used to indicate the CAPSLOCK key state, and for the F1 help key.
** Multiple concurrent yelp() calls seem to bring down Nautilus (2.16.3),
** so a lock has been added for exclusion.
*/
static gint snoop_handler(GtkWidget *widget, GdkEventKey *event, gpointer data)
{
  GNGUI *gg = &gn_gui;
  static LOCK snoop_lock = LOCK_INIT;
  GError *error;
  int status = 0;
  static char *pon, *poff;

  DPRINT_V(("snoop_handler()\n"));

  /* Initialise these just once, since if we start as root we stay that way */
  if (pon == NULL) {
	pon = (gg->uid == 0) ? "elvis-rdbl" : "elvis-gnbl";
	poff = (gg->uid == 0) ? "elvis-rdgn" : "elvis-gngn";
  }

  /* Show the state of the capslock key */
  if (event->keyval == 0xFFE5 && event->type == GDK_KEY_PRESS) {
	if (! (event->state & GDK_LOCK_MASK) && ! gg->caps_set) {
		gtk_image_set_from_stock(GTK_IMAGE(gg->indicators), pon,
							GTK_ICON_SIZE_MENU);
		gg->caps_set = TRUE;
	}
	else if ((event->state & GDK_LOCK_MASK) && gg->caps_set) {
		gtk_image_set_from_stock(GTK_IMAGE(gg->indicators), poff,
							GTK_ICON_SIZE_MENU);
		gg->caps_set = FALSE;
	}
  }

  /* Use the F1 key to call for help */
  if (event->keyval == 0xFFBE && event->state == 0 &&
				event->type == GDK_KEY_PRESS) {
	status = 1;	/* Even if we can't get the mutex */
	if (TRYLOCK(snoop_lock)) {
		error = NULL;
		(void) gnome_help_display(
				YELP_ELVIS,
				NULL,
				&error
				);
		if (error != NULL) {
			msg(MSG_ERROR, error->message);
			msgflush();
			free(error);
		}
		UNLOCK(snoop_lock);
	}
  }

  return status;
}

#ifdef ADD_UNUSED
/*
** p o p u p _ h a n d l e r
**
** popup signal handler.
**
** Returns:	TRUE		Event handled.
**		FALSE		Event still needs handling.
*/
static gint popup_handler(GtkWidget *widget, GdkEvent *event, gpointer data)
{
  GdkEventButton *eb;
  int status = FALSE;

  DPRINT_F(("popup_handler()\n"));

  /* Assert macros */
  g_return_val_if_fail( widget != NULL, FALSE);
  g_return_val_if_fail( GTK_IS_MENU(widget), FALSE);
  g_return_val_if_fail( event != NULL, FALSE);

  /* The "widget" is the data that was supplied when
  ** g_signal_connect_swapped() was called, in this case
  ** mousemenu.  The event is the instance to which
  ** the handler is connected.
  */

  /* Button events:
  **
  ** GDK_BUTTON_PRESS
  ** GDK_2BUTTON_PRESS
  ** GDK_3BUTTON_PRESS
  ** GDK_BUTTON_RELEASE
  */

  if (event->type == GDK_BUTTON_PRESS) {
	eb = (GdkEventButton *) event;
	/* Button 3 click brings up the popup menu */
	if (eb->button == 3) {
		gtk_menu_popup(
			GTK_MENU(widget),
			NULL,
			NULL,
			NULL,
			NULL,
			eb->button,
			eb->time
			);
		status = TRUE;
	}
  }

  return status;
}
#endif

#ifdef ADD_UNUSED
/*
** m o t i o n _ h a n d l e r
**
** Handler for motion-notify-event (mouse) signals.
**
** Returns:	TRUE		Event handled.
**		FALSE		Event still needs handling.
*/
static gint motion_handler(GtkWidget *widget, GdkEvent *event, gpointer data)
{
  int done = FALSE;

  DPRINT_V(("motion_handler()\n"));

  if (event->type == GDK_MOTION_NOTIFY) {
	done = TRUE;
  }

  return FALSE;
}
#endif

/*
** s c r o l l _ h a n d l e r
**
** Handler for scroll-event (mouse) signals.
**
** Returns:	TRUE		Event handled.
**		FALSE		Event still needs handling.
*/
static gint scroll_handler(GtkWidget *widget, GdkEvent *event, gpointer data)
{
  GNGUI *gg = (GNGUI *) data;
  EDWIN *edwin = gg->edhead;
  GdkEventScroll *es;
  int done = FALSE;

  DPRINT_V(("scroll_handler()\n"));

  /* Apparently buttons 4 & 5 are already translated into SCROLL events.
  ** Check scrolling limits to prevent overrun. ???
  */
  if (event->type == GDK_SCROLL) {
	es = (GdkEventScroll *) event;
	if (es->direction == GDK_SCROLL_UP) {
		if (!o_scrollwheel)
			eventscroll((GUIWIN *) edwin, SCROLL_FWDLN, 1, 0L);
		else
			eventscroll((GUIWIN *) edwin, SCROLL_BACKLN, 1, 0L);
	}
	if (es->direction == GDK_SCROLL_DOWN) {
		if (!o_scrollwheel)
			eventscroll((GUIWIN *) edwin, SCROLL_BACKLN, 1, 0L);
		else
			eventscroll((GUIWIN *) edwin, SCROLL_FWDLN, 1, 0L);
	}
	/* Update the window */
	if (gg->edhead != NULL) (void) eventdraw(gg->edhead);
	done = TRUE;
  }

  return done;
}


/*
** m o u s e _ h a n d l e r
**
** Handler for mouse button and movement signals.
**
** Returns:	TRUE		Event handled.
**		FALSE		Event still needs handling.
**
** The data argument is the object to which the signal handler was
** attached, gg->edhead->vterm.
*/
static gint mouse_handler(GtkWidget *widget, GdkEvent *event, gpointer data)
{
  GNGUI *gg = &gn_gui;
  EDWIN *edwin = gg->edhead;
  GdkEventButton *eb;
  GdkEventMotion *em;
  glong col, row;
  long rc;
  int status;
  static LOCK mouse_lock = LOCK_INIT;
  struct Position {
	glong cursor_col;
	glong cursor_row;
	gdouble mouse_x;
	gdouble mouse_y;
	glong mouse_col;
	glong mouse_row;
	gint valid;
  };
  static struct Position posn[3] = { { 0 } }; /* BEGIN, CURRENT, END */
  static int selected = NO;	/* Characters selected: NO, PARTIAL, YES */
  static int marked = NO;	/* Block marked: NO, YES */
  static int moved = FALSE;	/* Pointer moved */
  static int displayed = TRUE;	/* Screen displayed (correctly) */
#ifdef DEBUG_MOUSE
  static int printed = NO;	/* Debug data printed: NO, MAYBE, YES */
  gint j;
#endif

  /* Elvis uses the following mouse commands (I think):
  **
  ** CLICK_CANCEL: row = col = -1, cancel selection.
  ** CLICK_MOVE: move cursor to pointer.
  ** CLICK_NONE: no action.
  ** CLICK_PASTE: row = col = 1, paste selection from clipboard to cursor.
  ** CLICK_SELCHAR: select characters from mark to cursor.
  ** CLICK_SELLINE: select lines from mark to cursor.
  ** CLICK_SELRECT: select rectangle from mark to cursor.
  ** CLICK_SSCHAR: row = col = -1, start selecting characters from cursor.
  ** CLICK_SSLINE: row = col = -1, start selecting lines from cursor.
  ** CLICK_TAG: row = col = -1, go to tag.
  ** CLICK_UNTAG: row = col = -1, go to previous tag.
  ** CLICK_YANK: row = col = -1, copy selection to clipboard.
  **
  ** eventclick() returns a long, -1, 0, or the offset (of the cursor ???)
  */

  /* The "widget" is the data that was supplied when
  ** g_signal_connect_swapped() was called, in this case
  ** mousemenu.  The event is the instance to which
  ** the handler is connected.
  */

  /* Assert macros */
  g_return_val_if_fail( widget != NULL, FALSE);
  g_return_val_if_fail( GTK_IS_MENU(widget), FALSE);
  g_return_val_if_fail( event != NULL, FALSE);

  /* Try to get a lock, and quit on failure */
  if (!TRYLOCK(mouse_lock)) {
	DPRINT_M(("m_h(): lock in use\n"));
	return FALSE;
  }

  /* Get the block selected state, which may have changed since
  ** the last call to this function.
  */
  selected = gg->block_sel;
  if (selected == NO && marked == YES) {
	marked = NO;
	if (gg->menu_state & MENU_MARK) {
		gg->menu_state &= ~MENU_MARK;
		config_menus(gg);
	}
  }

#ifdef DEBUG_MOUSE
  /* Re-mark on significant changes only */
  if (printed == NO) {
	if (event->type == GDK_MOTION_NOTIFY)
		printed = MAYBE;
  }
  else if (printed == MAYBE) {
	if (event->type == GDK_MOTION_NOTIFY)
		printed = YES;
	else
		printed = NO;
  }
  else if (printed == YES) {
	if (event->type != GDK_MOTION_NOTIFY)
		printed = NO;
  }
  /* Show the event on any transition into NO or MAYBE */
  if (printed != YES) {
	DPRINT_M(("mouse_handler(): event %s\n",
		event->type == GDK_BUTTON_PRESS ? "press" :
		(event->type == GDK_2BUTTON_PRESS ? "2press" :
		(event->type == GDK_3BUTTON_PRESS ? "3press" :
		(event->type == GDK_BUTTON_RELEASE ? "release" : "motion")))));
  }
#endif

  /* Pointer has moved: update position and selection data */
  eb = NULL;
  if (event->type == GDK_MOTION_NOTIFY) {
	em = (GdkEventMotion *) event;
	moved = FALSE;
	if (em->x != posn[CURRENT].mouse_x || em->y != posn[CURRENT].mouse_y) {
		posn[CURRENT].mouse_x = em->x;
		posn[CURRENT].mouse_y = em->y;
		posn[CURRENT].valid = FALSE;
		moved = TRUE;
		/* Update selection data */
		if (selected == PARTIAL) {
			posn[CURRENT].mouse_row = (glong) (em->y /
							gg->pango_height);
			posn[CURRENT].mouse_col = (glong) (em->x /
							gg->pango_width);
			rc = 0;
			/* Move cursor to pointer */
			if ((rc = eventclick((GUIWIN *) edwin,
						posn[CURRENT].mouse_row,
						posn[CURRENT].mouse_col,
						CLICK_MOVE)) < 0) {
				DPRINT_M(("m_h(): err B %ld\n", rc));
				/* Cancel the failed selection */
				posn[BEGIN].valid = posn[END].valid = FALSE;
				if ((rc = eventclick((GUIWIN *) edwin,
						-1,
						-1,
						CLICK_CANCEL)) < 0) {
					DPRINT_M(("m_h(): err C %ld\n", rc));
					/* MT */ ;
				}
				selected = NO;
				moved = FALSE ;
			}
			displayed = FALSE;
		}
	}
  }

  /* Button events:
  **
  ** GDK_BUTTON_PRESS
  ** GDK_2BUTTON_PRESS
  ** GDK_3BUTTON_PRESS
  ** GDK_BUTTON_RELEASE
  */

  /* Mouse button state has changed: do something */
  if (event->type == GDK_BUTTON_PRESS ||
		event->type == GDK_2BUTTON_PRESS ||
		event->type == GDK_BUTTON_RELEASE) {
	eb = (GdkEventButton *) event;
	/* Get the cursor and pointer positions regardless */
	vte_terminal_get_cursor_position(
					VTE_TERMINAL(edwin->vterm),
					&col,
					&row
					);
	/* Convert mouse coords to characters using current font size */
	posn[CURRENT].cursor_col = col;
	posn[CURRENT].cursor_row = row;
	posn[CURRENT].mouse_x = eb->x;
	posn[CURRENT].mouse_y = eb->y;
	posn[CURRENT].mouse_col = (glong) (eb->x / gg->pango_width);
	posn[CURRENT].mouse_row = (glong) (eb->y / gg->pango_height);
	posn[CURRENT].valid = TRUE;

#ifdef DPRINT_MM
	{ /* Debugging block begin */
	int j;
	for (j = 0 ; j < 3 ; j++) {
		g_print("%s: col %ld row %ld event %f %f col %ld row %ld %s\n",
			j == BEGIN ? "BEG" : (j == CURRENT ? "CUR" : "END"),
			posn[j].cursor_col, posn[j].cursor_row,
			posn[j].mouse_x, posn[j].mouse_y,
			posn[j].mouse_col, posn[j].mouse_row,
			posn[j].valid ? "valid" : "invalid"
			);
	}
	} /* Debugging block end */
#endif

	/* Whatever the selection state, button 1 press updates the cursor.
	** If no selection is current, button 1 press starts a char selection.
	** If a selection is current, button 1 press removes any selection.
	** If a selection is in progress, button 1 release ends char selection.
	**
	** What is the difference between selecting a single character
	** and updating the cursor position. ???
	**
	** If no selection is current, button 2 press pastes from the clipboard.
	** If a selection is current, button 2 press does nothing.
	**
	** Button 3 press brings up the mouse menu.
	**
	** Double clicks are reserved for specialised marking and lookup.
	**
	** Button 1 double click runs the tag command.
	*/

	/* Consider button 1 */
	if (eb->button == 1) {
		if (event->type == GDK_BUTTON_PRESS) {
			/* Move the cursor to the pointer position */
			if ((rc = eventclick((GUIWIN *) edwin,
					posn[CURRENT].mouse_row,
					posn[CURRENT].mouse_col,
					CLICK_MOVE)) < 0) {
				DPRINT_M(("m_h(): err A %ld\n", rc));
				displayed = TRUE;
			}
			else
				displayed = FALSE;
		}

		if (event->type == GDK_2BUTTON_PRESS) {
			posn[END] = posn[CURRENT];
			/* Run the tag (^]) command.  Gnome doesn't
			** register a double click until after a single
			** click has been processed, so the cursor is
			** already in the correct position.
			*/
			(void) eventclick((GUIWIN *) edwin,
					-1,
					-1,
					CLICK_TAG);
			displayed = FALSE;
		}

		if (selected == NO && event->type == GDK_BUTTON_PRESS) {
			/* Start a character-based selection */
			posn[BEGIN] = posn[CURRENT];
			if ((rc = eventclick((GUIWIN *) edwin,
					posn[BEGIN].mouse_row,
					posn[BEGIN].mouse_col,
					CLICK_SELCHAR)) < 0) {
				DPRINT_M(("m_h(): err D %ld\n", rc));
				posn[BEGIN].valid = posn[CURRENT].valid = FALSE;
				displayed = TRUE;
			}
			else {
				selected = PARTIAL;
				displayed = FALSE;
			}
			DPRINT_M(("m_h(): start select\n"));
		}
		else if (selected == PARTIAL &&
					event->type == GDK_BUTTON_RELEASE) {
			/* If the pointer has moved, it's a selection */
			if (posn[CURRENT].mouse_row != posn[BEGIN].mouse_row ||
					posn[CURRENT].mouse_col !=
					posn[BEGIN].mouse_col) {
				/* End of selection, copy to the clipboard */
				posn[END] = posn[CURRENT];
				if ((rc = eventclick(
						(GUIWIN *) edwin,
						posn[END].mouse_row,
						posn[END].mouse_col,
						CLICK_YANK)) < 0) {
					DPRINT_M(("m_h(): err E %ld\n", rc));
					/* Cancel the failed selection */
					if ((rc = eventclick((GUIWIN *) edwin,
							-1,
							-1,
							CLICK_CANCEL)) < 0) {
						DPRINT_M(("m_h(): err F %ld\n",
									rc));
					}
					selected = NO;
				}
				else
					selected = YES;
				displayed = FALSE;
				DPRINT_M(("m_h(): partial select yes\n"));
			}
			else {
				/* No movement means no selection */
				posn[BEGIN].valid = posn[END].valid = FALSE;
				if ((rc = eventclick((GUIWIN *) edwin,
						-1,
						-1,
						CLICK_CANCEL)) < 0) {
					DPRINT_M(("m_h(): err G %ld\n", rc));
					/* MT */ ;
				}
				selected = NO;
				displayed = TRUE;
				DPRINT_M(("m_h(): partial select no\n"));
			}
		}
		else if (selected == YES && event->type == GDK_BUTTON_PRESS) {
			/* Remove the selection */
			posn[BEGIN].valid = posn[END].valid = FALSE;
			if ((rc = eventclick((GUIWIN *) edwin,
					-1,
					-1,
					CLICK_CANCEL)) < 0) {
				DPRINT_M(("m_h(): err H %ld\n", rc));
				/* MT */ ;
			}
			selected = NO;
			displayed = FALSE;
			DPRINT_M(("m_h(): clear select press\n"));
		}
	}

	/* Consider button 2 */
	if (eb->button == 2) {
		if (selected == NO && event->type == GDK_BUTTON_PRESS) {
			/* Paste from the clipboard at the cursor position.
			** Which clipboard is this using ??? ###
			*/
			eventclick(gg->edhead, -1, -1, CLICK_PASTE);
			displayed = FALSE;
		}
		else if (selected == YES && event->type == GDK_BUTTON_PRESS) {
			/* Should we do nothing here, or move the cursor,
			** or remove the current selection, or twiddle our
			** thumbs ???
			*/
#if 0
			/* Move the cursor to the pointer position */
			if ((rc = eventclick((GUIWIN *) edwin,
					posn[CURRENT].mouse_row,
					posn[CURRENT].mouse_col,
					CLICK_MOVE)) < 0) {
				DPRINT_M(("m_h(): err K %ld\n", rc));
				displayed = TRUE;
			}
			else
				displayed = FALSE;
#endif
#if 0
			/* Remove the current selection */
			posn[BEGIN].valid = posn[END].valid = FALSE;
			if ((rc = eventclick((GUIWIN *) edwin,
					-1,
					-1,
					CLICK_CANCEL)) < 0) {
				DPRINT_M(("m_h(): err K %ld\n", rc));
				/* MT */ ;
			}
			selected = NO;
			displayed = FALSE;
#endif
		}
	}

	/* Consider button 3 */
	if (eb->button == 3 && event->type == GDK_BUTTON_PRESS) {
		/* Display the mouse popup menu */
		gtk_menu_popup(
			GTK_MENU(widget),
			NULL,
			NULL,
			NULL,
			NULL,
			eb->button,
			eb->time
			);
	}
  }

  /* Update the screen if necessary */
  if (!displayed) {
	if (gg->edhead != NULL) (void) eventdraw(gg->edhead);
	displayed = TRUE;
  }

  /* Update the main and mouse menus when a block is selected */
  if (selected == NO) {
	if (marked == YES) {
		if (gg->menu_state & MENU_MARK) {
			marked = NO;
			gg->menu_state &= ~MENU_MARK;
			config_menus(gg);
		}
	}
  }
  else if (selected == YES) {
	if (marked == NO) {
		if (!(gg->menu_state & MENU_MARK)) {
			marked = YES;
			gg->menu_state |= MENU_MARK;
			config_menus(gg);
		}
	}
  }

  /* Update the global block flag */
  gg->block_sel = selected;

  /* Release the lock.  Can this protection be reduced ??? */
  UNLOCK(mouse_lock);

  if (event->type == GDK_MOTION_NOTIFY || eb->button != 3)
	status = TRUE;
  else
	status = FALSE;

  /* FALSE to continue propagation, TRUE if we've handled it */
  return status;
}


/*
** r a n g e _ h a n d l e r
**
** Handler for range change-value signals.
**
** Returns:	TRUE		Event handled.
**		FALSE		Event still needs handling.
*/
static gboolean range_handler(GtkRange *range, GtkScrollType scroll,
					gdouble value, gpointer data)
{
  GNGUI *gg = (GNGUI *) data;
  double d, e;

  DPRINT_F(("range_handler()\n"));

  /* This restores the functionality of the lower vertical stepper,
  ** but it can't handle the full sensitivity, which should be two
  ** step_increments, not one. ??? ###
  */
  if (scroll == GTK_SCROLL_STEP_FORWARD) {
	/* Setting gg->ver_adj->value to d moves the scrollbar, but
	** neither emitting the value-changed signal nor calling the
	** v_val_chg_handler() then updates the main window. ???
	v_val_chg_handler(gg->vert_adj, gg);
	g_print("range_handler(): new value %f\n", d);
	*/
	d = gg->vert_adj->value + gg->vert_adj->step_increment;
	e = gg->vert_adj->upper - gg->vert_adj->page_size;
	if (d > e) d = e;
	gg->vert_adj->value = d;
  }

#ifdef DEBUG_MENU
  { /* Debugging block begin */
  char *pc;
  switch (scroll) {
  case GTK_SCROLL_NONE: pc = "GTK_SCROLL_NONE"; break;
  case GTK_SCROLL_JUMP: pc = "GTK_SCROLL_JUMP"; break;
  case GTK_SCROLL_STEP_BACKWARD: pc = "GTK_SCROLL_STEP_BACKWARD"; break;
  case GTK_SCROLL_STEP_FORWARD: pc = "GTK_SCROLL_STEP_FORWARD"; break;
  case GTK_SCROLL_PAGE_BACKWARD: pc = "GTK_SCROLL_PAGE_BACKWARD"; break;
  case GTK_SCROLL_PAGE_FORWARD: pc = "GTK_SCROLL_PAGE_FORWARD"; break;
  case GTK_SCROLL_STEP_UP: pc = "GTK_SCROLL_STEP_UP"; break;
  case GTK_SCROLL_STEP_DOWN: pc = "GTK_SCROLL_STEP_DOWN"; break;
  case GTK_SCROLL_PAGE_UP: pc = "GTK_SCROLL_PAGE_UP"; break;
  case GTK_SCROLL_PAGE_DOWN: pc = "GTK_SCROLL_PAGE_DOWN"; break;
  case GTK_SCROLL_STEP_LEFT: pc = "GTK_SCROLL_STEP_LEFT"; break;
  case GTK_SCROLL_STEP_RIGHT: pc = "GTK_SCROLL_STEP_RIGHT"; break;
  case GTK_SCROLL_PAGE_LEFT: pc = "GTK_SCROLL_PAGE_LEFT"; break;
  case GTK_SCROLL_PAGE_RIGHT: pc = "GTK_SCROLL_PAGE_RIGHT"; break;
  case GTK_SCROLL_START: pc = "GTK_SCROLL_START"; break;
  case GTK_SCROLL_END: pc = "GTK_SCROLL_END"; break;
  default: pc = "UNKNOWN"; break;
  }
  DPRINT_F(("range_handler(): action %s value %f\n", pc, value));
  } /* Debugging block end */
#endif

  return FALSE;
  }


/*
** t h e m e _ h a n d l e r
**
** Handler for theme changed signal.
**
** Returns:	Nothing		Always
*/
static void theme_handler(GtkIconTheme *theme, gpointer data)
{
  GNGUI *gg = (GNGUI *) data;
  GtkSettings *settings;
  char *pc;

  DPRINT_F(("theme_handler()\n"));
#if 0
{
  GtkIconTheme *pt;
  /* Note the current theme */
  pt = gtk_icon_theme_get_default();

  g_print("theme_handler(): current 0x%08lx theme 0x%08lx default 0x%08lx\n",
			(ULONG) gg->theme, (ULONG) theme, (ULONG) pt);

  gg->theme = pt;
  }
#endif

  /* Get the current icon theme name */
  settings = gtk_widget_get_settings(GTK_WIDGET(gg->gwindow));
  pc = NULL;
  g_object_get((GObject *) settings, "gtk-icon-theme-name", &pc, NULL);
  if (pc != NULL) {
	strncpy(gg->theme_name, pc, F_MAX);
	g_free(pc);
  }
  else
	strcpy(gg->theme_name, "hicolor");

  /* Update the toolbar Tools menu icon */
  get_tools_icon();

  return;
}


/*
** g c o n f _ n o t i f y _ h a n d l e r
**
** Handler for GConfClient change notification.
**
** Returns:	Nothing		Always
*/
static void gconf_notify_handler(
			GConfClient *client,
			guint conn_id,
			GConfEntry *entry,
			gpointer data
			)
{
  GNGUI *gg = (GNGUI *) data;
  GError *gconf_err = NULL;
  gboolean flg;
  gint value;
  gchar *str, buff[64];
  const char *key;
  int status = OK;

  DPRINT_F(("gconf_notify_handler()\n"));

  key = gconf_entry_get_key(entry);

#if 0
  { /* Debugging block begin */
  struct Gconfkey *pg;
  int found = FALSE;
  for (pg = gconfkeys ; pg->key != NULL ; pg++) {
	if (strcmp(pg->key, key) == SAME) {
		found = TRUE;
		break;
	}
  }
  g_print("gconf_notify_handler(): GConfEntry key |%s| %s\n",
				key, found ? "found" : "not found");
  } /* Debugging block end */
#endif

  /* The startup position is the only GConf setting that really
  ** needs to be monitored once the program is running.
  */
  if (strcmp(key, GCONF_PATH "/print/printmenu") == SAME) {
	/* Returns FALSE on error, which is not the default */
	flg = gconf_client_get_bool(gg->gconf_client, key, &gconf_err);
	if (gconf_err == NULL) {
		/* Update the local database */
		if (o_printmenu != flg) {
			o_printmenu = flg;
			/* Update the menu entry */
			gtk_widget_set_sensitive(
					gtk_ui_manager_get_widget(
					gg->mainuimanager,
					"/ui/menubar/PrintMenu"),
					(gboolean) o_printmenu);
		}
	}
	else {
		g_error_free(gconf_err);
		status = FAILED;
	}
  }
  else if (strcmp(key, GCONF_PATH "/ui/dictionary") == SAME) {
	/* Returns NULL on error, which is not the default */
	str = gconf_client_get_string(gg->gconf_client, key, &gconf_err);
	if (gconf_err == NULL) {
		/* Update only on success */
		if (g_find_program_in_path(str) == NULL) {
			msg(MSG_ERROR,
				"[s]Could not find dictionary program $1", str);
			msgflush();
		}
		else {
			strncpy(gg->dictionary, str, sizeof(gg->dictionary)-1);
			o_dictionary = gg->dictionary;
			gtk_widget_set_sensitive(
					gtk_ui_manager_get_widget(
					gg->mainuimanager,
					"/ui/menubar/ToolsMenu/Dictionary"),
					TRUE);
		}
	}
	else {
		g_error_free(gconf_err);
		status = FAILED;
	}
  }
  else if (strcmp(key, GCONF_PATH "/ui/docscount") == SAME) {
	/* Returns 0 on error, which is not the default */
	value = gconf_client_get_int(gg->gconf_client, key, &gconf_err);
	if (gconf_err == NULL) {
		sprintf(buff, "--docscount=%d", value);
		status = (*parsers[O_DOCS].parse)(
				buff,
				O_DOCS,
				parsers[O_DOCS].idx
				);
	}
	else {
		g_error_free(gconf_err);
		status = FAILED;
	}
  }
  else if (strcmp(key, GCONF_PATH "/ui/hidemouse") == SAME) {
	/* Returns FALSE on error, which is the default */
	flg = gconf_client_get_bool(gg->gconf_client, key, &gconf_err);
	if (gconf_err == NULL) {
		o_hidemouse = flg;
		vte_terminal_set_mouse_autohide(
			VTE_TERMINAL(gg->edhead->vterm), (int) o_hidemouse);
	}
	else {
		g_error_free(gconf_err);
		status = FAILED;
	}
  }
  else if (strcmp(key, GCONF_PATH "/ui/manmode") == SAME) {
	/* Returns FALSE on error, which is not the default */
	flg = gconf_client_get_bool(gg->gconf_client, key, &gconf_err);

	DPRINT_G(("gconf_notify_handler(): error is %s flg is %s\n",
			gconf_err == NULL ? "NULL" : "not NULL",
			flg ? "TRUE" : "FALSE"));

	if (gconf_err == NULL) {
		/* Update the local database */
		if (o_manmode != flg) {
			o_manmode = flg;
			/* Update the menu entry */
			switch_manpage_menu();
		}
	}
	else {
		g_error_free(gconf_err);
		status = FAILED;
	}
  }
  else if (strcmp(key, GCONF_PATH "/ui/newmode") == SAME) {
	/* Returns FALSE on error, which is not the default */
	flg = gconf_client_get_bool(gg->gconf_client, key, &gconf_err);
	if (gconf_err == NULL) {
		/* Update the local database */
		if (o_newmode != flg) {
			o_newmode = flg;
			/* Update the menu entry */
			switch_docs_open_menu();
		}
	}
	else {
		g_error_free(gconf_err);
		status = FAILED;
	}
  }
  else if (strcmp(key, GCONF_PATH "/ui/scrollwheel") == SAME) {
	/* Returns FALSE on error, which is not the default */
	flg = gconf_client_get_bool(gg->gconf_client, key, &gconf_err);
	if (gconf_err == NULL) {
		o_scrollwheel = flg;
	}
	else {
		g_error_free(gconf_err);
		status = FAILED;
	}
  }
  else if (strcmp(key, GCONF_PATH "/ui/shellcols") == SAME) {
	/* Returns 0 on error, which is not the default */
	value = gconf_client_get_int(gg->gconf_client, key, &gconf_err);
	if (gconf_err == NULL) {
		sprintf(buff, "--shellcols=%d", value);
		status = (*parsers[O_SHCOLS].parse)(
				buff,
				O_SHCOLS,
				parsers[O_SHCOLS].idx
				);
	}
	else {
		g_error_free(gconf_err);
		status = FAILED;
	}
  }
  else if (strcmp(key, GCONF_PATH "/ui/shellrows") == SAME) {
	/* Returns 0 on error, which is not the default */
	value = gconf_client_get_int(gg->gconf_client, key, &gconf_err);
	if (gconf_err == NULL) {
		sprintf(buff, "--shellrows=%d", value);
		status = (*parsers[O_SHROWS].parse)(
				buff,
				O_SHROWS,
				parsers[O_SHROWS].idx
				);
	}
	else {
		g_error_free(gconf_err);
		status = FAILED;
	}
  }
  else if (strcmp(key, GCONF_PATH "/ui/shellmenu") == SAME) {
	/* Returns FALSE on error, which is the default */
	flg = gconf_client_get_bool(gg->gconf_client, key, &gconf_err);
	if (gconf_err == NULL) {
		o_shellmenu = flg;
	}
	else {
		g_error_free(gconf_err);
		status = FAILED;
	}
  }
  else if (strcmp(key, GCONF_PATH "/ui/showtoolbar") == SAME) {
	/* Returns FALSE on error, which is not the default */
	flg = gconf_client_get_bool(gg->gconf_client, key, &gconf_err);
	if (gconf_err == NULL) {
		/* Update the local database */
		if (o_showtoolbar != flg) {
			o_showtoolbar = flg;
			/* Update the menu entry */
			switch_toolbar_menu();
			/* Switch the toolbar in or out of the main window */
			switch_toolbar();
		}
	}
	else {
		g_error_free(gconf_err);
		status = FAILED;
	}
  }
  else if (strcmp(key, GCONF_PATH "/ui/spellcheck") == SAME) {
	/* Returns NULL on error, which is not the default */
	str = gconf_client_get_string(gg->gconf_client, key, &gconf_err);
	if (gconf_err == NULL) {
		/* Update only on success */
		if (g_find_program_in_path(str) == NULL) {
			msg(MSG_ERROR,
				"[s]Could not find spelling checker $1", str);
			msgflush();
		}
		else {
			strncpy(gg->spellcheck, str, sizeof(gg->spellcheck)-1);
			o_spellcheck = gg->spellcheck;
			gtk_widget_show(gtk_ui_manager_get_widget(
					gg->mainuimanager,
					"/ui/menubar/ToolsMenu/Spellcheck"));
		}
	}
	else {
		g_error_free(gconf_err);
		status = FAILED;
	}
  }
  else if (strcmp(key, GCONF_PATH "/ui/startposition") == SAME) {
	/* Returns FALSE on error, which is the default */
	flg = gconf_client_get_bool(gg->gconf_client, key, &gconf_err);
	if (gconf_err == NULL) {
		/* Update the local database */
		if (o_startposition != flg) {
			o_startposition = flg;
			/* Update the menu entry */
			switch_startpos_menu();
			/* Save the position if required */
			if (o_startposition)
				store_settings(TRUE);
		}
	}
	else {
		g_error_free(gconf_err);
		status = FAILED;
	}
  }

  /* Having got a failure code, what do we do with it ??? */
  if (status == FAILED) {
	msg(MSG_ERROR, "[s]Could not parse GConf key $1", key);
	msgflush();
  }

  return;
}

#ifdef SESSION_MGR
/*
** c l i e n t _ s a v e _ h a n d l e r
**
** Handler for session manager save-yourself signal.
**
** Returns:	TRUE		System saved!!!
**		FALSE		Uh-oh
**
** Most parameters can be ignored.
*/
static int client_save_handler(
		GnomeClient *client,
		int phase,
		GnomeSaveStyle save_style,
		int shutdown,
		GnomeInteractStyle interact_style,
		int fast,
		gpointer data
		)
{
  GNGUI *gg = (GNGUI *) data;
  const char *prefix;
  gchar **argv;
  guint argc;
  int j;

#if 0
  /* Get the config prefix - not used */
  prefix = gnome_client_get_config_prefix(gg->gnome_client);

  /* Discard the old state and save the new state */
#endif

  /* Set commands to clone and restart this application.
  ** Note that we use the same values for both - the
  ** session management code will automatically add
  ** whatever magic option is required to set the session
  ** id on startup. The client_data was set to the
  ** command used to start this application when
  ** save_yourself handler was connected.
  ** Where should argv[] be freed ???
  */
  if ((argv = g_try_malloc(sizeof(gchar *) * 4) != NULL) {
	for (j = 0 ; j < 4 ; j++)
		argv[j] = NULL;
	argc = 1;

	argv[0] = gg->prog_name;

	gnome_client_set_clone_command (client, argc, argv);
	gnome_client_set_restart_command (client, argc, argv);
  }

  return (argv == NULL) ? FALSE : TRUE;
}


/*
** c l i e n t _ d i e _ h a n d l e r
**
** Handler for session manager die signal.
**
** Returns:	Nothing		Always
*/
static gint client_die_handler(GnomeClient *client, gpointer data)
{
  GNGUI *gg = (GNGUI *) data;

  /* Shut down everything in sight */
  msgflush();
  eventex(gg->edhead, "quit!", ElvFalse);

  return TRUE;
}
#endif

/*
** d r a g _ d a t a _ h a n d l e r
**
** Handler for drag-data-get signal (outgoing data).
**
** Returns:	Nothing		Always
**
** The drag-data-get signal is emitted when the destination requests
** data from the source via gtk_drag_get_data().
** The source should either supply the data, or an error string of a
** "safe" (text) default type, via a call to gtk_selection_data_set().
** This sets up the selection_data object and passes it to the
** destination, which handles it via the drag-data-received signal.
*/
static void drag_data_handler(
			GtkWidget *widget,
			GdkDragContext *drag_context,
			GtkSelectionData *data,
			guint info,
			guint time,
			gpointer user_data)
{
  GNGUI *gg = &gn_gui;
  GdkAtom txt;
  guchar *err_str = "Drag error: unknown target type";

  DPRINT_F(("drag_data_handler()\n"));

  txt = gdk_atom_intern("text/plain", TRUE);
  if (data->target != txt) {
	DPRINT_D(("drag_data_handler(): bad type\n"));
	data->target = txt;
	gtk_selection_data_set(
			data,		/* Allocated object */
			data->target,	/* target type */
			8,		/* number of bits per unit */
			err_str,	/* pointer to data to be sent */
			(gint) strlen(err_str)	/* length of data in units */
			);
  }
  else {
	DPRINT_D(("drag_data_handler(): good type\n"));

	/* Hand the task to Elvis */
	eventclick(gg->edhead, -1, -1, CLICK_YANK);

	/* Tell the client we're ready */
	gtk_selection_data_set(
			data,		/* Allocated object */
			data->target,	/* target type */
			8,		/* number of bits per 'unit' */
			gg->drop_text,	/* pointer to data to be sent */
			(gint) gg->drop_len	/* length of data in units */
			);
  }

  /* NB. drag_end_handler() deals with cleanup */

  return;
}


/*
** d r a g _ b e g i n _ h a n d l e r
**
** Handler for drag-begin signal.
**
** Returns:	Nothing		Always
*/
static void drag_begin_handler(
			GtkWidget *widget,
			GdkDragContext *drag_context,
			gpointer user_data)
{
  /* GNGUI *gg = &gn_gui */

  DPRINT_F(("drag_begin_handler()\n"));

  return;
}


/*
** d r a g _ e n d _ h a n d l e r
**
** Handler for drag-end signal.
**
** Returns:	Nothing		Always
*/
static void drag_end_handler(
			GtkWidget *widget,
			GdkDragContext *drag_context,
			gpointer user_data)
{
  GNGUI *gg = &gn_gui;

  DPRINT_F(("drag_end_handler()\n"));

  if (gg->drop_type == PUT_TXT)
	g_free(gg->drop_text);
  gg->drop_text = NULL;
  gg->drop_type = NONE;
  gg->drop_len = 0;

  return;
}


/*
** d r o p _ d a t a _ h a n d l e r
**
** Handler for drag-data-received signal (incoming data).
**
** Returns:	Nothing		Always
*/
static void drop_data_handler(
			GtkWidget *widget,
			GdkDragContext *drag_context,
			gint x,
			gint y,
			GtkSelectionData *data,
			guint info,
			guint time,
			gpointer user_data)
{
  GNGUI *gg = &gn_gui;
  int valid;
  char *pc;
  char name[PATH_MAX];
  char buff[PATH_MAX + 6];

  DPRINT_F(("drop_data_handler()\n"));

#ifdef DPRINT_DD
  { /* Debugging block begin */
  g_print("drop_data_handler(): found actions");
  if (drag_context->action & GDK_ACTION_DEFAULT)
	g_print(" GDK_ACTION_DEFAULT");
  if (drag_context->action & GDK_ACTION_COPY)
	g_print(" GDK_ACTION_COPY");
  if (drag_context->action & GDK_ACTION_MOVE)
	g_print(" GDK_ACTION_MOVE");
  if (drag_context->action & GDK_ACTION_LINK)
	g_print(" GDK_ACTION_LINK");
  if (drag_context->action & GDK_ACTION_PRIVATE)
	g_print(" GDK_ACTION_PRIVATE");
  if (drag_context->action & GDK_ACTION_ASK)
	g_print(" GDK_ACTION_ASK");
  g_print("\n");
  } /* Debugging block end */
#endif

  /* Apparently some drag sources supply MOVE instead of COPY ??? */
  valid = FALSE;
  if (data->length >= 0 && data->format == 8 &&
			drag_context->action == GDK_ACTION_COPY) {
	/* Set up a file drop; read the file contents, or edit the file.
	** File editing is the reasonable expectation for multiple drops.
	** If multiple files are dropped, should each have a new window ???.
	*/
	if (gg->drop_type == GET_URI) {
#if 0
		/* Don't edit the file, paste its contents */
		if ((fd = open(name, O_RDONLY)) < 0)
			valid = FALSE;
		else {
			gg->drop_text = NULL;
			gg->drop_len = 0;
			gg->drop_fd = fd;
			valid = TRUE;
			/* and treat as a text drop */
		}
#endif
		for (pc = data->data ; *pc != '\0' ; pc++) {
			if (url_to_path(name, pc, sizeof(name)) == OK) {
				/* Ask Elvis to load the new filename */
				DPRINT_D((
				"drop_data_handler(): filename(s) |%s|\n",
				name));
				strcpy(buff, "snext ");
				strcat(buff, name);

				/* Update the arglist */
				msgflush();
				eventex(gg->edhead, buff, ElvFalse);

				/* Force a screen update */
				(void) eventdraw(gg->edhead);
				valid = TRUE;
			}
			/* Move to the start of the next item on the list */
			while (*pc != '\0' && *pc != '\n') pc++;
			if (*pc == '\0') pc--;
		}
	}
	/* Set up a text drop */
	else if (gg->drop_type == GET_TXT) {
		gg->drop_text = data->data;
		gg->drop_len = data->length;
		gg->drop_fd = -1;

		/* Call back into Elvis to start the process */
		eventclick(gg->edhead, -1, -1, CLICK_PASTE);

		/* Force a screen update */
		if (gg->edhead != NULL)
			(void) eventdraw(gg->edhead);

		valid = TRUE;
	}
	/* Huh ? */
	else
		valid = FALSE;

	DPRINT_D(("drop_data_handler(): valid task, data length %d\n",
							data->length));
  }

  /* Tell GTK we're done */
  gtk_drag_finish(drag_context, valid, FALSE, time);

  /* Tidy up */
  gg->drop_type = NONE;
  gg->drop_text = NULL;
  gg->drop_len = 0;
  if (gg->drop_fd >=0)
	close(gg->drop_fd);
  gg->drop_fd = -1;

  return;
}


/*
** d r a g _ d r o p _ h a n d l e r
**
** Handler for drag-drop signal.
**
** Returns:	TRUE		Accept drop
**		FALSE		Otherwise
**
** Notes on drag and drop from the PHP/GTK+ site.
**
** text/uri-list is the standard file listing format for drag and drop.
**
** Specification:
** A list of files, one file per line, each line ending with a \r\n,
** carriage return and newline. File names must be valid, url-encoded URIs.
**
** Examples:
** file://localhost/usr/bin/opera[\r\n]
** file://localhost/c:/programs/myfile.txt[\r\n]
** file://server/data/software.xml[\r\n]
** http://gtk.php.net/apps/[\r\n]
**
** However, many programs don't follow this specification.
** KDE programs like Konqueror, Kate and so on use:
**   file:/path/to/file.
** Only one "/", totally incorrect.
** Nautilus, Gnome's file manager, uses:
**   file:///path/to/file.
** Correct protocol (file://), hostname left out (not really correct,
** but nearly), and the full path with a leading "/".
** The Opera browser uses:
**   file://path/to/file.
** What is that? A relative filename?
** Firefox uses the same specification as Nautilus:
**   file:///path/to/file
** This is nearly correct. However, Firefox doesn't drop files to
** targets accepting text/uri-list, only to text/plain ones.
** Microsoft Explorer is also nearly correct, using:
**   file:///c:/path/to/file.
** However, all these programs seem to _accept_ correct URIs in a drop.
**
** Experimenting, an icon drop from Nautilus gives the following types:
**
** x-special/gnome-icon-list
** text/uri-list
** _NETSCAPE_URL
** UTF8_STRING
** COMPOUND_TEXT
** TEXT
** STRING
** text/plain;charset=utf-8
** text/plain;charset=ISO-8859-1
** text/plain
**
** A text drop from Gedit gives the following types:
** GTK_TEXT_BUFFER_CONTENTS
** UTF8_STRING
** COMPOUND_TEXT
** TEXT
** STRING
** text/plain;charset=utf-8
** text/plain;charset=ISO-8859-1
** text/plain
**
** Dealing with various sources, the best approach seems to be to
** first look for a text/uri-list, and treat the data as a filename,
** and then look for text/plain, and treat the data as a text buffer.
*/
static gboolean drag_drop_handler(
			GtkWidget *widget,
			GdkDragContext *drag_context,
			gint x,
			gint y,
			guint time,
			gpointer user_data)
{
  GNGUI *gg = &gn_gui;
  GdkAtom uri, txt, *pa;
  GList *pg;
  int found;

  DPRINT_F(("drag_drop_handler\n"));

  /* Sequence:
  ** Check that the drop is over a valid part of the widget.
  ** Choose the target type.
  ** Call gtk_drag_get_data to emit drag-data-get on the source.
  */
  uri = gdk_atom_intern("text/uri-list", TRUE);
  txt = gdk_atom_intern("text/plain", TRUE);
  found = 0;
  for (pg = drag_context->targets ; pg != NULL ; pg = pg->next) {
	if (pg->data == uri) {
		found |= 1;
		DPRINT_D(("drag_drop_handler(): found uri\n"));
	}
	else if (pg->data == txt) {
		found |= 2;
		DPRINT_D(("drag_drop_handler(): found txt\n"));
	}
	DPRINT_D((
	"drag_drop_handler(): targets 0x%08lx next 0x%08lx data 0x%08lx |%s|\n",
	(ULONG) pg, (ULONG) pg->next, (ULONG) pg->data, gdk_atom_name(pg->data)
	));
  }

  /* If we recognise a datatype, request it from the source */
  if (found != 0) {
	/* Choose text/uri-list over text/plain */
	if (found & 1) {
		gg->drop_type = GET_URI;
		pa = &uri;
	}
	else {
		gg->drop_type = GET_TXT;
		pa = &txt;
	}
	gtk_drag_get_data(
			widget,		/* receives 'drag-data-received' */
			drag_context,	/* the current state of the DnD */
			*pa,		/* the target type we want */
			time		/* time stamp */
			);
  }

  return found == 0 ? FALSE : TRUE;
}


/*
** e n t r y _ h a n d l e r
**
** Handler for activate signal from entry box.
**
** Returns:	Nothing		Always
**
** The entry box is used for search, replace, and move commands
** and, since it's available, for entering filenames.
*/
static void entry_handler(GtkEditable *entry, gpointer data)
{
  GNGUI *gg = (GNGUI *) data;
  /* It's easier to keep a buffer for each repeated command type */
  static struct Command {
	char buff[64];
	int flg;
  } cmds[3] = {{{ '\0' }, 0 }};
  int state;
  const gchar *text = NULL;
  char *pc, delim[2];
  char buff[PATH_MAX];

  DPRINT_F(("entry_handler()\n"));

  /* Clear the message bar in preparation for any resulting messages */
  gtk_statusbar_push(GTK_STATUSBAR(gg->msgbar), gg->edhead->ms_id, "");

  /* Check for a new command */
  if (entry != NULL) {
	/* Remove the entry widget */
	g_signal_handler_block(G_OBJECT(gg->entry), gg->entry_id);
	gtk_widget_reparent(gg->entry, gg->entrybox);
	gtk_widget_hide(gg->entry);

	/* Replace the message widget */
	gtk_widget_reparent(gg->msgbar, gg->hpaned);
	gtk_widget_show(gg->msgbar);
	gtk_paned_set_position(GTK_PANED(gg->hpaned), STATUS_BAR_SIZE);

	/* Revert the focus to the main window */
	gtk_widget_grab_focus(gg->edhead->vterm);

	/* Release the lock */
	UNLOCK(entry_lock);

	/* Check for a result */
	text = gtk_entry_get_text(
		GTK_ENTRY(gnome_entry_gtk_entry(GNOME_ENTRY(gg->entry))));
	if (text == NULL)
		gg->move_cmd = MOVE_ERR;
  }

  /* Carry out the command */
  if (gg->move_cmd == MOVE_ERR) {
	/* MT */ ;
  }
  else if (gg->move_cmd == FIND_NEXT || gg->move_cmd == FIND_PREV) {
	cmds[F_CMD].flg = FALSE;
	/* Sort out the command delimiter */
	delim[0] = gg->move_cmd == FIND_NEXT ? '/' : '?';
	delim[1] = '\0';
	/* The user may have overcompensated */
	if (*text == '/' || *text == '?') {
		delim[0] = *text;
		buff[0] = '\0';
	}
	else {
		/* Initialise the buffer with the delimiter */
		strcpy(buff, delim);
	}

	/* A leading '/' or '?' must match the selected command */
	if ((delim[0] == '/' && gg->move_cmd != FIND_NEXT) ||
			(delim[0] == '?' && gg->move_cmd != FIND_PREV)) {
		gg->move_cmd = MOVE_ERR;
	}
	else {
		strncat(buff, text, sizeof(buff) - 4);
		strcat(buff, delim);
		strcat(buff, "\n");
		strcpy(cmds[F_CMD].buff, buff);
		cmds[F_CMD].flg = TRUE;
		msgflush();
		/* This is a visual mode command */
		eventkeys(gg->edhead, toCHAR(buff), (int) strlen(buff));
		(void) eventdraw(gg->edhead);
	}
  }
  else if (gg->move_cmd == FIND_AGAIN && cmds[F_CMD].flg) {
	strcpy(buff, cmds[F_CMD].buff);
	msgflush();
	eventkeys(gg->edhead, toCHAR(buff), (int) strlen(buff));
	(void) eventdraw(gg->edhead);
  }
  else if (gg->move_cmd == REPLACE) {
	cmds[R_CMD].flg = FALSE;
	/* Replace commands can be so complicated that it seems best
	** just to pass the whole thing over to Elvis unchanged.
	*/
	strncpy(buff, text, sizeof(buff) - 1);
	strcpy(cmds[R_CMD].buff, buff);
	cmds[R_CMD].flg = TRUE;
	msgflush();
	eventex(gg->edhead, buff, ElvFalse);
	(void) eventdraw(gg->edhead);
  }
  else if (gg->move_cmd == REPLACE_AGAIN && cmds[R_CMD].flg) {
	strcpy(buff, cmds[R_CMD].buff);
	msgflush();
	eventex(gg->edhead, buff, ElvFalse);
	(void) eventdraw(gg->edhead);
  }
  else if (gg->move_cmd == GOTO) {
	cmds[G_CMD].flg = FALSE;
	strncpy(buff, text, sizeof(buff) - sizeof(" goto") - 1);

	/* The goto command is approximately [$][+-][0-9*] (I think) */
	state = 0;
	pc = buff;
	while (state < 4) {
		switch (*pc++) {
		case '$':	state = (state != 0) ? 10 : 1 ; break;
		case '+':
		case '-':	state = (state > 1) ? 10 : 2 ; break;
		case '0': case '1': case '2': case '3': case '4':
		case '5': case '6': case '7': case '8': case '9':
				state = (state > 3) ? 10 : 3 ; break;
		case '\0':	state = (state < 1) ? 10 : 4 ; break;
		default:	state = 10 ; break;
		}
	}
	if (state == 10)
		gg->move_cmd = MOVE_ERR;
	else {
		strcat(buff, " goto");
		strcpy(cmds[G_CMD].buff, buff);
		cmds[G_CMD].flg = TRUE;
		msgflush();
		eventex(gg->edhead, buff, ElvFalse);
		(void) eventdraw(gg->edhead);
	}
  }
  else if (gg->move_cmd == GOTO_AGAIN && cmds[G_CMD].flg) {
	strcpy(buff, cmds[G_CMD].buff);
	msgflush();
	eventex(gg->edhead, buff, ElvFalse);
	(void) eventdraw(gg->edhead);
  }
  /* The file open function just shares the text entry widget */
  else if (gg->move_cmd == FILE_NAME) {
	strcpy(buff, "snext ");
	strncat(buff, text, sizeof(buff) - 7);
	msgflush();
	eventex(gg->edhead, buff, ElvFalse);
	(void) eventdraw(gg->edhead);
  }
  else {
	msg(MSG_ERROR, "Unknown command in search, replace or move");
	msgflush();
  }

  /* Report any failure */
  if (gg->move_cmd == MOVE_ERR) {
	msg(MSG_ERROR, "Invalid string for search, replace or move");
	msgflush();
  }

  return;
}


/*
** f i l e n a m e _ h a n d l e r
**
** Handler for activate signal in File menu.
**
** Returns:	Nothing		Always
*/
static void filename_handler(GtkWidget *widget, gpointer data)
{
  GNGUI *gg = &gn_gui;
  struct Enode *pe, *pn = (ENODE *) data;
  unsigned int u;
  int here, there, count, dirn;
  char *pc, buff[PATH_MAX + 64];

  DPRINT_F(("filename_handler()\n"));

  /* Scan the argslist array and get the appropriate argument */
  buff[0] = '\0';
  if (gg->edhead == NULL || gg->edhead->argshead == NULL ||
			undot_path(ioabsolute(gg->edhead->argshead->name),
					buff, sizeof(buff) - 1) == FAILED) {
	DPRINT_A(("filename_handler(): bad path |%s|\n", buff));
	return;
  }

  /* Find the start of the current argslist by searching for the node with
  ** the lowest id, which will be the first one on the list.
  */
  u = gg->edhead->argshead->id;
  for (pe = gg->edhead->argshead->next ; pe != gg->edhead->argshead ;
							pe = pe->next) {
	if (u > pe->id) u = pe->id;
  }
  for (pe = gg->edhead->argshead->next ; pe->id != u ; pe = pe->next)
	/* MT */ ;

  /* Find out where we are and where we want to go */
  here = find_node(&pe, buff, TRUE , FALSE);
  there = find_node(&pe, pn->name, TRUE, FALSE);

  /* Calculate the new window's offset */
  count = 0;
  if (here < 0 || there < 0) {
	here = there = -1;
	count = 0;
	pc = NULL;
  }
  else if (here < there) {
	dirn = NEXT;
	count = there - here;
	pc = "n";
  }
  else if (here > there) {
	dirn = PREV;
	count = here - there;
	pc = "N";
  }
  else {
	dirn = HERE;
	count = 0;
	pc = NULL;
  }

  DPRINT_A(("filename_handler(): cmd %c here %d there %d count %d\n",
				pc == NULL ? '#' : *pc, here, there, count));

  /* Flush all message buffers and call back into Elvis */
  if (pc != NULL) {
	msgflush();
	while (count--)
		eventex(gg->edhead, pc, ElvFalse);
	/* Force a screen update */
	(void) eventdraw(gg->edhead);
  }

  return;
}


/*
** d o c u m e n t s _ h a n d l e r
**
** Handler for activate signal in Documents menu.
**
** Returns:	Nothing		Always
**
** The data argument is a pointer to the relevant ENODE.
**
** This should add a file to the current arglist, rather than
** loading an entirely new window, but Elvis has no easy way
** to update the arglist. ###
*/
static void documents_handler(GtkWidget *widget, gpointer data)
{
  GNGUI *gg = &gn_gui;
  struct Enode *pn = (ENODE *) data;
  char buff[PATH_MAX];

  DPRINT_F(("documents_handler()\n"));

  DPRINT_P(("documents_handler(): new window mode %s\n",
					o_newmode ? "snext" : "next"));

  /* Open either a new window or a new ring, depending */
  strcpy(buff, o_newmode ? "snext " : "next");
  strncat(buff, pn->name, sizeof(buff) - 6);

  /* Pass the command to Elvis */
  msgflush();
  eventex(gg->edhead, buff, ElvFalse);

  /* Force a screen update */
  (void) eventdraw(gg->edhead);

  return;
}


/*
** v _ v a l _ c h g _ h a n d l e r
**
** Handler for value-changed signal in the vertical scroll bar widget.
**
** Returns:	TRUE		Event handled.
**		FALSE		Event still needs handling.
**
** This tells the main Elvis code that the text has been scrolled.
** Elvis tells the GUI that it wants the text scrolled via the
** scrollbar() function.
*/
static gint v_val_chg_handler(GtkAdjustment *adj, gpointer data)
{
  GNGUI *gg = (GNGUI *) data;
  double scale, b, c, d, e, f, g;
  long int position, slider;
  int status;

  DPRINT_V(("v_val_chg_handler()\n"));

  /* This may be called before complete window initialisation;
  ** Elvis is then given invalid arguments and blows up.
  ** There seems no reason to propagate the signal further.
  */
  if (gg == NULL || gg->edhead == NULL) return TRUE;

  /* Range of values for a GTK+ scrollbar is from adjustment->lower
  ** to adjustment->upper - adjustment->page_size, ie. from
  ** 0 to (1 - page_size).
  ** Elvis uses the range from zero to (bar height - slider height).
  **
  ** Since the call back into Elvis is made with a scaled value and not
  ** exact line numbers it's difficult to position the slider exactly over
  ** the full range for all file sizes.  This is especially noticeable
  ** when dragging the slider to the last line of the file.
  **
  ** It would be easier if Elvis' eventscroll() had a "go to line" mode. ###
  */
  scale = 1000.0;
  /* gg->vert_adj->lower */
  b = scale * gg->vert_adj->upper;	/* Pixels per slider trough */
  c = scale * gg->vert_adj->step_increment;	/* Pixels per line move */
  /* gg->vert_adj->page_increment */
  d = scale * gg->vert_adj->page_size;	/* Slider height */
  e = scale * gg->vert_adj->value;	/* Slider position */
  f = b - d;
  if (e >= f) e = f + c / 100.0;
  f /= b;
  (void) modf(e * f, &g);
  position = (long int) g;
  (void) modf(b * f, &g);
  slider = (long int) g;

  DPRINT_S(("v_val_chg_handler(): position %ld slider %ld\n",
						position, slider));

  status = eventscroll(			/* eventscroll() uses VI commands */
	(GUIWIN *) gg->edhead,		/* window to be scrolled */
	SCROLL_PERCENT,			/* type of scrolling */
	position,			/* amount to scroll */
	slider				/* scrollbar height */
	);

  /* Cursor is updated by eventscroll() calling setcursor() */

  /* Double check for valid windows */
  if (gg != NULL && gg->edhead != NULL)
	(void) eventdraw(gg->edhead);

  /* FALSE to continue propagation, TRUE if we've handled it */
  return FALSE;
}

#ifdef ADD_UNUSED
/*
** d e l _ h a n d l e r
**
** Handler for delete-event signal.
**
** Returns:	TRUE		Always
*/
static gint del_handler(GtkWidget *widget, GdkEvent *event, gpointer data)
{
  int status;

  DPRINT_F(("del_handler()\n"));

  /* This signal closes the current editing window.
  ** On delete, either hide or destroy the vterm;
  ** if it's the last vterm, kill the whole program.
  */
  status = destroy_edwin((GNGUI *) data);

  /* FALSE can be used to force a destroy signal;
  ** however, this is not the default handler
  ** and the return value is ignored.
  */
  return TRUE;
}
#endif

/*
** g l o g _ h a n d l e r
**
** Handler for GLib message logger.
**
** Returns:	Nothing		Always
**
** This is used to supress a known warning, but display all other messages.
*/
static void glog_handler(const gchar *domain, GLogLevelFlags level,
				const gchar *message, gpointer data)
{
  DPRINT_F(("glog_handler()\n"));

  if (!(level == G_LOG_LEVEL_WARNING &&
		strncmp(message, "gtk_scrolled_window_add():", 25) == SAME))
	g_print("%s\n", message);

  return;
}


/*
** g i o _ h a n d l e r
**
** Handler for a GIOChannel watcher.
**
** Returns:	TRUE		Source should not be removed
**		FALSE		Otherwise
*/
static gboolean gio_handler(GIOChannel *src, GIOCondition cond, gpointer data)
{
  GNGUI *gg = &gn_gui;
  GIOStatus gs = G_IO_STATUS_ERROR;
  GError *error = NULL;
  gsize len;
  guint flags;
  int refresh, status = TRUE;
  char *pc, *pd;

  DPRINT_F(("gio_handler()\n"));

  flags = *(guint *) data;

  if (cond & G_IO_IN) {
	pd = "G_IO_IN";
	status = TRUE;
	/* Read a block of data */
	pc = NULL;
	gs = g_io_channel_read_line(
				src,	/* GIOChannel */
				&pc,	/* Line read, NULL on zero length */
				&len,	/* Length of read data */
				NULL,	/* Position of line terminator */
				&error	/* Error returned */
				);
	if (gs == G_IO_STATUS_EOF) {
		status = FALSE;
	}
	else if (gs == G_IO_STATUS_ERROR) {
		msg(MSG_ERROR, "Error reading data");
		msgflush();
	}
	else if (pc != NULL) {
		/* Write to the current file */
		if (flags & 1) {
			DPRINT_Y(("gio_handler(): writing data |%s|\n", pc));
			eventkeys(gg->edhead, toCHAR(pc), (int) len);
			(void) eventdraw(gg->edhead);
		}
		/* Write to the current window */
		else {
			/* Write out the data via Elvis, so that the
			** "Hit <Enter> to continue" message works.
			*/
			refresh = o_exrefresh;
			o_exrefresh = 1;
			drawextext(winofgw(gg->edhead), pc, (int) len);
			o_exrefresh = refresh;
		}
	}
  }
  else if (cond & G_IO_ERR) {
	pd = "G_IO_ERR";
	status = FALSE;
  }
  else if (cond & G_IO_HUP) {
	pd = "G_IO_HUP";
	status = FALSE;
  }
  else if (cond & G_IO_NVAL) {
	pd = "G_IO_NVAL";
	status = FALSE;
  }
  else {
	pd = "G_IO_UNKNOWN";
	status = FALSE;
  }

  /* This displays the "Hit <Enter> to continue" message,
  ** but why is ESC needed to force an update ???
  */
  if (!(cond & G_IO_IN)) {
	/* Channel already closed
	(void) g_io_channel_shutdown(gg->gio_out, TRUE, NULL);
	*/
	eventkeys(gg->edhead, "\033", 1);
	(void) eventdraw(gg->edhead);
  }

#if 0
  { /* Debugging block begin */
  char *pe;
  if (gs == G_IO_STATUS_ERROR)
	pe = "G_IO_STATUS_ERROR";
  else if (gs == G_IO_STATUS_NORMAL)
	pe = "G_IO_STATUS_NORMAL";
  else if (gs == G_IO_STATUS_EOF)
	pe = "G_IO_STATUS_EOF";
  else if (gs == G_IO_STATUS_AGAIN)
	pe = "G_IO_STATUS_AGAIN";
  else
	pe = "G_IO_STATUS_UNKNOWN";

  if (gs == G_IO_STATUS_ERROR && error != NULL)
	g_print("gio_handler(): error |%s|\n", error->message);

  g_print("gio_handler(): gio_cond %s gio_status %s\n", pd, pe);
  } /* Debugging block end */
#endif

  return status;
}


/*
** k e y _ p r e s s _ h a n d l e r
**
** Handler for key-press-event signal.
**
** Returns:	TRUE		Event handled.
**		FALSE		Event still needs handling.
**
** The data argument is a pointer to a Gnome window instance structure.
** The main body of the Elvis code does almost all the key remapping and
** conversion.  The full list of event->keyval values is in gdkkeysyms.h.
*/
static gint key_press_handler(GtkWidget *widget, GdkEventKey *event,
								gpointer data)
{
  GNGUI *gg = (GNGUI *) data;
  KEY * pkey;
  int ctrl, alt;
  char buff[2];

  DPRINT_V(("key_press_handler()\n"));

  /* Something is wrong, and there's no point in continuing */
  if (gg == NULL || gg->edhead == NULL) {
	DPRINT_K(("key_press_handler(): invalid widget\n"));
	return TRUE;
  }

  /* Break out of a broken clipboard read ###
  clip_keypress();
  */

  ctrl = (event->state & GDK_CONTROL_MASK) != 0;
  alt = (event->state & GDK_MOD1_MASK) != 0;

  DPRINT_K(("key_press_handler(): step 1\n"));

  /* A keyval is an unsigned int */
  if ((!ctrl) && ((event->keyval > 0) && (event->keyval < 0xff))) {
	/* We have a normal key */
	DPRINT_K(("key_press_handler(): step 2\n"));
	buff[0] = event->keyval;
	buff[1] = '\0';
	eventkeys(gg->edhead, toCHAR(buff), 1);
	DPRINT_K(("key_press_handler(): step 2a\n"));
	/* eventkeys() may remove our window, so re-check */
	if (gg != NULL && gg->edhead != NULL)
		(void) eventdraw(gg->edhead);
  }
  else {
	DPRINT_K(("key_press_handler(): step 3\n"));
	/* We have some sort of special key; what is it ? */
	for (pkey = &keyvalues[0] ; pkey->nchars != 0 ; pkey++) {
		DPRINT_K(("key_press_handler(): step 3a\n"));
		if ((event->keyval == pkey->keyval)
				&& ((ctrl == pkey->ctrlstate)
				|| (pkey->ctrlstate == -1))
				&& ((alt == pkey->altstate)
				|| (pkey->altstate == -1))) {
			DPRINT_K(("key_press_handler(): step 3b\n"));
			eventkeys(gg->edhead, toCHAR(pkey->keystr),
							pkey->nchars);
			/* eventkeys() may remove our window, so re-check */
			if (gg != NULL && gg->edhead != NULL)
				(void) eventdraw(gg->edhead);
			DPRINT_K(("key_press_handler(): step 3c\n"));
			break;
		}
	}
	/* Some keys, eg: shift, aren't valid alone.
	** Show them only while debugging.
	*/
	if (pkey->nchars == 0) {
		/* Stop GCC complaining */ ;
		/* gdk_keyval_name() returns NULL for an invalid key */
		DPRINT_K(("key_press_handler(): 0x%08x \\%03o %s%s%s\n",
			event->keyval, event->keyval & 0xff,
			gdk_keyval_name(event->keyval),
			ctrl ? " ctl" : "", alt ? " alt" : ""));
	}
  }

  DPRINT_K(("key_press_handler(): 4\n"));

  return FALSE;
}


/*
** d e s t r o y _ h a n d l e r
**
** Handler for Gnome main window destroy signal.
**
** Returns:	Nothing		Always
**
** The standard quit method for Elvis is term().
*/
static void destroy_handler(GtkWidget *widget, gpointer data)
{
  GNGUI *gg = &gn_gui;

  DPRINT_F(("destroy_handler()\n"));

  /* Flush all message buffers and call back into Elvis */
  msgflush();

  /* There are problems with :quall! (see quit_action()),
  ** and the work-around is to close each window in turn.
  */
  while (gg->edhead != NULL) {
	eventex(gg->edhead, "q!", ElvFalse);
	/* Give GTK+ time to stabilise */
	while (gtk_events_pending()) gtk_main_iteration();
  }

  return;
}


/*
** r e s i z e _ h a n d l e r
**
** Handler for a main window configure-event signal.
**
** Returns:	TRUE		Event handled.
**		FALSE		Event still needs handling.
**
** This resets the vterm size to fit the new main window.
**
** There are problems on the first call of this function, where vterm
** sizes and main window sizes don't match.  This looks as if the
** interface hasn't stabilised, but waiting for gtk_events_pending()
** doesn't show any improvement. ??? ###
*/
static gint resize_handler(GtkWidget *widget, GdkEventConfigure *event,
								gpointer data)
{
  GNGUI *gg = (GNGUI *) data;
  EDWIN *edwin = gg->edhead;
  static UINT flg = TRUE;
  glong char_width;
  glong char_height;
  glong rows, cols;
  int old_width, old_height;
  int rdelta, cdelta;

  DPRINT_R(("resize_handler(): event height %d width %d vt %s flg %u\n",
				event->height, event->width,
				(edwin == NULL) ? "null" :
				(edwin->vterm == NULL) ? "null" : "valid",
				flg
				));

  if (edwin == NULL) return FALSE;

  /* Get the last known size of the main window */
  old_height = gg->main_height;
  old_width = gg->main_width;
  /* And of the terminal and its font */
  rows = edwin->rows;
  cols = edwin->cols;
  char_width = edwin->pango_width;
  char_height = edwin->pango_height;

  /* If height is 0, use the current window size */
  if (old_height == 0) {
	rdelta = cdelta = 0;
  }
  /* Otherwise, find the changes in height or width and convert them
  ** to rows or columns. The widget hints ensure that the widget size
  ** is changed only in increments of char_height or char_width.
  */
  else {
	rdelta = (event->height - old_height) / char_height;
	cdelta = (event->width - old_width) / char_width;
  }
  gg->main_height = event->height;
  gg->main_width = event->width;

  DPRINT_R(("resize_handler(): cur_rows %ld delta %d\n",
					edwin->rows, rdelta));
  DPRINT_R(("resize_handler(): cur_cols %ld bar_cols %d delta %d\n",
					edwin->cols, gg->bar_cols, cdelta));

  /* Re-size the vterm */
  edwin->rows = rows + rdelta;
  edwin->cols = cols + cdelta;
  if (flg) {
	flg = FALSE;
	edwin->cols -= gg->bar_cols;
	edwin->cols--;
	edwin->rows--;
  }

  /* And tell Elvis about it */
  eventresize(edwin, edwin->rows, edwin->cols);

  /* Force a screen update */
  (void) eventdraw(edwin);

  DPRINT_R(("resize_handler(): final rows vterm %ld vte %ld\n",
	edwin->rows,vte_terminal_get_row_count(VTE_TERMINAL(edwin->vterm))));
  DPRINT_R(("resize_handler(): final cols vterm %ld vte %ld\n",
	edwin->cols,vte_terminal_get_column_count(VTE_TERMINAL(edwin->vterm))));

  return FALSE;
}


/*
** s t a t e _ h a n d l e r
**
** Handler for a main window window-state-event signal.
**
** Returns:	TRUE		Event handled.
**		FALSE		Event still needs handling.
*/
static gboolean state_handler(GtkWidget *widget,
				GdkEventWindowState *event, gpointer data)
{
  GNGUI *gg = (GNGUI *) data;
  EDWIN *edwin = gg->edhead;

  DPRINT_F(("state_handler()\n"));

#if 0
  { /* Debugging block begin */
  static UINT pass = 0;
  if (pass < 5) {
	pass++;
	DPRINT_R(("state_handler(): pass %u\n", pass));
  }
  } /* Debugging block end */
#endif

  /* Track the current state */
  gg->window_state = event->new_window_state;

  /* If we've maximised or unmaximised the window, record the new size */
  if (event->changed_mask & GDK_WINDOW_STATE_MAXIMIZED) {
	/* Give GTK+ time to stabilise */
	while (gtk_events_pending()) gtk_main_iteration();
	edwin->rows = vte_terminal_get_row_count(
					VTE_TERMINAL(edwin->vterm));
	edwin->cols = vte_terminal_get_column_count(
					VTE_TERMINAL(edwin->vterm));
	/* Adjustment for window scrollbar */
	edwin->cols -= gg->bar_cols;
  }

  return FALSE;
}


/*
** d e f a u l t _ f o n t _ h a n d l e r
**
** Handler for font dialog default button.
**
** Returns:	Nothing		Always
**
** The data argument is an array of struct Select.
*/
static void default_font_handler(GtkWidget *widget, gpointer data)
{
  GNGUI *gg = &gn_gui;
  struct Select *sels = (struct Select *) data;
  int state;

  DPRINT_F(("default_font_handler()\n"));

  state = gtk_toggle_button_get_active(GTK_TOGGLE_BUTTON(widget));
  gtk_button_set_label(GTK_BUTTON(widget), state ? "Yes" : "No");

  if (state) {
	gtk_font_button_set_font_name(GTK_FONT_BUTTON(sels[0].button),
							gg->def_fontname);
#ifdef DEBUG_MENU
	gtk_font_button_set_font_name(GTK_FONT_BUTTON(sels[1].button),
							gg->def_dbg_fontname);
#endif
  }

  return;
}


/*
** d e f a u l t _ c o l o r _ h a n d l e r
**
** Handler for color dialog default button.
**
** Returns:	Nothing		Always
**
** The data argument is an array of struct Select.
*/
static void default_color_handler(GtkWidget *widget, gpointer data)
{
  GNGUI *gg = &gn_gui;
  struct Select *sels = (struct Select *) data;
  int state;

  DPRINT_F(("default_color_handler()\n"));

  state = gtk_toggle_button_get_active(GTK_TOGGLE_BUTTON(widget));
  gtk_button_set_label(GTK_BUTTON(widget), state ? "Yes" : "No");

  if (state) {
	gtk_color_button_set_color(GTK_COLOR_BUTTON(sels[BACK].button),
							&gg->def_backcolor);
	gtk_color_button_set_color(GTK_COLOR_BUTTON(sels[FORE].button),
							&gg->def_forecolor);
  }

  return;
}


/*
** d e f a u l t _ c u r s o r _ h a n d l e r
**
** Handler for cursor dialog default button.
**
** Returns:	Nothing		Always
**
** The data argument is an array of struct Select.
*/
static void default_cursor_handler(GtkWidget *widget, gpointer data)
{
  GNGUI *gg = &gn_gui;
  struct Select *sels = (struct Select *) data;
  int state;

  DPRINT_F(("default_cursor_handler()\n"));

  state = gtk_toggle_button_get_active(GTK_TOGGLE_BUTTON(widget));
  gtk_button_set_label(GTK_BUTTON(widget), state ? "Yes" : "No");

  if (state) {
	gtk_color_button_set_color(GTK_COLOR_BUTTON(sels[0].button),
						&gg->def_cursorcolor);
	gtk_toggle_button_set_active(GTK_TOGGLE_BUTTON(sels[1].button),
						gg->def_cursorblink);
  }

  return;
}


/*
** t e r m i n a l _ f o n t _ h a n d l e r
**
** Handle the terminal font-set signal.
**
** Returns:	Nothing		Always
**
** The user data is the default font button.
*/
static void terminal_font_handler(GtkWidget *widget, gpointer data)
{
  GNGUI *gg = &gn_gui;
  const gchar *name, *fontname;
  char buff[F_SIZ];

  DPRINT_F(("terminal_font_handler()\n"));

  name = gtk_font_button_get_title(GTK_FONT_BUTTON(widget));
  fontname = gtk_font_button_get_font_name(GTK_FONT_BUTTON(widget));

  /* Revert to the current font if the chosen font is not monospace.
  ** Annoying, but there seems no way to filter the dialog names.
  */
  if (check_mono_family((char *) fontname) != OK) {
	strncpy(buff, fontname, F_MAX);
	if (strcmp(name, "Current") == SAME) {
		gtk_font_button_set_font_name(GTK_FONT_BUTTON(widget),
							gg->edhead->fontname);
	}
#ifdef DEBUG_MENU
	else if (strcmp(name, "Debug") == SAME) {
		gtk_font_button_set_font_name(GTK_FONT_BUTTON(widget),
							gg->cur_dbg_fontname);
	}
#endif
	else {
		gtk_font_button_set_font_name(GTK_FONT_BUTTON(widget),
							gg->def_fontname);
	}
	fontname = gtk_font_button_get_font_name(GTK_FONT_BUTTON(widget));
	msg(MSG_ERROR, "[s]Font $1 is not a monospace font", buff);
	msgflush();
  }

  /* Clear the default font button when the font changes */
  if (strcmp(name, "Current") == SAME) {
	if (strcmp(fontname, gg->def_fontname) != SAME)
		gtk_toggle_button_set_active(GTK_TOGGLE_BUTTON(data), FALSE);
  }
#ifdef DEBUG_MENU
  else if (strcmp(name, "Debug") == SAME) {
	if (strcmp(fontname, gg->def_dbg_fontname) != SAME)
		gtk_toggle_button_set_active(GTK_TOGGLE_BUTTON(data), FALSE);
  }
#endif

  return;
}


/*
** c o l o r _ h a n d l e r
**
** Handle the color-set signal.
**
** Returns:	Nothing		Always
**
** The user data is the default color button.
*/
static void color_handler(GtkWidget *widget, gpointer data)
{
  GNGUI *gg = &gn_gui;
  GdkColor color;
  const gchar *name;

  DPRINT_F(("color_handler()\n"));

  name = gtk_color_button_get_title(GTK_COLOR_BUTTON(widget));
  gtk_color_button_get_color(GTK_COLOR_BUTTON(widget), &color);

  /* Clear the default color button when the forecolor or backcolor changes */
  if (strcmp(name, "Background") == SAME) {
	if (!gdk_color_equal(&gg->def_backcolor, &color))
		gtk_toggle_button_set_active(GTK_TOGGLE_BUTTON(data), FALSE);
  }
  else if (strcmp(name, "Foreground") == SAME) {
	if (!gdk_color_equal(&gg->def_forecolor, &color))
		gtk_toggle_button_set_active(GTK_TOGGLE_BUTTON(data), FALSE);
  }
  else if (strcmp(name, "Color is") == SAME) {
	if (!gdk_color_equal(&gg->def_cursorcolor, &color))
		gtk_toggle_button_set_active(GTK_TOGGLE_BUTTON(data), FALSE);
  }

  return;
}


/*
** c u r s o r _ h a n d l e r
**
** Handler for toggled signal.
**
** Returns:	Nothing		Always
**
** The user data is the default cursor button.
*/
static void cursor_handler(GtkWidget *widget, gpointer data)
{
  GNGUI *gg = &gn_gui;
  int state;

  DPRINT_F(("cursor_handler()\n"));

  state = gtk_toggle_button_get_active(GTK_TOGGLE_BUTTON(widget));
  gtk_button_set_label(GTK_BUTTON(widget), state ? "On" : "Off");

  /* Clear the default cursor button when the cursor blink state changes */
  if ((state && !gg->def_cursorblink) || (!state && gg->def_cursorblink))
	gtk_toggle_button_set_active(GTK_TOGGLE_BUTTON(data), FALSE);

  return;
}


/*
** b e g i n _ p r i n t _ h a n d l e r
**
** Handler for GtkPrintOperation begin-print signal.
**
** Returns:	Nothing		Always
*/
static void begin_print_handler(GtkPrintOperation *operation,
				GtkPrintContext *context, gpointer data)
{
  GNGUI *gg = &gn_gui;
  struct Printjob *print_job = (struct Printjob *) data;
  GtkPrintPages print_pages;
  GtkPageRange *page_range;
  PangoFontDescription *pango_head_desc;
  PangoFontDescription *pango_body_desc;
  PangoFontMetrics *pango_head_metrics;
  PangoFontMetrics *pango_body_metrics;
  FILE *fptmp;
  double d, e, page_height, head_char_height, body_char_height;
  unsigned int max_count, char_count, line_count;
  gint range_count;
  int j, k, y, index, status;
  char *pc, *pt;

  DPRINT_F(("begin_print_handler()\n"));

  /* Deactivate the Print menu "Print Setup" and "Print File" entries */
  gtk_widget_set_sensitive(gtk_ui_manager_get_widget(
				gg->mainuimanager,
				"/ui/menubar/PrintMenu/Print Setup"),
				FALSE);
  gtk_widget_set_sensitive(gtk_ui_manager_get_widget(
				gg->mainuimanager,
				"/ui/menubar/PrintMenu/Print File"),
				FALSE);

  /* The print operation settings are available only after they have been
  ** explicitly set, or gtk_print_operation_run() has been called.
  */
  gg->print_settings = gtk_print_operation_get_print_settings(operation);
  print_pages = GTK_PRINT_PAGES_ALL;

  if (gg->print_settings == NULL) {
	gtk_print_operation_cancel(gg->print_operation);
	msg(MSG_ERROR, "Cannot obtain print settings");
	msgflush();
	return;
  }
  else {
	print_pages = gtk_print_settings_get_print_pages(gg->print_settings);
	switch (print_pages) {
	case GTK_PRINT_PAGES_ALL:
		pc = "All";
		break;
	case GTK_PRINT_PAGES_CURRENT:
		pc = "Current";
		break;
	case GTK_PRINT_PAGES_RANGES:
		pc = "Ranges";
		break;
	default:
		pc = "Default";
		break;
	}
	range_count = 0;
	page_range = gtk_print_settings_get_page_ranges(gg->print_settings,
								&range_count);
#ifdef DPRINT_ZZ
	{ /* Debugging block begin */
	g_print(
		"begin_print_handler(): print_pages %d %s"
		" page_range 0x%08lx range_count %d\n",
		print_pages, pc, (unsigned long) page_range, range_count
		);
	if (*pc == 'R' && range_count != 0) {
		for (j = 0 ; j < range_count ; j++) {
			g_print(
			"begin_print_handler(): range %d start %d end %d\n",
			j, page_range[j].start, page_range[j].end
			);
		}
	}
	} /* Debugging block end */
#endif
  }

  /* If any ranges are set, fail if they are reversed or overlapping */
  status = OK;
  if (print_pages == GTK_PRINT_PAGES_RANGES && range_count != 0) {
	index = -1;
	for (j = 0 ; j < range_count && status == OK ; j++) {
		if (index >= page_range[j].start)
			status = FAILED;
		index = page_range[j].start;
		if (index > page_range[j].end)
			status = FAILED;
	}
  }
  if (status == FAILED) {
	gtk_print_operation_cancel(gg->print_operation);
	msg(MSG_ERROR, "Reversed or overlapping page ranges");
	msgflush();
	return;
  }

  /* Get a pango language; will this ever change after startup ??? */
  if ((print_job->pango_lang = gtk_get_default_language()) == NULL)
	print_job->pango_lang = pango_language_from_string(DEFAULT_PANGO_LANG);

  /* Get the printing pango context */
  print_job->pango_context = pango_cairo_font_map_create_context(
		(PangoCairoFontMap *) pango_cairo_font_map_get_default());

  pango_head_desc = pango_font_description_from_string(gg->cur_hdr_print_font);
  d = (double) pango_font_description_get_size(pango_head_desc) / PANGO_SCALE;
  print_job->head_font_size = d + 0.5;

  /* Get the header font metrics */
  pango_head_metrics = pango_context_get_metrics(print_job->pango_context,
					pango_head_desc, print_job->pango_lang);
  y = pango_font_metrics_get_ascent(pango_head_metrics);
  y += pango_font_metrics_get_descent(pango_head_metrics);
  /* Convert from Pango units to points */
  head_char_height = (gdouble) y / PANGO_SCALE;
  pango_font_metrics_unref(pango_head_metrics);
  pango_head_metrics = NULL;

  /* Get the body font metrics */
  pango_body_desc = pango_font_description_from_string(gg->cur_print_font);
  e = (double) pango_font_description_get_size(pango_body_desc) / PANGO_SCALE;
  print_job->body_font_size = e + 0.5;
  pango_body_metrics = pango_context_get_metrics(print_job->pango_context,
					pango_body_desc, print_job->pango_lang);
  y = pango_font_metrics_get_ascent(pango_body_metrics);
  y += pango_font_metrics_get_descent(pango_body_metrics);
  /* Convert from Pango units to points */
  body_char_height = (gdouble) y / PANGO_SCALE;
  pango_font_metrics_unref(pango_body_metrics);
  pango_body_metrics = NULL;

  /* Put the font descriptions somewhere accessible */
  print_job->pango_head_desc = pango_head_desc;
  print_job->head_char_height = head_char_height;
  print_job->pango_body_desc = pango_body_desc;
  print_job->body_char_height = body_char_height;
  print_job->head_font_size = d;
  print_job->body_font_size = e;

  /* Copy the file (which Elvis may be holding open) to a scratch file
  ** in the system's temporary directory.  During the copy process, get
  ** the number of lines to be printed, and the maximum line length.
  */
  fptmp = NULL;
  max_count = line_count = char_count = 0;
  if ((fptmp = fdopen(g_file_open_tmp("elvprnXXXXXX", &pt, NULL), "w+"))
								== NULL) {
	fclose(print_job->fpin);
	print_job->fpin = NULL;
	gtk_print_operation_cancel(gg->print_operation);
	msg(MSG_ERROR, "Cannot open temporary file for printing");
	msgflush();
	return;
  }
  else {
	k = EOF;
	while ((j = fgetc(print_job->fpin)) != EOF) {
		k = j;
		char_count++;
		/* pango_layout_set_text() needs the space ??? */
		if (j == '\t')
			char_count += 7;
		else if (j == '\n') {
			line_count++;
			if (max_count < char_count)
				max_count = char_count;
			char_count = 0;
		}
		if (fputc(j, fptmp) == EOF) {
			fclose(print_job->fpin);
			print_job->fpin = NULL;
			fclose(fptmp);
			remove(pt);
			gtk_print_operation_cancel(gg->print_operation);
			msg(MSG_ERROR,
				"[s]Cannot write to temporary file $1", pt);
			msgflush();
			return;
		}
	}
	/* Files may not always have a terminating newline */
	if (k != '\n') {
		(void) fputc('\n', fptmp);	/* Risk it */
		line_count++;
		if (max_count < char_count)
			max_count = char_count;
	}
  }
  fclose(print_job->fpin);
  print_job->fpin = NULL;

  /* Allocate a buffer capable of holding the longest line found */
  if ((print_job->buff = g_try_malloc(max_count + 1)) == NULL) {
	fclose(fptmp);
	remove(pt);
	gtk_print_operation_cancel(gg->print_operation);
	msg(MSG_ERROR, "Cannot allocate memory for printing");
	msgflush();
	return;
  }

  /* Update the print job structure */
  rewind(fptmp);
  print_job->fpin = fptmp;
  print_job->tmpname = pt;
  print_job->line_size = max_count;
  print_job->line_count = line_count;
  print_job->current_line = 0;

  /* Get the page height available for text, in pixels */
  page_height = gtk_print_context_get_height(context);

  /* Allow for header and footer space */
  if (gg->cur_hdrprint)
	page_height -= HEADER_LINES * print_job->head_char_height;
  else
	page_height -= print_job->body_font_size;
  page_height -= FOOTER_LINES * print_job->body_font_size;

  /* Get the total number of lines, and the total number of pages */
  print_job->lines_per_page = page_height / print_job->body_font_size;
  print_job->page_count = (print_job->line_count - 1) /
					print_job->lines_per_page + 1;
  gtk_print_operation_set_n_pages(operation, print_job->page_count);

#ifdef DPRINT_ZZ
  { /* Debugging block begin */
  g_print("begin_print_handler():\n");
  g_print("filename:     |%s|\n", print_job->filename);
  g_print("upper header: |%s|\n", print_job->upper_header);
  g_print("head font:     %d\n", print_job->head_font_size);
  g_print("body font:     %d\n", print_job->body_font_size);
  g_print("lines/page:    %d\n", print_job->lines_per_page);
  g_print("line count:    %d\n", print_job->line_count);
  g_print("page count:    %d\n", print_job->page_count);
  } /* Debugging block end */
#endif

  return;
}


/*
** d r a w _ p a g e _ h a n d l e r
**
** Handler for GtkPrintOperation draw-page signal.
**
** Returns:	Nothing		Always
**
** Note that the header text uses a line height matching the font height,
** whereas the body text uses the font size.
** The function argument page_number counts from 0, not 1.
*/
static void draw_page_handler(GtkPrintOperation *operation,
			GtkPrintContext *context, gint page_num, gpointer data)
{
  GNGUI *gg = &gn_gui;
  struct Printjob *print_job = (struct Printjob *) data;
  cairo_t *cairo = NULL;
  PangoLayout *layout = NULL;
  gdouble page_width, page_height;
  gint layout_width, layout_height;
  char buff[64];
  char *pb;
  int j;

  DPRINT_F(("draw_page_handler()\n"));

  /* If we're printing a range, then some pages may have been skipped */
  while (print_job->current_line < page_num * print_job->lines_per_page) {
	fgets(print_job->buff, print_job->line_size + 1, print_job->fpin);
	print_job->current_line++;
  }

  /* Get the cairo context */
  cairo = gtk_print_context_get_cairo_context(context);
  if (cairo == NULL) {
	msg(MSG_ERROR, "No Cairo context available for printing");
	msgflush();
	return;
  }

  /* The context sizes are in pixels ??? */
  page_width = gtk_print_context_get_width(context);
  page_height = gtk_print_context_get_height(context);
  DPRINT_Z(("draw_page_handler(): cairo context is %s, width %f height %f\n",
			cairo == NULL ? "NULL" : "valid",
			page_width, page_height));

  /* Create the page header block */
  if (gg->cur_hdrprint) {

	/* Draw a rectangular outline for emphasis; the vertical offset
	** allows for the broad line width.
	*/
	cairo_rectangle(cairo, 0.0, 1.0, page_width,
				HEADER_LINES * print_job->head_char_height);
	cairo_set_line_width(cairo, 2.0);
	cairo_stroke(cairo);

	/* Create the page header layout and set the header font */
	layout = gtk_print_context_create_pango_layout(context);
	pango_layout_set_font_description(layout, print_job->pango_head_desc);

	/* Add the upper header to the layout */
	pango_layout_set_text(layout, print_job->upper_header, -1);
	pango_layout_get_size(layout, &layout_width, &layout_height);
	pango_layout_set_alignment(layout, PANGO_ALIGN_LEFT);
	cairo_move_to(cairo, HEADER_MARGIN, print_job->head_char_height / 2.0);
	pango_cairo_show_layout(cairo, layout);

	/* Add the page number to the layout */
	sprintf(buff, "%d/%d", page_num + 1, print_job->page_count);
	pango_layout_set_text(layout, buff, -1);
	pango_layout_get_size(layout, &layout_width, &layout_height);
	pango_layout_set_alignment(layout, PANGO_ALIGN_RIGHT);
	cairo_move_to(cairo, page_width -
			(HEADER_MARGIN + (gdouble) layout_width / PANGO_SCALE),
			print_job->head_char_height / 2.0);
	pango_cairo_show_layout(cairo, layout);

	/* Add the lower header to the layout */
	pango_layout_set_text(layout, print_job->lower_header, -1);
	pango_layout_set_alignment(layout, PANGO_ALIGN_LEFT);
	cairo_move_to(cairo, HEADER_MARGIN, 1.5 * print_job->head_char_height);
	pango_cairo_show_layout(cairo, layout);

	/* Tidy up */
	g_object_unref(layout);
  }

  /* Set up for the body of the text */
  layout = gtk_print_context_create_pango_layout(context);
  pango_layout_set_font_description(layout, print_job->pango_body_desc);

  /* Move past the header block */
  if (gg->cur_hdrprint) {
	cairo_move_to(cairo, 0.0,
			(1.0 + HEADER_LINES) * print_job->head_char_height);
  }
  else {
	cairo_move_to(cairo, 0.0, (double) print_job->body_font_size);
  }

  /* Print the body of the text */
  for (j = 0 ; j < print_job->lines_per_page &&
		print_job->current_line < print_job->line_count ;
		j++, print_job->current_line++) {
	if (fgets(print_job->buff, print_job->line_size + 1, print_job->fpin)
								!= NULL) {
		pb = print_job->buff;
		/* Clear any trailing newline */
		while (*pb != '\0') pb++;
		if (pb != print_job->buff && *(pb - 1) == '\n')
			*(pb - 1) = '\0';
		pango_layout_set_text(layout, print_job->buff, -1);
		pango_cairo_show_layout(cairo, layout);
		cairo_rel_move_to(cairo, 0.0,
					(double) print_job->body_font_size);
	}
  }

  /* Tidy up */
  g_object_unref(layout);
  layout = NULL;

  return;
}


/*
** e n d _ p r i n t _ h a n d l e r
**
** Handler for GtkPrintOperation end-print signal.
**
** Returns:	Nothing		Always
**
** This function is called even if the print job fails.
*/
static void end_print_handler(GtkPrintOperation *operation,
		GtkPrintContext *context, gpointer data)
{
  GNGUI *gg = &gn_gui;
  struct Printjob *print_job = (struct Printjob *) data;

  DPRINT_F(("end_print_handler()\n"));

  /* Free or dereference any resources that the print job holds */
  if (print_job->fpin != NULL) {
	fclose(print_job->fpin);
	print_job->fpin = NULL;
  }

  if (print_job->tmpname != NULL) {
	unlink(print_job->tmpname);
	g_free(print_job->tmpname);
	print_job->tmpname = NULL;
  }

  if (print_job->buff != NULL) {
	g_free(print_job->buff);
	print_job->buff = NULL;
  }

  if (print_job->pango_head_desc != NULL) {
	pango_font_description_free(print_job->pango_head_desc);
	print_job->pango_head_desc = NULL;
  }

  if (print_job->pango_body_desc != NULL) {
	pango_font_description_free(print_job->pango_body_desc);
	print_job->pango_body_desc = NULL;
  }

  /* Reactivate the Print menu "Print Setup" and "Print File" entries */
  gtk_widget_set_sensitive(gtk_ui_manager_get_widget(
				gg->mainuimanager,
				"/ui/menubar/PrintMenu/Print Setup"),
				TRUE);
  gtk_widget_set_sensitive(gtk_ui_manager_get_widget(
				gg->mainuimanager,
				"/ui/menubar/PrintMenu/Print File"),
				TRUE);
  return;
}


/*
** d o n e _ p r i n t _ h a n d l e r
**
** Handler for GtkPrintOperation done signal.
**
** Returns:	Nothing		Always
**
** Contrary to the documentation, the result argument is an enum,
** not a pointer to an enum.
*/
static void done_print_handler(GtkPrintOperation *operation,
		GtkPrintOperationResult *result, gpointer data)
{
  GNGUI *gg = &gn_gui;
  struct Printjob *print_job = (struct Printjob *) data;
  GtkWidget *error_dialog;
  GError *error;
  int status = OK;

  DPRINT_F(("done_print_handler()\n"));

  /* Check the result and tidy up accordingly */
  switch ((unsigned int) result) {

  case GTK_PRINT_OPERATION_RESULT_ERROR:
	gtk_print_operation_get_error(operation, &error);
	DPRINT_Z(("done_print(): error %s\n", error->message));
	error_dialog = gtk_message_dialog_new(
			GTK_WINDOW(gg->gwindow),
			GTK_DIALOG_DESTROY_WITH_PARENT,
			GTK_MESSAGE_ERROR,
			GTK_BUTTONS_CLOSE,
			"%s",
			error->message
			);
	g_error_free(error);

	g_signal_connect(
			error_dialog,
			"response",
			G_CALLBACK(gtk_widget_destroy),
			NULL
			);

	gtk_widget_show(error_dialog);
	break;

  case GTK_PRINT_OPERATION_RESULT_APPLY:
	/* unref'ing the print_settings fails; apparently print_settings
	** is no longer an object.  Why ??? ###
	if (gg->print_settings != NULL)
		g_object_unref(gg->print_settings);
	*/
	gg->print_settings = g_object_ref(
		gtk_print_operation_get_print_settings(gg->print_operation));
	break;

  case GTK_PRINT_OPERATION_RESULT_CANCEL:
	/* Do not update the print settings */
	msg(MSG_ERROR, "Current print job cancelled by user");
	msgflush();
	break;

  case GTK_PRINT_OPERATION_RESULT_IN_PROGRESS:
	/* This is the only case which doesn't require a clean up */
	msg(MSG_INFO, "Current print job still in progress");
	msgflush();
	status = NK;
	break;

  default:
	DPRINT_Z(("done_print_handler(): result UNKNOWN\n"));
	break;
  }

  if (status == OK) {
	/* Once again, unref'ing the print_operation traps.  Why ??? ###
	if (gg->print_operation != NULL)
		g_object_unref(gg->print_operation);
	*/
	g_free(print_job);
	gg->print_operation = NULL;
  }

  return;
}


/*
** p r i n t _ n a m e _ h a n d l e r
**
** Handler for print file name dialog file chooser button
** selection-changed signal.
**
** Returns:	Nothing		Always
**
** The data argument is an array of struct Select.
*/
static void print_name_handler(GtkWidget *widget, gpointer data)
{
  GNGUI *gg = &gn_gui;
  struct Select *sels = (struct Select *) data;
  gchar *pg;

  DPRINT_F(("print_name_handler()\n"));

  pg = gtk_file_chooser_get_filename((GtkFileChooser *) widget);

  if (pg != NULL) {
	/* We have a valid filename, so the OK button can be re-sensitised */
	gtk_dialog_set_response_sensitive(GTK_DIALOG(gg->print_dialog),
							GTK_RESPONSE_OK, TRUE);
	/* If we've chosen a non-default file, then clear the default button */
	if (strcmp(pg, gg->edhead->edname) != SAME)
		gtk_toggle_button_set_active(GTK_TOGGLE_BUTTON(sels[1].button),
									FALSE);
  }
  g_free(pg);

  return;
}


/*
** d e f a u l t _ p r i n t _ n a m e _ h a n d l e r
**
** Handler for print file name dialog default button.
**
** Returns:	Nothing		Always
**
** The data argument is an array of struct Select.
*/
static void default_print_name_handler(GtkWidget *widget, gpointer data)
{
  GNGUI *gg = &gn_gui;
  struct Select *sels = (struct Select *) data;

  DPRINT_F(("default_print_name_handler()\n"));

  /* Set the main dialog OK button sensitivity to match the default */
  gtk_dialog_set_response_sensitive(GTK_DIALOG(gg->print_dialog),
					GTK_RESPONSE_OK, sels[0].setting);

  if (gtk_toggle_button_get_active(GTK_TOGGLE_BUTTON(widget))) {
	if (sels[0].setting) {
		/* We have an actual default file */
		gtk_file_chooser_set_filename(GTK_FILE_CHOOSER(sels[0].button),
							gg->edhead->edname);
		gtk_button_set_label(GTK_BUTTON(widget), "Yes");
	}
	else if (gg->lastprint[0] != '\0') {
		/* We have a default file from the last print operation */
		gtk_file_chooser_set_filename(GTK_FILE_CHOOSER(sels[0].button),
							gg->lastprint);
		gtk_button_set_label(GTK_BUTTON(widget), "Yes");
	}
	else {
		/* We haven't yet opened an actual file */
		gtk_file_chooser_unselect_all(GTK_FILE_CHOOSER(sels[0].button));
		gtk_toggle_button_set_active(GTK_TOGGLE_BUTTON(widget), FALSE);
		gtk_button_set_label(GTK_BUTTON(widget), "No");
	}
  }
  else {
	gtk_button_set_label(GTK_BUTTON(widget), "No");
  }

  return;
}


/*
** p r i n t _ h e a d e r _ h a n d l e r
**
** Handler for print setup dialog toggle button.
**
** Returns:	Nothing		Always
**
** The user data is the default font button.
*/
static void print_header_handler(GtkWidget *widget, gpointer data)
{
  GNGUI *gg = &gn_gui;
  int state;

  DPRINT_F(("print_handler()\n"));

  state = gtk_toggle_button_get_active(GTK_TOGGLE_BUTTON(widget));
  gtk_button_set_label(GTK_BUTTON(widget), state ? "Yes" : "No");

  /* Clear the default print setup button when the header print state changes */
  if ((state && !gg->def_hdrprint) || (!state && gg->def_hdrprint))
	gtk_toggle_button_set_active(GTK_TOGGLE_BUTTON(data), FALSE);

  return;
}


/*
** p r i n t _ f o n t _ h a n d l e r
**
** Handler for print setup dialog font button font-set signal.
**
** Returns:	Nothing		Always
**
** The user data is the default font button.
*/
static void print_font_handler(GtkWidget *widget, gpointer data)
{
  GNGUI *gg = &gn_gui;
  const gchar *name, *fontname;

  DPRINT_F(("print_font_handler()\n"));

  /* Unlike the terminal, printers are not limited to monospace fonts */
  name = gtk_font_button_get_title(GTK_FONT_BUTTON(widget));
  fontname = gtk_font_button_get_font_name(GTK_FONT_BUTTON(widget));

  /* Clear the default setup button when a font changes */
  if (strcmp(name, "Body font") == SAME) {
	if (strcmp(fontname, gg->def_print_font) != SAME)
		gtk_toggle_button_set_active(GTK_TOGGLE_BUTTON(data), FALSE);
  }
  else if (strcmp(name, "Header font") == SAME) {
	if (strcmp(fontname, gg->def_hdr_print_font) != SAME)
		gtk_toggle_button_set_active(GTK_TOGGLE_BUTTON(data), FALSE);
  }
  else {
	gtk_font_button_set_font_name(GTK_FONT_BUTTON(widget),
							gg->def_print_font);
  }

  return;
}


/*
** d e f a u l t _ p r i n t _ s e t u p _ h a n d l e r
**
** Handler for print setup dialog default button.
**
** Returns:	Nothing		Always
**
** The data argument is an array of struct Select.
*/
static void default_print_setup_handler(GtkWidget *widget, gpointer data)
{
  GNGUI *gg = &gn_gui;
  struct Select *sels = (struct Select *) data;
  int state;

  DPRINT_F(("default_print_setup_handler()\n"));

  state = gtk_toggle_button_get_active(GTK_TOGGLE_BUTTON(widget));
  gtk_button_set_label(GTK_BUTTON(widget), state ? "Yes" : "No");

  if (state) {
	gtk_font_button_set_font_name(GTK_FONT_BUTTON(
			sels[P_HEAD_INDEX].button), gg->def_hdr_print_font);
	gtk_font_button_set_font_name(GTK_FONT_BUTTON(
			sels[P_BODY_INDEX].button), gg->def_print_font);
	gtk_toggle_button_set_active(GTK_TOGGLE_BUTTON(
			sels[P_FLG_INDEX].button), gg->def_hdrprint);
  }

  return;
}

#ifdef ADD_UNUSED
/*
** h e l l o
**
** Placeholder callback.
**
** Returns:	Nothing		Always
*/
static void hello(GtkWidget *widget, gpointer data)
{
  /* write a message describing the source */
  g_print("Hello from %s\n", (char *) data);

  return;
}
#endif

/*
** i n i t _ d a n d
**
** Initialise the drag and drop interface.
**
** Returns	OK	Always
**
** This function bales out early on error,
** and thus has multiple return points.
*/
static int init_dand(GNGUI *gg)
{
  GtkTargetEntry targets[] = {
	{ "text/plain", 0, 1 },
	{ "text/uri-list", 0, 2 }
  };
  int status = OK;

  DPRINT_F(("init_dand()\n"));

  /* Not much point in continuing */
  if (gg == NULL || gg->holder == NULL) return FAILED;

  /* Set up (global) access to the default clipboard.
  ** We use the higher-level clipboard functions, and not the
  ** low-level clipboard data selection stuff.
  */
  if ((Clipboard = gtk_clipboard_get_for_display(gdk_display_get_default(),
					GDK_SELECTION_CLIPBOARD)) == NULL)
	return FAILED;

  /* Drag and drop protocol in GTK+.
  **
  ** "signal-name"
  ** [function_call]
  ** (user action)
  **
  ** Source                        GTK+                         Destination
  ** ------                        ----                         -----------
  **                               (user begins drag)
  ** [begin_cb]<-------------------"drag-begin"
  **                               (user drops)
  **                               "drag-leave"---------------->[leave_cb]
  **                               "drag-drop"----------------->[drop_cb]
  **                               [gtk_drag_get_data]<---------[drop_cb]
  ** [data_cb]<--------------------"drag-data-get"
  ** [data_cb]-------------------->[gtk_selection_data_set]
  **                               "drag-data-received"-------->[received_cb]
  **                               [gtk_drag_finish]<-----------[received_cb]
  ** [end_cb]<---------------------"drag-end"
  **
  ** All possible source signals.
  **  "drag-data-get"  (implemented)
  **  "drag-data-delete"
  **  "drag-begin"
  **  "drag-end" (implemented)
  ** All possible destination signals.
  **  "drag-data-received"  (implemented)
  **  "drag-leave"
  **  "drag-motion"
  **  "drag-drop"  (implemented)
  */

  /* Clear all the local drop source data */
  gg->drop_type = NONE;
  gg->drop_text = NULL;
  gg->drop_len = 0;
  gg->drop_fd = -1;

  /* Set up the viewport window (which surrounds the vterm)
  ** as a drag source. This re-uses the clipread() function.
  ** Supply only text/plain, not text/uri-list.
  */
  gtk_drag_source_set(
		gg->holder,
		GDK_BUTTON3_MASK,
		targets,
		1,
		GDK_ACTION_COPY
		);
  g_signal_connect(
		G_OBJECT(gg->holder),
		"drag-data-get",
		G_CALLBACK(drag_data_handler),
		(void *) gg
		);
  g_signal_connect(
		G_OBJECT(gg->holder),
		"drag-begin",
		G_CALLBACK(drag_begin_handler),
		(void *) gg
		);
  g_signal_connect(
		G_OBJECT(gg->holder),
		"drag-end",
		G_CALLBACK(drag_end_handler),
		(void *) gg
		);

  /* Set up the viewport window as a drop destination.
  ** Accept both text/plain and text/uri-list.
  ** This re-uses the clipread() function.
  */
  gtk_drag_dest_set(
		gg->holder,
		GTK_DEST_DEFAULT_ALL,		/* flags */
		targets,			/* array of GtkTargetEntry */
		2,				/* number of targets */
		GDK_ACTION_COPY			/* actions bitmask */
		);
  g_signal_connect(
		G_OBJECT(gg->holder),
		"drag-drop",
		G_CALLBACK(drag_drop_handler),
		(void *) gg
		);
  g_signal_connect(
		G_OBJECT(gg->holder),
		"drag-data-received",
		G_CALLBACK(drop_data_handler),
		(void *) gg
		);

  return status;
}


/*
** i n i t _ d e f a u l t s
**
** Initialise application settings and parse arguments.
**
** Returns:	Modified argc	Success
**		-1		Failure
**
** The options are used to initialise and update the gg->cur_xxx values,
** and at the end of the initialisation checks valid values are copied to
** the program defaults (gg->def_xxx).
**
** Command line options override GConf options which in turn override
** compiled-in defaults.
*/
static int init_defaults(int argc, char *argv[])
{
  GNGUI *gg = &gn_gui;
  size_t t;
  int j;
  char raw[16], buff[S_BUF_SIZ];

  DPRINT_F(("init_defaults()\n"));

  /* Start with all defaults in use */
  gg->use_def_font = TRUE;
  gg->use_def_color = TRUE;
  gg->use_def_cursor = TRUE;
  gg->use_def_print = TRUE;

  /* Load a default terminal font name. If we don't have the default font
  ** family on this machine select the first monospace font found.
  ** (init_interface() has already checked for the existence of at least
  ** one such font.)
  */
  if ((j = get_mono_fonts(DEFAULT_FONT_NAME, buff)) == 0) {
	strncpy(gg->def_fontname, buff, F_MAX);
	strncpy(gg->cur_fontname, buff, F_MAX);
  }
  else {
	strncpy(gg->def_fontname, DEFAULT_FONT_NAME, F_MAX);
	strncpy(gg->cur_fontname, DEFAULT_FONT_NAME, F_MAX);
  }

  /* Set up the default printer fonts regardless */
  strncpy(gg->def_print_font, DEFAULT_PRINT_FONT_NAME, F_MAX);
  strncpy(gg->cur_print_font, DEFAULT_PRINT_FONT_NAME, F_MAX);
  strncpy(gg->def_hdr_print_font, DEFAULT_PRINT_FONT_NAME, F_MAX);
  strncpy(gg->cur_hdr_print_font, DEFAULT_PRINT_FONT_NAME, F_MAX);

  /* Set the default debugging font regardless */
  strncpy(gg->def_dbg_fontname, DEFAULT_DBG_FONT_NAME, F_MAX);
  strncpy(gg->cur_dbg_fontname, DEFAULT_DBG_FONT_NAME, F_MAX);

  /* Set up a fixed default colormap for terminal and cursor */
  gg->scheme_flg = (set_default_colors(gg) == OK) ? TRUE : FALSE;

  /* Check the current GConf settings, which set some default options.
  ** A failure isn't fatal, but is reported once the main window is
  ** available.
  */
  if (gg->gconf_client != NULL && check_gconf() == FAILED) {
	gconf_client_notify_remove(gg->gconf_client, gg->gconf_id);
	gconf_client_remove_dir(gg->gconf_client, GCONF_PATH, NULL);
	/* g_free(GCONF_PATH) */
	g_object_unref(gg->gconf_client);
	gg->gconf_client = NULL;
	gg->gconf_id = 0;
  }

  /* Saving and restoring the current window size, current rings and files,
  ** and current colors, fonts and cursors would certainly be useful, but
  ** Gnome deprecates the saving of actual window positions.  Nor is GConf
  ** supposed to be used for state information.  The default window position
  ** is therefore stored in a file in the configuration directory.
  ** o_startposition itself is read from the GConf database on startup.
  */

  /* Get the startup data from the GTKRC file */
  (void) load_settings();

  /* Elvis calls the GUI init() function _before_ parsing its own
  ** options; the -G option will have been removed, but no others.
  ** Any options that the GUI doesn't understand must be left alone,
  ** which means that standard popt() parsing is impractical.
  ** Do our own parsing, removing Gnome GUI arguments from argv.
  */
  argc = parse_opts(argc, argv);

  /* See if we've set useable options, and if the range check fails,
  ** reset the option value to its defaults.  This is a pretty low-level
  ** approach, but it's quick and easy and keeps most of the option data
  ** in one place, the OPTDESC structure.
  **
  ** It doesn't check any string options, since they have no real default.
  */
  for (t = 0 ; t < (sizeof(optvals) / sizeof(optvals[0])) ; t++) {
	if (optdescs[t].dflt == NULL) continue;
	sprintf(buff, "%ld", optvals[t].value.number);
	if (optisnumber(&optdescs[t], &optvals[t], buff) == -1)
		optvals[t].value.number = strtol(optdescs[t].dflt, NULL, 0);
  }

#if 0
  int optisstring(
		OPTDESC	*desc,	/* description of the option */
		OPTVAL	*val,	/* value of the option */
		CHAR	*newval,/* value the option should have (as a string) */
		)
#endif

  /* Make sure that Elvis has access to the defaults for the :set command */
  optisstring(&optdescs[O_TFONT], &optvals[O_TFONT], gg->cur_fontname);
  optisstring(&optdescs[O_FCOLOR], &optvals[O_FCOLOR], gg->cur_forename);
  optisstring(&optdescs[O_BCOLOR], &optvals[O_BCOLOR], gg->cur_backname);
  optisstring(&optdescs[O_CCOLOR], &optvals[O_CCOLOR], gg->cur_cursorname);
#ifdef DEBUG_MENU
  optisstring(&optdescs[O_DFONT], &optvals[O_DFONT], gg->cur_dbg_fontname);
#endif
  /* def_cursorblink is set to the default option value regardless */
  gg->cur_cursorblink = gg->def_cursorblink = o_cursorblink;
  /* def_hdrprint is set to the default option value regardless */
  gg->cur_hdrprint = gg->def_hdrprint = o_printheader;
  /* o_showtoolbar is set up when the first editing window is created */
  optisstring(&optdescs[O_PFONT], &optvals[O_PFONT], gg->cur_print_font);
  optisstring(&optdescs[O_HFONT], &optvals[O_HFONT], gg->cur_hdr_print_font);

  /* Register our new options for use by Elvis' :set command.
  ** Function call is:
  ** void optinsert(char *domain, int nopts, OPTDESC desc[], OPTVAL val[])
  */
  optinsert("gnome", QTY(optdescs), optdescs, optvals);

  /* Set the current fonts using a pango font description, eg:
  ** "Terminal regular 10"
  ** "Monospace italic 12"
  ** It's difficult to see how we can continue without a font,
  ** so an error is fatal.
  */
  if (set_current_font(gg, gg->cur_fontname, TRUE) == FAILED)
	return -1;

  /* Note taken from Elvis' map.c:
  **
  ** mapinsert adds a map to the map table, or replaces an existing map.
  ** New maps will be added to the end of the list of maps.
  **
  ** "rawin" is the first argument to a ":map" command and "cooked" is the
  ** second.  "rawlen" and "cooklen" are their lengths; the strings don't need
  ** to be NUL terminated.  The "rawin" string may be either the characters
  ** to be mapped, or a key label which is supported by the GUI; this function
  ** will call the GUI's keylabel() function to perform the conversion.  This
  ** function will allocate copies of both of these strings, so the calling
  ** function can discard its own copies as soon as this function returns.
  **
  ** "label" is either NULL or a key label. The :map command always passes NULL,
  ** and the GUI's init will usually call this function with a label string. The
  ** label should be NUL-terminated, and its storage space cannot be discarded
  ** after the map because this function does *NOT* make a copy of the label.
  **
  ** "flags" indicates when the map should be effective.
  */
  for (t = 0 ; t < (sizeof(mapkeys) / sizeof(mapkeys[0])) ; t++) {
	if (mapkeys[t].sym == GDK_Delete || mapkeys[t].sym == GDK_KP_Delete)
		sprintf(raw, "%c", '\177');
	else
		sprintf(raw, "%c%04lx", '\013', (long) mapkeys[t].sym);
	mapinsert(
		raw,			/* Actual characters sent */
		(int) strlen(raw),	/* Number of actual characters sent */
		mapkeys[t].cooked,	/* Apparent characters sent */
		(int) strlen(mapkeys[t].cooked), /* Number of characters sent */
		mapkeys[t].label,	/* The key label */
		mapkeys[t].flags,	/* When the map should be active */
		NULL			/* mapmode value, NULL for any */
		);
  }

  return argc;
}


/*
** i n i t _ g n o m e _ w i n
**
** Initialise the Gnome application main window.
**
** Returns:	OK		Success
**		FAILED		Otherwise
**
** This function is called only from init().
** It sets up a series of widgets, one inside the other, in the order:
** vterm => viewport => scrollwin => gnomeapp
** The viewport widget is needed to allow the vterm to be removed and
** replaced as the main body of the code switches between different files.
** The scrollwin widget is needed to permit controlled scrolling of the
** vterm.
**
** Note that as far as the main body of the Elvis code is concerned the
** "editing window" is a struct Edwin; a series of these structures is
** linked off the main struct GnomeGui, with each EDWIN having its own
** vterm.  If Elvis opens a new window, what it gets is the same GNGUI
** with another set of linked EDWINs.
**
** Data that deals with the Gnome GUI is attached to GNGUI, while EDWIN
** structures hold the data specific to a given editing window.
**
** Other set of structures (linked lists) handle the file editing rings
** and the file arglists.
*/
static int init_gnome_win(GNGUI *gg)
{
  GtkSettings *settings;
  GdkModifierType state;
  GtkRequisition req;
  gint size;
  int status = OK;
  char *pc;

  DPRINT_F(("init_gnome_win()\n"));

  /* Set the unused 'next' pointer */
  gg->next = NULL;

  /* Initialise some Gnome variables.  The default rows and columns
  ** are hard-wired into VTE's initialisation code, so our defaults
  ** have to be set later when everything has been started and can
  ** be re-configured.
  */
  gg->def_rows = DEF_ROWS;
  gg->def_cols = DEF_COLS;

  /* Reserve these palette indices for vte_terminal's default
  ** background and foreground colors.
  */
  gg->font_codes[0] = gg->font_codes[7] = 1;

  /* Create the main Gnome widget and get the topmost widget
  ** in its hierarchy. (Should be the same.)
  */
  gg->gwindow = gnome_app_new("Elvis", "Editor");
  gg->toplevel = gtk_widget_get_toplevel(GTK_WIDGET(gg->gwindow));

  /* Don't automatically save the dock's layout configuration */
  gnome_app_enable_layout_config(GNOME_APP(gg->gwindow), FALSE);

  /* Connect the basic signals */
  g_signal_connect(
		gg->gwindow,
		"destroy",
		G_CALLBACK(destroy_handler),
		NULL
		);

  /* Connect a signal handler to track window resize events */
  g_signal_connect(
		gg->toplevel,
		"configure-event",
		G_CALLBACK(resize_handler),
		(void *) gg
		);

  /* Connect a signal handler to track window state events */
  g_signal_connect(
		gg->toplevel,
		"window-state-event",
		G_CALLBACK(state_handler),
		(void *) gg
		);

  /* Get the current icon theme name */
  settings = gtk_widget_get_settings(GTK_WIDGET(gg->gwindow));
  pc = NULL;
  g_object_get((GObject *) settings, "gtk-icon-theme-name", &pc, NULL);
  DPRINT_I(("init_gnome_win(): icon theme is |%s|\n",
					pc == NULL ? "NULL" : pc));
  if (pc != NULL) {
	strncpy(gg->theme_name, pc, F_MAX);
	g_free(pc);
  }
  else
	strcpy(gg->theme_name, "hicolor");

  /* Load the various program icons */
  if (get_icons() == FAILED)
	status = FAILED;
  /* Install any non-stock icons */
  else if (get_tools_icon() == FAILED)
	status = FAILED;

  /* Don't stop for failed icon loads */
  status = OK;

  /* Build the main menus */
  if (create_main_menu(gg) == FAILED)
	status = FAILED;
  /* Build the mouse menu */
  else if (create_mouse_menu(gg) == FAILED)
	status = FAILED;
  /* Build the toolbar */
  else if (create_toolbar(gg) == FAILED)
	status = FAILED;

  /* No point in going on */
  if (status == FAILED) {
	DPRINT_I(("init_gnome_win(): early failure\n"));
	goto End;
  }

#ifdef PRINT_MENU
  /* De-sensitise the main menu Print entry, depending */
  gtk_widget_set_sensitive(gtk_ui_manager_get_widget(
				gg->mainuimanager,
				"/ui/menubar/PrintMenu"),
				(gboolean) o_printmenu);
#endif

  /* Record the initial capslock key state and set the initial indicator icon */
  gdk_window_get_pointer(gtk_widget_get_parent_window(GTK_WIDGET(gg->gwindow)),
							NULL, NULL, &state);
  gg->caps_set = (state & GDK_LOCK_MASK) ? TRUE : FALSE;
  if (gg->caps_set)
	pc = (gg->uid == 0) ? "elvis-rdbl" : "elvis-gnbl";
  else
	pc = (gg->uid == 0) ? "elvis-rdgn" : "elvis-gngn";
  gg->indicators = gtk_image_new_from_stock(pc, GTK_ICON_SIZE_MENU);

  /* Install a key snooper to track the capslock key and catch the
  ** F1 help key.  Other uses will no doubt become apparent in time.
  ** (We don't care if this fails).
  */
  gg->snoop_id = gtk_key_snooper_install(snoop_handler, NULL);

  /* Check the availability of the dictionary and spellcheck programs */
  pc = (gg->dictionary[0] == '\0') ? DICT_UTIL : gg->dictionary;
  if (g_find_program_in_path(pc) == NULL) {
	pc = "";
	gtk_widget_set_sensitive(gtk_ui_manager_get_widget(
		gg->mainuimanager, "/ui/menubar/ToolsMenu/Dictionary"), FALSE);
  }
  optisstring(&optdescs[O_DICT], &optvals[O_DICT], pc);
  pc = (gg->spellcheck[0] == '\0') ? SPELL_UTIL : gg->spellcheck;
  if (g_find_program_in_path(pc) == NULL) {
	pc = "";
	gtk_widget_hide(gtk_ui_manager_get_widget(
		gg->mainuimanager, "/ui/menubar/ToolsMenu/Spellcheck"));
  }
  optisstring(&optdescs[O_SPELL], &optvals[O_SPELL], pc);

  /* Load any known notepad file names into the Notepads menu */
  (void) load_fileslist(NOTEPADS);
  /* Load any known document file names into the Documents menu */
  (void) load_fileslist(DOCUMENTS);

  /* Set up the window status bar components */
  gg->statusbar = gtk_statusbar_new();
  /* Status bar resize grip is TRUE by default */
  gtk_statusbar_set_has_resize_grip(GTK_STATUSBAR(gg->statusbar), FALSE);
  gg->msgbar = gtk_statusbar_new();
  gtk_statusbar_set_has_resize_grip(GTK_STATUSBAR(gg->msgbar), TRUE);

  /* Use this for parking later */
  gg->msgbarbox = gtk_event_box_new();
  gtk_event_box_set_visible_window(GTK_EVENT_BOX(gg->msgbarbox), FALSE);

  /* Allow Elvis' status updates reasonable space, and leave the
  ** rest for Elvis' message output.
  */
  gg->hpaned = gtk_hpaned_new();
  gtk_paned_set_position(GTK_PANED(gg->hpaned), STATUS_BAR_SIZE);

  /* Don't expand the statusbar, but do allow it to shrink */
  gtk_paned_pack1(GTK_PANED(gg->hpaned), gg->statusbar, FALSE, TRUE);
  gtk_paned_pack2(GTK_PANED(gg->hpaned), gg->msgbar, TRUE, TRUE);

  /* Get and pack a container for the indicator icon and the hbox */
  gg->barbox = gtk_hbox_new(FALSE, 2);
  gtk_box_pack_start(GTK_BOX(gg->barbox), gg->indicators, FALSE, FALSE, 2);
  gtk_box_pack_start(GTK_BOX(gg->barbox), gg->hpaned, TRUE, TRUE, 2);
  /* statusbar is the Gnome window status bar; it is packed into hpaned,
  ** which in turn is packed into barbox.
  */
  gnome_app_set_statusbar_custom(GNOME_APP(gg->gwindow), gg->barbox,
							gg->statusbar);

  /* Create the text entry widget and park it. The event box is used
  ** to catch events for widgets which don't have their own window.
  ** Since several commands use the entry widget, it needs to be
  ** protected by a lock.
  */
  INITLOCK(entry_lock);
  /* The gnome_entry widget is documented as "deprecated", but for no
  ** apparent reason and with no known replacement.  Use it until it
  ** breaks, and then revert to GtkEntry.  The only similar widget, a
  ** GtkEntry with gtk_entry_completion_set_popup_completion(), isn't
  ** appropriate here.
  */
#ifdef DEPRECATED
  gg->history_name[0] = '\0';
  strncat(gg->history_name, HISTORY_ELVIS, sizeof(gg->history_name) - 1);
  gg->entry = gnome_entry_new(gg->history_name);
  gnome_entry_set_max_saved(GNOME_ENTRY(gg->entry), HIST_COUNT);
#else
  gg->entry = gtk_entry_new();
#endif
  /* Fix the entry window height to the height of the msgbar,
  ** to prevent screen size jumps when they are interchanged.
  */
  gtk_widget_size_request(GTK_WIDGET(gg->msgbar), &req);
  gtk_widget_set_size_request(GTK_WIDGET(gg->entry), -1, req.height);
  /* Park the entry widget somewhere handy */
  gg->entrybox = gtk_event_box_new();
  gtk_event_box_set_visible_window(GTK_EVENT_BOX(gg->entrybox), FALSE);
  gtk_container_add(GTK_CONTAINER(gg->entrybox), gg->entry);
  gg->entry_id = g_signal_connect(
			gg->entry,
			"activate",	/* triggered by <CR> in entry */
			G_CALLBACK(entry_handler),
			(void *) gg
			);
  g_signal_handler_block(G_OBJECT(gg->entry), gg->entry_id);
  gg->move_cmd = NO_MOVE;

  /* Create a dummy edit window to set the initial size of the vbox */
  gg->vterm = vte_terminal_new();

  /* Set the terminal's font and rows and columns so that the surrounding
  ** windows can all be sized correctly.  The font name is converted
  ** internally into a PangoFontDescription before being passed to vte.
  **
  ** vte_terminal uses a fixed 24x80 size when created, and changes
  ** have to be made later when the window has been displayed.  The
  ** bar_cols value is an estimate, set in set_current_font().
  */
  vte_terminal_set_font_from_string(VTE_TERMINAL(gg->vterm), gg->cur_fontname);
  vte_terminal_set_size(VTE_TERMINAL(gg->vterm),
			gg->def_cols + gg->bar_cols, gg->def_rows);

  DPRINT_I(("init_gnome_win(): initial terminal size rows %d cols %d\n",
				gg->def_rows, gg->def_cols + gg->bar_cols));

  /* Set up a scrolling window, which contains a vbox which in turn
  ** contains a vterm.  Scrolling windows usually contain a viewport
  ** widget, but in this case the Elvis/VTE combination does all its
  ** own scrolling.
  */

  /* Create the scrolling window */
  gg->scrollwin = gtk_scrolled_window_new(NULL, NULL);
  gg->vert_range = gg->horiz_range = NULL;

  /* Get the various scrolling widgets */
  gg->horiz_adj = gtk_scrolled_window_get_hadjustment(
				GTK_SCROLLED_WINDOW(gg->scrollwin));
  gg->vert_adj = gtk_scrolled_window_get_vadjustment(
				GTK_SCROLLED_WINDOW(gg->scrollwin));
  gg->horiz_bar = gtk_scrolled_window_get_hscrollbar(
				GTK_SCROLLED_WINDOW(gg->scrollwin));
  gg->vert_bar = gtk_scrolled_window_get_vscrollbar(
				GTK_SCROLLED_WINDOW(gg->scrollwin));
  gg->horiz_range = GTK_RANGE(GTK_SCROLLBAR(gg->horiz_bar));
  gg->vert_range = GTK_RANGE(GTK_SCROLLBAR(gg->vert_bar));

  /* Monitor scrolling actions for debugging; for some reason the
  ** vertical scrollbar lower stepper doesn't change the GtkAdjustment
  ** value, although the upper stepper works fine.  (Both steppers have
  ** the correct sensitivity.) The change-value handler is used to update
  ** the value specifically for GTK_SCROLL_STEP_FORWARD. ??? ###
  */
  g_signal_connect(
		gg->vert_range,
		"change-value",
		G_CALLBACK(range_handler),
		(void *) gg
		);

  /* Configure the scrolling window */
  gtk_scrolled_window_set_shadow_type(GTK_SCROLLED_WINDOW(gg->scrollwin),
							GTK_SHADOW_NONE);
#if 0
  { /* Debugging block begin */
  /* Check the stepper sensitivities */
  g_print("init_gnome_win(): upper stepper sensitivity %d\n",
	gtk_range_get_upper_stepper_sensitivity(gg->vert_range));
  g_print("init_gnome_win(): lower stepper sensitivity %d\n",
	gtk_range_get_lower_stepper_sensitivity(gg->vert_range));
  /* Modify the scrolling window "slider-width" style property */
  gboolean w = 1;
  g_object_get((gpointer) gg->scrollwin, "window-placement-set", &w, NULL);
  g_print("init_gnome_win(): \"window-placement-set\" is %d\n", w);
  gtk_widget_style_get(GTK_WIDGET(gg->vert_range), "slider-width", &w, NULL);
  g_print("init_gnome_win(): \"slider-width\" is %d\n", w);
  } /* Debugging block end */
#endif

  /* If the scroll bars are shown here the final window won't be full size */
  gtk_scrolled_window_set_policy(
		GTK_SCROLLED_WINDOW(gg->scrollwin),
		GTK_POLICY_NEVER, GTK_POLICY_NEVER);

  /* Connect only the value-changed signal; other GtkAdjustment stuff is
  ** set by a gtk_adjustment_changed() call in Elvis' scrollbar function.
  */
  gtk_signal_connect(
		GTK_OBJECT(gg->vert_adj),
		"value-changed",
		G_CALLBACK(v_val_chg_handler),
		(void *) gg
		);

  /* A viewport sets all the GtkAdjustment _except_ the value, and is the
  ** preferred way to use a scrolling window.  However, there seems no way
  ** to disconnect a scrolling window's scrollbars from the viewport, so in
  ** this case a non-scrolling vbox is used instead.  This causes GTK+ to
  ** throw a run-time warning, which is trapped by glog_handler().
  gg->holder = gtk_viewport_new(NULL, NULL);
  */
  gg->holder = gtk_vbox_new(0, 0);
  /* Put the temporary vterm in the viewport */
  gtk_container_add(GTK_CONTAINER(gg->holder), GTK_WIDGET(gg->vterm));
  /* Put the viewport in the scrolling window */
  gtk_container_add(GTK_CONTAINER(gg->scrollwin), gg->holder);
  /* Finally! put the scrolling window in the main Gnome window */
  gnome_app_set_contents(GNOME_APP(gg->gwindow), gg->scrollwin);

#ifdef DPRINT_HH
  { /* Debugging block begin */
  g_print("init_gnome_win(): vterm alloc h %d w %d\n",
	gg->vterm->allocation.height, gg->vterm->allocation.width);
  g_print("init_gnome_win(): viewport alloc h %d w %d\n",
	gg->holder->allocation.height, gg->holder->allocation.width);
  g_print("init_gnome_win(): scrollwin alloc h %d w %d\n",
	gg->scrollwin->allocation.height, gg->scrollwin->allocation.width);
  g_print("init_gnome_win(): gwindow alloc h %d w %d\n",
	gg->gwindow->allocation.height, gg->gwindow->allocation.width);
  g_print("init_gnome_win(): toplevel alloc h %d w %d\n",
	gg->toplevel->allocation.height, gg->toplevel->allocation.width);
  } /* Debugging block end */
#endif

  /* Update the main window geometry */
  update_geometry(gg);

  /* tooltips are enabled by default
  gtk_tooltips_enable(gg->tooltips);
  */

  /* Display the newly created widgets.
  ** This draws the screen in its initial position, rather than
  ** exposing it when all the initial size and position updates are
  ** complete, but there seems no other way to get GTK to use the
  ** correct proportions.
  ** The result looks pretty clumsy, and needs investigating. ###
  */
  gtk_widget_show(gg->menubar);
  gtk_widget_show(gg->statusbar);
  gtk_widget_show(gg->msgbar);
  gtk_widget_show(gg->hpaned);
  gtk_widget_show(gg->indicators);
  gtk_widget_show(gg->barbox);
  gtk_widget_show(gg->vterm);
  gtk_widget_show(gg->holder);
  gtk_widget_show(gg->scrollwin);

  /* Display the entire window */
  gtk_widget_show(gg->gwindow);

  /* Is it worth using this to increase window redrawing performance ???
  gtk_widget_set_redraw_on_allocate(gg->gwindow, FALSE);
  */

  /* Now the windows are visible it's safe to update the scroll bars */
  gtk_scrolled_window_set_policy(
		GTK_SCROLLED_WINDOW(gg->scrollwin),
		GTK_POLICY_NEVER, GTK_POLICY_ALWAYS);

  /* Get the exact scrollbar stepper height and width */
  gtk_widget_style_get(GTK_WIDGET(GTK_RANGE(
		gtk_scrolled_window_get_vscrollbar(
		GTK_SCROLLED_WINDOW(gg->scrollwin)))),
		"stepper-size",
		&size,
		NULL
	);
  gg->stepper_height = (size >= 0) ? size : 0;
  gtk_widget_style_get(GTK_WIDGET(GTK_RANGE(
		gtk_scrolled_window_get_vscrollbar(
		GTK_SCROLLED_WINDOW(gg->scrollwin)))),
		"slider-width",
		&size,
		NULL
	);
  gg->stepper_width = (size >= 0) ? size : 0;

#ifdef DPRINT_HH
  { /* Debugging block begin */
  g_print("init_gnome_win(): vterm final alloc h %d w %d\n",
	gg->vterm->allocation.height, gg->vterm->allocation.width);
  g_print("init_gnome_win(): viewport final alloc h %d w %d\n",
	gg->holder->allocation.height, gg->holder->allocation.width);
  g_print("init_gnome_win(): scrollwin final alloc h %d w %d\n",
	gg->scrollwin->allocation.height, gg->scrollwin->allocation.width);
  g_print("init_gnome_win(): gwindow final alloc h %d w %d\n",
	gg->gwindow->allocation.height, gg->gwindow->allocation.width);
  g_print("init_gnome_win(): toplevel final alloc h %d w %d\n",
	gg->toplevel->allocation.height, gg->toplevel->allocation.width);
  } /* Debugging block end */
#endif

  /* Come here on failure */
  End:

  return status;
}


/*
** i n i t _ i n t e r f a c e
**
** Initialise GLib, GTK+ and Gnome.
**
** Returns:	OK		Success
**		FAILED		Otherwise
**
** This function bales out early on error,
** and thus has multiple return points.
*/
static int init_interface(GNGUI *gg)
{
  const gchar * const * langlist;
  char *pc;
  /* Build an argument list for the main Gnome application window */
  int child_argc = 1;
  char *child_argv[] = { NULL, NULL };
  child_argv[0] = gg->prog_name;

  DPRINT_F(("init_interface()\n"));

#ifdef DPRINT_II
  { /* Debugging block begin */
  g_print( "init_interface(): size optdescs %lu optvals %lu menu_entries %lu\n",
		(ULONG)(sizeof(optdescs) / sizeof(optdescs[0])),
		(ULONG)(sizeof(optvals) / sizeof(optvals[0])),
		(ULONG)(sizeof(menu_entries) / sizeof(menu_entries[0]))
		);
  } /* Debugging block end */
#endif

  /* Check that these structure and array sizes match */
  g_assert((sizeof(optdescs) / sizeof(optdescs[0])) ==
			(sizeof(optvals) / sizeof(optvals[0])));

#ifdef G_THREADS_ENABLED
  /* Initialise the threaded GLib (thread_supported => initialized).
  ** Threads aren't used directly in this code, since semaphores are
  ** used instead of g_static_mutex().  Does the GTK+ core use threads ???
  */
  if (!g_thread_supported()) g_thread_init(NULL);
#endif

  /* Add the .elvis.rc file to the GTK rc file path.  Leaks ### */
  gtk_rc_add_default_file(gnome_util_prepend_user_home("." GTKRC_ELVIS));

  /* Initialise i18n stuff.  However, no further use is made of i18n.
  ** Note that gtk_set_locale() is called automatically by gtk_init().
  */
  langlist = g_get_language_names();

  /* Initialise the current GNOME libraries for use by the application.
  **
  ** From the Gnome documentation:
  ** Every GNOME application will need to call this function
  ** (or gnome_program_initv()) early in its lifetime to initialize the
  ** various GNOME libraries in a consistent fashion. This function is
  ** very flexible in allowing the user to specify which modules should be
  ** initialised and any special parameter values that should be passed to
  ** these modules (along with processing commandline options).
  ** It loads the specified module_info, which is normally LIBGNOME_MODULE
  ** or LIBGNOMEUI_MODULE and pulls in all the dependencies. Programs that
  ** are not running in setuid or setgid mode will also load modules
  ** specified in the --load-modules and also in the GNOME_MODULES
  ** environment variable.
  ** After setting up the module loading, this function then calls
  ** (in order) gnome_program_preinit(), gnome_program_parse_args()
  ** and gnome_program_postinit().
  **
  ** Despite the sterling recommendation above, there's no way of adding
  ** any sort of brief descriptive message to the page or so of largely
  ** irrelevant options that "prog --usage" produces.
  */
  gg->gnome_program = gnome_program_init(
			"elvis",
			VERSION,
			LIBGNOMEUI_MODULE,
			child_argc,
			child_argv,
			GNOME_PROGRAM_STANDARD_PROPERTIES,
			GNOME_PARAM_HUMAN_READABLE_NAME, "Elvis",
			GNOME_PARAM_ENABLE_SOUND, FALSE,
			NULL
			);

  /* Check that the environment is accessible */
  pc = gnome_program_locate_file(
		NULL,				/* GnomeProgram */
		GNOME_FILE_DOMAIN_PIXMAP,	/* Gnome domain relative to */
						/* GNOME_PARAM_APP_DATADIR */
		"elvis",			/* Filename to find */
		FALSE,				/* Return a pathname */
		NULL				/* List of pathnames */
		);
  DPRINT_I(("init_interface(): Gnome path is |%s|\n", pc));
  if (pc != NULL) g_free(pc);

#ifdef CFGDIR_ELVIS
  /* Check that the configuration directory exists. The lack of a
  ** directory just means the use of defaults; the error is reported
  ** in loop() once the main GUI window is up and running.
  */
  pc = gnome_util_prepend_user_home(CFGDIR_ELVIS);
  gg->cfgdir_flg = g_file_test(pc, G_FILE_TEST_IS_DIR) ? TRUE : FALSE;
  g_free(pc);
  pc = NULL;
#endif

  /* Session management initialisation; since gnome-config is deprecated
  ** it's not clear how configuration files (apart from GConf stuff)
  ** should be handled. One alternative would be to use gtk_init_add()
  ** and gtk_quit_add() to register functions to read and save preferred
  ** settings.
  */
#ifdef SESSION_MGR
  /* Get the master client, set up by gnome_program_init() */
  gg->gnome_client = gnome_master_client();

  if (GNOME_CLIENT_CONNECTED(gg->gnome_client)) {
	/* We are connected, and can recover our configuration.
	** Note that gnome_config_push_prefix() and _pop_prefix()
	** are both now deprecated.
	*/
	gnome_config_push_prefix(
		gnome_client_get_config_prefix(gg->gnome_client));
	gnome_config_pop_prefix();

	/* Recover configuration ??? */
  }
  else {
	/* We are on our own */
	DPRINT_I(("init_interface(): Gnome Master Client not found\n"));
  }

  /* The two basic session management handlers */
  g_signal_connect(
		gg->gnome_client,
		"save-yourself",
		G_CALLBACK(client_save_handler),
		(void *) gg
		);
  g_signal_connect(
		gg->gnome_client,
		"die",
		G_CALLBACK(client_die_handler),
		(void *) gg
		);
#endif

  /* Set up GLib error logging, used to catch a known warning generated
  ** by the scrolling window widget.
  */
  gg->glog_id = g_log_set_handler(
		"Gtk",
		G_LOG_LEVEL_MASK |
		G_LOG_FLAG_FATAL |
		G_LOG_FLAG_RECURSION,
		glog_handler,
		NULL
		);

  /* Gnome can't run without a GConf server, so something's badly wrong */
  if ((gg->gconf_client = gconf_client_get_default()) == NULL)
	return FAILED;
  else if (!gconf_client_dir_exists(gg->gconf_client, GCONF_PATH, NULL))
	gg->gconf_client = NULL;
  else {
	/* Namespace (directory tree) for GConfClient to watch */
	gconf_client_add_dir(
			gg->gconf_client,
			GCONF_PATH,
			GCONF_CLIENT_PRELOAD_RECURSIVE,
			NULL
			);

	/* Set up GConf key change monitoring */
	gg->gconf_id = gconf_client_notify_add(
			gg->gconf_client,	/* Client */
			GCONF_PATH,		/* Attribute key */
			gconf_notify_handler,	/* Notify function */
			(gpointer) gg,		/* User data */
			NULL,			/* Remove function */
			NULL			/* Error return */
			);

	/* What can be done about GConf errors anyway ??? */
	gconf_client_set_error_handling(gg->gconf_client,
					GCONF_CLIENT_HANDLE_NONE);
  }

  /* Find out what colors are available (ie. test palette)
  if (get_default_colormap() == FAILED)
	return FAILED;
  */

  /* Check display and pango contexts; the Gnome window doesn't yet exist */
  if ((gg->display = gdk_display_get_default()) == NULL)
	return FAILED;
  if ((gg->screen = gdk_display_get_default_screen(gg->display)) == NULL)
	return FAILED;
  if ((gg->pango_context =
		gdk_pango_context_get_for_screen(gg->screen)) == NULL)
	return FAILED;

  /* Check that there's at least one monospace font available */
  if (check_mono_family(NULL) == FAILED)
	return FAILED;

  /* This is owned by GTK+ and should never be free'd.  However,
  ** this GUI doesn't at present use styles at all.
  */
  gg->style = gtk_style_copy(gtk_widget_get_default_style());

  return OK;
  }


/*
** s h o w _ w a r n i n g
**
** Show warning messages from initialisation.
**
** Returns:	TRUE		More messages
**		FALSE		No more messages
*/
static gboolean show_warning(gpointer data)
{
  static int count = 0;
  UINT flags = *(UINT *) data;
  int j;
  char *pc;

  DPRINT_F(("show_warning()\n"));

  /* If there's nothing to do, don't do it */
  if (flags == 0)
	return FALSE;

  /* For each bit set in data, pass the relevant message to Elvis */
  for (j = count ; j < 5 ; j++) {
	if (flags & (1 << j)) {
		if (j == 0) {
			pc = gnome_util_prepend_user_home(CFGDIR_ELVIS);
			msg(MSG_ERROR,
			"[s]Configuration directory $1 not found", pc);
		}
		else if (j == 1) {
			msg(MSG_ERROR,
			"[s]Configuration file $1 not found", GTKRC_ELVIS);
		}
		else if (j == 2) {
			msg(MSG_ERROR, "GConf user settings not found");
		}
		else if (j == 3) {
			msg(MSG_ERROR, "Default color scheme not parsed");
		}
		else if (j == 4) {
			msg(MSG_ERROR, "[s]Could not parse filename $1",
								BadFile);
			BadFile[0] = '\0';
		}
		msgflush();
		break;
	}
  }

  /* Move to the next available message */
  count = ++j;

  return (count > 4) ? FALSE : TRUE;
}


/*
** f i r s t c m d
**
** Execute Elvis' mainfirstcmd() function.
**
** Returns:	FALSE		Always
**
** The FALSE return ensures this functions removal from the GTK idle loop.
*/
static gboolean firstcmd(gpointer data)
{
  GNGUI *gg = &gn_gui;

  /* Execute -c command or -t tag, and if mainfirstcmd() returns TRUE
  ** then quit the editor.
  */
  if (mainfirstcmd(windefault))
	gtk_main_quit();
  else
	(void) eventdraw(gg->edhead);

  return FALSE;
}


/*
** l o a d _ s e t t i n g s
**
** Load the startup settings.
**
** Returns:	OK		Success
**		FAILED		Otherwise
*/
static int load_settings(void)
{
  GNGUI *gg = &gn_gui;
  FILE *fpin;
  UINT u;
  int status;
  int x, y, w, h;
  char *pb, *pf, buff[PATH_MAX] = { '\0' };

  DPRINT_F(("load_settings()\n"));

#ifdef CFGDIR_ELVIS
  if (gg->cfgdir_flg)
	pf = gnome_util_prepend_user_home(CFGDIR_ELVIS "/" GTKRC_ELVIS);
  else
	pf = gnome_util_prepend_user_home("." GTKRC_ELVIS);
#else
  pf = gnome_util_prepend_user_home("." GTKRC_ELVIS);
#endif

  /* Try and open the config file. Accept that the file may not be
  ** there, and that some lines may not parse correctly, simply
  ** returning an error status along with any actual data found.
  */
  status = FAILED;
  if ((fpin = fopen(pf, "r")) != NULL) {
	while (fgets(buff, PATH_MAX, fpin) != NULL) {
		pb = buff;
		/* Clear leading whitespace */
		while (*pb == ' ' && *pb == '\t') pb++;
		/* Lose comments and empty lines */
		if (*pb == '#' || *pb == '\n' || *pb == '\0') continue;
		/* Clear any trailing newline */
		while (*pb != '\0') pb++;
		if (pb != buff && *(pb - 1) == '\n')
			*(pb - 1) = '\0';

		/* Check for a valid option string */
		for (u = 0 ; u < sizeof(rcoptions) / sizeof(rcoptions[0]) ;
								u++) {
			if (strncmp(buff, rcoptions[u].name, rcoptions[u].len)
								== SAME) {
				pb = buff + rcoptions[u].len;
				break;
			}
		}

		/* Parse the result */
		switch (u) {
		case 0:
			if (sscanf(pb, "%dx%d%*c%d%*c%d",
						&w, &h, &x, &y) == 4) {
				gg->main_start_x = x;
				gg->main_start_y = y;
			}
			else {
				gg->main_start_x = -1;
				gg->main_start_y = -1;
			}
			status = OK;
			break;
		case 1:
			/* Note that this file may no longer exist */
			gg->lastprint[0] = '\0';
			strncat(gg->lastprint, pb, sizeof(gg->lastprint) - 1);
			status = OK;
			break;
		default:
			status = FAILED;
			break;
		}
	}
  }
  if (fpin != NULL) fclose(fpin);
  g_free(pf);
  pf = NULL;

  return status;
}


/*
** s t o r e _ s e t t i n g s
**
** Store the startup settings.
**
** Returns:	OK		Success
**		FAILED		Otherwise
*/
static int store_settings(int posn)
{
  GNGUI *gg = &gn_gui;
  FILE *fpout = NULL;
  GError *err = NULL;
  gsize len;
  gchar *data;
  gboolean rc;
  int w, h, x, y;
  int j, flg, linecount, status = OK;
  char *pc, *pf, **ppc, **lines;

  DPRINT_F(("store_settings()\n"));

#ifdef CFGDIR_ELVIS
  if (gg->cfgdir_flg)
	pf = gnome_util_prepend_user_home(CFGDIR_ELVIS "/" GTKRC_ELVIS);
  else
	pf = gnome_util_prepend_user_home("." GTKRC_ELVIS);
#else
  pf = gnome_util_prepend_user_home("." GTKRC_ELVIS);
#endif

  /* Read the whole gtrc file into memory */
  rc = g_file_get_contents(
			pf,		/* Filename */
			&data,		/* Buffer */
			&len,		/* Bytes read */
			&err		/* Error code */
			);

  /* The GTKrc file may not in fact exist, so don't report failure */
  if (!rc)
	return FAILED;

  /* Break the data up into lines */
  for (linecount = 0, pc = data ; *pc != '\0' ; pc++)
	if (*pc == '\n') linecount++;
  /* The file may not have a terminating newline */
  if (pc != data && *(pc - 1) != '\n')
	linecount++;

  /* Get memory for the line header buffer */
  if ((lines = g_try_malloc((linecount + 1) * sizeof(*lines))) == NULL) {
	msg(MSG_ERROR, "Cannot allocate memory for configuration file");
	return FAILED;
  }

  /* And initialise the header */
  for (flg = TRUE, j = 0, pc = data ; *pc != '\0' ; pc++) {
	if (flg) {
		lines[j++] = pc;
		flg = FALSE;
	}
	if (*pc == '\n') {
		*pc = '\0';
		flg = TRUE;
	}
  }
  lines[j] = NULL;

  /* Only update the main window position values if we need them */
  if (posn) {
	gtk_window_get_position(GTK_WINDOW(gg->toplevel), &x, &y);

	/* There's no point in trying to save the height and width;
	** the window manager won't use them.
	gtk_window_get_size(GTK_WINDOW(gg->toplevel), &w, &h);
	*/
	w = h = 0;
  }

  /* Clear the option set flags */
  for (j = 0 ; j < (sizeof(rcoptions) / sizeof(rcoptions[0])) ; j++)
	rcoptions[j].set = FALSE;

  /* Write out an updated data file */
  if ((fpout = fopen(pf, "w")) == NULL)
	status = FAILED;
  else {
	for (ppc = lines ; *ppc != NULL ; ppc++) {
		/* Drop empty lines */
		if (*ppc[0] == '\0') continue;
		/* Update the start position only when explicitly commanded */
		if (posn) {
			/* Change only the start position */
			if (strncmp(*ppc, rcoptions[0].name, rcoptions[0].len)
								== SAME) {
				rcoptions[0].set = TRUE;
				fprintf(fpout,
					"geometry=%dx%d+%d+%d\n", w, h, x, y);
			}
			else
				fprintf(fpout, "%s\n", *ppc);
		}
		else {
			/* Change anything but the start position */
			if (strncmp(*ppc, rcoptions[1].name, rcoptions[1].len)
								== SAME) {
				rcoptions[1].set = TRUE;
				/* Update the lastprint name if necessary */
				if (gg->lastprint[0] != '\0') {
					fprintf(fpout, "%s%s\n",
						rcoptions[1].name,
						gg->lastprint);
				}
				else {
					/* Just drop the line on floor */
					continue;
				}
			}
			else
				fprintf(fpout, "%s\n", *ppc);
		}
	}

	/* Add the initial position, if it hasn't been set */
	if (!rcoptions[0].set && posn) {
		fprintf(fpout, "geometry=%dx%d+%d+%d\n", w, h, x, y);
	}

	/* Add the last file printed, if it hasn't been set */
	if (!rcoptions[1].set && !posn && gg->lastprint[0] != '\0') {
		fprintf(fpout, "%s%s\n", rcoptions[1].name, gg->lastprint);
	}

	/* Tidy up */
	fclose(fpout);
	fpout = NULL;
  }

  /* More tidying up */
  g_free(lines);
  lines = NULL;
  g_free(pf);
  pf = NULL;

  return status;
}


#ifdef DEBUG_MENU
/*
** g e t _ p r i n t _ s e t t i n g s
**
** Get the current print settings.
**
** Returns	Nothing		Always
*/
static void get_print_settings(void)
{
#if 0
  #define GTK_PRINT_SETTINGS_PRINTER        "printer"
  #define GTK_PRINT_SETTINGS_ORIENTATION    "orientation"
  #define GTK_PRINT_SETTINGS_PAPER_FORMAT   "paper-format"
  #define GTK_PRINT_SETTINGS_PAPER_WIDTH    "paper-width"
  #define GTK_PRINT_SETTINGS_PAPER_HEIGHT   "paper-height"
  #define GTK_PRINT_SETTINGS_USE_COLOR      "use-color"
  #define GTK_PRINT_SETTINGS_COLLATE        "collate"
  #define GTK_PRINT_SETTINGS_REVERSE        "reverse"
  #define GTK_PRINT_SETTINGS_DUPLEX         "duplex"
  #define GTK_PRINT_SETTINGS_QUALITY        "quality"
  #define GTK_PRINT_SETTINGS_N_COPIES       "n-copies"
  #define GTK_PRINT_SETTINGS_NUMBER_UP      "number-up"
  #define GTK_PRINT_SETTINGS_RESOLUTION     "resolution"
  #define GTK_PRINT_SETTINGS_SCALE          "scale"
  #define GTK_PRINT_SETTINGS_PRINT_PAGES    "print-pages"
  #define GTK_PRINT_SETTINGS_PAGE_RANGES    "page-ranges"
  #define GTK_PRINT_SETTINGS_PAGE_SET       "page-set"
  #define GTK_PRINT_SETTINGS_DEFAULT_SOURCE "default-source"
  #define GTK_PRINT_SETTINGS_MEDIA_TYPE     "media-type"
  #define GTK_PRINT_SETTINGS_DITHER         "dither"
  #define GTK_PRINT_SETTINGS_FINISHINGS     "finishings"
  #define GTK_PRINT_SETTINGS_OUTPUT_BIN     "output-bin"
  #define GTK_PRINT_SETTINGS_OUTPUT_FILE_FORMAT  "output-file-format"
  #define GTK_PRINT_SETTINGS_OUTPUT_URI          "output-uri"
#endif

  /* Possible printer setting names and their types.
  **
  ** The enumerated types are:
  ** N None
  ** B gboolean
  ** C gchar *
  ** D gdouble
  ** E enum
  ** I gint
  ** R range (struct)
  */
  enum typ { N, B, C, D, E, I, R };
  struct Setting {
	const gchar *name;
	enum typ type;
	union {
		int j;
		gchar * s;
		double f;
	} value;
	char * format;
  } settings[] = {
	{GTK_PRINT_SETTINGS_PRINTER, C, { 0 }, "%s" }, /* gchar * */
	{GTK_PRINT_SETTINGS_ORIENTATION, E, { 0 }, "%s" }, /* enum */
	{GTK_PRINT_SETTINGS_PAPER_FORMAT, E, { 0 }, "%s" }, /* enum */
	{GTK_PRINT_SETTINGS_PAPER_WIDTH, D, { 0 }, "%s" }, /* gdouble */
	{GTK_PRINT_SETTINGS_PAPER_HEIGHT, D, { 0 }, "%s" }, /* gdouble */
	{GTK_PRINT_SETTINGS_USE_COLOR, B, { 0 }, "%s" }, /* gboolean */
	{GTK_PRINT_SETTINGS_COLLATE, B, { 0 }, "%s" }, /* gboolean */
	{GTK_PRINT_SETTINGS_REVERSE, B, { 0 }, "%s" }, /* gboolean */
	{GTK_PRINT_SETTINGS_DUPLEX, E, { 0 }, "%s" }, /* enum */
	{GTK_PRINT_SETTINGS_QUALITY, E, { 0 }, "%s" }, /* enum */
	{GTK_PRINT_SETTINGS_N_COPIES, I, { 0 }, "%s" }, /* gint */
	{GTK_PRINT_SETTINGS_NUMBER_UP, I, { 0 }, "%s" }, /* gint */
	{GTK_PRINT_SETTINGS_RESOLUTION, I, { 0 }, "%s" }, /* gint */
	{GTK_PRINT_SETTINGS_SCALE, D, { 0 }, "%s" }, /* gdouble */
	{GTK_PRINT_SETTINGS_PRINT_PAGES, E, { 0 }, "%s" }, /* enum */
	{GTK_PRINT_SETTINGS_PAGE_RANGES, R, { 0 }, "%s" }, /* struct */
	{GTK_PRINT_SETTINGS_PAGE_SET, E, { 0 }, "%s" }, /* enum */
	{GTK_PRINT_SETTINGS_DEFAULT_SOURCE, C, { 0 }, "%s" }, /* gchar * */
	{GTK_PRINT_SETTINGS_MEDIA_TYPE, C, { 0 }, "%s" }, /* gchar * */
	{GTK_PRINT_SETTINGS_DITHER, C, { 0 }, "%s" }, /* gchar * */
	{GTK_PRINT_SETTINGS_FINISHINGS, C, { 0 }, "%s" }, /* gchar * */
	{GTK_PRINT_SETTINGS_OUTPUT_BIN, C, { 0 }, "%s" }, /* gchar * */
	{GTK_PRINT_SETTINGS_OUTPUT_FILE_FORMAT, C, { 0 }, "%s" }, /* gchar * */
	{GTK_PRINT_SETTINGS_OUTPUT_URI, C, { 0 }, "%s" } /* gchar * */
  };

  GNGUI *gg = &gn_gui;
  size_t t;
  const char *pcc;
  char buff[80];

  if (gg->print_settings != NULL)
	g_print("get_print_settings(): print_settings is NULL\n");
  else {
	for (t = 0 ; t < (sizeof(settings) / sizeof(settings[0])) ; t++) {
		/* gtk_print_settings() returns the value as a string */
		pcc = gtk_print_settings_get(gg->print_settings,
						settings[t].name);
		if (pcc == NULL) {
			g_print("get_print_settings(): %s is NULL\n",
						settings[t].name);
		}
		else {
			sprintf(buff, "%s%s is %s\n", "get_print_settings(): ",
					settings[t].name, settings[t].format);
			g_print(buff, pcc);
		}
#if 0
		g_object_get((GObject *) settings,
			settings[t].name, &(settings[t].value), NULL);
		g_print(buff, settings[t].value);
#endif
	}
  }

  return;
}
#endif

/*
** l o a d _ f i l e s l i s t
**
** Load a linked list of filenames.
**
** Returns:	OK		Success
**		FAILED		Otherwise
*/
static int load_fileslist(int type)
{
  GNGUI *gg = &gn_gui;
  FILE *fpin = NULL;
  UINT index;
  int flg, status = OK;
  char *pb, *pc, buff[PATH_MAX] = { '\0' };

  DPRINT_F(("load_fileslist()\n"));

#ifndef CFGDIR_ELVIS
  gg->cfgdir_flg = FALSE;
#endif

  /* Sort out the type of linked list we are building */
  if (gg->cfgdir_flg) {
	if (type == DOCUMENTS)
		pc = gnome_util_prepend_user_home(
				CFGDIR_ELVIS "/" DOCSLIST_ELVIS);
	else if (type == NOTEPADS)
		pc = gnome_util_prepend_user_home(
				CFGDIR_ELVIS "/" NOTESLIST_ELVIS);
	else
		pc = gnome_util_prepend_user_home(FILE_ELVIS);
  }
  else {
	if (type == DOCUMENTS)
		pb = DOCSLIST_ELVIS;
	else if (type == NOTEPADS)
		pb = NOTESLIST_ELVIS;
	else
		pb = FILE_ELVIS;
	pc = gnome_util_prepend_user_home(pb);
  }

  /* Try and open the data file; if successful, update the local list */
  flg = TRUE;
  if ((fpin = fopen(pc, "r")) != NULL) {
	g_free(pc);
	pc = NULL;
	index = 0;
	/* Ignore errors - we just get no file list */
	while (index < o_docscount && fgets(buff, PATH_MAX, fpin) != NULL) {
		pb = buff;
		/* Clear leading whitespace */
		while (*pb == ' ' && *pb == '\t') pb++;
		/* Lose comments and empty lines */
		if (*pb == '#' || *pb == '\n' || *pb == '\0') continue;
		/* Clear any trailing newline */
		while (*pb != '\0') pb++;
		if (pb != buff && *(pb - 1) == '\n')
			*(pb - 1) = '\0';

		if (buff[0] == '/') {
			pc = buff;	/* Absolute path */
			flg = FALSE;
		}
		else
			pc = gnome_util_prepend_user_home(buff); /* Relative */

		/* If it's not on the list, add it anyway */
		if (type == DOCUMENTS && !check_node(gg->docshead, pc)) {
			if ((status = add_node(&gg->docshead, pc, NULL,
							BEFORE)) == OK) {
				init_enode(gg->docshead);
				add_docsmenu(gg->docshead);
				index++;
				DPRINT_P(("load_fileslist(): added |%s|\n",
						gg->docshead->name));
			}
		}
		else if (type == NOTEPADS && !check_node(gg->noteshead, pc)) {
			if ((status = add_node(&gg->noteshead, pc, NULL,
							BEFORE)) == OK) {
				init_enode(gg->noteshead);
				add_notesmenu(gg->noteshead);
				index++;
				DPRINT_P(("load_fileslist(): added |%s|\n",
						gg->noteshead->name));
			}
		}
		/* Failures are simply ignored */
	}
  }
  if (fpin != NULL) fclose(fpin);
  if (flg) g_free(pc);
  pc = NULL;

  return status;
}


/*
** s t o r e _ d o c s l i s t
**
** Store the Documents list.
**
** Returns:	OK		Success
**		FAILED		Otherwise
**
** Note that if the list is empty (cleared) DOCSLIST_ELVIS
** is truncated to zero length.
*/
static int store_docslist(void)
{
  GNGUI *gg = &gn_gui;
  FILE *fpout = NULL;
  struct Enode *pn;
  int index, status = OK;
  char *pc;

  DPRINT_F(("store_docslist()\n"));

#ifdef CFGDIR_ELVIS
  if (gg->cfgdir_flg)
	pc = gnome_util_prepend_user_home(CFGDIR_ELVIS "/" DOCSLIST_ELVIS);
  else
	pc = gnome_util_prepend_user_home(DOCSLIST_ELVIS);
#else
  pc = gnome_util_prepend_user_home(DOCSLIST_ELVIS);
#endif

  /* Try and open the data file; if successful, write out the local list */
  if ((fpout = fopen(pc, "w")) != NULL && gg->docshead != NULL) {
	for (index = 0, pn = gg->docshead->prev ;
					pn != NULL && index < o_docscount ;
					pn = pn->prev, index++) {
		DPRINT_P(("store_docslist(): entry %d is |%s| |%s|\n",
						index, pn->name, pn->label));
		fprintf(fpout, "%s\n", pn->name);
		if (pn->prev == gg->docshead->prev) break;
	}
  }
  if (fpout != NULL) fclose(fpout);
  g_free(pc);
  pc = NULL;

  return status;
}


/*
** c h e c k _ g c o n f
**
** Check the GConf settings database.
**
** Returns:	OK		Success
**		FAILED		Otherwise
*/
static int check_gconf(void)
{
  GNGUI *gg = &gn_gui;
  GConfValue *pv;
  struct Gconfkey *pg;
  int j, status = OK;
  char buff[64];

  DPRINT_F(("check_gconf()\n"));

  /* Find out if GConf has a value for the key, and
  ** if it has update the local value accordingly.
  */
  for (pg = gconfkeys ; pg->key != NULL ; pg++) {
	pv = gconf_client_get_without_default(
			gg->gconf_client,
			(gchar *) pg->key,
			NULL
			);

	DPRINT_G(("check_gconf(): key value |%s|\n",
			pv == NULL ? "NULL" : gconf_value_to_string(pv)));

	/* If there's a value, update the local setting */
	if (pv != NULL) {
		switch (pv->type) {
		case GCONF_VALUE_INT:
			/* Use the existing parsers[] parser */
			DPRINT_G(("check_gconf(): setting %s to %d\n",
					pg->key, gconf_value_get_int(pv)));
			sprintf(buff, "%s=%d",
				optdescs[parsers[pg->index].idx].longname,
				gconf_value_get_int(pv));
			status = (*parsers[pg->index].parse)(
					buff,
					pg->index,
					parsers[pg->index].idx
					);
			DPRINT_G((
				"check_gconf(): index %d value %ld status %s\n",
				pg->index,
				optvals[pg->index].value.number,
				status == OK ? "OK" :
				(status == NK) ? "NK" : "FAILED"
				));
			break;
		case GCONF_VALUE_BOOL:
			DPRINT_G(("check_gconf(): setting %s to %d\n",
					pg->key, gconf_value_get_bool(pv)));
			optvals[pg->index].value.boolean =
						gconf_value_get_bool(pv);
			break;
		case GCONF_VALUE_STRING:
			/* Use the existing parsers[] parser; note that
			** unlike gconfkeys[], optvals[] and optdescs[]
			** the parsers[] table is of variable size.
			*/
			if (pg->option == NULL)
				buff[0] = '\0';
			else
				strcpy(buff, pg->option);
			j = strlen(buff);
			strncat(buff, gconf_value_get_string(pv),
							sizeof(buff) - j);
			DPRINT_G(("check_gconf(): got string |%s|\n", buff));
			status = (*parsers[pg->index].parse)(
					buff,
					pg->index,
					parsers[pg->index].idx
					);
#if 0
			/* Should this message be buffered up until the main
			** window is available ???
			*/
			if (status == FAILED) {
				msg(MSG_ERROR,
					"[s]Could not parse argument $1", buff);
				msgflush();
			}
#endif
			break;
		default:
			DPRINT_G((
				"check_gconf(): unknown case %d\n", pv->type));
			status = FAILED;
			break;
		}
	}
  }

  return status;
}


/*
** r e q u e s t _ c l i p
**
** Request a selection from the clipboard.
**
** Returns:	Bytes read	Success
**		0		Otherwise
**
** As a side effect the target of the first argument is updated
** to point to the clipboard data.  It is a temporary buffer
** which must be free'd with g_free().
*/
static size_t request_clip(char **ppt, size_t maxlen)
{
  size_t len;
  char *pb;

  DPRINT_F(("request_clip()\n"));

  if (Clipboard == NULL) return 0;

  if (gtk_clipboard_wait_is_text_available(Clipboard))
	pb = (char *) gtk_clipboard_wait_for_text(Clipboard);
  else
	pb = NULL;

  if (pb == NULL)
	len = 0;
  else
	len = strlen(pb);
  *ppt = pb;

  DPRINT_D(("request_clip(): len %lu maxlen %d buff %s\n",
		(ULONG) len, maxlen, pb == NULL ? "invalid" : "valid"));

  return len;
}


/*
**  u r l _ t o _ p a t h
**
** Convert a URL to a local path.
**
** Returns:	OK		Success
**		FAILED		Otherwise
**
** Dropped names are of the form (at least with Nautilus):
** file:///path/file\r\n
** For some of the other versions of path strings, see the
** drag_drop_handler() function notes.
*/
static int url_to_path(char *path, char *data, size_t len)
{
  size_t t;
  int status = FAILED;
  char *pb, *pc, *pd;

  DPRINT_F(("url_to_path()\n"));

  pb = path;
  for (t = 0, pc = data ; t < (len - 1) ; t++, pc++) {
	if (*pc == '\0' || *pc == '\r' || *pc == '\n') break;
	if (status == OK)
		*pb++ = *pc;
	/* We've found the URL separator; now find the pathname */
	if (*pc == ':') {
		for (pd = ++pc ; *pd != '\0' && *pd == '/' ; pd++)
			/* MT */ ;
		if (pd > pc + 2) pc = pd - 2;
		status = OK;
	}
  }
  *pb = '\0';

  DPRINT_D(("url_to_path(): path |%s| data |%s|\n", path, data));

  return status;
}


/*
** u n d o t _ p a t h
**
** Remove "." and ".." entries from a path.
**
** Returns:	OK		Success
**		FAILED		Otherwise
**
** This has a maximum directory depth of 63, after which it fails.
** The cleanup really needs to be done in the main body of the Elvis
** code.
*/
static int undot_path(char *oldpath, char *newpath, size_t len)
{
  int cur, state, status = OK;
  char *newdirs[64] = { NULL };
  char *pold, *pnew;

  DPRINT_F(("undot_path()\n"));

  newpath[0] = '\0';
  if (oldpath == NULL) return FAILED;
  if (strlen(oldpath) >= len) return FAILED;

  pnew = newpath;
  for (cur = 0, state = 1, pold = oldpath ;
			cur < 64 && state != 10 && *pold != '\0' ; pold++) {
	*pnew++ = *pold;
	switch (state) {
	case 1:
		if (*pold == '/') {
			newdirs[cur] = pnew;
			cur++;
			state = 2;
		}
		break;
	case 2:
		if (*pold == '.')
			state = 3;
		else
			state = 1;
		break;
	case 3:
		if (*pold == '/') {
			/* Found /./ */
			pnew = newdirs[cur - 1];
			state = 2;
		}
		else if (*pold == '.')
			state = 4;
		else
			state = 1;
		break;
	case 4:
		if (*pold == '/') {
			/* Found /../ */
			cur--;
			if (cur < 1) cur = 1;
			pnew = newdirs[cur - 1];
			state = 2;
		}
		else if (*pold == '.') {
			/* Found /... (ie. error) */
			state = 9;
		}
		else {
			/* But eg. ./..foo is permitted */
			state = 1;
		}
		break;
	case 9:
	default:
		state = 10;
		BadFile[0] = '\0';
		strncat(BadFile, oldpath, sizeof(BadFile) - 1);
		newpath[0] = '\0';
		status = FAILED;
		msg(MSG_ERROR, "[s]Could not parse filename $1", BadFile);
		msgflush();
		break;
	}
  }
  *pnew = '\0';

  DPRINT_Q(("undot_path(): changed |%s| to |%s|\n", oldpath, newpath));

  return status;
}


/*
** g e t _ i d
**
** Get a unique id number.
**
** Returns:	The next id number	Always
**
** The id rolls over at UINT_MAX; a session won't call for anywhere
** near that number of ids, and getting around the problem is slow.
*/
static UINT get_id(void)
{
  static UINT id = 0;

  DPRINT_V(("get_id()\n"));

  if (++id == UINT_MAX)
	id = 1;

  return id;
}


/*
** g e t _ e x n a m e
**
** Get the name of an EXCMD enum.
**
** Returns:	A pointer to the name	Success
**		NULL			Otherwise
*/
static char * get_exname(int cmd)
{
#ifdef DPRINT_AA
  /* Names of EXCMD enums from ex.h, used for debugging */
  static struct Ex_enums {
	int value;
	char *name;
  } ee[] = {
	{ 0, "EX_ABBR" },
	{ 1, "EX_ALIAS" },
	{ 2, "EX_ALL" },
	{ 3, "EX_APPEND" },
	{ 4, "EX_ARGS" },
	{ 5, "EX_AT" },
	{ 6, "EX_AUTOCMD" },
	{ 7, "EX_AUEVENT" },
	{ 8, "EX_AUGROUP" },
	{ 9, "EX_BANG" },
	{ 10, "EX_BBROWSE" },
	{ 11, "EX_BREAK" },
	{ 12, "EX_BROWSE" },
	{ 13, "EX_BUFFER" },
	{ 14, "EX_CALC" },
	{ 15, "EX_CASE" },
	{ 16, "EX_CC" },
	{ 17, "EX_CD" },
	{ 18, "EX_CHANGE" },
	{ 19, "EX_CHECK" },
	{ 20, "EX_CHREGION" },
	{ 21, "EX_CLOSE" },
	{ 22, "EX_COLOR" },
	{ 23, "EX_COMMENT" },
	{ 24, "EX_COPY" },
	{ 25, "EX_DEFAULT" },
	{ 26, "EX_DELETE" },
	{ 27, "EX_DIGRAPH" },
	{ 28, "EX_DISPLAY" },
	{ 29, "EX_DO" },
	{ 30, "EX_DOALIAS" },
	{ 31, "EX_DOAUTOCMD" },
	{ 32, "EX_DOPROTO" },
	{ 33, "EX_ECHO" },
	{ 34, "EX_EDIT" },
	{ 35, "EX_ELSE" },
	{ 36, "EX_EQUAL" },
	{ 37, "EX_ERRLIST" },
	{ 38, "EX_ERROR" },
	{ 39, "EX_EVAL" },
	{ 40, "EX_FILE" },
	{ 41, "EX_FOLD" },
	{ 42, "EX_FOR" },
	{ 43, "EX_GOTO" },
	{ 44, "EX_GLOBAL" },
	{ 45, "EX_GUI" },
	{ 46, "EX_HELP" },
	{ 47, "EX_IF" },
	{ 48, "EX_INSERT" },
	{ 49, "EX_JOIN" },
	{ 50, "EX_LAST" },
	{ 51, "EX_LET" },
	{ 52, "EX_LIST" },
	{ 53, "EX_LOCAL" },
	{ 54, "EX_LPR" },
	{ 55, "EX_MAKE" },
	{ 56, "EX_MAP" },
	{ 57, "EX_MARK" },
	{ 58, "EX_MESSAGE" },
	{ 59, "EX_MKEXRC" },
	{ 60, "EX_MOVE" },
	{ 61, "EX_NEXT" },
	{ 62, "EX_NOHLSEARCH" },
	{ 63, "EX_NORMAL" },
	{ 64, "EX_NUMBER" },
	{ 65, "EX_ONLY" },
	{ 66, "EX_OPEN" },
	{ 67, "EX_POP" },
	{ 68, "EX_PRESERVE" },
	{ 69, "EX_PREVIOUS" },
	{ 70, "EX_PRINT" },
	{ 71, "EX_PUSH" },
	{ 72, "EX_PUT" },
	{ 73, "EX_QALL" },
	{ 74, "EX_QUIT" },
	{ 75, "EX_READ" },
	{ 76, "EX_REDO" },
	{ 77, "EX_REGION" },
	{ 78, "EX_REWIND" },
	{ 79, "EX_SALL" },
	{ 80, "EX_SAFELY" },
	{ 81, "EX_SBBROWSE" },
	{ 82, "EX_SBROWSE" },
	{ 83, "EX_SET" },
	{ 84, "EX_SHELL" },
	{ 85, "EX_SHIFTL" },
	{ 86, "EX_SHIFTR" },
	{ 87, "EX_SLAST" },
	{ 88, "EX_SNEW" },
	{ 89, "EX_SNEXT" },
	{ 90, "EX_SOURCE" },
	{ 91, "EX_SPELL" },
	{ 92, "EX_SPLIT" },
	{ 93, "EX_SPREVIOUS" },
	{ 94, "EX_SREWIND" },
	{ 95, "EX_STAG" },
	{ 96, "EX_STACK" },
	{ 97, "EX_STOP" },
	{ 98, "EX_SUBAGAIN" },
	{ 99, "EX_SUBRECENT" },
	{ 100, "EX_SUBSTITUTE" },
	{ 101, "EX_SUSPEND" },
	{ 102, "EX_SWITCH" },
	{ 103, "EX_TAG" },
	{ 104, "EX_THEN" },
	{ 105, "EX_TRY" },
	{ 106, "EX_UNABBR" },
	{ 107, "EX_UNALIAS" },
	{ 108, "EX_UNBREAK" },
	{ 109, "EX_UNDO" },
	{ 110, "EX_UNFOLD" },
	{ 111, "EX_UNMAP" },
	{ 112, "EX_UNREGION" },
	{ 113, "EX_VERSION" },
	{ 114, "EX_VGLOBAL" },
	{ 115, "EX_VISUAL" },
	{ 116, "EX_WARNING" },
	{ 117, "EX_WHILE" },
	{ 118, "EX_WINDOW" },
	{ 119, "EX_WNEXT" },
	{ 120, "EX_WORDFILE" },
	{ 121, "EX_WORDS" },
	{ 122, "EX_WQUIT" },
	{ 123, "EX_WRITE" },
	{ 124, "EX_XIT" },
	{ 125, "EX_YANK" },
	{ 126, "EX_Z" }
  };

  DPRINT_F(("get_exname()\n"));

  if (cmd >= 0 && cmd < (int) (sizeof(ee) / sizeof(ee[0])))
	return ee[cmd].name;
  else
	return NULL;
#else
  if (cmd >= 0 && cmd < 127)
	return "valid";
  else
	return NULL;
#endif
}


/*
** g e t _ i c o n s
**
** Get the program logo and any icons not from stock.
**
** Supplies the program logo from the current theme, and if
** that can't be found supplies a logo from an internal buffer.
**
** Loads other non-standard program items from pixmap buffers,
** and sets up an icon factory.
**
** Returns:	OK		Success
**		FAILED		Otherwise
*/
static int get_icons(void)
{
  struct Icondata {
	char *name;
	const GdkPixdata *pixdata;
	GdkPixbuf *pixbuf;
	GtkIconSet *iconset;
  };

  static struct Icondata indic_icons[] = {
	{ "elvis-gnbl", &gnbl_pixdata, NULL, NULL },
	{ "elvis-rdbl", &rdbl_pixdata, NULL, NULL },
	{ "elvis-gngn", &gngn_pixdata, NULL, NULL },
	{ "elvis-rdgn", &rdgn_pixdata, NULL, NULL }
  };
  static struct Icondata down_icons[] = {
	{ "elvis-down-lge", &down_lge_pixdata, NULL, NULL },
	{ "elvis-down-wht", &down_wht_pixdata, NULL, NULL },
	{ "elvis-down-blk", &down_blk_pixdata, NULL, NULL },
	{ "elvis-down-std", &down_std_pixdata, NULL, NULL }
  };
  GNGUI *gg = &gn_gui;
  GtkIconFactory *iconfact = NULL;
  GError *error = NULL;
  GtkIconInfo *iconinfo = NULL;
  GdkPixbuf *elvispixbuf = NULL;
  gchar **pgpath;
  gint gcount;
  size_t t;
  int status = OK;

  DPRINT_F(("get_icons()\n"));

  /* Get the default icon theme */
  gg->theme = gtk_icon_theme_get_default();
  DPRINT_L(("get_icons(): icon theme is %s\n",
				gg->theme == NULL ? "NULL" : "valid"));

  /* Connect a signal to monitor theme changes */
  g_signal_connect(
		gg->theme,
		"changed",
		G_CALLBACK(theme_handler),
		(void *) gg
		);

  /* Add the current directory to the theme search path for testing
  gtk_icon_theme_append_search_path(gg->theme, "~/elvis");
  */

  /* Get the theme search path */
  gcount = 0;
  gtk_icon_theme_get_search_path(gg->theme, &pgpath, &gcount);
  DPRINT_L(("get_icons(): search path is %s, depth %d\n",
		pgpath == NULL ? "NULL" : "valid", gcount));
#ifdef DPRINT_LL
  { /* Debugging block begin */
  int j;
  for (j = 0 ; j < gcount ; j++)
	g_print("get_icons(): search path %d is %s\n", j, pgpath[j]);
  } /* Debugging block end */
#endif

  /* Search the theme paths for a given name.
  ** If we don't add the right (current) directory, this will fail.
  ** Avoid GTK_ICON_LOOKUP_USE_BUILTIN, since the external icon will
  ** then be used in preference to the internal.
  ** Why does this always succeed ???
  */
  iconinfo = gtk_icon_theme_lookup_icon(gg->theme, LOGO_ELVIS, 48, 0);
  DPRINT_L(("get_icons(): theme lookup iconinfo is %s\n",
				iconinfo == NULL ? "NULL" : "valid"));

  /* If the name isn't found, load a pixbuf from embedded pixdata */
  if (iconinfo == NULL) {
	/* elvis_pixdata is defined in elvis-icon.h.  The flag controls
        ** copying of raw pixel data; we use run-length encoded data.
	*/
	elvispixbuf = gdk_pixbuf_from_pixdata(&elvis_pixdata, FALSE, &error);
	DPRINT_L(("get_icons(): theme lookup pixbuf is %s\n",
				elvispixbuf == NULL ? "NULL" : "valid"));

	/* Add the icon directly to the current theme */
	gtk_icon_theme_add_builtin_icon(LOGO_ELVIS, 48, elvispixbuf);
  }

  /* Get a new icon factory */
  iconfact = gtk_icon_factory_new();
  DPRINT_L(("get_icons(): iconfact is %s\n",
			iconfact == NULL ? "NULL" : "valid"));
  status = (iconfact == NULL) ? FAILED : OK;

  /* Load the set of down icons for the toolbar.
  ** Do we need an icon factory _and_ themed icons ???
  */
  for (t = 0 ; status == OK && t < sizeof(down_icons) /
						sizeof(down_icons[0]) ; t++) {
	error = NULL;
	down_icons[t].pixbuf = gdk_pixbuf_from_pixdata(down_icons[t].pixdata,
							FALSE, &error);
	if (error == NULL) {
		gtk_icon_theme_add_builtin_icon(down_icons[t].name, 24,
							down_icons[t].pixbuf);
		down_icons[t].iconset = gtk_icon_set_new_from_pixbuf(
							down_icons[t].pixbuf);
		gtk_icon_factory_add(iconfact, down_icons[t].name,
							down_icons[t].iconset);
	}
	else
		status = FAILED;
  }

  /* Load a set of indicators for the status bar */
  for (t = 0 ; status == OK && t < sizeof(indic_icons) /
						sizeof(indic_icons[0]) ; t++) {
	error = NULL;
	indic_icons[t].pixbuf = gdk_pixbuf_from_pixdata(indic_icons[t].pixdata,
							FALSE, &error);
	if (error == NULL) {
		gtk_icon_theme_add_builtin_icon(indic_icons[t].name, 24,
							indic_icons[t].pixbuf);
		indic_icons[t].iconset = gtk_icon_set_new_from_pixbuf(
							indic_icons[t].pixbuf);
		gtk_icon_factory_add(iconfact, indic_icons[t].name,
							indic_icons[t].iconset);
	}
	else
		status = FAILED;
  }

  /* Add the factory to the default factory list */
  if (status == OK)
	gtk_icon_factory_add_default(iconfact);

  return status;
}

/*
** g e t _ t o o l s _ i c o n
**
** Get the toolbar Tools menu icon most suited to the current theme.
**
** Returns:	OK		Success
**		FAILED		Otherwise
*/
static int get_tools_icon(void)
{
  GNGUI *gg = &gn_gui;
  int status = OK;
  char *pc;

  if (gg == NULL)
	pc = "elvis-down-std";
  else if (strcmp(gg->theme_name, "gnome") == SAME)
	pc = "elvis-down-std";
  else if (strcmp(gg->theme_name, "HighContrast") == SAME)
	pc = "elvis-down-blk";
  else if (strcmp(gg->theme_name, "HighContrastInverse") == SAME)
	pc = "elvis-down-wht";
  else if (strcmp(gg->theme_name, "LargePrint") == SAME)
	pc = "elvis-down-lge";
  else if (strcmp(gg->theme_name, "hicolor") == SAME)
	pc = "elvis-down-std";
  else
	pc = "elvis-down-std";

  if (tools[LIST_RINGS].icon == NULL) {
	tools[LIST_RINGS].icon = (const gchar *)
		gtk_image_new_from_stock(pc, GTK_ICON_SIZE_SMALL_TOOLBAR);
  }
  else {
	gtk_image_set_from_icon_name((GtkImage *) tools[LIST_RINGS].icon, pc,
						GTK_ICON_SIZE_SMALL_TOOLBAR);
  }
  gtk_widget_show((GtkWidget *) tools[LIST_RINGS].icon);

  return status;
}


/*
** s i z e _ m a i n _ w i n d o w
**
** Show and set the main window size, based on the vterm size.
**
** Returns:	Nothing		Always
*/
static void size_main_window(GNGUI *gg, int actual)
{
  EDWIN *edwin;
  glong rows, cols;
  glong width, height, wdelta, hdelta;
  glong char_width, char_height;
  int xpad, ypad;

  DPRINT_F(("size_main_window()\n"));

  /* Just checking */
  if (gg == NULL || gg->edhead == NULL) return;

  /* Get the current terminal geometry */
  edwin = gg->edhead;
  rows = edwin->rows;
  cols = edwin->cols;
  vte_terminal_get_padding(VTE_TERMINAL(edwin->vterm), &xpad, &ypad);
  edwin->xpad = xpad;
  edwin->ypad = ypad;

  /* Get the current window sizes */
  wdelta = gg->toplevel->allocation.width - gg->scrollwin->allocation.width;
  hdelta = gg->toplevel->allocation.height - gg->scrollwin->allocation.height;
  char_width = edwin->pango_width;
  char_height = edwin->pango_height;
  width = wdelta + WIN_SHIM + xpad + char_width * (cols + gg->bar_cols);
  height = hdelta + WIN_SHIM + ypad + char_height * rows;

#ifdef DPRINT_HH
  { /* Debugging block begin */
  g_print(
  "size_m_w(): hdelta %ld ypad %d c_height %ld X rows %ld = %ld height %ld\n",
	hdelta, ypad, char_height, rows, char_height * rows, height);
  g_print(
  "size_m_w(): wdelta %ld xpad %d c_width %ld X cols %ld = %ld width %ld\n",
	wdelta, xpad, char_width, cols, char_width * cols, width);
  g_print(
  "size_m_w(): allocation top w %d h %d scroll w %d h %d vterm w %d h %d\n",
	gg->toplevel->allocation.width, gg->toplevel->allocation.height,
	gg->scrollwin->allocation.width, gg->scrollwin->allocation.height,
	edwin->vterm->allocation.width, edwin->vterm->allocation.height
	);
  g_print(
  "size_m_w(): requisition top w %d h %d scroll w %d h %d vterm w %d h %d\n",
	gg->toplevel->requisition.width, gg->toplevel->requisition.height,
	gg->scrollwin->requisition.width, gg->scrollwin->requisition.height,
	edwin->vterm->requisition.width, edwin->vterm->requisition.height
	);
  g_print(
  "size_m_w(): request w %ld h %ld term char w %ld h %ld row %ld col %ld\n",
	width, height,
	char_width, char_height,
	rows, cols
	);
  } /* Debugging block end */
#endif

  /* If the main window exists, set its size */
  if (GTK_WIDGET_MAPPED(gg->toplevel)) {
	if (actual) {
		DPRINT_H(("size_main_window(): setting actual size\n"));
		gtk_window_resize(GTK_WINDOW(gg->toplevel), width, height);
	}
	else {
		DPRINT_H(("size_main_window(): setting default size\n"));
		gtk_window_set_default_size(GTK_WINDOW(gg->toplevel),
							width, height);
	}
  }

  return;
}


/*
** u p d a t e _ g e o m e t r y
**
** Update the VTE terminal geometry.
**
** Returns:	Nothing		Always
*/
static void update_geometry(GNGUI *gg)
{
  EDWIN *edwin;
  GdkGeometry hints;
  GtkWidget *vt;
  glong char_width;
  glong char_height;
  glong rows, cols;
  int xpad, ypad;

  DPRINT_F(("update_geometry()\n"));

  /* Just in case */
  if (gg == NULL || (gg->edhead == NULL && gg->vterm == NULL))
	return;

  /* Find the terminal window to match */
  if ((edwin = gg->edhead) == NULL) {
	vt = gg->vterm;
	char_width = gg->pango_width;
	char_height = gg->pango_height;
	rows = gg->def_rows;
	cols = gg->def_cols;
  }
  else {
	vt = edwin->vterm;
	char_width = edwin->pango_width;
	char_height = edwin->pango_height;
	rows = edwin->rows;
	cols = edwin->cols;
  }

  DPRINT_R(("update_geometry(): rows %ld cols %ld\n", rows, cols));

  /* Applies to the vte_terminal scroll bar - not used.
  adjust = vte_terminal_get_adjustment(VTE_TERMINAL(vt));
  */
  vte_terminal_get_padding(VTE_TERMINAL(vt), &xpad, &ypad);

  /* To scroll correctly the vterm hints must be applied to the
  ** scrolling window.  The scrollwin widget frames are set to
  ** zero border thickness, but there still seems to be a finite
  ** frame thickness generated somehow, possibly by the viewport.
  ** Use 2 as a rough size correction.
  */
  hints.base_width = xpad + WIN_SHIM;
  hints.base_height = ypad + WIN_SHIM;
  hints.width_inc = char_width;
  hints.height_inc = char_height;

  /* The minimum size is that of the target geometry widget */
  hints.min_width = hints.base_width + hints.width_inc * MIN_COLS;
  hints.min_height = hints.base_height + hints.height_inc * MIN_ROWS;

  /* The usual gravity setting leaves a gap at the top and left edges.
  ** Gravity CENTER and STATIC don't seem to clear this, and nor
  ** does applying the correction to the vterm window itself.
  */
  hints.win_gravity = GDK_GRAVITY_NORTH_WEST;

  /* Once we're here, we may as well update everything anyway */

  /* Update the current values */
  if (edwin != NULL) {
	edwin->cols = cols;
	edwin->rows = rows;
	edwin->hints = hints;
	edwin->xpad = xpad;
	edwin->ypad = ypad;
  }

  /* Update the top-level window */
  gtk_window_set_geometry_hints(
			GTK_WINDOW(gg->toplevel), /* Target window */
			gg->scrollwin,		/* Target widget */
			&hints,			/* Geometry */
			GDK_HINT_MIN_SIZE |	/* Mask */
			GDK_HINT_BASE_SIZE |
			GDK_HINT_RESIZE_INC |
			GDK_HINT_WIN_GRAVITY
			);

  return;
}


/*
** w r i t e _ b u f f
**
** Write text to a window.
**
** Returns:	Nothing		Always
*/
static void write_buff(EDWIN *edwin, char *buff, size_t num)
{
  DPRINT_W(("write_buff(): size %u buff[0] |%c| num %u\n",
			sizeof(edwin->keybuff), buff[0], num));

  /* This happens somehow on shutdown ??? */
  if (edwin == NULL || edwin->vterm == NULL || buff == NULL) return;

  /* Buffering seems to _slow_ the interface a great deal,
  ** since up to 1024 bytes can be in storage at any point.
  ** Even a 32-byte buffer seems twitchy; should write_buff()
  ** just be replaced by direct calls. ??? ###
  */
#if 1
  vte_terminal_feed(VTE_TERMINAL(edwin->vterm), buff, (glong) num);
#else
  static int len = 0;
  int size = sizeof(edwin->keybuff);	/* Used for debugging only */
  char *pc;

  size = 32;
  g_assert(size <= sizeof(edwin->keybuff));

  /* If there's space in the buffer, fine */
  if ((edwin->keycount + num) < (size - 1)) {
	strncpy(edwin->keybuff + edwin->keycount, buff, num - 1);
	edwin->keycount += num;
  }
  else {
	/* Stop messing around and write out everything we've got */
	vte_terminal_feed(VTE_TERMINAL(edwin->vterm), edwin->keybuff,
							edwin->keycount);
	edwin->keycount = 0;
	vte_terminal_feed(VTE_TERMINAL(edwin->vterm), buff, num);
  }
#endif

  return;
}


/*
** t e r m i n a l
**
** Run a terminal shell.
**
** Returns:	OK		Success
**		FAILED		Otherwise
**
** There's some sort of bug in VTE; it becomes confused when a second
** terminal is opened (ie. Elvis runs a shell), _unless_ a terminal is
** already running.  It looks like an environment problem, but the
** working and failing environments appear identical to 'set'.  The
** symptoms are a doubling of characters in the parent terminal. ###
*/
static int terminal(char *cmd)
{
  GNGUI *gg = &gn_gui;
  char shell[64], command[64], geom[64];
  char home[PATH_MAX];
  char *argv[8] = { NULL };
  char *envv[8] = { NULL };
  pid_t pid;
  int status = OK;
  gchar *pc;

  DPRINT_F(("terminal()\n"));

  /* gnome-terminal is not easy to configure.
  ** The problem is that any xterm window title set in .bashrc is used
  ** automatically, and overrides command line arguments.  Furthermore,
  ** the --window-with-profile argument will accept only the default
  ** profile, but must be given for the --geometry argument to work.
  **  All in all, this seems the best, tho' uneasy, compromise.
  **
  ** Unfortunately, if gnome-terminal is given a command to run inside
  ** the shell, it runs the command and then immediately _closes_ the
  ** shell, so it can't easily be used for eg: the man command.
  */
  if ((pc = g_find_program_in_path("gnome-terminal")) == NULL)
	return FAILED;

  /* Set up argument strings */
  shell[0] = '\0';
  strncat(shell, pc, sizeof(shell) - 1);	/* "/usr/bin/xterm" ??? */
  strcpy(home, g_get_home_dir());
  sprintf(geom, "--geometry=%ldx%ld", o_shellcols, o_shellrows);

  /* Load the argument vector */
  argv[0] = "--window-with-profile=Default";
  argv[1] = geom;
  argv[2] = "--title=Elvis";
  argv[3] = o_shellmenu ? "--show-menubar" : "--hide-menubar";
  if (cmd != NULL) {
	strcpy(command, "--command=");
	strncat(command, cmd, 64 - sizeof("--command=") - 1);
	argv[4] = command;
  }

  /*  These should be the default environment.
  envv[0] = "TERM=xterm";
  envv[1] = "LINES=24";
  */

  /* Starts the specified command under a newly-allocated control
  ** pseudo-terminal. TERM is automatically set to reflect the
  ** terminal widget's emulation setting. If lastlog, utmp, or wtmp
  ** are TRUE, logs the session to the specified system log files.
  */
  pid = vte_terminal_fork_command(
		VTE_TERMINAL(gg->edhead->vterm), /* a vte terminal */
		shell,			/* the name of a binary */
		argv,			/* the argument list */
		envv,			/* the environment variables */
		home,			/* the start directory or NULL */
		FALSE,			/* log to the lastlog */
		FALSE,			/* log to the utmp/utmpx log */
		FALSE			/* log to the wtmp/wtmpx log */
		);

  return status;
}


/*
** t r a p _ h a n d l e r
**
** Handle Unix signals requesting a controlled shutdown,
** at present only SIGINT and SIGTERM.
**
** The signal handler function prototype is:
** typedef void (*sighandler_t)(int);
*/
static void trap_handler(int signum)
{
  signal(signum, SIG_IGN);	/* Ignore a repeated signal */

  quit_action(NULL, (gpointer) &gn_gui);

  return;
}


/*
** c r e a t e _ e d i t _ r i n g
**
** Create an edit ring.
**
** Returns:	number of nodes	Success
**		-1		Otherwise
*/
static int create_edit_ring(ENODE **ppn, char **list, int next)
{
  struct Enode *pt;
  int j, count, status;
  char **ppc, buff[PATH_MAX + 64];

  DPRINT_F(("create_edit_ring()\n"));

  /* Build a linked list of the new filenames, preserving their order and
  ** generating both long and short versions of the names. Call back into
  ** the main body of the Elvis code with ioabsolute() in order to
  ** get the current working directory.
  */
  status = OK;
  count = 0;
  pt = *ppn;
  for (ppc = list ; ppc != NULL && *ppc != NULL && status == OK ; ppc++) {
	buff[0] = '\0';
	strncat(buff, ioabsolute(*ppc), sizeof(buff) - 1);
	if ((status = add_node(&pt, buff, NULL, AFTER)) == OK) {
		init_enode(pt);
		count++;
	}
	else
		status = FAILED;
  }

  if (status == OK) {
	if (pt != NULL) {
		/* Move to the start of the list */
		pt = pt->next;
		/* Move to the required index */
		for (j = 1 ; j < next ; j++)
			pt = pt->next;
	}
	*ppn = pt;
  }

  /* If the status isn't OK, how do we clean up the mess ??? */
  return (status == OK) ? count : -1;
}


/*
** d e s t r o y _ e d i t _ r i n g
**
** Destroy an edit ring.
**
** Returns:	OK		Success
**		FAILED		Otherwise
*/
static int destroy_edit_ring(struct Enode **ppn)
{
  struct Enode *pe;
  int status = OK;

  if (ppn == NULL || *ppn == NULL)
	return FAILED;

  for (pe = *ppn ; pe != NULL && pe->next != *ppn ; /* MT */)
	status = del_node(pe->next, NULL, NULL);
  status = del_node(pe, NULL, NULL);

  *ppn = NULL;

  return status;
}


/*
** g o t o _ e d i t _ r i n g
**
** Move to another edit ring.
**
** Returns:	OK		Success
**		FAILED		Otherwise
*/
static int goto_edit_ring(GNGUI *gg, int count, int dirn)
{
  int status = OK;

  DPRINT_F(("goto_edit_ring()\n"));

  /* Check for limiting cases */
  if (count < 1) return FAILED;
  if (gg->edhead == NULL) return FAILED;
  if (gg->edcount == 1) return OK;
  if (dirn == HERE) return OK;

  /* Hide the current vterm */
  gtk_widget_hide(gg->edhead->vterm);

  /* Block the current vterm's signal handlers */
  g_signal_handler_block(G_OBJECT(gg->edhead->vterm), gg->edhead->key_id);
  g_signal_handler_block(G_OBJECT(gg->edhead->vterm), gg->edhead->mouse_up_id);
  g_signal_handler_block(G_OBJECT(gg->edhead->vterm), gg->edhead->mouse_dn_id);
  g_signal_handler_block(G_OBJECT(gg->edhead->vterm), gg->edhead->mouse_mv_id);
  g_signal_handler_block(G_OBJECT(gg->edhead->vterm), gg->edhead->scroll_id);

  /* Move the old vterm to a parking container: reparent(widget, new_parent) */
  gtk_event_box_set_visible_window(GTK_EVENT_BOX(gg->edhead->eventbox), FALSE);
  gtk_widget_reparent(GTK_WIDGET(gg->edhead->vterm), gg->edhead->eventbox);

  /* Unload the current Files menu */
  unload_files_menu(gg->edhead->argshead);
  /* Move to the new editing window */
  while (count--) {
	if (dirn == PREV) gg->edhead = gg->edhead->prev;
	else if (dirn == NEXT) gg->edhead = gg->edhead->next;
  }
  /* Reload the new Files menu */
  load_files_menu(gg->edhead->argshead);

  /* Recover the new vterm from the parking container */
  gtk_widget_reparent(GTK_WIDGET(gg->edhead->vterm), gg->holder);

  /* Unblock the new vterm's signal handlers */
  g_signal_handler_unblock(G_OBJECT(gg->edhead->vterm),
						gg->edhead->key_id);
  g_signal_handler_unblock(G_OBJECT(gg->edhead->vterm),
						gg->edhead->mouse_up_id);
  g_signal_handler_unblock(G_OBJECT(gg->edhead->vterm),
						gg->edhead->mouse_dn_id);
  g_signal_handler_unblock(G_OBJECT(gg->edhead->vterm),
						gg->edhead->mouse_mv_id);
  g_signal_handler_unblock(G_OBJECT(gg->edhead->vterm),
						gg->edhead->scroll_id);

  /* Display the new vterm */
  gtk_widget_grab_focus(gg->edhead->vterm);
  gtk_widget_show(gg->edhead->vterm);

  /* Retitle the main Gnome window */
  gtk_window_set_title(GTK_WINDOW(gg->gwindow), gg->edhead->edtitle);

  /* Update the main window geometry */
  update_geometry(gg);

  return status;
}


/*
** u p d a t e _ e d i t _ f i l e s
**
** Update the edit ring files.
**
** Returns:	0		Always
**
** The "next" argument is equal to the current argslist file index + 1.
*/
static int update_edit_files(char **list, int next)
{
  GNGUI *gg = &gn_gui;
  struct Enode *pe, *first;
  unsigned int u;
  int j, count, match, index;
  char buff[PATH_MAX + 64];
  char **ppc, *pcmd;

  DPRINT_F(("update_edit_files()\n"));

  /* List not yet initialised */
  if (list == NULL || gg->edhead == NULL || gg->edhead->argshead == NULL) {
	DPRINT_A((
		"update_edit_files(): list %s edhead %s argshead %s\n",
		list == NULL ? "NULL" : "valid",
		gg->edhead == NULL ? "NULL" : "valid",
		(gg->edhead == NULL || gg->edhead->argshead == NULL)
							? "NULL" : "valid"
		));
	return 0;
  }

  /* Workaround for :sNext bug */
  if (gg->cur_cmd == EX_MAIN)
	pcmd = "EX_MAIN";
  else if (gg->cur_cmd == EX_NAME)
	pcmd = "EX_NAME";
  else if (gg->cur_cmd == EX_WINUP)
	pcmd = "EX_WINUP";
  else if (gg->cur_cmd == EX_WINDN)
	pcmd = "EX_WINDN";
  else
	pcmd = get_exname(gg->cur_cmd);
  DPRINT_A(("update_edit_files(): command |%s|\n",
					pcmd == NULL ? "NULL" : pcmd));
  if (pcmd == NULL) return 0;

  /* Get the size of the given file list */
  for (count = 0, ppc = list ; *ppc != NULL ; count++, ppc++)
	/* MT */ ;

  /* Check for an unchanged argslist; compare the list sizes */
  index = -1;
  match = FALSE;
  if (count == gg->edhead->argscount) {
	match = TRUE;
	/* Compare the list contents; an index < 0 indicates a missing file */
	pe = gg->edhead->argshead;
	for (ppc = list ; *ppc != NULL ; ppc++) {
		/* ioabsolute() returns an absolute file name */
		strncpy(buff, ioabsolute(*ppc), sizeof(buff) - 1);
		DPRINT_A(("update_edit_files(): list |%s| ioabsolute |%s|\n",
							*ppc, buff));
		if ((index = find_node(&pe, buff, TRUE, FALSE)) < 0) {
			match = FALSE;
			break;
		}
	}
  }

#ifdef DPRINT_AA
  { /* Debugging block begin */
  char *results[] = { "After", "Before", "Found", "Failed",
			"zero", "one", "two", "three" };
  if (list[0] == NULL)
	g_print("update_edit_files(): list[0] is NULL\n");
  else {
	g_print("update_edit_files(): next %d list search %s\n",
						next, results[index + 4]);
	for (ppc = list ; ppc != NULL && *ppc != NULL ; ppc++)
		g_print("                     filename |%s|\n", *ppc);
  }
  } /* Debugging block end */
#endif

  /* If we have a new arglist, load it */
  if (list[0] != NULL && !match) {
	/* Remove the current Files menu */
	unload_files_menu(gg->edhead->argshead);
	/* Destroy the current arglist */
	pe = gg->edhead->argshead;
	destroy_edit_ring(&pe);
	gg->edhead->argscount = 0;
	/* Rebuild the argslist */
	count = create_edit_ring(&pe, list, next);
	/* Reload the Files menu */
	load_files_menu(pe);
	gg->edhead->argshead = pe;
  }

  /* Find the start of the current argslist by searching for the node with
  ** the lowest id, which will be the first one on the list.
  */
  u = gg->edhead->argshead->id;
  for (pe = gg->edhead->argshead->next ; pe != gg->edhead->argshead ;
							pe = pe->next) {
	if (u > pe->id) u = pe->id;
  }
  for (pe = gg->edhead->argshead->next ; pe->id != u ; pe = pe->next)
	/* MT */ ;
  first = pe;

  /* Carry out command-specific actions */
  switch (gg->cur_cmd) {
  case EX_MAIN:		/* Create a new ring on startup */
  case EX_QUIT:		/* The original ring is reloaded by destroygw() */
  case EX_WINUP:	/* Move to the next ring (wi+) */
  case EX_WINDN:	/* Move to the previous ring (wi-) */
  case EX_SNEXT:	/* Create a new ring and switch to it */
  case EX_SPREVIOUS:	/* Move to the previous ring */
  case EX_SREWIND:	/* Move to the first ring */
  case EX_SLAST:	/* Move to the last ring */
	break;

  case EX_LAST:		/* Move to the last entry in the argslist */
	gg->edhead->argshead = first->prev;
	break;

  case EX_NEXT:		/* Move to the next entry in the argslist */
  case EX_WNEXT:
	for (pe = first, j = 0 ; j < (next - 1) ; pe = pe->next, j++)
		/* MT */ ;
	gg->edhead->argshead = pe;
	break;

  case EX_PREVIOUS:	/* Move to the previous entry in the argslist */
	for (pe = first, j = 0 ; j < (next - 1) ; pe = pe->next, j++)
		/* MT */ ;
	gg->edhead->argshead = pe;
	break;

  case EX_REWIND:	/* Move to the first entry in the argslist */
	gg->edhead->argshead = first;
	break;

  case EX_NAME:		/* Update the argslist */
	if (match) {
		pe = gg->edhead->argshead;
		/* We need to move to a new file in the argslist */
		u = gg->edhead->argshead->id;
		for (pe = gg->edhead->argshead->next ;
				pe != gg->edhead->argshead ;
				pe = pe->next) {
			if (u > pe->id) u = pe->id;
		}
		for (pe = gg->edhead->argshead->next ;
				pe->id != u ;
				pe = pe->next) {
			/* MT */ ;
		}
		for (j = 1 ; j < next ; j++)
			pe = pe->next;
		gg->edhead->argshead = pe;
	}
	break;

  default:
	/* Are any of the other commands relevant ??? */
	DPRINT_A(("update_edit_files(): irrelevant command %s\n", pcmd));
	break;
  }

  /* Update the Edwin interface as necessary */
  DPRINT_A(("update_edit_files(): match %s index %d count %d argscount %d\n",
		match ? "TRUE" : "FALSE", index, count, gg->edhead->argscount));
  if (!match && count > 0)
	gg->edhead->argscount = count;
  if (gg->cur_cmd == EX_LAST ||
			gg->cur_cmd == EX_NAME ||
			gg->cur_cmd == EX_NEXT ||
			gg->cur_cmd == EX_PREVIOUS ||
			gg->cur_cmd == EX_REWIND ||
			gg->cur_cmd == EX_WNEXT) {
	gg->edhead->edname = gg->edhead->argshead->name;
	gg->edhead->edtitle = gg->edhead->argshead->label;
  }

  /* Clear the command once it's been executed */
  gg->cur_cmd = -1;

  return 0;
}


/*
** l o a d _ f i l e s _ m e n u
**
** Add filenames to the Files menu.
**
** Returns:	OK		Success
**		FAILED		Otherwise
*/
static int load_files_menu(ENODE *pn)
{
  GNGUI *gg = &gn_gui;
  GtkUIManager *self = NULL;
  GtkAction *act = NULL;
  struct Enode *pe, *pf;
  GList *groups_list = NULL;
  GList *actions_list = NULL;
  size_t t;
  int index, flg, status = OK;
  static char new[64];
  char path[] = "/ui/menubar/FileMenu/Empty/";	/* NB. trailing '/' */
  char *pb, *pc, label[64], buff[PATH_MAX + 64];

  DPRINT_F(("load_files_menu()\n"));

  if (pn == NULL || pn->name[0] == '\0') return FAILED;

#ifdef CFGDIR_ELVIS
  if (gg->cfgdir_flg)
	pb = gnome_util_prepend_user_home(CFGDIR_ELVIS "/" NOTEPAD_ELVIS);
  else
	pb = gnome_util_prepend_user_home(NOTEPAD_ELVIS);
#else
  if (pb == NULL)
	pb = gnome_util_prepend_user_home(NOTEPAD_ELVIS);
#endif

  /* Get the UI Manager */
  self = gg->mainuimanager;

  /* Insert the file names into the menu, using the short names as labels.
  ** Add a maximum of ENTRY_COUNT files.
  */
  for (pe = pn, index = 0 ; (pe != pn || index == 0) && index < ENTRY_COUNT ;
						pe = pe->next, index++) {
	DPRINT_A(("load_files_menu(): loading |%s|\n", pe->name));

	pc = pe->name;

	/* Ignore the default file name, less the terminating %d */
	if (strncmp(UNTITLED_BUF, pc, sizeof(UNTITLED_BUF) - 3) == SAME)
		continue;

	/* Ignore the Elvis notepad file, a full path */
	if (strcmp(pb, pc) == SAME)
		continue;

	/* Ignore the list of notepads, which are full paths */
	flg = FALSE;
	for (pf = gg->noteshead ; pf != NULL ; pf = pf->next) {
		if (pf->name != NULL && strcmp(pf->name, pc) == SAME) {
			flg = TRUE;
			break;
		}
		if (pf->next == gg->noteshead) break;
	}
	if (flg) continue;

#ifdef DPRINT_AA
	{ /* Debugging block begin */
	g_print("load_files_menu(): adding name |%s| label |%s|\n",
					pe->name, pe->label);
	} /* Debugging block end */
#endif
	/* There seems no easy way to turn off mnemonics,
	** so underlines must be doubled while space allows.
	*/
	for (t = 0, pb = new, pc = pe->label ; *pc != '\0' && t < 63 ;
							pc++, pb++, t++) {
		*pb = *pc;
		if (*pb == '_') *(++pb) = '_';
	}
	*pb = '\0';

	/* Set up the menu item with the adjusted label string */
	act = gtk_action_new(
			pe->name,	/* Name */
			new,		/* Label */
			NULL,		/* Tooltip */
			NULL		/* Icon */
			);

	/* Connect the filename_handler() function to the menu item */
	pe->signal = g_signal_connect(
				G_OBJECT(act),
				"activate",
				G_CALLBACK(filename_handler),
				(void *) pe
				);

	/* Get the main menus action groups */
	groups_list = gtk_ui_manager_get_action_groups(self);

	/* Generate a unique label - well, almost */
	strcpy(buff, path);
	pc = buff + sizeof(path) - 1;
	sprintf(label, "%s.%04x", pe->label, get_id() & 0xffff);
	strncat(pc, label, sizeof(buff) - sizeof(path) - 6);

	/* Add the new action to the first group */
	gtk_action_group_add_action((GtkActionGroup *) groups_list->data, act);

	/* Get the actions of the first action group */
	actions_list = gtk_action_group_list_actions(
				(GtkActionGroup *) groups_list->data);
#ifdef DPRINT_AA
	{ /* Debugging block begin */
	GList *pa = NULL;
	/* List all the available actions - new action is included */
	for (pa = actions_list ; pa != NULL ; pa = pa->next) {
		g_print("load_files_menu(): actions |%s|\n",
				gtk_action_get_name(GTK_ACTION(pa->data)));
	}
	} /* Debugging block end */
#endif
	pe->uiid = gtk_ui_manager_new_merge_id(self);
	gtk_ui_manager_add_ui(			/* NB. no error code */
			self,
			pe->uiid,		/* Merge id */
			path,			/* Path */
			label,			/* Name */
			pe->name,		/* Action */
			GTK_UI_MANAGER_AUTO,	/* Item type */
			0			/* Below path */
			);
	gtk_ui_manager_ensure_update(self);

	/* Make a note of the relevant Edwin window ID */
	gg->edwin_id = gg->edhead->id;
  }

  return status;
}


/*
** u n l o a d _ f i l e s _ m e n u
**
** Remove filenames from the Files menu.
**
** Returns:	OK		Success
**		FAILED		Otherwise
*/
static int unload_files_menu(ENODE *pz)
{
  GNGUI *gg = &gn_gui;
  GtkUIManager *self = NULL;
  GtkAction *act = NULL;
  struct Edwin *pd;
  struct Enode *pt;
  GList *groups_list = NULL;
  int index, status;

  DPRINT_F(("unload_files_menu()\n"));

  /* Find out which edit ring loaded the menu; it may not be current */
  for (pd = gg->edhead ; pd->id != gg->edwin_id ; pd = pd->next) {
	if (pd->next == gg->edhead)
		break;
  }
  if (pd->id != gg->edwin_id) {
	DPRINT_A(("unload_files_menu(): bad Edwin ID %u\n", pd->id));
	return FAILED;
  }

  /* Get the UI Manager */
  self = gg->mainuimanager;

  /* Remove the old filenames from the menu */
  status = OK;
  for (index = 0, pt = pd->argshead ; pt != NULL && index < ENTRY_COUNT ;
						pt = pt->next, index++) {
	DPRINT_A(("unload_files_menu(): unloading |%s|\n", pt->name));
	if (pt->uiid != 0) {
		gtk_ui_manager_remove_ui(self, pt->uiid);
		gtk_ui_manager_ensure_update(self);
		pt->uiid = 0;
	}

	/* Get the main menus action groups */
	if ((groups_list = gtk_ui_manager_get_action_groups(self)) == NULL) {
		DPRINT_A(("unload_files_menu(): No group\n"));
		status = FAILED;
	}
	/* Get an action from the group : why is it sometimes not
	** initialised ??? ###
	*/
	else if ((act = gtk_action_group_get_action((GtkActionGroup *)
					groups_list->data, pt->name)) == NULL) {
		DPRINT_A(("unload_files_menu(): no action |%s|\n", pt->name));
		status = FAILED;
	}

	if (status == OK) {
		/* Remove the action from the group */
		gtk_action_group_remove_action(
			(GtkActionGroup *) groups_list->data, act);

		/* Disconnect the signal: why is it sometimes not
		** connected ??? ###
		*/
		if (g_signal_handler_is_connected(G_OBJECT(act), pt->signal))
			g_signal_handler_disconnect(G_OBJECT(act), pt->signal);

		/* Destroy the action */
		g_object_unref(act);
	}
	if (pt->next == pd->argshead) break;
  }

  return status;
}


/*
** p a r s e _ i n t
**
** Parse a single integer option.
**
** Returns:	OK		Success
**		NK		Value out of range
**		FAILED		Otherwise
**
** Arguments are those of the standard options parsing functions.
*/
static int parse_int(char *str, size_t index, size_t opt_index)
{
  long int value, upper, lower;
  int status = OK;
  char *pc, *pb;

  DPRINT_F(("parse_int()\n"));

  upper = lower = value = 0;
  for (pc = str ; *pc != '\0' && *pc != '=' ; pc++)
	/* MT */ ;
  if (*pc != '\0' && *(++pc) != '\0') {
	value = strtol(pc, NULL, 0);
	lower = strtol(optdescs[index].limit, NULL, 0);
	for (pb = optdescs[index].limit ; *pb != '\0' ; pb++)
		if (*pb == ':') { pb++; break; }
	upper = strtol(pb, NULL, 0);
	DPRINT_Q(("parse_int(): lower %ld upper %ld value %ld\n",
				lower, upper, value));
	if (upper >= lower) {
		status = NK;
		if (value >= lower && value <= upper) {
			optvals[opt_index].value.number = value;
			status = OK;
		}
	}
  }
  else
	status = FAILED;

  return status;
}


/*
** p a r s e _ t w o _ i n t s
**
** Parse a double integer option.
**
** Returns:	OK		Success
**		NK		Value out of range
**		FAILED		Otherwise
**
** Arguments are those of the standard options parsing functions.
*/
static int parse_two_ints(char *str, size_t index, size_t opt_index)
{
  long int upper, lower, one, two;
  int status;
  char *pc;

  DPRINT_F(("parse_two_ints()\n"));

  /* The string is of the form --geometry=RxC, in the same
  ** format as that used by Gnome terminal.  The code relies
  ** on the two arguments being adjacent in the optdescs array.
  */
  status = FAILED;
  upper = lower = 0;
  for (pc = str ; *pc != '\0' && *pc != '=' ; pc++)
	/* MT */ ;
  if (*pc != '\0' && *(++pc) != '\0') {
	status = NK;
	if (sscanf(pc, "%ldx%ld", &one, &two) == 2
			&& sscanf(optdescs[opt_index].limit, "%ld:%ld",
				&lower, &upper) == 2
			&& lower > 0 && upper > lower
			&& one >= lower && one <= upper
			&& sscanf(optdescs[opt_index + 1].limit, "%ld:%ld",
				&lower, &upper) == 2
			&& lower > 0 && upper > lower
			&& two >= lower && two <= upper) {
		optvals[opt_index].value.number = one;
		optvals[opt_index + 1].value.number = two;
		status = OK;
	}
  }

  return status;
}


/*
** p a r s e _ b o o l
**
** Parse a boolean option.
**
** Returns:	OK		Success
**		NK		Value out of range
**		FAILED		Otherwise
**
** Arguments are those of the standard options parsing functions.
*/
static int parse_bool(char *str, size_t index, size_t opt_index)
{
  size_t t, count;
  int status;
  char *pb, *pc, buff[S_BUF_SIZ];
  static struct Args {
	char *name;
	int boolean;
  } args[] = {
	{"one", 1}, {"all", 0},
	{"new", 1}, {"cur", 0},
	{"on", 1}, {"off", 0},
	{"true", 1}, {"false", 0},
	{"up", 1}, {"dn", 0},
	{"yes", 1}, {"no", 0}
  };

  DPRINT_F(("parse_bool()\n"));

  /* The input string is of the form --option=bool, where bool can be any
  ** element of a number of opposed pairs in any variety of capitalisation.
  */
  status = FAILED;
  strncpy(buff, str, S_BUF_MAX);
  count = sizeof(args) / sizeof(args[0]);
  for (pb = pc = buff ; *pc != '\0' ; pc++) {
	if (*pc == '=')
		pb = pc;
	*pc = tolower(*pc);
  }
  if (*pb == '=' && *(++pb) != '\0') {
	for (t = 0 ; t < count ; t++) {
		if (strcmp(pb, args[t].name) == SAME) {
			optvals[opt_index].value.boolean = args[t].boolean;
			status = OK;
			break;
		}
	}
  }

  return status;
}


/*
** p a r s e _ c o l o r
**
** Parse a color (string) option.
**
** Returns:	OK		Success
**		NK		Value out of range
**		FAILED		Otherwise
**
** Arguments are those of the standard options parsing functions.
*/
static int parse_color(char *str, size_t index, size_t opt_index)
{
  GNGUI *gg = &gn_gui;
  GdkColor color = { 0 };
  int status = OK;
  char *pc;

  DPRINT_F(("parse_color()\n"));

  /* The input string is of the form --option=name, where name can be any
  ** valid color string.  The string can be in any of the forms accepted
  ** by XParseColor; these include a name for a color from rgb.txt, such
  ** as DarkSlateGray, or a hex specification such as #3050b2 or #35b.
  */

  /* Get the actual option and color name */
  for (pc = str ; *pc && *pc != '=' ; pc++)
	/* MT */ ;
  if (*pc == '\0' || *(++pc) == '\0')
	status = FAILED;
  else if (strncmp(str, "--backcolor", sizeof("--backcolor") - 1) == SAME) {
	if (gdk_color_parse(pc, &color)) {
		gg->cur_backcolor = color;
		*pc = toupper(*pc);
		strncpy(gg->cur_backname, pc, F_SIZ - 1);
		if (!InitDone) {
			gg->def_backcolor = color;
			strncpy(gg->def_backname, pc, F_SIZ - 1);
		}
		optisstring(&optdescs[O_BCOLOR], &optvals[O_BCOLOR],
							gg->cur_backname);
	}
	else
		status = FAILED;
  }
  else if (strncmp(str, "--forecolor", sizeof("--forecolor") - 1) == SAME) {
	if (gdk_color_parse(pc, &color)) {
		gg->cur_forecolor = color;
		*pc = toupper(*pc);
		strncpy(gg->cur_forename, pc, F_SIZ - 1);
		if (!InitDone) {
			gg->def_forecolor = color;
			strncpy(gg->def_forename, pc, F_SIZ - 1);
		}
		optisstring(&optdescs[O_FCOLOR], &optvals[O_FCOLOR],
							gg->cur_forename);
	}
	else
		status = FAILED;
  }
  else if (strncmp(str, "--cursorcolor", sizeof("--cursorcolor") - 1) == SAME) {
	if (gdk_color_parse(pc, &color)) {
		gg->cur_cursorcolor = color;
		*pc = toupper(*pc);
		strncpy(gg->cur_cursorname, pc, F_SIZ - 1);
		if (!InitDone) {
			gg->def_cursorcolor = color;
			strncpy(gg->def_cursorname, pc, F_SIZ - 1);
		}
		optisstring(&optdescs[O_CCOLOR], &optvals[O_CCOLOR],
							gg->cur_cursorname);
	}
	else
		status = FAILED;
  }
  else
	status = FAILED;

  return status;
}


/*
** p a r s e _ f o n t
**
** Parse a font (string) option.
**
** Returns:	OK		Success
**		NK		Value out of range
**		FAILED		Otherwise
**
** Arguments are those of the standard options parsing functions.
*/
static int parse_font(char *str, size_t index, size_t opt_index)
{
  GNGUI *gg = &gn_gui;
  PangoFontDescription *pango_desc = NULL;
  PangoFontMask mask;
  const char *pango_family_name = NULL;
  int j, type, valid;
  /* DEBUG_FONT is the highest font index */
  static unsigned char fonts[DEBUG_FONT + 1][F_SIZ];
  char *pc;

  DPRINT_F(("parse_font()\n"));

  for (pc = str ; *pc && *pc != '=' ; pc++)
	/* MT */ ;
  /* If the argument is bad, just bale out */
  if (*(++pc) == '\0')
	return FAILED;

  /* Find out which option we're checking.  Option strings are of the
  ** form "--fonttype=fontname" where --fonttype is:
  ** --normalfont
  ** --boldfont
  ** --italicfont
  ** --debugfont
  ** --headerfont
  ** --printerfont
  ** --terminalfont
  */
#ifdef DEBUG_MENU
  if (str[4] == 'b')
	type = DEBUG_FONT;
  else
#endif
  if (str[4] == 'a')
	type = HEADER_FONT;
  else if (str[4] == 'i')
	type = PRINTER_FONT;
  else if (str[4] == 'r')
	type = TERMINAL_FONT;
  else
	type = TERMINAL_FONT;	/* A reasonable default */

  /* The input string is of the form --option=name, where name can be
  ** any valid font string.
  **
  ** This string has the form "[FAMILY-LIST] [STYLE-OPTIONS] [SIZE]",
  ** where FAMILY-LIST is a comma separated list of families optionally
  ** terminated by a comma, STYLE_OPTIONS is a whitespace separated list
  ** of words where each WORD describes one of style, variant, weight,
  ** or stretch, and SIZE is a decimal number (size in points). Any one
  ** of the options may be absent. If FAMILY-LIST is absent, then the
  ** family_name field of the resulting font description will be
  ** initialized to NULL. If STYLE-OPTIONS is missing, then all style
  ** options will be set to the default values. If SIZE is missing,
  ** the size in the resulting font description will be set to 0.
  */
  pango_desc = pango_font_description_from_string((const char *) pc);
  pango_family_name = pango_font_description_get_family(pango_desc);
  mask = pango_font_description_get_set_fields(pango_desc);

  DPRINT_N(("parse_font(): parse check family |%s|\n", pango_family_name));

  for (j = 0 ; j < 6 ; j++) {
	switch (mask & (1 << j)) {
	case 0:
			DPRINT_N(("Pango attribute %d invalid\n", j));
			valid = FALSE;
			break;
	case (1 << 0):
			DPRINT_N(("PANGO_FONT_MASK_FAMILY valid\n"));
			break;
	case (1 << 1):
			DPRINT_N(("PANGO_FONT_MASK_STYLE valid\n"));
			break;
	case (1 << 2):
			DPRINT_N(("PANGO_FONT_MASK_VARIANT valid\n"));
			break;
	case (1 << 3):
			DPRINT_N(("PANGO_FONT_MASK_WEIGHT valid\n"));
			break;
	case (1 << 4):
			DPRINT_N(("PANGO_FONT_MASK_STRETCH valid\n"));
			break;
	case (1 << 5):
			DPRINT_N(("PANGO_FONT_MASK_SIZE valid\n"));
			break;
	default:
			DPRINT_N(("Pango attribute %d unknown\n", j));
			valid = FALSE;
			break;
	}
  }

  /* Unlike the terminal, the printer doesn't require a monospaced font.
  ** Check that the font family is available, and if so park it in
  ** permanent storage.
  */
  valid = FALSE;
  if (((type == HEADER_FONT || type == PRINTER_FONT) &&
					check_font_family(pc, type) == OK) ||
					check_mono_family(pc) == OK) {
	valid = TRUE;
	if (type == TERMINAL_FONT) {
		strncpy(fonts[TERMINAL_FONT], pc, F_MAX);
		strcpy(gg->cur_fontname, fonts[TERMINAL_FONT]);
		if (!InitDone)
			strcpy(gg->def_fontname, fonts[TERMINAL_FONT]);
		optisstring(&optdescs[O_TFONT], &optvals[O_TFONT],
							gg->cur_fontname);
	}
	else if (type == PRINTER_FONT) {
		strncpy(fonts[PRINTER_FONT], pc, F_MAX);
		strcpy(gg->cur_print_font, fonts[PRINTER_FONT]);
		if (!InitDone)
			strcpy(gg->def_print_font, fonts[PRINTER_FONT]);
		optisstring(&optdescs[O_PFONT], &optvals[O_PFONT],
							gg->cur_print_font);
	}
	else if (type == HEADER_FONT) {
		strncpy(fonts[HEADER_FONT], pc, F_MAX);
		strcpy(gg->cur_hdr_print_font, fonts[HEADER_FONT]);
		if (!InitDone)
			strcpy(gg->def_hdr_print_font, fonts[HEADER_FONT]);
		optisstring(&optdescs[O_HFONT], &optvals[O_HFONT],
							gg->cur_hdr_print_font);
	}
#ifdef DEBUG_MENU
	else if (type == DEBUG_FONT) {
		strncpy(fonts[DEBUG_FONT], pc, F_MAX);
		strcpy(gg->cur_dbg_fontname, fonts[DEBUG_FONT]);
		if (!InitDone)
			strcpy(gg->def_dbg_fontname, fonts[DEBUG_FONT]);
		optisstring(&optdescs[O_DFONT], &optvals[O_DFONT],
							gg->cur_dbg_fontname);
	}
#endif
	else
		valid = FALSE;
  }

  if (!valid)
	msg(MSG_ERROR, "[s]Couldn't parse fontname $1", pc);

  DPRINT_N(("parse_font():%smatch for |%s|\n", valid ? " " : " no ",
						pango_family_name));

  return valid ? OK : FAILED;
}


/*
** p a r s e _ s t r i n g
**
** Parse a string option.
**
** Returns:	OK		Success
**		NK		Value out of range
**		FAILED		Otherwise
**
** Arguments are those of the standard options parsing functions.
** In fact, there's not much to be done except check for length.
*/
static int parse_string(char *str, size_t index, size_t opt_index)
{
  GNGUI *gg = &gn_gui;
  int status = OK;

  DPRINT_F(("parse_string()\n"));

  /* The two string options aren't set from the command line,
  ** and hence have no leading "--name" string.  Note that the
  ** actual availability of the programs isn't checked here.
  */
  if (strcmp("dictionary", optdescs[opt_index].longname) == SAME) {
	strncpy(gg->dictionary, str, sizeof(gg->dictionary) - 1);
	optisstring(&optdescs[O_SPELL], &optvals[O_SPELL], gg->dictionary);
  }
  else if (strcmp("spellcheck", optdescs[opt_index].longname) == SAME) {
	strncpy(gg->spellcheck, str, sizeof(gg->spellcheck) - 1);
	optisstring(&optdescs[O_SPELL], &optvals[O_SPELL], gg->spellcheck);
  }
  else
	status = FAILED;

  return status;
}


/*
** p a r s e _ o p t s
**
** Parse the global argument string for Gnome GUI options.
**
** Returns:	Modified argc	Always
**
** argv is updated as a side-effect.
*/
static int parse_opts(int argc, char **argv)
{
  size_t index;
  int status = OK;
  char **ppb, **ppc;

  DPRINT_F(("parse_opts()\n"));

#ifdef DPRINT_QQ
  for (ppc = argv ; *ppc != NULL ; ppc++)
	g_print("parse_opts(): initial argv |%s|\n", *ppc);
  g_print("parse_opts(): initial argc %d\n", argc);
#endif

  /* Scan the given arguments and see if any strings are familiar */
  for (ppc = argv ; *ppc != NULL ; ppc++) {
	for (index = 0 ; index < (sizeof(parsers) / sizeof(parsers[0])) ;
								index++) {
		if (strncmp(*ppc, parsers[index].name,
				parsers[index].len) == SAME) {
			status = (*parsers[index].parse)(
					*ppc,
					index,
					parsers[index].idx
					);
			if (status == FAILED) {
				DPRINT_Q(("parse_opts(): parse failure\n"));
				msg(MSG_ERROR,
					"[s]Could not parse argument $1", *ppc);				msgflush();
			}
			else {
				/* value has been changed */
				optvals[parsers[index].idx].flags |= OPT_SET;
			}
			/* Remove the argument from the list anyway */
			for (ppb = ppc ; *ppb != NULL ; ppb++)
				*ppb = *(ppb + 1);
			if (ppc != argv) ppc--;
			argc--;

			/* If the argument has been parsed, quit */
			break;
		}
	}
  }

#ifdef DPRINT_QQ
  for (ppc = argv ; *ppc != NULL ; ppc++)
	g_print("parse_opts(): final argv |%s|\n", *ppc);
  g_print("parse_opts(): final argc %d\n", argc);
#endif

  return argc;
}

#ifdef ADD_UNUSED
/*
** g e t _ d e f a u l t _ c o l o r m a p
**
** Get the default (system) color map.
**
** Returns:	OK		Success
**		FAILED		Otherwise
**
** Converting Elvis' idea of colors to VTE's idea of colors is
** too complicated; better to just let Elvis do it.
*/
static int get_default_colormap(void)
{
  GtkStyle *style;
  GdkVisual *visual;
  GdkColormap *sys_colormap, *colormap;
  GdkColor colors[VT_CUR_BG + 1] = { { 0 } };
  gint ok_colors[VT_CUR_BG + 1] = { 0 };
  int j, k, status = OK;

  DPRINT_F(("get_default_colormap()\n"));

  /* The GdkColor structure is used to describe an allocated
  ** or unallocated color.
  ** typedef struct {
  ** guint32 pixel;	For allocated colors, the value used to draw
  **			this color on the screen.
  ** guint16 red;	The red component of the color. A value between
  **			0 and 65535, with 65535 indicating full intensity.
  ** guint16 green;	The blue component of the color.
  ** guint16 blue;	The green component of the color.
  ** } GdkColor;
  **
  ** The colormap structure contains the following public fields.
  ** typedef struct {
  ** gint size;		For pseudo-color colormaps, the number of
  **			colors in the colormap.
  ** GdkColor *colors;	An array containing the current values in the
  **			colormap. This can be used to map from pixel
  **			values back to RGB values. This is only meaningful
  **			for pseudo-color colormaps.
  ** } GdkColormap;
  */

  /* Get the system color map */
  if ((sys_colormap = gdk_colormap_get_system()) == NULL) return FAILED;

  DPRINT_C((
	"get_default_colormap(): system color map has %d entries, pointer %s\n",
	sys_colormap->size, sys_colormap->colors == NULL ? "NULL" : "valid"
	));

  /* Get the system visual and build an appropriate colormap.
  **
  ** The vte_terminal docs describe a 28-color palette, made up of
  ** three 8-color sections (???).  The maximum size is given by the
  ** VTE library's internal define VT_CUR_BG + 1,  where VT_CUR_BG
  ** is set to 29, but this isn't exposed anywhere.  Since the array
  ** has indices 0-27, why is VT_CUR_BG 29 ???
  **
  ** According to the docs the default background color is taken from
  ** palette[0], and the default foreground color from palette[7].
  ** Why are are different indices used below ???
  **
  ** It looks as if VTE _might_ be using 29 colors:
  ** The colors 0-7 are the standard colors.
  ** The colors 8-15 (at VTE_COLOR_BRIGHT_OFFSET) are the bright versions
  **   of the 8 standard colors.
  ** The colors 16-23 (at VTE_COLOR_DIM_OFFSET) are the dim versions
  **   of the standard colors.
  ** The colors 24-29 (VTE_DEF_FG, VTE_DEF_BG, VTE_BOLD_FG, VTE_DIM_FG
  ** VTE_DEF_HL, VTE_CUR_BG) are the default foreground, background, and
  ** highlight colors.  It isn't clear how the 29/30 indices fit with a
  ** 28-color palette ???
  **
  ** See also the comments in the create_edwin() function.
  */
  if ((visual = gdk_visual_get_system()) == NULL)
	status = FAILED;
  else if ((colormap = gdk_colormap_new(visual, TRUE)) == NULL) /* editable */
	status = FAILED;

  DPRINT_C(("get_default_colormap(): visual is %s colormap is %s\n",
			visual == NULL ? "null" : "valid",
			colormap == NULL ? "null" : "valid"
			));

  /* This ought to work, but there's no easy way to find where a
  ** color is in the resulting colormap, or to remove duplicates
  ** from the colormap; and it doesn't play well with Elvis' form
  ** of palette.
  */
  k = gdk_colormap_alloc_colors(
		colormap,		/* GdkColormap */
		colors,			/* Colors to allocate */
		sizeof(colors) / sizeof(colors[0]), /* Number of colors */
		TRUE,			/* Writeable, private */
		TRUE,			/* Flag for matching */
		ok_colors		/* Success/failure flags */
		);
  if (colormap != NULL) {
	for (j = 0 ; j < (VT_CUR_BG + 1) ; j++)
		ok_colors[j] = FALSE;
  }
  DPRINT_C((
	"get_default_colormap(): failed colours %d map size given %d used %d\n",
	k,
	sizeof(colors) / sizeof(colors[0]),
	colormap == NULL ? (-1) : colormap->size
	));

  return status;
}
#endif

/*
** s e t _ d e f a u l t _ c o l o r s
**
** Set the default colors.
**
** Returns:	OK		Success
**		NK		Using fallback as default
**		FAILED		Otherwise
*/
static int set_default_colors(GNGUI *gg)
{
  GdkColor fc, bc, cc;
  size_t t;
  int j, status;
  char *pb, *pc, *fn, *bn, buff[64];

  DPRINT_F(("set_default_colors()\n"));

  /* Set the fallback scheme to black-on-white */
  gg->vte_colors[0] = color_white;	/* VTE's default background */
  gg->vte_colors[1] = color_black;	/* Elvis' default foreground */
  gg->vte_colors[7] = color_black;	/* VTE's default foreground */
  gg->def_backcolor = color_white;
  gg->def_forecolor = color_black;
  gg->def_cursorcolor = color_grey;
  strcpy(gg->def_backname, "White");
  strcpy(gg->def_forename, "Black");
  strcpy(gg->def_cursorname, "Grey");

  /* Update the current values */
  gg->cur_backcolor = gg->def_backcolor;
  gg->cur_forecolor = gg->def_forecolor;
  gg->cur_cursorcolor = gg->def_cursorcolor;
  strcpy(gg->cur_backname, gg->def_backname);
  strcpy(gg->cur_forename, gg->def_forename);
  strcpy(gg->cur_cursorname, gg->def_cursorname);

  /* We're using hardcoded fallbacks */
  status = NK;

  /* And updating them if possible */
  buff[0] = '\0';
  strncat(buff, DEFAULT_SCHEME, sizeof(buff) - 1);
  /* The scheme string is of the form "forecolor-on-backcolor",
  ** where either color name may be capitalised.
  */
  for (pb = pc = buff ; *pc != '\0' ; pc++) {
	if (isupper(*pc)) *pc = tolower(*pc);
	if (*pc == '-') {
		*pc = '\0';
		pb = pc;
	}
  }
  fn = buff;
  bn = ++pb;
  /* Rudimentary check for successful parsing */
  if (bn[0] != '\0' && fn[0] != '\0') {
	j = 0;
	fc = bc = cc = color_black;
	/* Find the colors equivalent to the foreground and background names */
	for (t = 0 ; t < (sizeof(fixedcolors)/sizeof(fixedcolors[0])) ; t++) {
		if (strcmp(fixedcolors[t].name, bn) == SAME) {
			bc = fixedcolors[t].color;
			j++;
		}
		if (strcmp(fixedcolors[t].name, fn) == SAME) {
			fc = fixedcolors[t].color;
			j++;
		}
	}
	/* If successful, update the GUI */
	if (j == 2) {
		gg->vte_colors[0] = bc;
		gg->vte_colors[1] = fc;
		gg->vte_colors[7] = fc;
		gg->def_backcolor = bc;
		gg->cur_backcolor = bc;
		gg->def_forecolor = fc;
		gg->cur_forecolor = fc;
		/* The cursor is set to an intermediate color */
		cc.red = (bc.red + fc.red) / 2;
		cc.green = (bc.green + fc.green) / 2;
		cc.blue = (bc.blue + fc.blue) / 2;
		gg->cur_cursorcolor = cc;
		gg->def_cursorcolor = cc;
		bn[0] = toupper(bn[0]);
		strcpy(gg->def_backname, bn);
		strcpy(gg->cur_backname, bn);
		fn[0] = toupper(fn[0]);
		strcpy(gg->def_forename, fn);
		strcpy(gg->cur_forename, fn);
		/* The color is derived, and so has no real name */
#if 1
		strcpy(gg->cur_cursorname, "Derived");
#else
		sprintf(gg->cur_cursorname, "#%04x%04x%04x",
					cc.red, cc.green, cc.blue);
#endif
		strcpy(gg->def_cursorname, gg->cur_cursorname);
		status = OK;
	}
	else
		status = FAILED;
  }

  /* Elvis and vte-terminal have different ideas about the position
  ** of the default foreground and background codes. Comply with both.
  */
  gg->font_codes[0] = gg->font_codes[1] = gg->font_codes[7] = 1;

  /* Update the :set options */
  optisstring(&optdescs[O_FCOLOR], &optvals[O_FCOLOR], gg->cur_forename);
  optisstring(&optdescs[O_BCOLOR], &optvals[O_BCOLOR], gg->cur_backname);
  optisstring(&optdescs[O_CCOLOR], &optvals[O_CCOLOR], gg->cur_cursorname);

  /* We always have the fallbacks */
  return (status == OK) ? OK : NK;
}


/*
** c h e c k _ f o n t _ f a m i l y
**
** Check that a font family is available.
**
** Returns:	OK		The named family exists
**		NK		The named family is not available
**		FAILED		Otherwise
*/
static int check_font_family(char *fontname, int type)
{
  GNGUI *gg = &gn_gui;
  PangoContext *pango_context = NULL;
  PangoFontFamily **pango_families = NULL;
  gint family_count = 0;
  int j, valid, status = OK;
  char *pf;

  DPRINT_F(("check_font_family()\n"));

  if (type == PRINTER_FONT) {
	pango_context = pango_cairo_font_map_create_context(
		(PangoCairoFontMap *) pango_cairo_font_map_get_default());
	DPRINT_N(("check_font_family(): pango_context is %s\n",
			pango_context == NULL ? "NULL" : "valid"));
  }
  else
	pango_context = gg->pango_context;

  if (pango_context == NULL) return FAILED;

  /* Get a list of all font families */
  pango_context_list_families(pango_context, &pango_families, &family_count);
  for (valid = FALSE, j = 0 ; !valid && (j < family_count) ; j++) {
	pf = (char *) pango_font_family_get_name(pango_families[j]);
	if (strncmp(fontname, pf, strlen(pf)) == SAME)
		valid = TRUE;
  }

  if (pango_families != NULL) g_free(pango_families);
  if (type == PRINTER_FONT) g_object_unref(pango_context);

  return (status == OK && valid) ? OK : FAILED;
}


/*
** c h e c k _ m o n o _ f a m i l y
**
** Check that a monospace font family is available.
**
** Returns:	OK		The named family, or a family, exists
**		NK		The named family is not available
**		FAILED		Otherwise
**
** If the argument is NULL, this function checks for the existence
** of any monospace family; if the argument is not NULL, the check
** is made for the named _family_.
*/
static int check_mono_family(char *fontname)
{
  GNGUI *gg = &gn_gui;
  PangoFontFamily **pango_families = NULL;
  gint family_count = 0;
  int j, valid, status = OK;
  char *pf;

  DPRINT_F(("check_mono_family()\n"));

  if (gg->pango_context == NULL) return FAILED;

  /* Standard Gnome has the following mono fonts and faces,
  ** though these seem to change with each release:
  **
  ** Andale Mono: Regular, Italic, Bold, Bold Italic
  ** Bitstream Vera Sans Mono: Roman, Bold, Oblique, Bold Oblique
  ** Courier New:  Regular, Italic, Bold, Bold Italic
  ** Fixed: Regular, Bold, Oblique, SemiCondensed,
  **       Bold SemiCondensed, Oblique SemiCondensed
  ** Monospace: Regular, Bold, Italic, Bold Italic
  ** Nimbus Mono L: Regular, Bold, Regular Oblique, Bold Oblique
  */

  /* Get a list of all font families and filter out unwanted names */
  pango_context_list_families(gg->pango_context,
					&pango_families, &family_count);
  for (valid = FALSE, j = 0 ; !valid && (j < family_count) ; j++) {
	if (pango_font_family_is_monospace(pango_families[j])) {
		pf = (char *) pango_font_family_get_name(pango_families[j]);
		DPRINT_N(("check_mono_family(): comparing |%s| |%s|\n",
				(fontname == NULL) ? "NULL" : fontname, pf));
		if (fontname == NULL || strncmp(fontname,
						pf, strlen(pf)) == SAME) {
			valid = TRUE;
		}
	}
  }

  if (pango_families != NULL) g_free(pango_families);

  return (status == OK && valid) ? OK : FAILED;
}


/*
** g e t _ m o n o _ f o n t s
**
** Get a monospace font, or a default.
**
** Returns:	Number of fonts found	Success
**		0			Requested font not available
**		-1			No monospace font available
*/
static int get_mono_fonts(char *fontname, char *newfont)
{
  GNGUI *gg = &gn_gui;
  PangoFontFamily **pango_families = NULL;
  gint family_count = 0;
  int j, count, valid, status = OK;
  const char *pf;

  DPRINT_F(("get_mono_fonts()\n"));

  /* Get a list of all font families and filter out unwanted names */
  pango_context_list_families(gg->pango_context,
				&pango_families, &family_count);
  for (valid = FALSE, count = 0, j = 0 ; !valid && (j < family_count) ; j++) {
	if (pango_font_family_is_monospace(pango_families[j])) {
		pf = pango_font_family_get_name(pango_families[j]);
		DPRINT_N(("get_mono_fonts(): mono font family |%s|\n", pf));
#ifdef DPRINT_NN
		list_fonts(pango_families[j]);
#endif
		/* Update the new font buffer in case the default isn't found.
		** Don't parse out a full Pango description, just use the
		** current font family and a reasonable size. ###
		*/
		if (newfont != NULL && count++ == 0) {
			strncpy(newfont, pf, F_MAX - 3);
			strcat(newfont, " 10");
		}
		if (strncmp(DEFAULT_FONT_NAME, pf, strlen(pf)) == SAME)
			valid = TRUE;
	}
  }

  if (pango_families != NULL) g_free(pango_families);

  return status == OK ? (valid ? count : 0) : -1;
}

#ifdef DPRINT_NN
/*
** l i s t _ f o n t s
**
** List the available monospace fonts for debugging.
**
** Returns:	Nothing		Always
*/
static void list_fonts(PangoFontFamily *family)
{
  PangoFontFace **pango_faces;
  int *pango_sizes, size_count, face_count;
  int j, k;

  pango_font_family_list_faces(family, &pango_faces, &face_count);
  /* List the faces */
  for (j = 0 ; j < face_count ; j++) {
	g_print("        font face |%s|\n",
				pango_font_face_get_face_name(pango_faces[j]));
	pango_font_face_list_sizes(pango_faces[j], &pango_sizes, &size_count);
	/* List the sizes */
	for (k = 0 ; k < size_count ; k++) {
		g_print("                font size %d\n",
				pango_sizes[k] / PANGO_SCALE); /* Points */
	}
	if (pango_sizes != NULL) g_free(pango_sizes);
  }
  if (pango_faces != NULL) g_free(pango_faces);

  return;
}
#endif

/*
** s e t _ c u r r e n t _ f o n t
**
** Set the current font.
**
** Returns:	OK		Success
**		FAILED		Otherwise
**
** Note: at present this handles only one font, and not bold or italic.
*/
static int set_current_font(GNGUI *gg, char *fontname, int global)
{
  PangoContext *pango_context = NULL;
  PangoLanguage *pango_lang = NULL;
  PangoFontDescription *pango_desc = NULL;
  static PangoFontMetrics *pango_metrics = NULL;
  int x, y, valid, status = OK;

  DPRINT_F(("set_current_font()\n"));

  /* Get the default pango context. Is the vterm context relevant ??? ### */
  if (gg->edhead == NULL)
	pango_context = gg->pango_context;
  else
	pango_context = gtk_widget_get_pango_context(gg->edhead->vterm);

  /* Bale out early on obvious failure */
  if (pango_context == NULL || fontname == NULL || fontname[0] == '\0')
	return FAILED;

  /* Get a pango language; will this ever change after startup ??? */
  if ((pango_lang = gtk_get_default_language()) == NULL)
	pango_lang = pango_language_from_string(DEFAULT_PANGO_LANG);

  /* Check that we have this actual font family */
  valid = (check_mono_family(fontname) == OK);
  if (valid) {
	/* Get the font metrics. pango_metrics must be cleaned up with
	** pango_font_metrics_unref(), but it stays in use until the
	** font changes.
	*/
	if (pango_metrics != NULL) {
		pango_font_metrics_unref(pango_metrics);
		pango_metrics = NULL;
	}
	pango_desc = pango_font_description_from_string(fontname); /* Leak ? */
	pango_metrics = pango_context_get_metrics(
			pango_context, pango_desc, pango_lang);

	/* Get the current font metrics for pointer/char conversion.
	** Digits are said to be a better width approximation.
	x = pango_font_metrics_get_approximate_char_width(gg->pango_metrics);
	*/
	x = pango_font_metrics_get_approximate_digit_width(pango_metrics);
	y = pango_font_metrics_get_ascent(pango_metrics);
	y += pango_font_metrics_get_descent(pango_metrics);
	x /= PANGO_SCALE;
	y /= PANGO_SCALE;

	/* Update the global settings if required */
	if (global) {
		strncpy(gg->cur_fontname, fontname, F_MAX);
		optisstring(&optdescs[O_TFONT], &optvals[O_TFONT],
							gg->cur_fontname);
		gg->pango_desc = pango_desc;
		gg->pango_lang = pango_lang;
		/* gg->pango_context = pango->context; see below */
		gg->pango_metrics = pango_metrics;
		gg->pango_width = x;
		gg->pango_height = y;
	}
	/* Update any local settings (if available) anyway */
	if (gg->edhead != NULL) {
		strncpy(gg->edhead->fontname, fontname, F_MAX);
		gg->edhead->pango_desc = pango_desc;
		gg->edhead->pango_lang = pango_lang;
		/* gg->edhead->pango_context = pango->context; see below */
		gg->edhead->pango_metrics = pango_metrics;
		gg->edhead->pango_width = x;
		gg->edhead->pango_height = y;
	}
	/* A scrolling window widget overlaps its child and its scrollbar,
	** so the vte_terminal size must be increased over the size used
	** by Elvis to allow all Elvis' columns to be displayed.  The
	** default seems a reasonable bet, until the window is initialised.
	*/
	if (gg->scrollwin == NULL || x == 0)
		gg->bar_cols = 2;
	else
		gg->bar_cols = gg->stepper_width / x +
				((gg->stepper_width % x) ? 1 : 0);
	DPRINT_N(("set_current_font(): bar_cols %d\n", gg->bar_cols));
  }

  /* pango_context is temporary, and must be free'd; but while this works
  ** on Gnome 2.12, it crashes on 2.16. ###
  g_free(pango_context);
  pango_context = NULL;
  */

  /* The pango_font_metrics object must be unref'd when no longer used,
  ** but this instance lasts the life of the program. ???
  */

  if (pango_desc == NULL || pango_lang == NULL || pango_metrics == NULL)
	status = FAILED;

  DPRINT_N(("set_current_font(): fontname |%s|%sfound.\n"
		"            pango_desc is%svalid\n"
		"            pango_lang is%svalid\n"
		"            pango_metrics is%svalid\n"
		"            global%supdated, local%supdated\n",
		fontname,
		valid ? " " : " not ",
		pango_desc == NULL ? " in" : " ",
		pango_lang == NULL ? " in" : " ",
		pango_metrics == NULL ? " in" : " ",
		global ? " " : " not ",
		gg->edhead == NULL ? " not " : " "
	));
  DPRINT_N((
	"set_current_font(): metrics\n"
	"            ref_count %d\n"
	"            ascent %d\n"
	"            descent %d\n"
	"            approx char width %d\n"
	"            approx digit width %d\n",
	pango_metrics->ref_count,
	pango_metrics->ascent,
	pango_metrics->descent,
	pango_metrics->approximate_char_width,
	pango_metrics->approximate_digit_width
	));

  return status;
}


/*
** s e t _ b l i n k
**
** Update the cursor blink using :set.
**
** Returns:	1		Success
**		0		Otherwise
**
** The types of the arguments have to match the prototype in optdescs[].
*/
static int set_blink(struct optdesc_s *desc, OPTVAL *val, CHAR *bp)
{
  GNGUI *gg = &gn_gui;
  int status = OK;

  DPRINT_F(("set_blink()\n"));

  /* Update the main cursor blink flags.  The default setting
  ** is changed only during the initialisation process.
  */
  if (bp == NULL)
	status = FAILED;
  else if (*bp == 'F') {
	o_cursorblink = FALSE;
	gg->cur_cursorblink = FALSE;
	if (!InitDone)
		gg->def_cursorblink = FALSE;
  }
  else if (*bp == 'T') {
	o_cursorblink = TRUE;
	gg->cur_cursorblink = TRUE;
	if (!InitDone)
		gg->def_cursorblink = TRUE;
  }
  else
	status = FAILED;

  /* Update the current editing window cursor blink flag */
  if (status == OK && gg->edhead != NULL) {
	if (o_cursorblink && !gg->edhead->cursorblink) {
		gg->edhead->cursorblink = TRUE;
		vte_terminal_set_cursor_blinks(VTE_TERMINAL(gg->edhead->vterm),
						gg->edhead->cursorblink);
		}
	else if (!o_cursorblink && gg->edhead->cursorblink) {
		gg->edhead->cursorblink = FALSE;
		vte_terminal_set_cursor_blinks(VTE_TERMINAL(gg->edhead->vterm),
						gg->edhead->cursorblink);
	}
  }

  DPRINT_Q(("set_blink(): o_blink %s def_blink %s cur_blink %s\n",
		o_cursorblink ? "TRUE" : "FALSE",
		gg->def_cursorblink ? "TRUE" : "FALSE",
		gg->cur_cursorblink ? "TRUE" : "FALSE"
	));

  /* Update the use_default_cursor flags */
  if ((o_cursorblink && !gg->def_cursorblink) ||
				(!o_cursorblink && gg->def_cursorblink)) {
		gg->use_def_cursor = FALSE;
		if (gg->edhead != NULL)
			gg->edhead->use_def_cursor = FALSE;
  }
  /* To revert from defaults requires the use of the default button */

  return status;
}


/*
** s e t _ c u r s o r
**
** Update the cursor color using :set.
**
** Returns:	1		Success
**		0		Otherwise
*/
static int set_cursor(struct optdesc_s *desc, OPTVAL *val, CHAR *bp)
{
  GNGUI *gg = &gn_gui;
  size_t index;
  int status = OK;
  char buff[64];

  DPRINT_F(("set_cursor()\n"));

  /* Look up the option in parsers[] */
  strcpy(buff, "--cursorcolor");
  for (index = 0 ; index < (sizeof(parsers) / sizeof(parsers[0])) ; index++) {
	if (strcmp(parsers[index].name, buff) == SAME)
		break;
  }

  /* We know this is a valid option string, so add the color name */
  strcat(buff, "=");
  strncat(buff, val->value.string, 64 - strlen(buff) - 1);

  /* Parse what is now a standard option */
  if (parse_color(buff, index, parsers[index].idx) != OK) {
	/* Reset the failed option */
	optisstring(&optdescs[O_CCOLOR], &optvals[O_CCOLOR],
					gg->cur_cursorname);
	status = FAILED;
  }

  if (status == OK && gg->edhead != NULL)  {
	gg->edhead->use_def_cursor = FALSE;
	gg->edhead->cursorcolor = gg->cur_cursorcolor;
	vte_terminal_set_color_cursor(VTE_TERMINAL(gg->edhead->vterm),
					&gg->edhead->cursorcolor);
  }
  else
	status = FAILED;

  return (status == OK) ? 1 : 0;
}


/*
** s e t _ f o r e
**
** Update the foreground color using :set.
**
** Returns:	1		Success
**		0		Otherwise
**
** The types of the arguments have to match the prototype in optdescs[].
*/
static int set_fore(struct optdesc_s *desc, OPTVAL *val, CHAR *bp)
{
  GNGUI *gg = &gn_gui;
  size_t index;
  int status = OK;
  char buff[64];

  DPRINT_F(("set_fore()\n"));

  /* Look up the option in parsers[] */
  strcpy(buff, "--forecolor");
  for (index = 0 ; index < (sizeof(parsers) / sizeof(parsers[0])) ; index++) {
	if (strcmp(parsers[index].name, buff) == SAME)
		break;
  }

  /* We know this is a valid option string, so add the color name */
  strcat(buff, "=");
  strncat(buff, val->value.string, 64 - strlen(buff) - 1);

  /* Parse what is now a standard option */
  if (parse_color(buff, index, parsers[index].idx) != OK) {
	/* Reset the failed option */
	optisstring(&optdescs[O_FCOLOR], &optvals[O_FCOLOR], gg->cur_forename);
	status = FAILED;
  }

  if (status == OK && gg->edhead != NULL) {
	gg->edhead->use_def_color = FALSE;
	gg->edhead->forecolor = gg->cur_forecolor;
	vte_terminal_set_color_foreground(VTE_TERMINAL(gg->edhead->vterm),
						&gg->edhead->forecolor);
  }
  else
	status = FAILED;

  return (status == OK) ? 1 : 0;
}


/*
** s e t _ b a c k
**
** Update the background color using :set.
**
** Returns:	1		Success
**		0		Otherwise
**
** The types of the arguments have to match the prototype in optdescs[].
*/
static int set_back(struct optdesc_s *desc, OPTVAL *val, CHAR *bp)
{
  GNGUI *gg = &gn_gui;
  size_t index;
  int status = OK;
  char buff[64];

  DPRINT_F(("set_back()\n"));

  /* Look up the option in parsers[] */
  strcpy(buff, "--backcolor");
  for (index = 0 ; index < (sizeof(parsers) / sizeof(parsers[0])) ; index++) {
	if (strcmp(parsers[index].name, buff) == SAME)
		break;
  }

  /* We know this is a valid option string, so add the color name */
  strcat(buff, "=");
  strncat(buff, val->value.string, 64 - strlen(buff) - 1);

  /* Parse what is now a standard option */
  if (parse_color(buff, index, parsers[index].idx) != OK) {
	/* Reset the failed option */
	optisstring(&optdescs[O_BCOLOR], &optvals[O_BCOLOR], gg->cur_backname);
	status = FAILED;
  }

  if (status == OK && gg->edhead != NULL) {
	gg->edhead->use_def_color = FALSE;
	gg->edhead->backcolor = gg->cur_backcolor;
	vte_terminal_set_color_background(VTE_TERMINAL(gg->edhead->vterm),
						&gg->edhead->backcolor);
  }
  else
	status = FAILED;

  return (status == OK) ? 1 : 0;
}


/*
** s e t _ f o n t
**
** Update the default terminal fonts using :set.
**
** Returns:	1		Success
**		0		Otherwise
**
** The types of the arguments have to match the prototype in optdescs[].
** This function currently handles only one terminal font.
*/
static int set_font(struct optdesc_s *desc, OPTVAL *val, CHAR *bp)
{
  GNGUI *gg = &gn_gui;
  size_t index;
  int status = OK;
  char buff[64];

  DPRINT_F(("set_font()\n"));

  /* Look up the option in parsers[] */
  strcpy(buff, "--");
  strcat(buff, desc->longname);
  for (index = 0 ; index < (sizeof(parsers) / sizeof(parsers[0])) ; index++) {
	if (strcmp(parsers[index].name, buff) == SAME)
		break;
  }

  /* We know this is a valid option name string, so add the font name */
  strcat(buff, "=");
  strncat(buff, val->value.string, 64 - strlen(buff) - 1);

  /* Parse what is now a standard option */
  if (parse_font(buff, index, parsers[index].idx) != OK) {
	DPRINT_N(("set_font(): failed |%s| def |%s| cur |%s| mac |%s|\n",
		buff, gg->def_fontname, gg->cur_fontname, o_terminalfont));
	/* Reset the failed option */
	optisstring(&optdescs[O_TFONT], &optvals[O_TFONT], gg->cur_fontname);
	status = FAILED;
  }

  /* Update the current editing window */
  if (status == OK && gg->edhead != NULL) {
	gg->edhead->use_def_font = FALSE;
	strcpy(gg->edhead->fontname, gg->cur_fontname);
	(void) set_current_font(gg, gg->edhead->fontname, FALSE);
	vte_terminal_set_font(VTE_TERMINAL(gg->edhead->vterm),
					gg->edhead->pango_desc);
	/* Refresh the main window size */
	gg->main_height = 0;
	gg->main_width = 0;
	update_geometry(gg);
	size_main_window(gg, TRUE);
  }
  else
	status = FAILED;

  return (status == OK) ? 1 : 0;
}


/*
** s e t _ c o l s
**
** Update the terminal columns using :set.
**
** Returns:	1		Success
**		0		Otherwise
**
** The types of the arguments have to match the prototype in optdescs[].
*/
static int set_cols(struct optdesc_s *desc, OPTVAL *val, CHAR *bp)
{
  GNGUI *gg = &gn_gui;
  long int min, max, cols;
  int status = OK;

  DPRINT_F(("set_cols()\n"));

  cols = val->value.number;
  if (sscanf(desc->limit, "%ld:%ld", &min, &max) == 2 &&
				(cols >= min) && (cols <= max)) {
	o_defaultcols = cols;
	if (gg->edhead != NULL) {
		gg->edhead->cols = cols;

		/* Refresh the main window size */
		gg->main_height = 0;
		gg->main_width = 0;
		update_geometry(gg);
		size_main_window(gg, TRUE);
	}
  }
  else
	status = FAILED;

  return (status == OK) ? 1 : 0;
}


/*
** s e t _ r o w s
**
** Update the terminal rows using :set.
**
** Returns:	1		Success
**		0		Otherwise
**
** The types of the arguments have to match the prototype in optdescs[].
*/
static int set_rows(struct optdesc_s *desc, OPTVAL *val, CHAR *bp)
{
  GNGUI *gg = &gn_gui;
  long int min, max, rows;
  int status = OK;

  DPRINT_F(("set_rows()\n"));

  rows = val->value.number;
  if (sscanf(desc->limit, "%ld:%ld", &min, &max) == 2 &&
				(rows >= min) && (rows <= max)) {
	o_defaultrows = rows;
	if (gg->edhead != NULL) {
		gg->edhead->rows = rows;

		/* Refresh the main window size */
		gg->main_height = 0;
		gg->main_width = 0;
		update_geometry(gg);
		size_main_window(gg, TRUE);
	}
  }
  else
	status = FAILED;

  return (status == OK) ? 1 : 0;
}


/*
** s e t _ t o o l
**
** Update the toolbar display using :set.
**
** Returns:	1		Success
**		0		Otherwise
**
** The types of the arguments have to match the prototype in optdescs[].
*/
static int set_tool(struct optdesc_s *desc, OPTVAL *val, CHAR *bp)
{
  GNGUI *gg = &gn_gui;
  int here, there,  status = OK;

  DPRINT_F(("set_tool()\n"));

  if (bp == NULL)
	status = FAILED;
  else if (*bp == 'F')
	here = FALSE;
  else if (*bp == 'T')
	here = TRUE;
  else
	status = FAILED;

  /* Pass the information to GConf */
  there = gconf_client_get_bool(gg->gconf_client,
					GCONF_PATH "/ui/showtoolbar", NULL);

  if (status == OK && ((there && !here) || (here && !there))) {
	gconf_client_set_bool(gg->gconf_client,
			GCONF_PATH "/ui/showtoolbar", here, NULL);
	/* Give GTK+ time to stabilise */
	while (gtk_events_pending()) gtk_main_iteration();
  }

  return (status == OK) ? 1 : 0;
}


/*
** s e t _ p r i n t
**
** Update the printer fonts using :set.
**
** Returns:	1		Success
**		0		Otherwise
**
** The types of the arguments have to match the prototype in optdescs[].
** This function currently handles only one printer font.
*/
static int set_print(struct optdesc_s *desc, OPTVAL *val, CHAR *bp)
{
  GNGUI *gg = &gn_gui;
  size_t index;
  int status = OK;
  char buff[64];

  DPRINT_F(("set_print()\n"));

  /* Look up the option in parsers[] */
  strcpy(buff, "--");
  strcat(buff, desc->longname);
  for (index = 0 ; index < (sizeof(parsers) / sizeof(parsers[0])) ; index++) {
	if (strcmp(parsers[index].name, buff) == SAME)
		break;
  }

  /* We know this is a valid option name string, so add the font name */
  strcat(buff, "=");
  strncat(buff, val->value.string, 64 - strlen(buff) - 1);

  /* Parse what is now a standard option */
  if (parse_font(buff, index, parsers[index].idx) != OK) {
	DPRINT_N(("set_print(): failed |%s| def |%s| cur |%s| mac |%s|\n",
		buff, gg->def_print_font, gg->cur_print_font, o_printerfont));
	/* Reset the failed option */
	optisstring(&optdescs[O_PFONT], &optvals[O_PFONT], gg->cur_print_font);
	status = FAILED;
  }
  else
	gg->use_def_print = FALSE;

  return (status == OK) ? 1 : 0;
}


/*
** a d d _ d o c s m e n u
**
** Add an item to the Documents menu.
**
** Returns:	OK		Success
**		FAILED		Otherwise
*/
static int add_docsmenu(ENODE *pn)
{
  GNGUI *gg = &gn_gui;
  GList *groups_list = NULL;
  GtkAction *act = NULL;
  GtkUIManager *self = NULL;
  size_t t;
  int status = OK;
  static char new[64];
  char *pb, *pc, label[64], buff[PATH_MAX + 64];
  char path[] = "/ui/menubar/DocsMenu/Empty/";	/* NB. trailing '/' */

  DPRINT_P(("add_docsmenu()\n"));

  /* There seems no easy way to turn off mnemonics,
  ** so underlines must be doubled while space allows.
  */
  for (t = 0, pb = new, pc = pn->label ; *pc != '\0' && t < 63 ;
						pc++, pb++, t++) {
	*pb = *pc;
	if (*pb == '_') *(++pb) = '_';
  }
  *pb = '\0';

  /* Set up the menu item with the adjusted label string */
  act = gtk_action_new(
		pn->name,	/* Name */
		new,		/* Label */
		"tooltip",	/* Tooltip won't show: why not ??? ### */
		GTK_STOCK_FILE	/* Icon will show */
		);

  /* Connect the documents_handler() function to the menu item */
  pn->signal = g_signal_connect(
			G_OBJECT(act),
			"activate",
			G_CALLBACK(documents_handler),
			(void *) pn
			);

  /* Get the UI Manager */
  self = gg->mainuimanager;
  DPRINT_P(("add_docsmenu(): self is %s\n",
				self == NULL ? "NULL" : "valid"));

  /* Get the main menu action groups */
  groups_list = gtk_ui_manager_get_action_groups(self);
  DPRINT_P(("add_docsmenu(): groups_list is %s\n",
				groups_list == NULL ? "NULL" : "valid"));

  /* Generate a unique label - well, almost */
  strcpy(buff, path);
  pc = buff + sizeof(path) - 1;
  sprintf(label, "%s.%04x", pn->label, get_id() & 0xffff);
  strncat(pc, label, sizeof(buff) - sizeof(path) - 6);

  /* Add the new action to the group */
  gtk_action_group_add_action((GtkActionGroup *) groups_list->data, act);
  pn->uiid = gtk_ui_manager_new_merge_id(self);
  gtk_ui_manager_add_ui(		/* NB. no error code */
		self,
		pn->uiid,		/* Merge id */
		path,			/* Path */
		label,			/* Name */
		pn->name,		/* Action */
		GTK_UI_MANAGER_AUTO,	/* Item type */
		0			/* Below path */
		);
  gtk_ui_manager_ensure_update(self);

  /* This seems the only way to add a tooltip to a GtkUIManager item ??? */
  gtk_tooltips_set_tip(
		gg->menutips,		/* Tooltips */
		gtk_ui_manager_get_widget(self, buff), /* Widget */
		pn->name,		/* Tip text */
		NULL			/* Private text */
		);

  /* Track the number of entries */
  gg->docscount++;

  return status;
}


/*
** r m v _ d o c s m e n u
**
** Remove an item from the Documents menu.
**
** Returns:	OK		Success
**		FAILED		Otherwise
*/
static int rmv_docsmenu(ENODE *pn)
{
  GNGUI *gg = &gn_gui;
  int status = OK;
  GList *groups_list = NULL;
  GtkAction *act = NULL;
  GtkUIManager *self = NULL;

  DPRINT_P(("rmv_docsmenu()\n"));

  if (pn == NULL) return FAILED;

  self = gg->mainuimanager;
  DPRINT_P(("rmv_docsmenu(): self is %s\n",
				self == NULL ? "NULL" : "valid"));

  /* Should g_object_unref() be called here ??? ### */
  if (pn->uiid != 0) {
	gtk_ui_manager_remove_ui(self, pn->uiid);
	gtk_ui_manager_ensure_update(self);
	pn->uiid = 0;
  }

  /* Get the main menus action groups */
  if ((groups_list = gtk_ui_manager_get_action_groups(self)) == NULL) {
	DPRINT_P(("rmv_docsmenu(): no group found\n"));
	status = FAILED;
  }
  /* Get an action from the group : why is it not initialised ??? ### */
  else if ((act = gtk_action_group_get_action((GtkActionGroup *)
				groups_list->data, pn->name)) == NULL) {
	DPRINT_P(("rmv_docsmenu(): no action found for |%s|\n", pn->name));
	status = FAILED;
  }

  if (status == OK) {
	DPRINT_P(("rmv_docsmenu(): removing action |%s|\n", pn->name));
	/* Remove the action from the group */
	gtk_action_group_remove_action(
		(GtkActionGroup *) groups_list->data, act);

	/* Disconnect the signal: why isn't it still connected ??? ### */
	if (g_signal_handler_is_connected(G_OBJECT(act), pn->signal))
		g_signal_handler_disconnect(G_OBJECT(act), pn->signal);

	/* Destroy the action */
	g_object_unref(act);
  }

  /* Track the number of entries */
  if (gg->docscount > 0) gg->docscount--;

  return status;
}


/*
** a d d _ n o t e s m e n u
**
** Add an item to the Notebooks submenu.
**
** Returns:	OK		Success
**		FAILED		Otherwise
**
** The initial setup for this menu, using NOTEPAD_ELVIS, is done
** in init_gnome_win().
*/
static int add_notesmenu(ENODE *pn)
{
  GNGUI *gg = &gn_gui;
  GtkWidget *note;
  int status = OK;

  DPRINT_P(("add_notesmenu()\n"));

  /* Get a new menu item */
  note = gtk_menu_item_new_with_label(pn->label);

  /* Connect the handler() function to the menu item */
  pn->signal = g_signal_connect(
			G_OBJECT(note),
			"activate",
			G_CALLBACK(note_action),
			(void *) pn->name
			);

  /* This seems the only way to add a tooltip to a GtkUIManager item ??? */
  gtk_tooltips_set_tip(
		gg->menutips,		/* Tooltips */
		note,			/* Widget */
		pn->name,		/* Tip text */
		NULL			/* Private text */
		);

  /* Add the menu item to the Notepad submenu */
  gtk_menu_shell_append(GTK_MENU_SHELL(gg->notesmenu), note);

  gtk_widget_show(note);

  /* Track the number of entries */
  gg->docscount++;

  return status;
}


/*
** c r e a t e _ e d w i n
**
** Create a file editing window.
**
** Returns:	OK		Success
**		FAILED		Otherwise
**
** Called by the Gnome GUI code, not by the main Elvis code.
** The global struct GnomeGui gn_gui is updated as a side effect.
*/
static int create_edwin(EDWIN **ppw)
{
  GNGUI *gg = &gn_gui;
  EDWIN *edwin = *ppw;
  GdkGeometry nullhints = { 0 };
  GtkWidget *vterm;
  glong rows, cols;
  int status = OK;
  char buff[S_BUF_SIZ];

  DPRINT_F(("create_edwin()\n"));

#ifdef DPRINT_CC
  { /* Debugging block begin */
  int j;
  /* List the allocated colors by name for debugging */
  g_print("create_edwin():\n");
  for (j = 0 ; j < sizeof(color_names) / sizeof(color_names[0]) ; j++) {
	if (gg->font_codes[j] != 0) {
		g_print("    index %2d font %2d %s |%s| %04x:%04x:%04x\n",
			j,
			gg->font_codes[j],
			(gg->color_types[j] == FORE_COLOR) ? "fg" : "bg",
			(char *)(color_names + j),
			gg->vte_colors[j].red,
			gg->vte_colors[j].green,
			gg->vte_colors[j].blue
			);
	}
  }
  } /* Debugging block end */
#endif

  /* Gnome seems to need a unique ID for each status line message,
  ** so this roundabout method is used to create one.  However, it
  ** runs out of scope if more than UINT_MAX file windows are opened
  ** in one instantiation. This is not a significant limitation.
  */
  sprintf(buff, "%016x", get_id());

  /* Start with the default size */
  rows = gg->def_rows;
  cols = gg->def_cols;

  /* Get a new terminal widget; initial size is fixed at 24x80 */
  if ((vterm = vte_terminal_new()) == NULL)
	status = FAILED;
  /* Add the first editing window, removing the vterm placeholder */
#if 1
  else if (edwin == NULL) {
	if ((edwin = g_try_malloc(sizeof(EDWIN))) != NULL) {
		edwin->prev = edwin;
		edwin->next = edwin;
		gtk_widget_destroy(gg->vterm);
		gg->vterm = NULL;
	}
  }
#else
  else if (gg->vterm != NULL) {
	if ((edwin = g_try_malloc(sizeof(EDWIN))) != NULL) {
		gg->edhead = edwin;
		gg->edhead->prev = gg->edhead;
		gg->edhead->next = gg->edhead;
		gtk_widget_destroy(gg->vterm);
		gg->vterm = NULL;
	}
	else
		status = FAILED;
  }
#endif
  /* Add a new editing widget */
  else if (add_edwin(edwin, AFTER) == FAILED)
	status = FAILED;
  /* Park the current editing widget */
  else {
	/* Block the current vterm's signal handlers */
	g_signal_handler_block(G_OBJECT(edwin->vterm),
						edwin->key_id);
	g_signal_handler_block(G_OBJECT(edwin->vterm),
						edwin->mouse_up_id);
	g_signal_handler_block(G_OBJECT(edwin->vterm),
						edwin->mouse_dn_id);
	g_signal_handler_block(G_OBJECT(edwin->vterm),
						edwin->mouse_mv_id);
	g_signal_handler_block(G_OBJECT(edwin->vterm),
						edwin->scroll_id);
	/* Move the current vterm to a spare container */
	gtk_event_box_set_visible_window(
			GTK_EVENT_BOX(edwin->eventbox), FALSE);
	gtk_widget_reparent(
			GTK_WIDGET(edwin->vterm), edwin->eventbox);
  }

  if (status == FAILED) {
	DPRINT_E(("create_edwin(): early failure\n"));
	goto End;
  }

  /* Update the edit window indices */
  gg->edcount++;
  gg->edhead = edwin->next;

  DPRINT_E(("create_edwin(): edhead 0x%08lx, new vterm 08%08lx\n",
					(ULONG) gg->edhead, (ULONG) vterm));

  /* Initialise the new EDWIN and make it current */
  gg->edhead->gn_gui = gg;
  gg->edhead->argshead = NULL;
  gg->edhead->argscount = 0;
  if ((gg->edhead->eventbox = gtk_event_box_new()) == NULL)
	status = FAILED;
  gg->edhead->vterm = vterm;
  gg->edhead->id = get_id();
  gg->edhead->key_id = 0;
  gg->edhead->mouse_up_id = 0;
  gg->edhead->mouse_dn_id = 0;
  gg->edhead->mouse_mv_id = 0;
  gg->edhead->scroll_id = 0;
  gg->edhead->attr_mask = 0;	/* UINT_MAX; */
  gg->edhead->clr_mask = 0;	/* COLOR_FG | COLOR_BG; */
  gg->edhead->fg_clr = 0;
  gg->edhead->bg_clr = 0;
  gg->edhead->keybuff[0] = '\0';
  gg->edhead->keycount = 0;
  gg->edhead->hints = nullhints;
  gg->edhead->xpad = 0;
  gg->edhead->ypad = 0;
  gg->edhead->rows = rows;
  gg->edhead->cols = cols;
  gg->edhead->use_def_font = gg->use_def_font;
  gg->edhead->use_def_color = gg->use_def_color;
  gg->edhead->use_def_cursor = gg->use_def_cursor;
#if 0
  gg->edhead->forecolor = gg->vte_colors[7];	/* VTE's palette */
  gg->edhead->backcolor = gg->vte_colors[0];
#else
  gg->edhead->forecolor = gg->cur_forecolor;	/* Main palette */
  gg->edhead->backcolor = gg->cur_backcolor;
#endif
  gg->edhead->cursorcolor = gg->cur_cursorcolor;
  gg->edhead->cursorblink = gg->cur_cursorblink;

  DPRINT_E(("create_edwin: new cursor RGB 0x%04x 0x%04x 0x%04x\n",
		gg->edhead->cursorcolor.red,
		gg->edhead->cursorcolor.green,
		gg->edhead->cursorcolor.blue
		));
  /* Deal with the colormap on a per-window basis ???
  gg->edhead->colormap = gg->colormap;
  */

  /* Update the local Pango data */
  (void) set_current_font(gg, gg->cur_fontname, FALSE);

  /* Set the vterm's font */
  vte_terminal_set_font(VTE_TERMINAL(gg->edhead->vterm),
					gg->edhead->pango_desc);
  /* Allow the use of bold */
  vte_terminal_set_allow_bold(VTE_TERMINAL(gg->edhead->vterm), TRUE);
  /* But don't enforce it
  vte_terminal_set_color_bold(VTE_TERMINAL(gg->edhead->vterm), &red);
  */

  /* Make a note of the status and menu bar ids */
  gg->edhead->st_id = gtk_statusbar_get_context_id(
		GTK_STATUSBAR(gg->edhead->gn_gui->statusbar), buff);
  gg->edhead->ms_id = gtk_statusbar_get_context_id(
		GTK_STATUSBAR(gg->edhead->gn_gui->msgbar), buff);
  /* Set the event mask _before_ the widget is realised */
  gtk_widget_set_events(gg->edhead->vterm, GDK_KEY_PRESS_MASK);
  /* Set the audible and visible bells; Elvis consider them alternatives */
  if (o_errorbells || o_warningbells) {
	vte_terminal_set_visible_bell(VTE_TERMINAL(gg->edhead->vterm),
						(gboolean) o_flash);
	vte_terminal_set_audible_bell(VTE_TERMINAL(gg->edhead->vterm),
						(gboolean) !o_flash);
  }
  else {
	vte_terminal_set_audible_bell(VTE_TERMINAL(gg->edhead->vterm), FALSE);
	vte_terminal_set_visible_bell(VTE_TERMINAL(gg->edhead->vterm), FALSE);
  }
  /* Unset the vte_terminal's scrollback, since Elvis has its own */
  vte_terminal_set_scrollback_lines(VTE_TERMINAL(gg->edhead->vterm), 0);
  /* Set standard sizes and colours */
  vte_terminal_set_size(VTE_TERMINAL(gg->edhead->vterm),
			gg->edhead->cols + gg->bar_cols, gg->edhead->rows);

  /* Set the VTE color map size to the default maximum of 16; VTE
  ** defaults to colors[0] as background and colors[7] as foreground.
  **
  ** The terminal widget uses a 28-color model comprised of the default
  ** foreground and background colors, the bold foreground color, the dim
  ** foreground color, an eight color palette, and bold versions of the
  ** eight color palette, and a dim version of the the eight color palette.
  **
  ** palette_size must be either 0, 8, 16, or 24. If foreground is NULL
  ** palette_size is documented as either 0, 8, 16, or 24, but 24 doesn't
  ** seem to work. If foreground is NULL and palette_size is greater than 0,
  ** the new foreground color is taken from palette[7]. If background is NULL
  ** and palette_size is greater than 0, the new background color is taken
  ** from palette[0]. If palette_size is 8 or 16, the second (bold) and the
  ** third (dim) 8-color palettes are extrapolated if necessary from the new
  ** background and the available palette entries.
  */
  if (gg->edhead->use_def_color) {
	vte_terminal_set_colors(VTE_TERMINAL(gg->edhead->vterm),
			&gg->edhead->forecolor,	/* foreground */
			&gg->edhead->backcolor,	/* background */
			gg->vte_colors,		/* palette */
			16			/* palette size */
			);
  }
  else {
	vte_terminal_set_colors(VTE_TERMINAL(gg->edhead->vterm),
			NULL,
			NULL,
			gg->vte_colors,
			16
			);
  }

  /* This has red letters on a black background
  vte_terminal_set_default_colors(VTE_TERMINAL(gg->edhead->vterm));
  */

  /* The cursor color has to be set after the rest of the colors */
  vte_terminal_set_color_cursor(VTE_TERMINAL(gg->edhead->vterm),
					&gg->edhead->cursorcolor);
  /* Set the cursor blink */
  vte_terminal_set_cursor_blinks(VTE_TERMINAL(gg->edhead->vterm),
					gg->edhead->cursorblink);

  /* Set the mouse pointer autohide mode */
  vte_terminal_set_mouse_autohide(VTE_TERMINAL(gg->edhead->vterm),
						(int) o_hidemouse);

  /* Trap keypresses */
  gg->edhead->key_id = g_signal_connect(
			G_OBJECT(gg->edhead->vterm),
			"key-press-event",
			G_CALLBACK(key_press_handler),
			(void *) gg
			);
  /* Trap mouse buttons and movement */
  gg->edhead->mouse_dn_id = g_signal_connect_swapped(
			G_OBJECT(gg->edhead->vterm),
			"button-press-event",
			G_CALLBACK(mouse_handler),
			(void *) gg->mousemenu
			);
  gg->edhead->mouse_up_id = g_signal_connect_swapped(
			G_OBJECT(gg->edhead->vterm),
			"button-release-event",
			G_CALLBACK(mouse_handler),
			(void *) gg->mousemenu
			);
  gg->edhead->mouse_mv_id = g_signal_connect_swapped(
			G_OBJECT(gg->edhead->vterm),
			"motion-notify-event",
			G_CALLBACK(mouse_handler),
			(void *) gg->mousemenu
			);
  /* Scrolling (mouse buttons 4 & 5) has its own signal */
  gg->edhead->scroll_id = g_signal_connect(
			G_OBJECT(gg->edhead->vterm),
			"scroll-event",
			G_CALLBACK(scroll_handler),
			(void *) gg
			);

  /* Show the current (new) vterm */
  gtk_container_add(GTK_CONTAINER(gg->holder), GTK_WIDGET(gg->edhead->vterm));
  gtk_widget_show(gg->edhead->vterm);
  gtk_widget_grab_focus(gg->edhead->vterm);

  /* Set the cursor type. Can VTE be persuaded to use this ???
  vterm->pvt->mouse_default_cursor = gg->cursor;
  */
#if 0
  gg->cursor = gdk_cursor_new_for_display(
	gtk_widget_get_display(GTK_WIDGET(gg->edhead->vterm)), GDK_CIRCLE);
  gdk_window_set_cursor(GTK_WIDGET(gg->edhead->vterm)->window, gg->cursor);
#endif

  /* Refresh the main window size */
  gg->main_height = 0;
  gg->main_width = 0;
  update_geometry(gg);
  size_main_window(gg, TRUE);

  /* Debugging only
  gtk_widget_show(gg->gwindow);
  */

  /* Come here on failure */
  End:

  DPRINT_E(("create_edwin(): return status %s\n",
			status == OK ? "OK" : "FAILED"));

  return status;
}


/*
** d e s t r o y _ e d w i n
**
** Destroy a file editing window.
**
** Returns:	OK		Success
**		FAILED		Otherwise
**
** The GnomeGui structure pointed to by gg is updated as a side effect.
*/
static int destroy_edwin(EDWIN **ppw)
{
  GNGUI *gg = &gn_gui;
  EDWIN *edwin = *ppw;
  int status = OK;

  DPRINT_F(("destroy_edwin()\n"));

  /* Just checking */
  if (edwin == NULL || edwin->vterm == NULL)
	return FAILED;

  /* Hide the current vterm */
  gtk_widget_hide(edwin->vterm);

  /* Disconnect the current vterm's signal handlers */
  g_signal_handler_disconnect(
		G_OBJECT(edwin->vterm), edwin->key_id);
  g_signal_handler_disconnect(
		G_OBJECT(edwin->vterm), edwin->mouse_up_id);
  g_signal_handler_disconnect(
		G_OBJECT(edwin->vterm), edwin->mouse_dn_id);
  g_signal_handler_disconnect(
		G_OBJECT(edwin->vterm), edwin->mouse_mv_id);
  g_signal_handler_disconnect(
		G_OBJECT(edwin->vterm), edwin->scroll_id);

  /* Disconnect the vterm and destroy it */
  gtk_event_box_set_visible_window(GTK_EVENT_BOX(edwin->eventbox), FALSE);
  gtk_widget_reparent(GTK_WIDGET(edwin->vterm), edwin->eventbox);
  /* A vterm isn't a widget, but you can destroy it once it's in a container */
  gtk_widget_destroy(edwin->eventbox);
  /* Just in case */
  edwin->vterm = NULL;
  edwin->eventbox = NULL;

  /* Remove any associated statusbar message */
  gtk_statusbar_pop(GTK_STATUSBAR(gg->statusbar), edwin->st_id);
  gtk_statusbar_pop(GTK_STATUSBAR(gg->msgbar), edwin->ms_id);
  edwin->st_id = 0;
  edwin->ms_id = 0;

  /* Destroy the Files menu entries and the argslist nodes.
  ** No, destroygw() does that.
  */

  /* Destroy Elvis' current window.
  eventdestroy(edwin);
  ** No, apparently Elvis does that.
  */

  /* Remove the current EDWIN from the linked list */
  gg->edhead = (edwin == edwin->prev) ? NULL : edwin->prev;
  status = rmv_edwin(edwin);
  gg->edcount--;

  /* Load the new EDWIN */
  if (gg->edhead != NULL && gg->edcount > 0) {

	DPRINT_E(("destroy_edwin(): new vterm is %s\n",
			edwin->vterm == NULL ? "null" : "valid"));

	/* Restore the current (new) vterm to the scrolling window */
	gtk_widget_reparent(GTK_WIDGET(gg->edhead->vterm), gg->holder);

	/* Unblock the vterm signal handlers */
	g_signal_handler_unblock(
			G_OBJECT(gg->edhead->vterm), gg->edhead->key_id);
	g_signal_handler_unblock(
			G_OBJECT(gg->edhead->vterm), gg->edhead->mouse_up_id);
	g_signal_handler_unblock(
			G_OBJECT(gg->edhead->vterm), gg->edhead->mouse_dn_id);
	g_signal_handler_unblock(
			G_OBJECT(gg->edhead->vterm), gg->edhead->mouse_mv_id);
	g_signal_handler_unblock(
			G_OBJECT(gg->edhead->vterm), gg->edhead->scroll_id);

	/* Show the reloaded vterm */
	gtk_widget_grab_focus(gg->edhead->vterm);
	/* This segfaults when quall! is run - why ??? ###
	gtk_widget_show(gg->edhead->vterm);
	*/

	/* Refresh the main window size */
	gg->main_height = 0;
	gg->main_width = 0;
	update_geometry(gg);
	size_main_window(gg, TRUE);

	/* Retitle the main Gnome window */
	gtk_window_set_title(GTK_WINDOW(gg->gwindow), gg->edhead->edtitle);
  }

  DPRINT_E(("destroy_edwin(): return status %s\n",
			status == OK ? "OK" : "FAILED"));

  return status;
}


/*
** a d d _ e d w i n
**
** Add an EDWIN structure to a linked list,
** either BEFORE or AFTER the given node.
**
** Returns:	OK		Success
**		FAILED		Otherwise
*/
static int add_edwin(EDWIN *edwin, int posn)
{
  EDWIN *pe;
  int status = OK;

  DPRINT_F(("add_edwin()\n"));

  if (edwin == NULL || edwin->prev == NULL || edwin->next == NULL)
	status = FAILED;
  else if ((pe = g_try_malloc(sizeof(EDWIN))) == NULL)
	status = FAILED;
  else if (posn == BEFORE) {
	pe->prev = edwin->prev;
	edwin->prev->next = pe;
	pe->next = edwin;
	edwin->prev = pe;
  }
  else if (posn == AFTER) {
	pe->prev = edwin->next->prev;
	edwin->next->prev = pe;
	pe->next = edwin->next;
	edwin->next = pe;
  }
  else {
	g_free(pe);
	status = FAILED;
  }

  return status;
}


/*
** r m v _ e d w i n
**
** Remove a EDWIN structure from a linked list and destroy it.
**
** Returns:	OK		Success
**		FAILED		Otherwise
*/
static int rmv_edwin(EDWIN *edwin)
{
  int status = OK;

  DPRINT_F(("rmv_edwin()\n"));

  if (edwin == NULL || edwin->prev == NULL || edwin->next == NULL)
	status = FAILED;
  else {
	edwin->next->prev = edwin->prev;
	edwin->prev->next = edwin->next;
	g_free(edwin);
  }

  return status;
}


/*
** i n i t _ e n o d e
**
** Initialise an ENODE's type-specific fields.
**
** Returns:	OK		Success
**		FAILED		Otherwise
*/
static int init_enode(ENODE *pn)
{
  char *pb, *pc;

  DPRINT_X(("init_enode()\n"));

  if (pn == NULL || pn->name == NULL) return FAILED;

  /* Initialise the uiid and signal fields */
  pn->uiid = 0;
  pn->signal = 0;

  /* Update the node.label field with the simple filename */
  pc = pn->name;
  for (pb = pc ; *pc != '\0' ; pc++) {
	if (*pc == SEPCHR)
		pb = pc;
  }
  if (*pb == SEPCHR)
	pn->label = ++pb;
  else
	pn->label = pn->name;

  return OK;
}

/* Code to handle generic doubly-linked lists */

/*
** a d d _ n o d e
**
** Add a node to a linked list, either BEFORE or AFTER
** the given node.
**
** Returns:	OK		Success
**		FAILED		Otherwise
**
** The node pointer argument is updated to point at the new node.
*/
static int add_node(ENODE **ppn, char *name, void *data, int posn)
{
  ENODE *curr, *pt;
  size_t len, size;
  int status = OK;
  char newname[PATH_MAX];

  /* Load a default name for debugging
  if (name == NULL) name = "@";
  */

  DPRINT_X(("add_node()\n"));

  /* Remove any "." and ".." entries from the path.  Bad paths
  ** have already been removed in newfiles().
  */
  (void) undot_path(name, newname, sizeof(newname));

  /* Build and populate the node */
  size = sizeof(ENODE);
  len = (newname[0] == '\0') ? 0 : strlen(newname) + 1;
  curr = *ppn;
  pt = NULL;
  if (curr != NULL && (curr->prev == NULL || curr->next == NULL))
	status = FAILED;
  else if ((pt = g_try_malloc(size + len)) == NULL)
	status = FAILED;
  else if (curr == NULL) {
	DPRINT_X(("add_node(): new pointer 0x%08lx\n", (ULONG) pt));
	pt->prev = pt;
	pt->next = pt;
	pt->data = data;
	pt->name = NULL;
	pt->label = NULL;
	curr = pt;
   }
  else if (posn == BEFORE) {
	pt->prev = curr->prev;
	curr->prev->next = pt;
	pt->next = curr;
	curr->prev = pt;
	pt->data = data;
	pt->name = NULL;
	pt->label = NULL;
	curr = pt;
  }
  else if (posn == AFTER) {
	pt->prev = curr->next->prev;
	curr->next->prev = pt;
	pt->next = curr->next;
	curr->next = pt;
	pt->data = data;
	pt->name = NULL;
	pt->label = NULL;
	curr = pt;
  }
  else {
	g_free(pt);
	pt = NULL;
	status = FAILED;
  }

  if (pt != NULL) {
	pt->id = get_id();
	if (newname[0] != '\0') {
		pt->name = (char *) pt + size;
		strcpy(pt->name, newname);
	}
	pt->uiid = 0;
	pt->signal = 0;
  }

  DPRINT_X(("add_node(): added string %s between %s and %s\n",
			pt->name == NULL ? "NULL" : pt->name,
			pt->prev->name == NULL ? "NULL" : pt->prev->name,
			pt->next->name == NULL ? "NULL" : pt->next->name
			));

  /* Update the value of the current (new) node */
  if (status == OK)
	*ppn = curr;

  return status;
}


/*
** c h e c k _ n o d e
**
** Check for a node in a linked list.
**
** Returns:	TRUE		Found
**		FALSE		Otherwise
**
** This checks the whole list, which can therefore be unsorted.
*/
static int check_node(ENODE *pn, char *name)
{
  ENODE *pt;
  int status = FALSE;

  DPRINT_X(("check_node()\n"));

  if (pn == NULL)
	return FALSE;
  else if (pn->next == pn)
	return (strcmp(pn->name, name) == SAME) ? TRUE : FALSE;
  else {
	for (pt = pn ; /* MT */ ; pt = pt->next) {
		if (strcmp(pt->name, name) == SAME) {
			status = TRUE;
			break;
		}
		if (pt->next == pn)
			break;
	}
  }

  return status;
}


/*
** d e l _ n o d e
**
** Delete a node from a linked list.
**
** Returns:	OK		Success
**		FAILED		Otherwise
**
** Also deletes any linked list data attached to the node.
** Updates the values pointed to by the name and data arguments as
** side effects.
*/
static int del_node(ENODE *pn, char *name, void **pdata)
{
  ENODE *pe;
  int status = OK;

  DPRINT_X(("del_node()\n"));

  if (pn == NULL || pn->prev == NULL || pn->next == NULL)
	status = FAILED;
  else {
	if (pn->data != NULL) {
		/* Stop the compiler complaining */
		pe = (ENODE *) pn->data;
		del_list(&pe);
		pn->data = (void *) pe;
	}
	DPRINT_X(("del_node(): deleting |%s|\n",
			pn->name == NULL ? "NULL" : pn->name));
	if (name != NULL && pn->name != NULL) strcpy(name, pn->name);
	if (pdata != NULL) *pdata = pn->data;
	DPRINT_X(("del_node(): freeing 0x%08lx next 0x%08lx\n",
					(ULONG) pn, (ULONG) pn->next));
	pn->next->prev = pn->prev;
	pn->prev->next = pn->next;
	g_free(pn);
  }

  return status;
}


/*
** f i n d _ n o d e
**
** Find a node in a linked list.
**
** Returns:	Index		Present
**		BEFORE		Not found, insert before
**		AFTER		Not found, insert after
**		FAILED		Otherwise
**
** The #defines BEFORE, AFTER, FAILED are all known to be less than 0.
** If full is TRUE, the whole list is checked (ie. an unsorted list).
** If update is TRUE, the target of the ppn argument is updated.
*/
static int find_node(ENODE **ppn, char *name, int full, int update)
{
  ENODE *pt;
  int index, result, status;

  DPRINT_X(("find_node()\n"));

  index = 0;
  status = FAILED;
  if (*ppn == NULL || (*ppn)->prev == NULL || (*ppn)->next == NULL) {
	DPRINT_X(("find_node(): head is NULL\n"));
	status = FAILED;
  }
  /* If there's only one element, no point in looking further */
  else if (*ppn == (*ppn)->next) {
	DPRINT_X(("find_node(): single element list\n"));
	if ((result = strcmp(name, (*ppn)->name)) == SAME)
		status = FOUND;
	else if (result < SAME)
		status = BEFORE;
	else
		status = AFTER;
  }
  else {
	for (index = 0, pt = *ppn ; index == 0 || pt != *ppn ;
						index++, pt = pt->next) {
		DPRINT_X(("find_node(): comparing |%s| and |%s|\n",
							name, pt->name));
		if ((result = strcmp(name, pt->name)) == SAME) {
			status = FOUND;
			break;
		}
		else if (!full && result < SAME) {	/* Sorted list */
			status = BEFORE;
			break;
		}
		else
			status = AFTER;
	}
	/* Move back if an unsuccessful search has wrapped around,
	** and not simply remained on the first element.
	*/
	if (index != 0 && pt == *ppn) pt = pt->prev;

	/* Update pointer if required */
	if (update) *ppn = pt;
  }

  if (status == FOUND) status = index;

  DPRINT_X(("find_node(): given |%s|, current %s, index %d status %d\n",
  name, (ppn != NULL && *ppn != NULL) ? (*ppn)->name : "NULL", index, status));

  return status;
}


/*
** s h o w _ n o d e
**
** Show a node from a linked list.
**
** Returns:	OK		Success
**		FAILED		Otherwise
*/
static int show_node(ENODE *pn, char *header, int d_flg)
{
  int status = OK;
  char *pc;

  DPRINT_X(("show_node()\n"));

  if (pn == NULL) return FAILED;

  pc = pn->name == NULL ? "NULL" : pn->name;

  DPRINT_X((
	"Node prev 0x%08lx this 0x%08lx next 0x%08lx data 0x%08lx name |%s|\n",
	(ULONG) pn->prev, (ULONG) pn, (ULONG) pn->next, (ULONG) pn->data, pc
	));
#if 0
  if (d_flg) {
	if (header == NULL)
		DPRINT_X(("%s\n", pc));
	else
		DPRINT_X(("%s |%s|\n", header, pc));
  }
  else {
	if (header == NULL)
		DPRINT_X(("%s\n", pc));
	else
		DPRINT_X(("%s %s\n", header, pc));
  }
#endif

  return status;
}


/*
** d e l _ l i s t
**
** Delete a linked list.
**
** Returns:	OK		Success
**		FAILED		Otherwise
**
** The node pointer argument is updated as a side effect.
*/
static int del_list(ENODE **ppn)
{
  ENODE *pn;
  int status = OK;

  DPRINT_X(("del_list()\n"));

  if (ppn == NULL || *ppn == NULL)
	return FAILED;

  for (pn = *ppn ; pn != NULL && pn->next != *ppn ; /* MT */) {
	DPRINT_X(("del_list(): deleting 0x%08lx next 0x%08lx\n",
				(ULONG) pn, (ULONG) pn->next));
	status = del_node(pn->next, NULL, NULL);
  }
  DPRINT_X(("del_list(): deleting 0x%08lx next 0x%08lx\n",
				(ULONG) pn, (ULONG) pn->next));
  status = del_node(pn, NULL, NULL);
  *ppn = NULL;

  return status;
}


/*
** s h o w _ l i s t
**
** Show a linked list.
**
** Returns:	OK		Success
**		FAILED		Otherwise
*/
static int show_list(ENODE *pn)
{
  ENODE *pt;
  int index;

  DPRINT_X(("show_list()\n"));

  for (index = 0, pt = pn ; pt && (index == 0 || pt != pn) ;
						index++, pt = pt->next) {
	show_node(pt, NULL, FALSE);
	if (pt->data != NULL)
		show_list(pt->data);
  }

  return OK;
}

#ifdef ADD_UNUSED
/*
** o c t s t r
**
** Convert a string to printable characters and octal values.
**
** Returns:	Pointer to buffer	Always.
**
** Note that the return value points to a static local buffer.
*/
static char * octstr(char *pc)
{
  static char buff[BUF_SIZ];
  static char sbuff[16];
  int j, offset, index;

  for (index = 0, j = 0 ; j < (BUF_SIZ / 4) && *pc != '\0' ; j++) {

	if ((*pc & 0x80) || (iscntrl(*pc) && !isspace(*pc))) {
		sprintf(sbuff, "\\%03o", *pc);
		offset = 4;
	}
	else {
		sbuff[0] = *pc;
		if (sbuff[0] == '\r') sbuff[0] = '\n';
		sbuff[1] = '\0';
		offset = 1;
	}
	sprintf(buff + index, "%s", sbuff);
	index += offset;
	pc++;
  }

  return buff;
}
#endif

/*
** t e s t
**
** See if Gnome is running.
**
** Returns:	2		GUI may be available
**		1		GUI is available
**		0		GUI is not available
**
** This function can be called _before_ init() is run, so
** can't rely on the Exists flag.
**
** The best approach seems to be to first check that X-windows
** is running as a prerequisite, and then look for one of Gnome's
** ORB servers.  However, the current Gnome ORB, ORBit, doesn't
** handle the CORBA "NameService" request, so the next best idea
** is to find a server which is activated only by the Gnome desktop.
** This is harder than it looks; the likely servers seem to change
** with each release of Gnome. ???
*/
static int test(void)
{
  Bonobo_ServerInfoList *pb_list = NULL;
  int j, status;
  char *pc, **ppq;
  /* values of the $TERM environment variable that are Xterms.
  ** "linux" is the string provided by the Gnome desktop.
  */
  static char *knownterm[] = {
	"xterm", "xterm-color", "dtterm", "iris-ansi", "aixterm",
	"hpterm", "Eterm", "linux", "gnome", "kvt", NULL
  };
  /* Use a server tied to a running Gnome desktop */
  char *bonobo_servers[] = {
		BONOBO_QUERY_1, BONOBO_QUERY_2, BONOBO_QUERY_3, NULL };

  DPRINT_F(("test()\n"));

  /* First check if X is running; Gnome won't be available otherwise.
  **
  ** This is taken from Steve Kirkendall's original X-windows test code:
  **
  ** "I've had some problems in which DISPLAY gets set to the valid name
  ** of a functioning X server, even though I'm personally not the user
  ** who's using that X server.
  **
  ** To prevent my elvis window from appearing on his X screen, I check
  ** a couple of environment variables.  If WINDOWID is set then I assume
  ** x11 is available; else I'll check the TERM variable against a list
  ** of known X terminals, and reject x11 unless TERM is in that list."
  */
  status = FAILED;	/* Ever the optimist */
  if (getenv("WINDOWID") != NULL)
	status = OK;
  else {
	if ((pc = getenv("TERM")) != NULL) {
		for (j = 0 ; knownterm[j] != NULL ; j++) {
			if (strcmp(knownterm[j], pc) == SAME) {
				status = OK;
				DPRINT_T(("test(): TERM found as %s.\n", pc));
				break;
			}
		}
	}
	else {
		status = FAILED;
		DPRINT_T(("test(): bad TERM\n"));
	}
  }

  /* Try to contact the X server. If we can contact it, great! */
  if (status == FAILED)
	/* MT */ ;
  else if (getenv("DISPLAY") == NULL) {
	status = FAILED;
	DPRINT_T(("test(): bad DISPLAY\n"));
  }
  else if (XOpenDisplay(NULL) == NULL) {
	status = FAILED;
	DPRINT_T(("test(): bad XOpenDisplay\n"));
  }
  else {
	DPRINT_T(("test(): X-status OK\n"));
  }

  /* X is running: is the Gnome desktop available ? */
  ppq = NULL;
  if (status == FAILED)
	/* MT */ ;
  else if (!bonobo_activation_is_initialized() && !bonobo_init(0, NULL)) {
	status = FAILED;		/* bonobo_init() Leak ### */
	DPRINT_T(("test(): bonobo initialization failure\n"));
  }
  else {
	/* There seems to be no server specific to Gnome ??? ### */
	for (ppq = bonobo_servers ; *ppq != NULL ; ppq++) {
		DPRINT_T(("test(): testing server |%s|\n", *ppq));
		pb_list = bonobo_activation_query(*ppq, NULL, NULL);
		/* We don't care if it's running, just that it's there */
		if (pb_list != NULL /* && pb_list->_length != 0 */) break;
	}
  }
  if (ppq != NULL && *ppq == NULL) {
	DPRINT_T(("test(): ppq is %s pb_list->_length is %d\n",
			*ppq == NULL ? "NULL" : "valid",
			pb_list == NULL ? 0 : pb_list->_length));
	status = FAILED;
  }

  if (pb_list != NULL) CORBA_free(pb_list);

  /* Update the global flag */
  if (status == OK) Exists = TRUE;

  /* Elvis uses 1 and 0, not TRUE and FALSE */
  return status == OK ? 1 : 0;
}


/*
** i n i t
**
** Initialise the editor's GUI.
**
** Returns:	Modified argc	Success
**		-1		Failure
**
** init() is called by main() before any other serious processing
** is done, and therefore gets first shot at the command line.
** It first initialises various bits of GLib, GTK+ and Gnome, and
** then parses the command line (handing back all the bits it doesn't
** understand), setting default fonts and colors, and so on.  Finally
** it calls init_gnome_win(), which builds the actual Gnome main
** window, which in turn contains the separate file editing windows
** which are initialised by create_edwin().
**
** Of the GUI interface fields, only "name" and test() are accessed
** before init().
**
** GTK+ returns no status from library calls, so there's
** no real way to check if these initialisations succeed.
*/
static int init(int argc, char *argv[])
{
  GNGUI *gg = &gn_gui;
  int status = OK;

  DPRINT_F(("init()\n"));

  /* Make it clear that we have no main window or editing window */
  gg->toplevel = NULL;
  gg->gwindow = NULL;
  gg->main_height = 0;
  gg->main_width = 0;
  gg->cur_cmd = -1;
  gg->edhead = NULL;
  gg->edcount = 0;
  gg->docshead = NULL;
  gg->docscount = 0;
  gg->noteshead = NULL;
  gg->notescount = 0;

  /* We also have no current print job or default file */
  gg->print_settings = NULL;
  gg->print_operation = NULL;
  gg->lastprint[0] = '\0';

  /* Park the name we were called by somewhere handy */
  gg->prog_name = argv[0];

  /* Find out who we actually are */
  gg->uid = getuid();

  /* Clear the dictionary and spell check utility names */
  gg->dictionary[0] = '\0';
  gg->spellcheck[0] = '\0';

  /* Clear the menu state */
  gg->menu_state = 0;

  /* No blocks are selected */
  gg->block_sel = NO;

  /* Clear the bad file name buffer */
  BadFile[0] = '\0';

  /* Try to handle the common signals gracefully.  Where else are they used?
  ** If Elvis is run from a Gnome-Terminal command line, do we want it to
  ** keep running when the terminal is closed ???
  **
  **   SIGHUP        1       Term    Hangup detected on controlling terminal
  **                                 or death of controlling process
  **   SIGINT        2       Term    Interrupt (^C) from keyboard
  **   SIGQUIT       3       Core    Quit from keyboard
  **   SIGILL        4       Core    Illegal Instruction
  **   SIGABRT       6       Core    Abort signal from abort(3)
  **   SIGFPE        8       Core    Floating point exception
  **   SIGKILL       9       Term    Kill signal
  **   SIGSEGV      11       Core    Invalid memory reference
  **   SIGPIPE      13       Term    Broken pipe: write to pipe with no readers
  **   SIGALRM      14       Term    Timer signal from alarm(2)
  **   SIGTERM      15       Term    Termination signal; kill(1) default value
  */
  signal(SIGHUP, SIG_IGN);	/* Ignore this; sent by closing terminal */
  signal(SIGINT, trap_handler);	/* Controlled shut down */
#if 0
  signal(SIGQUIT, SIG_DFL);	/* Failure, use default */
  signal(SIGILL, SIG_DFL);	/* Failure, use default */
  signal(SIGABRT, SIG_DFL);	/* Failure, use default */
  signal(SIGFPE, SIG_DFL);	/* Failure, use default */
  /* signal(SIGKILL, SIG_DFL);	** Can't be caught or ignored */
  signal(SIGSEGV, SIG_DFL);	/* Failure, use default */
  signal(SIGPIPE, SIG_DFL);	/* Leave as default; reset locally */
  signal(SIGALRM, SIG_DFL);	/* Leave as default; reset locally ? */
#endif
  signal(SIGTERM, trap_handler);/* Controlled shut down */

  /* Is this call necessary, or is it handled by gtk_init() ??? */
  g_type_init();

  /* This must be the first call to GTK+.
  ** It might be done earlier in test(), but there seems no
  ** way of undoing it if the interface isn't used after all.
  */
  if (!gtk_init_check(&argc, &argv))
	status = FAILED;
  /* Initialise GLib, GTK+ and Gnome */
  else if (init_interface(gg) == FAILED)
	status = FAILED;
  /* Sort out arguments and options */
  else if ((argc = init_defaults(argc, argv)) < 0)
	status = FAILED;
  /* Initialise the Gnome main window */
  else if (init_gnome_win(gg) == FAILED)
	status = FAILED;
  /* Initialise drag and drop */
  else if (init_dand(gg) == FAILED)
	status = FAILED;

  /* Configure the main menu and mouse menu */
  config_menus(gg);

  return (status == OK ) ? argc : -1;
}


/*
** u s a g e
**
** Display GUI dependant options.
**
** Returns:	Nothing		Always
*/
static void usage(void)
{
  size_t index;

  for (index = 0 ; index < sizeof(parsers) / sizeof(parsers[0]) ; index++) {
	if (parsers[index].cmd) {
		msg(MSG_INFO, "[ss]      $1$2",
				parsers[index].name, parsers[index].usage);
	}
  }

  return;
}


/*
** l o o p
**
** Run the main loop.
**
** Returns:	Nothing		Always
*/
static void loop(void)
{
  GNGUI *gg = &gn_gui;
  static UINT data = 0;

  DPRINT_F(("*** LOOP() ***\n"));

  /* These messages have to wait until the GUI window is fully established.
  ** Since there is only a single message line, there has to be some sort
  ** of delay between messages.  show_warning() is therefore loaded into
  ** the main loop with a timeout attached, and runs at two-second intervals
  ** displaying messages on the status bar until there are no more messages.
  ** However, only the last unparseable filename is reported.
  **
  ** It would also be possible for GTK to switch messages itself, using the
  ** status bar message ids, but it seems better to work through Elvis.
  **
  ** Why does Elvis call beep() for only the first in a series of messages ???
  */
#ifdef CFGDIR_ELVIS
  if (!gg->cfgdir_flg)
	data |= 1;
#endif
  if (o_startposition && (gg->main_start_x < 0 || gg->main_start_y < 0))
	data |= 2;
  if (gg->gconf_client == NULL)
	data |= 4;
  if (!gg->scheme_flg)
	data |= 8;
  if (BadFile[0] != '\0')
	data |= 16;

  /* Add a function to the idle loop to handle Elvis' mainfirstcmd() */
  g_idle_add(firstcmd, NULL);

  /* Add a function to the idle loop to handle startup error messages */
  g_timeout_add(2000, show_warning, (gpointer) &data);

  DPRINT_F(("loop(): running gtk_main()\n"));

  gtk_main();

  return;
}


/*
** p o l l
**
** Check for user interaction.
**
** Returns:	TRUE		Abort
**		FALSE		Continue processing
**
** Exactly what does Elvis want done with the reset argument?
** This function is generally called with reset TRUE. ???
*/
static ELVBOOL poll(ELVBOOL reset)
{
  DPRINT_V(("poll()\n"));

  if (StopThat) {
	StopThat = FALSE;
	return TRUE;
  }
  else
	return FALSE;
}


/*
** t e r m
**
** Shut down the Gnome GUI.
**
** Returns:	Nothing		Always
**
** Elvis usually calls this after the GUI has quit the main loop,
** so check the loop is still running before stopping it.
*/
static void term(void)
{
  GNGUI *gg = &gn_gui;

  DPRINT_F(("term(): loop level %u\n", gtk_main_level()));

  /* Tidier to unload the snooper first */
  if (gg->snoop_id != 0)
	gtk_key_snooper_remove(gg->snoop_id);
  gg->snoop_id = 0;

  /* Close down any GConf connection */
  if (gg->gconf_client != NULL) {
	gconf_client_notify_remove(gg->gconf_client, gg->gconf_id);
	gconf_client_remove_dir(gg->gconf_client, GCONF_PATH, NULL);
	/* g_free(GCONF_PATH) */
	g_object_unref(gg->gconf_client);
	gg->gconf_client = NULL;
  }

  /* Save anything in the docslist */
  (void) store_docslist();

   /* Save the gtrc settings, barring the start position */
   store_settings(FALSE);

#ifdef DEBUG_MENU
  /* If there's a child, kill it */
  if (Child != 0) {
	kill(Child, SIGINT);
	Child = 0;
  }
#endif

  /* Probably not needed */
  if (gtk_main_level() > 0) gtk_main_quit();

  return;
}


/*
** c r e a t e g w
**
** Create an editing window.
**
** Returns:	TRUE		Window created
**		FALSE		Window not created
**
** Called by the main Elvis code, not by the Gnome GUI code.
** The global struct GnomeGui gn_gui is updated as a side effect.
*/
static ELVBOOL creategw(char *name, char *attr)
{
  GNGUI *gg = &gn_gui;
  struct Enode *pn;
  gint height, width;
  int first, status;
  char *pc, *pf, buff[64];

  DPRINT_F(("creategw()\n"));

  /* The VTE code sets the terminal to a fixed 80x24 size on startup,
  ** so if those values aren't our defaults we have to resize the
  ** window.  Furthermore, it seems we must resize after the widget
  ** has been displayed, or the settings won't take effect.  So we
  ** update the first editing window here.
  */
  first = (gg->edhead == NULL) ? TRUE : FALSE;

  /* Mark initialisation complete, so default options won't be updated */
  InitDone = first;

  /* Park the current main window sizes */
  gtk_window_get_size((GtkWindow *) gg->toplevel, &width, &height);

  DPRINT_H(("creategw(): name |%s| attr |%s|\n", name, attr));
  DPRINT_H(("creategw(): height %d width %d edhead %s\n",
				height, width,
				gg->edhead == NULL ? "NULL" : "valid"
				));

  /* Create the GUI's editing window for this file */
  if ((status = create_edwin(&gg->edhead)) == OK) {
	/* Add an Enode for the (first file on the) current argslist */
	pn = gg->edhead->argshead;
	if (add_node(&pn, name, NULL, AFTER) == OK) {
		for (pc = pf = pn->name ; *pc != '\0' ; pc++)
			if (*pc == SEPCHR) pf = pc;
		if (*pf == SEPCHR) pf++;
		pn->label = pf;

		/* Update the current EDWIN and Files menu */
		gg->edhead->argscount = 1;
		gg->edhead->edname = pn->name;
		gg->edhead->edtitle = pn->label;
		unload_files_menu(gg->edhead->argshead);
		gg->edhead->argshead = pn;
		load_files_menu(gg->edhead->argshead);

		/* Retitle the main Gnome window */
		gtk_window_set_title(GTK_WINDOW(gg->gwindow),
					gg->edhead->edtitle);

		/* Create Elvis' window. eventcreate() is part of the main
		** Elvis code, and returns TRUE for success, FALSE otherwise.
		** If it returns FALSE, the new window must be destroyed.
		*/
		if (eventcreate(gg->edhead, NULL, name, gg->edhead->rows,
							gg->edhead->cols)) {
			(void) eventdraw(gg->edhead);
		}
		else {
			/* If eventcreate() fails, destroy the new window */
			status = destroy_edwin(&gg->edhead);
			/* Now shut down the interface (see below) */
			status = FAILED;
		}
	}
	else
		status = FAILED;
  }

  /* If the --showtoolbar option is in use, then update the GConf database */
  DPRINT_H(("creategw(): o_showtoolbar %d"
	" optdescs[O_TOOL] %ld optvals[0_TOOL].flags %u\n",
	o_showtoolbar, strtol(optdescs[O_TOOL].dflt, NULL, 0),
						optvals[O_TOOL].flags
	));

  if (optvals[O_TOOL].flags &= OPT_SET) {
	if (o_showtoolbar != gconf_client_get_bool(gg->gconf_client,
				GCONF_PATH "/ui/showtoolbar", NULL)) {
		gconf_client_set_bool(gg->gconf_client,
			GCONF_PATH "/ui/showtoolbar",
			(gboolean)  o_showtoolbar, NULL);
		/* Give GTK+ time to stabilise */
		while (gtk_events_pending()) gtk_main_iteration();
	}
  }

  /* Set the first editing window shown to the correct (default) size,
  ** and with it the toplevel window.
  ** This is a work-around for the fixed VTE default terminal size;
  ** it causes an annoying delay on startup, so a better work-around
  ** would be a good idea. ???
  ** There are also problems suppressing the main window's tooldock.
  ** The work-around is to wait for all the window sizes to stabilise,
  ** and then remove the toolbar if the command line option is set. ???
  */
  if (first && status == OK && !(o_defaultrows == DEF_ROWS &&
					o_defaultcols == DEF_COLS &&
					o_showtoolbar && !o_startposition)) {
	/* Update the rows and colums if required */
	if (!(o_defaultrows == DEF_ROWS && o_defaultcols == DEF_COLS)) {
		/* Give GTK+ time to stabilise */
		while (gtk_events_pending()) gtk_main_iteration();
		gg->def_cols = o_defaultcols;
		gg->def_rows = o_defaultrows;
		gg->edhead->cols = gg->def_cols;
		gg->edhead->rows = gg->def_rows;
		gg->main_height = 0;
		gg->main_width = 0;
		size_main_window(gg, TRUE);
	}

	/* Hide the tooldock and toolbar if required */
	if (!o_showtoolbar) {
		/* Give GTK+ time to stabilise */
		while (gtk_events_pending()) gtk_main_iteration();
		switch_toolbar();
	}

	/* Move the window to a fixed starting position if required */
	if (o_startposition && gg->main_start_x >= 0 && gg->main_start_y >= 0) {
		/* Give GTK+ time to stabilise */
		while (gtk_events_pending()) gtk_main_iteration();
		sprintf(buff, "0x0+%d+%d", gg->main_start_x, gg->main_start_y);
		(void) gtk_window_parse_geometry(GTK_WINDOW(gg->toplevel),
									buff);
	}
  }

  DPRINT_H((
	"creategw(): rows %ld vadj lower %f upper %f\n"
	"           step %f page %f size %f value %f\n",
	gg->edhead->rows,
	gg->vert_adj->lower, gg->vert_adj->upper,
	gg->vert_adj->step_increment, gg->vert_adj->page_increment,
	gg->vert_adj->page_size, gg->vert_adj->value
  ));

  /* On complete failure, shut down Elvis' Gnome interface.
  ** Should this be left for the main body of the code ???
  */
  if (status == FAILED) {
	pc = "Cannot create new Gnome window - quitting.";
	msg(MSG_ERROR, pc);
	msgflush();
	/* Try writing to the default output as well */
	g_print("%s\n", pc);
	gtk_main_quit();
  }

  return status == OK ? ElvTrue : ElvFalse;
}


/*
** d e s t r o y g w
**
** Destroy an editing window.
**
** Returns:	Nothing		Always
**
** What is the "force" argument expected to do ??? ###
*/
static void destroygw(GUIWIN *gw, ELVBOOL force)
{
  GNGUI *gg = &gn_gui;
  EDWIN *edwin = (EDWIN *) gw;
  struct Enode *pe;
  int status;

  DPRINT_F(("destroygw(): loop level %u\n", gtk_main_level()));

  pe = edwin->argshead;

  /* Remove the names from the Files menu */
  status = unload_files_menu(pe);

  /* Remove the argslist nodes */
  status = destroy_edit_ring(&pe);

  /* Destroy the edit window */
  status = destroy_edwin(&edwin);

  /* Revert the main window to its original size */
  gg->main_height = 0;
  gg->main_width = 0;
  size_main_window(gg, TRUE);

  /* If there's a window left, then we need to restore the File menu.
  ** If this is the last window Elvis thinks we're all done, but in fact
  ** there's still the GUI to shut down.  Break out of the main loop so
  ** that Elvis can call term() and then clean itself up.
  */
  if (gg->edcount > 0)
	load_files_menu(gg->edhead->argshead);
  else if (gtk_main_level() > 0)
	gtk_main_quit();

  return;
}


/*
** f o c u s g w
**
** Move the focus to another editing window.
**
** Returns:	TRUE		Always
**
** It might be worth returning FALSE if anyone seemed to use it.
*/
static ELVBOOL focusgw(GUIWIN *gw)
{
  GNGUI *gg = &gn_gui;
  EDWIN *pe, *head, *edwin;
  int flg, count, status;

  DPRINT_F(("focusgw()\n"));

  /* Get the current editing window, the head of the list */
  edwin = (EDWIN *) gw;
  head = gg->edhead;

  status = FAILED;
  /* Search the linked list for the required window */
  if (head == NULL || head->prev == NULL || head->next == NULL) {
	status = FAILED;
  }
  /* If there's only one element, don't try searching further */
  else if (head == edwin) {
	status = OK;
  }
  else {
	count = 0;
	for (flg = TRUE, pe = head ; flg || pe != head ;
					flg = FALSE, pe = pe->next) {
		if (pe == edwin) {
			status = FOUND;
			break;
		}
		count++;
	}
  }

  /* Update the master GUI structure */
  if (status == FOUND)
	status = goto_edit_ring(gg, count, NEXT);

  return TRUE;
}


/*
** r e t i t l e
**
** Retitle the main window.
**
** Returns:	Nothing		Always
*/
static void retitle(GUIWIN *gw, char *name)
{
  GNGUI *gg = &gn_gui;

  DPRINT_F(("retitle()\n"));

  /* Retitle the main Gnome window */
  gtk_window_set_title(GTK_WINDOW(gg->gwindow), name);

  return;
}


/*
** r e s e t
**
** Reset the Gnome GUI as far as is possible.
**
** Returns:	Nothing		Always
**
** Only a reset of the current editing window seems practical.
*/
static void reset(void)
{
  GNGUI *gg = &gn_gui;

  DPRINT_V(("reset()\n"));

  /* Reset the current vterm, resetting tabstops and
  ** clearing the scrollback buffer if required.
  ** This call is used only by the termcap GUI.
  ** The current EDWIN should also be updated. ###
  */
  if (gg != NULL && gg->edhead != NULL)
	vte_terminal_reset(VTE_TERMINAL(gg->edhead->vterm), TRUE, TRUE);

  return;
}


/*
** f l u s h
**
** Flush an editing window.
**
** Returns:	Nothing		Always
**
** With no buffering, there's little use for flush().
*/
static void flush(void)
{
  GNGUI *gg = &gn_gui;

  DPRINT_V(("flush()\n"));

  /* Just checking */
  if (gg->edhead == NULL) return;

  /* Update the scrollbar */
  gtk_widget_queue_draw(gg->vert_bar);

  return;
}


/*
** m o v e t o
**
** Move the cursor to a given character cell.
**
** Returns:	Nothing		Always
**
** The upper left character cell is designated column 0, row 0.
*/
static void moveto(GUIWIN *gw, int col, int row)
{
  EDWIN *edwin = (EDWIN *) gw;
  char buff[S_BUF_SIZ];

  DPRINT_W(("moveto() col %d row %d\n", col, row));

  /* Move the cursor with an escape code: ESC[YY;XXH
  ** The CSI is 0x9B on 8-bit terminals, and ESC[ on 7-bit.
  ** How does UTF-8 affect this?  Generally ESC[ seems to work.
  **
  ** Note that vte (xterm) row and column numbers start at 1,
  ** while Elvis uses 0.
  */
  sprintf(buff, "\033[%d;%dH", row + 1, col + 1);
  write_buff(edwin, buff, strlen(buff));

  return;
}


/*
** d r a w
**
** Put some text on the vte_terminal widget.
**
** Returns:	Nothing		Always
**
** By default the widget looks like an xterm or a VT220.  The
** text is written in the given font (color) starting at the
** cursor's current position.
**
** The color control codes description seems to envisage a fixed
** palette, but the vte_terminal widget uses the colors loaded by
** Elvis in the addcolor() function.
**
** Elvis uses colour code 0 to set to the terminal's default scheme,
** and then 1...16 as the user defined schemes.  Some documentation
** implies that the first eight are background colors and the last
** eight foreground, but this doesn't seem to be the case.  In fact,
** Elvis will allocate mostly foreground colors if it's allowed to.
*/
static void draw(
		GUIWIN	*gw,	/* the window where the text should be drawn */
		long	fg,	/* foreground color */
		long	bg,	/* background color */
		int	bits,	/* bitmap of other attributes */
		CHAR	*text,	/* the text to draw */
		int	len	/* number of characters in text */
		)
{
  EDWIN *edwin = (EDWIN *) gw;
  unsigned int mask;
  long fg_clr;
  long bg_clr;
  char buff[16];

  DPRINT_W(("draw(): fg %ld bg %ld bits %d text[0] |%c| len %d\n",
					fg, bg, bits, text[0], len));

#ifdef DPRINT_WW
  { /* Debugging block begin */
  DPRINT_W(("draw(): attributes are "));
  if (bits & COLOR_BOLD)
	DPRINT_W((" Bold"));
  if (bits & COLOR_ITALIC)
	DPRINT_W((" Italic"));
  if (bits & COLOR_UNDERLINED)
	DPRINT_W((" Underlined"));
  if (bits & COLOR_BOXED)
	DPRINT_W((" Boxed"));
  if (bits & (COLOR_PROP | COLOR_PROPSET))
	DPRINT_W((" Proportional"));
  if (bits & COLOR_GRAPHIC)
	DPRINT_W((" Graphic"));
  if (bits & COLOR_FG)
	DPRINT_W((" Foreground"));
  if (bits & COLOR_BG)
	DPRINT_W((" Background"));
  DPRINT_W(("\n"));
  } /* Debugging block end */
#endif

  /* vte_terminal has vte_terminal_set_color_XXX() for background,
  ** foreground, bold and dim.
  **
  ** For 8-bit systems CSI is 0x9b, but gnome-terminal seems to work with
  ** the 7-bit ESC [ sequence. ???
  */

  /* Check for changes in foreground colour, assuming a 16-colour xterm */
  if ((bits & COLOR_FG) && edwin->fg_clr != fg) {
	fg_clr = fg;
	DPRINT_W(("draw(): foreground changed\n"));
	/* If foreground not enabled then assume a default foreground */
	if ((fg < 0) || (fg > 16)) { /* Do something reasonable */
		fg_clr = 0;
		DPRINT_W(("draw(): has bad foreground code %ld\n", fg_clr));
	}
	if (fg_clr == 0)	/* Default foreground is xterm code 39 */
		sprintf(buff, "\033[%dm", 39);
	else if (fg_clr < 9)	/* Foregrounds 0..7 are xterm codes 30..37 */
		sprintf(buff, "\033[%ldm", fg_clr + 29);
	else			/* Foregrounds 8..15 are xterm codes 90..97 */
		sprintf(buff, "\033[%ldm", fg_clr + 90 - 9);

	/* Write out the mode change string */
	write_buff(edwin, buff, strlen(buff));
	/* Update the reference value */
	edwin->fg_clr = fg_clr;
  }

  /* Check for changes in background colour, assuming a 16-colour xterm */
  if ((bits & COLOR_BG) && edwin->bg_clr != bg) {
	bg_clr = bg;
	DPRINT_W(("draw(): background changed\n"));
	/* If background not enabled then assume a default background */
	if ((bg_clr < 0) || (bg_clr > 16)) { /* Do something reasonable */
		bg_clr = 0;
		DPRINT_W(("draw(): has bad background code %ld\n", bg_clr));
	}
	if (bg_clr == 0)	/* Default background is xterm code 49 */
		sprintf(buff, "\033[%dm", 49);
	else if (bg_clr < 9)	/* Backgrounds 0..7 are xterm codes 40..47 */
		sprintf(buff, "\033[%ldm", bg_clr + 39);
	else			/* Backgrounds 8..15 are xterm codes 100..107 */
		sprintf(buff, "\033[%ldm", bg_clr + 100 - 9);

	/* Write out the mode change string */
	write_buff(edwin, buff, strlen(buff));
	/* Update the reference value */
	edwin->bg_clr = bg_clr;
  }

  /* Send an ESC[VALm to VTE where VAL is a code to select features.
  ** VAL = 0 -> set normal (default)
  ** VAL = 1 -> set bold
  ** VAL = 2 -> set half
  ** VAL = 3
  ** VAL = 4 -> set underline
  ** VAL = 5 -> set blink
  ** VAL = 6
  ** VAL = 7 -> set reverse
  ** VAL = 8 -> set invisible
  ** VAL = 9 -> set strikethrough
  **
  ** VAL = 21 -> unset bold, half (bug)
  ** VAL = 22 -> unset bold, half
  ** VAL = 23
  ** VAL = 24 -> unset underline
  ** VAL = 25 -> unset blink
  ** VAL = 26
  ** VAL = 27 -> unset reverse
  ** VAL = 28 -> unset invisible
  ** VAL = 29 -> unset strikethrough
  */
  if ((mask = bits ^ edwin->attr_mask) != 0) {
	if (mask & COLOR_BOLD) {
		write_buff(edwin,
		(bits & COLOR_BOLD) ? "\033[1m" : "\033[22m",
		(bits & COLOR_BOLD) ? 4U : 5U);
	}
	if (mask & COLOR_ITALIC) {
		write_buff(edwin,
		(bits & COLOR_ITALIC) ? "\033[2m" : "\033[22m",
		(bits & COLOR_ITALIC) ? 4U : 5U);
	}
	if (mask & COLOR_UNDERLINED) {
		write_buff(edwin,
		(bits & COLOR_UNDERLINED) ? "\033[4m" : "\033[24m",
		(bits & COLOR_UNDERLINED) ? 4U : 5U);
	}
	if (mask & COLOR_BOXED) {
		write_buff(edwin,
		(bits & COLOR_BOXED) ? "\033[7m" : "\033[27m",
		(bits & COLOR_BOXED) ? 4U : 5U);
	}
	edwin->attr_mask = bits;
  }

  /* Write out the actual text */
  write_buff(edwin, text, (size_t) len);

  return;
}


/*
** s h i f t
**
** Shift the screen contents right.
**
** Returns:	FALSE		Always
**
** This command isn't enabled in the Gnome GUI interface.
*/
static ELVBOOL shift(GUIWIN *gw, int qty, int rows)
{
  DPRINT_F(("shift()\n"));

  return FALSE;
}


/*
** s c r o l l
**
** Scroll the editing window.
**
** Returns:	TRUE		Always
*/
static ELVBOOL scroll(GUIWIN *gw, int qty, ELVBOOL notlast)
{
  EDWIN *edwin = (EDWIN *) gw;
  int last;
  char buff[S_BUF_SIZ];

  DPRINT_V(("scroll(): notlast %d qty %d vt rows %ld\n",
		notlast, qty, edwin->rows));

  /* The scrolllast flag in guignome is set to FALSE,
  ** so Elvis should always call with notlast TRUE,
  ** ie. don't scroll the last line.
  **
  ** Note that xterm starts row numbers at 1.
  */
  if (notlast)
	last = edwin->rows - 1;
  else
	last = edwin->rows;

  /* Scroll lines up from the bottom, losing lines off the top.
  **
  ** Save the cursor, setup the scroll region (possibly not including
  ** the last line), delete some lines, restore the full scroll region,
  ** and then restore the cursor.
  **
  ** Save cursor Set scroll Delete lines Set scroll Restore cursor
  ** ESC7        ESC[1;Xr   ESC[XM       ESC[1;Xr   ESC8
  **
	sprintf(buff, "\0337\033[1;%dr\0338\033[%dM\033[1;%ldr",
  */
  if (qty < 0) {
	sprintf(buff, "\0337\033[1;%dr\033[%dM\033[1;%ldr\0338",
				last, -qty, edwin->rows);
  }

  /* Scroll lines down from the top, losing lines off the bottom.
  **
  ** Save the cursor, setup the scroll region (possibly not including
  ** the last line), insert some lines, restore the full scroll region,
  ** and then restore the cursor.
	sprintf(buff, "\0337\033[1;%dr\0338\033[%dL\033[1;%ldr",
  */
  else {
	sprintf(buff, "\0337\033[1;%dr\033[%dL\033[1;%ldr\0338",
				last, qty, edwin->rows);
  }

  /* Write out the result */
  if (qty != 0)
	write_buff(edwin, buff, strlen(buff));

  return TRUE;
}


/*
** c l r t o e o l
**
** Clear to end of line.
**
** Returns:	TRUE		Always
*/
static ELVBOOL clrtoeol(GUIWIN *gw)
{
  EDWIN *edwin = (EDWIN *) gw;

  DPRINT_W(("clrtoeol()\n"));

  write_buff(edwin, "\033[K", 3);

  return TRUE;
}


/*
** t e x t l i n e
**
** Write a line to the screen.
**
** Returns:	Nothing		Always
**
** This command isn't enabled in the Gnome GUI interface.
*/
static void textline(GUIWIN *gw, CHAR *text, int len)
{
  DPRINT_F(("textline()\n"));

  return;
}


/*
** b e e p
**
** Audible (and visual?) bell.
**
** Returns:	Nothing		Always
*/
static void beep(GUIWIN *gw)
{
  GNGUI *gg = &gn_gui;
  char buff[] = "\007";

  DPRINT_F(("beep()\n"));

  /* In some circumstances the main Elvis code calls this routine
  ** on shutdown with an invalid window pointer; it seems to be a
  ** race condition with the clean-up of the final editing window.
  ** The work-around is to use the gnomegui entry for the current
  ** window directly.
  */

  if (gg->edhead->vterm != NULL)
	vte_terminal_feed(VTE_TERMINAL(gg->edhead->vterm), buff, 1L);
  else
	write(2, buff, strlen(buff));	/* Used for debugging only */

#if 0
  /* This doesn't obey the VTE audible and visual bell flags */
  GdkDisplay * dp;
  dp = gtk_widget_get_display(GTK_WIDGET(vt));
  gdk_display_beep(dp);
#endif

  return;
}


/*
** g u i m s g
**
** Update the message bar of the Gnome window.
**
** Returns:	TRUE		Always
**
** Renamed from msg() in struct gui_s because of clash with
** functions in message.h.
**
** A MSG_DEBUG enum has been added to Elvis' message.h, and such
** messages are printed directly, not passed on to the GUI window.
**
** It would be nice if the color of these message strings could be
** set in elvis.clr. ???
*/
static ELVBOOL guimsg(GUIWIN *gw, MSGIMP imp, CHAR *text, int len)
{
  GNGUI *gg = &gn_gui;
  EDWIN *edwin = (EDWIN *) gw;
  char buff[256];

  /* Elvis (in message.c/msgflush()) can distinguish only between messages
  ** of type MSG_ERROR and MSG_INFO, using the highlighting marker in the
  ** output buffer. A MSG_DEBUG type marker is lost in the processing.
  */
  if (gg == NULL || gg->msgbar == NULL)
	g_print("%s\n", text);
  else {
	buff[0] = '\0';
	strncat(buff, text, sizeof(buff) - 1);
	gtk_statusbar_pop(GTK_STATUSBAR(gg->msgbar), edwin->ms_id);
	gtk_statusbar_push(GTK_STATUSBAR(gg->msgbar), edwin->ms_id, buff);
  }

  return TRUE;
}


/*
** s c r o l l b a r
**
** Set the size of the scroll bar.
**
** Returns:	Nothing		Always
**
** The scrollbar can control the paging of the input by itself,
** as can Elvis through its key commands, but the scrollbar must
** be made to track Elvis' key command inputs, and that is done
** through this routine.
**
** The vte_terminal widget has a GtkAdjustment element that deals
** with its scrollbar; however, in this case the terminal is being
** used with a scrollback buffer size of zero, and Elvis is handling
** the scrolling via a scrolling window containing the vte_terminal.
**
** The arguments top, bottom and total are the first line shown, the
** last line shown, and the total number of lines.
*/
static void scrollbar(GUIWIN *gw, long top, long bottom, long total)
{
  GNGUI *gg = &gn_gui;
  EDWIN *edwin = (EDWIN *) gw;
  double a, b, c, e;
  glong rows, cols, char_width, char_height;
  int xpad, ypad;

  DPRINT_V(("scrollbar()\n"));

  /* Get the current vterm size */
  rows = edwin->rows;
  cols = edwin->cols;
  /* Get the current vterm padding */
  xpad = edwin->xpad;
  ypad = edwin->ypad;
  /* Get the current window size */
  char_width = edwin->pango_width;
  char_height = edwin->pango_height;

  /* GtkAdjustment fields:
  **
  ** gdouble lower;		the minimum value.
  ** gdouble upper;		the maximum value.
  ** gdouble value;		the current value.
  ** gdouble step_increment;	the increment to use to make minor changes.
  **				In a GtkScrollbar this increment is used
  **				when the mouse is clicked on the arrows
  **				at the top and bottom of the scrollbar,
  **				to scroll by a small amount.
  ** gdouble page_increment;	the increment to use to make major changes.
  **				In a GtkScrollbar this increment is used
  **				when the mouse is clicked in the trough,
  **				to scroll by a large amount.
  ** gdouble page_size;	the page size.
  **				In a GtkScrollbar this is the size of the
  **				area which is currently visible.
  */

  /* The scrollbar widget (and some others) uses pixel dimensions */
  a = rows * char_height;		/* Pixels per screen */
					/* Pixels per slider trough */
  b = WIN_SHIM + ypad + a - 2 * gg->stepper_height;
  c = b / (double) (total);		/* Pixels per line move */
  if (c < 1) c = 1.0;			/* Minimum pixels move */

  /* Avoid infinities where possible */
  if (total == 0) {
	c = 0.0;
	total = rows - 1;
	top = 0;
  }

  /* Update the GtkAdjustment settings, apart from value */
  gg->vert_adj->lower = 0.0;
  gg->vert_adj->upper = b;
  gg->vert_adj->step_increment = c;
  gg->vert_adj->page_increment = c * (rows - 1);
  gg->vert_adj->page_size = b * ((double) (rows - 1) / (double) total);
  /* gg->vert_adj->value = b * (double) top / (double) total; */
  e = b * (double) top / (double) total;

  /* The screen has more lines than the file */
  if (total <= (edwin->rows - 1))
	gg->vert_adj->page_size = b;

  DPRINT_S((
	"scrollbar(): upper %f step_inc %f page_inc %f page_size %f value %f\n",
	gg->vert_adj->upper, gg->vert_adj->step_increment,
	gg->vert_adj->page_increment, gg->vert_adj->page_size,
	gg->vert_adj->value
	));

  /* Change the slider context values */
  gtk_adjustment_changed(gg->vert_adj);

  /* Update the current position.
  ** A signal handler deals with the actual value update.
  */
  gg->vert_adj->value = e;

  return;
}


/*
** s t a t u s b a r
**
** Update the status bar of the Gnome window.
**
** Returns:	TRUE		Always
**
** What is a "learn letter" ???
*/
static ELVBOOL statusbar(
	GUIWIN	*gw,	/* window whose status bar should be updated */
	CHAR	*cmd,	/* partial command keys */
	long	line,	/* line number */
	long	column,	/* column number */
	_CHAR_	learn,	/* learn letter, or '*' if modified, or ' ' otherwise */
	char	*mode	/* mode name, e.g. "Command" or "Input" */
	)
{
  GNGUI *gg = &gn_gui;
  EDWIN *edwin = (EDWIN *) gw;
  char buff[256];

  /* snprintf() needed here, but it's not in C89 ??? */
  sprintf(buff, "%ld/%ld %c %s", line, column, learn, mode);
  gtk_statusbar_pop(GTK_STATUSBAR(gg->statusbar), edwin->st_id);
  gtk_statusbar_push(GTK_STATUSBAR(gg->statusbar), edwin->st_id, buff);

  return TRUE;
}


/*
** k e y l a b e l
**
** Convert <label> to key code.
**
** Returns:	Length of raw codes	Success
**		0			Text not recognized
**
** Taken from Steve Kirkendall's guiX11 code.  Note that the keycodes used
** here are the GDK codes, whereas the guix11 interface (obviously) uses
** the X11 keycodes.  The main body of the Elvis code doesn't seem affected
** by this, but it would be nice to be certain. ???
**
** CHAR is typedef'd as unsigned char in elvis.h.
** ELVCTRL(ch) is defined as ((ch) ^ 0x40) in elvis.h.
*/
static int keylabel(
	CHAR	*given,		/* the string typed in by the user */
	int	givenlen,	/* length of the user's string */
	CHAR	**label,	/* pointer to (CHAR *) to set to key label */
	CHAR	**rawin		/* pointer to (CHAR *) to set to raw codes */
	)
{
  static UCHAR lookup[] =	/* Hex to decimal conversion */
	"0123456789abcdefABCDEF"
	"\000\001\002\003\004\005\006\007\010\011\012\013\014\015\016\017"
	"\012\013\014\015\016\017\177";
  static UCHAR rawbuff[16];	/* buffer, holds raw byte string */
  UCHAR labelbuff[32];		/* buffer, holds label string */
  char *name = NULL;
  guint keyvalue = 0;
  int j, k, status = OK;
  char modifier;		/* 'S' = shift, 'C' = ctrl, '\0' = normal */

  DPRINT_F(("keylabel()\n"));

  /* The lookup table has 16 hex input values, 6 duplicate (capital) hex
  ** input values, 22 decimal output values, an error output value, and
  ** a terminating '\0', making 21 valid entries and 46 elements in all.
  */
  assert(sizeof(lookup) == 46);

  /* No single character can be a key label */
  if (givenlen < 2)
	return 0;

  /* Apparently the user buffer may not contain an actual string */
  given[givenlen] = '\0';

  /* Inputs can have any of the following forms,
  ** where n is a decimal digit and h is a
  ** hexadecimal digit.
  ** #n
  ** #nn
  ** <#n>
  ** <#nn>
  ** <Fn>
  ** <Fnn>
  ** <hhhh>
  ** ^Khhhh
  ** ^Shhhh
  ** ^Chhhh
  ** <Keyname>
  ** <C-<Keyname>>
  ** <S-<Keyname>>
  */

  DPRINT_K(("keylabel(): given label |%s| length %d\n", given, givenlen));

  modifier = '\0';
  /* Is this a raw keycode ? */
  if (givenlen == 5 && (*given == ELVCTRL('K')
			|| *given == ELVCTRL('S')
			|| *given == ELVCTRL('C'))) {
	/* The leading char indicates the modifier */
	if (*given == ELVCTRL('S'))
		modifier = 'S';
	else if (*given == ELVCTRL('C'))
		modifier = 'C';
	else
		modifier = '\0';

	/* Convert a key string into a KeySym */
	for (j = 1, keyvalue = 0 ; j <= 4 ; j++) {
		keyvalue <<= 4;
		for (k = 0 ; k < 22 && given[j] != lookup[k] ; k++)
			/* MT */ ;
		if (lookup[k + 22] == '\177')
			status = FAILED;
		else
			keyvalue += lookup[k + 22];
	}

	/* Get the GDK keyname; the returned string can't be modified */
	if (status == OK && (name = gdk_keyval_name(keyvalue)) == NULL)
		status = FAILED;
  }
  /* Is this a label in #n or #nn format ? */
  else if (given[0] == '#' && isdigit(given[1]) &&
		(given[2] == '\0' || (isdigit(given[2]) && given[3] == '\0')) &&
		givenlen < (sizeof(labelbuff) - 1)) {
	strncpy((char *) labelbuff, (char *) given, sizeof(labelbuff) - 1);
	labelbuff[0] = 'F';
	labelbuff[givenlen] = '\0';
  }
  /* Is this a label in <#n> or <#nn> format ? */
  else if (given[1] == '#' && isdigit(given[2]) &&
		(given[3] == '>' || (isdigit(given[3]) && given[4] == '>')) &&
		givenlen < (sizeof(labelbuff) - 1)) {
	/* Standardise the format of the strings */
	given += 2;
	givenlen -= 2;
	strcpy((char *) labelbuff, "F");
	strncat((char *) labelbuff, (char *) given, sizeof(labelbuff) - 2);
	/* Lose the leading and trailing angle brackets */
	labelbuff[givenlen] = '\0';
  }
  /* Is this is a label in <foo> or <M-<foo>> format ? */
  else if (given[0] == '<' && given[givenlen - 1] == '>' &&
				givenlen < (sizeof(labelbuff) - 1)) {
	/* Look for "C-" or "S-" modifier */
	if (givenlen > 5
			&& (toupper(given[1]) == 'S'
			|| toupper(given[1]) == 'C')
			&& given[2] == '-') {
		modifier = toupper(given[1]);
		given += 3;
		givenlen -= 4;
	}
	given += 1;
	givenlen -= 2;
	labelbuff[0] = '\0';
	strncat((char *) labelbuff, (char *) given, sizeof(labelbuff) - 1);
	labelbuff[givenlen] = '\0';
  }
  else {
	/* This is too long to be a key label */
	status = FAILED;
  }

  DPRINT_K(("keylabel(): label is |%s| status %s\n",
				labelbuff, status == OK ? "OK" : "FAILED"));

  /* We now have either a valid keyvalue, or a labelbuff containing a name
  ** in canonical format.
  */
  if (status == OK) {
	if (name == NULL)
		name = (char *) labelbuff;

	/* Convert a label to a KeySym; gdk_keyval_from_name() is
	** documented as returning GDK_VoidSymbol, 0xffffff, on
	** failure.  However, in practice it seems to return 0.
	*/
	if ((keyvalue = gdk_keyval_from_name(name)) == GDK_VoidSymbol
							|| keyvalue == 0) {
		status = FAILED;
	}

	DPRINT_K(("keylabel(): conversion of name |%s| to 0x%04x %s\n",
			name, keyvalue, status == OK ? "OK" : "FAILED"));
  }

  /* Assume that "keyvalue" and "name" are the only valid variables */
  if (status == OK) {
	if (keyvalue >= GDK_F1 && keyvalue <= GDK_F12) {
		/* If this is an F-key then convert the label to #n format;
		** and yes, the GDK function key numbers are contiguous.
		*/
		if (modifier) {
			sprintf((char *) labelbuff, "#%ld%c",
				(long) (keyvalue - GDK_F1 + 1),
				tolower(modifier));
		}
		else {
			sprintf((char *) labelbuff, "#%ld",
				(long) (keyvalue - GDK_F1 + 1));
		}
	}
	else {
		/* Update the label to <foo> format */
		labelbuff[0] = '<';
		if (modifier) {
			labelbuff[1] = modifier;
			labelbuff[2] = '-';
			labelbuff[3] = '\0';
		}
		else {
			labelbuff[1] = '\0';
		}
		strcat(labelbuff + 1, name);
		strcat(labelbuff, ">");
	}

	/* Convert the KeySym into a raw code */
	sprintf((char *) rawbuff, "%c%04lx",
		modifier ? ELVCTRL(modifier) : ELVCTRL('K'), (long) keyvalue);

	/* Duplicate a NUL-terminated string */
	*label = safedup(labelbuff);
	*rawin = rawbuff;
  }

  return status == OK ? (int) strlen(rawbuff) : 0;
}


/*
** c l i p o p e n
**
** Open the clipboard.
**
** Returns:	TRUE		Success
**		FALSE		Otherwise
*/
static ELVBOOL clipopen(ELVBOOL forwrite)
{
  DPRINT_F(("clipopen()\n"));

  /* Get the default display, and open the (default) clipboard */
  if (Clipboard == NULL) {
	Clipboard = gtk_clipboard_get_for_display(gdk_display_get_default(),
					GDK_SELECTION_CLIPBOARD);
  }

  return Clipboard == NULL ? FALSE : TRUE;
}


/*
** c l i p w r i t e
**
** Write text to the clipboard.
**
** Returns:	0		Always
*/
static int clipwrite(CHAR *text, int len)
{
  GNGUI *gg = &gn_gui;
  guchar *pc;

  DPRINT_F(("clipwrite()\n"));

  /* Drag is handled in two stages, first by Elvis writing to the
  ** supposed clipboard, and then this buffer being passed to the
  ** client.  There seems no clean single-stage method.
  */
  if (gg->drop_type == PUT_TXT) {
	DPRINT_D(("clipwrite(): drop_type PUT_TXT\n"));
	if ((pc = g_try_malloc((size_t) len)) != NULL) {
		strcpy(pc, text);
		gg->drop_text = pc;
		gg->drop_len = len;
	}
  }
  /* Use the default clipboard set up by clipopen. If len is -1,
  ** the clipboard uses strlen() to get the copy size.
  */
  else {
	if (Clipboard == NULL)
		DPRINT_D(("clipwrite(): Clipboard is NULL\n"));
	else {
		DPRINT_D(("clipwrite(): writing %d to Clipboard\n", len));
		gtk_clipboard_set_text(Clipboard, text, len);
	}
  }

  return 0;
}


/*
** c l i p r e a d
**
** Read text from the clipboard.
**
** Returns:	Bytes read	Success
**		0		Otherwise
**
** Since it isn't necessary to specifically open or close the
** clipboard this function is re-used to handle data dropped
** on the main window.
*/
static int clipread(CHAR *text, int len)
{
  GNGUI *gg = &gn_gui;
  static size_t input_index = 0;
  static size_t input_len = 0;
  static size_t part_len = 0;
  static int state = EMPTY;
  static int more = FALSE;
  size_t maxoutlen;
  int amount;
  static char *pb = NULL;

  DPRINT_F(("clipread()\n"));

#if 0
  { /* Debugging block begin */
  char *names[] = {"EMPTY", "PARTIAL", "FULL", "FINAL"};
  DPRINT_B(("clipread(): state %s text 0x%08lx len %u fd %d\n",
	names[gg->drop_type], (ULONG) gg->drop_text, gg->drop_len, gg->drop_fd));
  } /* Debugging block end */
#endif

  /* Just can't cope with negativity */
  if (len < 0) return 0;

  /* If input_length is less than maxoutlength, all the data can be moved
  ** at once; otherwise we have to supply it in sections.  There
  ** can be repeated calls both to the data source and from the
  ** data sink.
  */
  maxoutlen = (size_t) len;
  amount = -1;
  while (amount < 0) {
	switch (state) {

	case EMPTY:
		input_index = input_len = part_len = 0;
		more = FALSE;
		/* Get all text drop data and hand it out in bits */
		if (gg->drop_type == GET_TXT && gg->drop_len > 0) {
			DPRINT_B(("clipread(): text drop\n"));
			if (maxoutlen > gg->drop_len) {
				input_len = gg->drop_len;
				more = FALSE;
			}
			else {
				input_len = maxoutlen;
				more = TRUE;
			}
			pb = gg->drop_text + input_index;
		}
		/* Get some of the file drop data and hand it out in bits */
		else if (gg->drop_type == GET_URI && gg->drop_fd >= 0) {
			DPRINT_B(("clipread(): file drop\n"));
			input_len = read(gg->drop_fd, text, maxoutlen);
			more = (input_len == maxoutlen) ? TRUE : FALSE;
			pb = text;
		}
		/* Get all the data from the clipboard */
		else {
			DPRINT_B(("clipread(): request clip\n"));
			input_len = request_clip(&pb, maxoutlen);
		}

		/* Deal with what we've got so far */
		if (pb == NULL || input_len == 0) {
			DPRINT_B(("clipread(): bad read ptr 0x%0l8x\n",
							(ULONG) pb));
			/* ie. give up */
			state = FINAL;
		}
		else {
			DPRINT_B(("clipread(): EMPTY got pointer 0x%08lx\n",
							(ULONG) pb));
			part_len = (input_len < maxoutlen) ?
							input_len : maxoutlen;
			state = FULL;
		}
		break;

	case FULL:		/* We have data; pass it to the program */
		DPRINT_B(("clipread(): FULL len %d index %d part_len %d\n",
					input_len, input_index, part_len));
		strncpy(text, pb + input_index, part_len);
		input_index += part_len;
		/* We have less than an output buffer's worth of data */
		if (input_index > input_len) {
			input_index -= part_len;
			part_len = input_len - input_index;
			state = FINAL;
		}
		/* But we have more data from the source */
		else if (more)
			state = EMPTY;
		amount = part_len;
		break;

	case FINAL:		/* Return 0 on completion */
	default:		/* Pretend this never happened */
#if 0
		if (state == FINAL)
			DPRINT_B(("clipread(): PARTIAL free pointer 0x%08lx\n",
						(ULONG) pb));
		else
			DPRINT_B(("clipread(): never happen!\n"));
#endif
		if (gg->drop_type == NONE) g_free(pb);
		pb = NULL;
		input_index = 0;
		input_len = 0;
		part_len = 0;
		state = EMPTY;
		more = FALSE;
		amount = 0;
		break;
	}
  }

  DPRINT_B(("clipread(): return amount %d\n", amount));

  return amount;
}


/*
** c l i p c l o s e
**
** Close the clipboard.
**
** Returns:	Nothing		Always
**
** The GTK+ clipboard is a persistent object,
** and shouldn't be free'd or unref'd.
*/
static void clipclose(void)
{
  DPRINT_F(("clipclose()\n"));

  return;
}


/*
** a d d c o l o r
**
** Add a color to the main color table.
**
** Returns:	Nothing		Always
**
** Convert a color name to an ANSI color value.  This function has the same
** signature as the gui->color() function, and is intended to be used by the
** termcap, open, script, and quit user interfaces.
**
** The RGB value is ignored, and the name simply parsed for its
** associated GdkColor.
**
** As a side effect this function updates the argument colorptr,
** which is an Elvis color index.
**
** Elvis works in color indices, which are generated by addcolor(),
** and are in fact the (index + 1) of the relevant color.
**
** What is the fontcode argument used for, and why does Elvis call
** this function with different colors with matching fontcodes ??? ###
**
** For the codes used to control the VTE terminal (xterm) see draw().
*/
static ELVBOOL addcolor(
		int fontcode,		/* index into colorinfo[] */
		CHAR *colorname,	/* name of the color */
		ELVBOOL isfg,		/* will this be a foreground color ? */
		long *colorptr,		/* where to store the color code */
		unsigned char rgb[3]	/* where to store the RGB value */
		)
{
  GNGUI *gg = &gn_gui;
  GdkColor color;
  int j, index, status;

  DPRINT_C(("addcolor(): font %02d %s |%s|\n",
			fontcode, isfg ? "fg" : "bg", colorname));

  /* Check that essential data items are available */
  if (gg == NULL || gg->vte_colors == NULL)
	return FALSE;

  if (!gdk_color_parse(colorname, &color)) {
	msg(MSG_ERROR,"[s]Couldn't parse color name $1", colorname);
	msgflush();
	return FALSE;
  }

  /* Default colors are no longer in use */
  gg->use_def_color = FALSE;

  /* Find a free palette entry; VTE uses a 0, 8, 16 or 30 color
  ** palette (and not the documented 24), with the option of setting
  ** foreground and background colors separately.  Given a 16-color
  ** palette, index 0 is the default background and index 7 the default
  ** foreground. (Elvis uses index 1 as the default foreground).
  **
  ** Elvis (always?) uses a 16-color palette with the zero code being
  ** reserved for the default (foreground or background) setting.
  ** However, it persists in adding more than the apparent maximum of
  ** 18 colors (the foreground, background, and 16-color palette. The
  ** answer is probably to use multiple font numbers for each palette
  ** slot, where only the numbers, and not the actual colors, differ.
  ** Postpone this improvement...
  **
  ** Does Elvis really split the palette into foreground and background
  ** colors?  draw() and the vte_terminal seem to expect a standardised
  ** VT220-like palette, with eight background and eight foreground
  ** colors, not the modified palette which Elvis builds here. ???
  */

  /* The color palette has a total of 16 elements to suit vte-terminal */
  for (status = FAILED, index = j = 0 ; j < 16 ; j++) {
	/* Always set any default background or foreground values */
	if (fontcode == 1) {
		DPRINT_C(("addcolor(): setting index %d font %d color |%s|\n",
						j, fontcode, colorname));
		/* Elvis uses indices 0 and 1 for background and foreground,
		** while VTE uses 0 and 7.  Cover all bases ???
		*/
		if (isfg) {
			index = 1;	/* Used to update *colorptr */
			/* Elvis */
			gg->font_codes[index] = fontcode;
			gg->color_types[index] = FORE_COLOR;
			gg->vte_colors[index] = color;
#ifdef DPRINT_CC
			strcpy((char *)(color_names + index), colorname);
#endif
			/* VTE */
			gg->font_codes[7] = fontcode;
			gg->color_types[7] = FORE_COLOR;
			gg->vte_colors[7] = color;
#ifdef DPRINT_CC
			strcpy((char *)(color_names + 7), colorname);
#endif
		}
		else {
			index = 0;
			/* Elvis and VTE */
			gg->font_codes[index] = fontcode;
			gg->color_types[index] = BACK_COLOR;
			gg->vte_colors[index] = color;
#ifdef DPRINT_CC
			strcpy((char *)(color_names + index), colorname);
#endif
		}
		/* Don't update the rest of the color array */
		status = OK;
		break;
	}
	/* We can have either a new color for an existing font code, or
	** the same color for a different fontcode; note that both the
	** foreground/background setting, stored in color_types[], and
	** the fontcode or color must match.  Update them both anyway.
	**
	** There seems no harm in updating the palette, but why
	** does Elvis do this ???
	*/
	else if (((isfg && (gg->color_types[j] == FORE_COLOR)) ||
			(!isfg && (gg->color_types[j] & BACK_COLOR))) &&
			(gg->font_codes[j] == fontcode ||
			gdk_color_equal(&color, &gg->vte_colors[j])) &&
			(j != 0 && j != 1 && j != 7)) {
#ifdef DPRINT_CC
		DPRINT_C((
		"addcolor(): updating index %2d to   %s font %2d color |%s|\n",
		j, isfg ? "fg" : "bg", fontcode, colorname));
		DPRINT_C((
		"                              from %s font %2d color |%s|\n",
		(gg->color_types[j] & FORE_COLOR) ? "fg" : "bg",
		gg->font_codes[j], (char *)(color_names + j)));
#endif
		gg->font_codes[j] = fontcode;
		gg->color_types[j] = isfg ? FORE_COLOR : BACK_COLOR;
		gg->vte_colors[j] = color;
#ifdef DPRINT_CC
		strcpy((char *)(color_names + j), colorname);
#endif
		index = j;
		status = OK;
		break;
	}
	/* Or if there's still space to load it, also good */
	else if (gg->font_codes[j] == 0) {
		DPRINT_C(("addcolor(): adding index %d font %d name |%s|\n",
						j, fontcode, colorname));
		gg->font_codes[j] = fontcode;
		gg->color_types[j] = isfg ? FORE_COLOR : BACK_COLOR;
		gg->vte_colors[j] = color;
#ifdef DPRINT_CC
		strcpy((char *)(color_names + j), colorname);
#endif
		index = j;
		status = OK;
		break;
	}
  }

  /* Update the caller */
  *colorptr = (status == OK) ? (long) (index + 1) : 0;

  /* Report outcome */
  if (status == FAILED) {
	msg(MSG_ERROR,
		"Unable to allocate color, out of terminal palette slots\n");
	msgflush();
	return FALSE;
  }
  else
	return TRUE;
}


/*
** f r e e c o l o r
**
** Remove a color from the main color table.
**
** Returns:	Nothing		Always
**
** Elvis works in color codes, reserving 0 for the default.
** Codes are in fact the (index + 1) of the relevant color.
*/
static void freecolor(long color, ELVBOOL isfg)
{
  GNGUI *gg = &gn_gui;

  DPRINT_C(("freecolor(): color code %ld fg %s\n",
				color, isfg ? "yes" : "no"));

  /* If this isn't a foreground color, then reload the background */
  if (!isfg && gg != NULL && gg->edhead != NULL) {
	/* Reload from the default and update the window */
	gg->edhead->backcolor = gg->def_backcolor;
	vte_terminal_set_color_background(VTE_TERMINAL(gg->edhead->vterm),
						&gg->edhead->backcolor);
  }

  if (gg == NULL || gg->vte_colors == NULL)
	/* MT */ ;
  else if (color < 1 || color > 16) {
	msg(MSG_ERROR, "[d]Couldn't free color reference %d", color);
	msgflush();
  }
  else {
	gg->font_codes[--color] = 0;
	gg->color_types[color] = NO_COLOR;
	gg->vte_colors[color] = color_black;
	DPRINT_C(("freecolor(): removed index %ld\n", color));
  }

  return;
}


/*
** s e t b g
**
** Set the terminal background color from the color palette.
**
** Returns:	Nothing		Always
*/
static void setbg(GUIWIN *gw, long bg)
{
  GNGUI *gg = &gn_gui;
  EDWIN *edwin = (EDWIN *) gw;

  DPRINT_F(("setbg()\n"));

  if (bg > 0 && bg < 17) {
	edwin->backcolor = gg->vte_colors[bg - 1];
	vte_terminal_set_color_background(VTE_TERMINAL(edwin->vterm),
						&edwin->backcolor);
  }

  return;
}


/*
** g u i c m d
**
** Update the X11 GUI toolbar.
**
** Returns:	FALSE		Always
**
** This command isn't enabled in the Gnome GUI interface.
** Could it be used as a general GUI-update function ???
*/
static ELVBOOL guicmd(GUIWIN *gw, char *extra)
{
  /* How are commands declared to Elvis ??? */
  DPRINT_F(("guicmd()\n"));

  return FALSE;
}


/*
** t a b c m d
**
** Used by the tags command, somehow.
**
** Returns:	FALSE		Always
**
** This command isn't enabled in the Gnome GUI interface.
*/
static ELVBOOL tabcmd(GUIWIN *gw, _CHAR_ key2, long count)
{
  DPRINT_F(("tabcmd()\n"));

  return FALSE;
}


/*
** s a v e
**
** Save the window position.
**
** Returns:	Nothing		Always
**
** This command isn't enabled in the Gnome GUI interface.
*/
void save(BUFFER buf, GUIWIN *gw)
{
  DPRINT_F(("save()\n"));

  return;
}


/*
** w i l d c a r d
**
** Filename wildcard expansion.
**
** Returns:	0		Always
**
** This command isn't enabled in the Gnome GUI interface.
*/
int wildcard(char *names, char *buf, int bufsize, ELVBOOL single)
{
  DPRINT_F(("wildcard()\n"));

  return 0;
}


/*
** p g o p e n
**
** Run an OS command connected to the Elvis process by pipes.
**
** Returns:	1		Success
**		0		Otherwise
**
** This command isn't enabled in the Gnome GUI interface.
** A generic version is defined in osunix/osprg.c.
**
** This version uses pipes for both input and output. If the willwrite
** flag is set it writes to the current file, and if not it writes to
** the current window.  The willread flag isn't used.
**
** However, the main body of the Elvis code uses the functions prgread()
** and prgwrite() to read and write from the child process, and these in
** turn use file descriptors local to the unix/osprg.c file.  It isn't
** therefore possible to use GUI-specific functions to handle I/O from
** child processes.  Should prgread/write() also be part of the GUI ???
*/
static ELVBOOL pgopen(char *command, ELVBOOL willwrite, ELVBOOL willread)
{
  GNGUI *gg = &gn_gui;
  GIOChannel *gio;
  GError *error = NULL;
  static guint io_flags;
  guint event_id;
  gboolean rc;
  int pid, fd_in, fd_out, fd_err, status = OK;
  int wait_pid, wait_status, wait_rc;
  char *argv[] = { NULL, NULL, NULL, NULL };

  DPRINT_F(("pgopen()\n"));

  argv[0] = (o_shell == NULL) ? SHELL_PRG : (char *) o_shell;
  argv[1] = "-c";
  argv[2] = command;

  io_flags = willwrite ? 1 : 0;
  io_flags |= willread ? 2 : 0;

  DPRINT_Y(("pgopen(): command |%s| shell |%s| Write %s Read %s\n",
	command, o_shell, willwrite ? "Yes" : "No", willread ? "Yes" : "No"));

  /*
  ** gboolean g_spawn_async_with_pipes(
  **   const gchar *directory,  ** child's CWD, or NULL to inherit parent's
  **   gchar **argv,            ** child's argument vector
  **   gchar **envp,            ** child's env, or NULL to inherit parent's
  **   GSpawnFlags flags,       ** flags from GSpawnFlags
  **   GSpawnChildSetupFunc child_setup, ** run this in the child before exec()
  **   gpointer user_data,      ** user data for child_setup()
  **   GPid *child_pid          ** child's pid
  **   gint *standard_input,    ** child's stdin fd
  **   gchar *standard_output,  ** child's stdout fd
  **   gchar *standard_error,   ** child's stderr fd
  **   GError **error           ** return location for error
  ** );
  ** Return code: TRUE on success, FALSE if an error was set.
  */
  rc = g_spawn_async_with_pipes(
			NULL,
			argv,
			NULL,
			G_SPAWN_SEARCH_PATH,
			NULL,
			NULL,
			&pid,
			&fd_in,
			&fd_out,
			&fd_err,
			&error
	);

  /* There must be every sort of race condition here.  Is there a
  ** cleaner way to run the shell process and the gio_handler ???
  */
  if (rc) {
	/* Connect a GIOChannel to the spawned process' standard output */
	gio = g_io_channel_unix_new(fd_out);
	gg->gio_out = gio;
#if 0
	/* Set up for binary data, just in case */
	if (g_io_channel_set_encoding(gio, NULL, &error) !=
							G_IO_STATUS_NORMAL) {
		status = FAILED;
		if (error != NULL) {
			DPRINT_Y(("pgopen(): error %s\n", error->message));
			g_clear_error(&error);
		}
	}
#endif
	/* Channel is buffered by default
	g_io_channel_set_buffered(gio, FALSE);
	*/

	/* Channel is left open by default */
	g_io_channel_set_close_on_unref(gio, TRUE);
#if 0
	/* Turn on non-blocking mode */
	g_io_channel_set_flags(gio,
		g_io_channel_get_flags(gio) | G_IO_FLAG_NONBLOCK, &error);
	if (error != NULL) {
		status = FAILED;
		DPRINT_Y(("pgopen(): error %s\n", error->message));
		g_clear_error(&error);
	}
#endif
	/* Add a watch function to the command's output.
	** Does an event_id of 0 indicate failure ???
	*/
	event_id = g_io_add_watch(
			gio,			/* Channel */
	#if 0
			G_IO_IN,		/* Watch condition */
	#else
			G_IO_IN | G_IO_ERR |
			G_IO_HUP | G_IO_NVAL,
	#endif
			gio_handler,		/* Handler */
			(gpointer) &io_flags	/* Handler data */
			);

	/* Try and get the status byte if it's available. However,
	** this wait doesn't return anything understandable:
	** status 65 pid -1 error ECHILD
	*/
	errno = 0;
	wait_pid = waitpid(pid, &wait_status, WNOHANG);
	if (WIFEXITED(wait_status))
		wait_rc = WEXITSTATUS(wait_status);
	else
		wait_rc = 127; /* Unknown command */
#if 0
	{ /* Debugging block begin */
	char *pc;
	if (errno == 0)
	pc = "NONE";
	else if (errno == ECHILD)
	pc = "ECHILD";
	else if (errno == EINTR)
	pc = "EINTR";
	else if (errno == EINVAL)
	pc = "EINVAL";
	else
	pc = "UNKNOWN";

	DPRINT_Y((
	"pgopen(): wait_status %d wait_pid %d wait_rc %d error %s\n",
				wait_status, wait_pid, wait_rc, pc));
	} /* Debugging block end */
#endif
  }
  else {
	if (error != NULL) {
		DPRINT_Y(("pgopen(): error %s\n", error->message));
		g_clear_error(&error);
	}
	gg->gio_out = NULL;
	status = FAILED;
	msg(MSG_ERROR, "[s]Failed to run command $1", command);
	msgflush();
  }

  /* Add the "Hit <Enter> ..." message */
  if (!willwrite)
	morepush(winofgw(gg->edhead), 0);

  return status == OK ? 1 : 0;
}


/*
** p g c l o s e
**
** Shut down an OS command.
**
** Returns:	The status byte		Always
**
** This command isn't enabled in the Gnome GUI interface.
** A generic version is defined in osunix/osprg.c.
*/
static int pgclose(void)
{
  GNGUI *gg = &gn_gui;

  DPRINT_F(("pgclose()\n"));

  /* Flush and close the channel, with automatic unref.
  ** However, closing the channel here cuts off the
  ** output.  The channel is supposed to clean itself
  ** up, and seems to do so.
  */
#if 0
  GIOStatus gio_status;
  GError *error = NULL;
  if (gg->gio_out != NULL) {
	gio_status = g_io_channel_shutdown(gg->gio_out, TRUE, &error);
	if (error != NULL) {
		DPRINT_Y(("pgclose(): error %s\n", error->message));
		g_clear_error(&error);
	}
	if (gio_status != G_IO_STATUS_NORMAL && gio_status != G_IO_STATUS_EOF)
		msg(MSG_ERROR, "Failed to shut down command");
  }
#endif

  /* Tidy up anyway */
  gg->gio_out = NULL;

  /* A zero status byte indicates success */
  return 0;
}


/*
** s t o p
**
** Run a command shell.
**
** Returns:	RESULT_COMPLETE		Success
**		RESULT_ERROR		Otherwise
**
** If alwaysfork is FALSE Elvis intends us to stop or suspend.
** We don't do that.
*/
static RESULT stop(ELVBOOL alwaysfork)
{
  DPRINT_F(("stop()\n"));

  if (!alwaysfork) return RESULT_ERROR;

  return (terminal(NULL) == OK) ? RESULT_COMPLETE : RESULT_ERROR;
}


/*
** n e w f i l e s
**
** Update the GUI's current filenames.
**
** Returns:	0		Always
**
** This is called by Elvis whenever the current arglist changes; note
** that the arglist is never partially altered, but only entirely replaced.
*/
int newfiles(char **list, int next)
{
  GNGUI *gg = &gn_gui;
  struct Enode *pn;
  size_t len;
  int j, flg, writeable, status;
  char *pb = NULL;
  char *pc, *pd, **ppc, **ppd, buff[PATH_MAX], name[PATH_MAX];

  DPRINT_F(("newfiles()\n"));

#ifdef DPRINT_AA
  { /* Debugging block begin */
  for (ppc = list ; ppc != NULL && *ppc != NULL ; ppc++)
	g_print("newfiles(): filename |%s|\n", *ppc);
  g_print("newfiles(): next %d\n", next);
  } /* Debugging block end */
#endif

  /* newfiles() isn't yet part of the GUI interface, so it has
  ** to (not) deal with being called by a non-Gnome Elvis.
  */
  if (!Exists) return 0;

#ifdef CFGDIR_ELVIS
  if (gg->cfgdir_flg)
	pb = gnome_util_prepend_user_home(CFGDIR_ELVIS "/" NOTEPAD_ELVIS);
  else
	pb = gnome_util_prepend_user_home(NOTEPAD_ELVIS);
#else
  if (pb == NULL)
	pb = gnome_util_prepend_user_home(NOTEPAD_ELVIS);
#endif

  /* Update the Documents menu */
  for (ppc = list ; ppc != NULL && *ppc != NULL ; ppc++) {
	pc = *ppc;
	flg = FALSE;

	/* Uncomment this to force Doxygen to parse this function correctly
	update_edit_files(list, next);
	*/

	/* Ignore the Elvis notepad file, a full path */
	if (strcmp(pb, pc) == SAME) continue;

	/* Ignore the list of notepads, which are full paths */
	for (pn = gg->noteshead ; pn != NULL ; pn = pn->next) {
		if (pn->name != NULL && strcmp(pn->name, pc) == SAME) {
			flg = TRUE;
			break;
		}
		if (pn->next == gg->noteshead) break;
	}
	if (flg) continue;

	/* The path may be absolute or relative */
	if (*pc != '/') {
		strcpy(buff, g_get_current_dir());
		len = PATH_MAX - 1 - strlen(buff);
		strncat(buff, "/", len);
		strncat(buff, pc, len);
	}
	else {
		buff[0] = '\0';
		strncat(buff, pc, sizeof(buff) - 1);
	}

	/* Remove any "." and ".." entries from the path */
	if (undot_path(buff, name, sizeof(name)) == FAILED) {
		/* Remove invalid filenames from the args list,
		** which is a NULL terminated array of pointers.
		** Apparently the main body of the Elvis code
		** can accept a changed array size.
		** First park the bad filename where the error
		** message handler can find it if necessary.
		*/
		BadFile[0] = '\0';
		strncat(BadFile, pc, sizeof(BadFile) - 1);
		for (ppd = ppc ; *ppd != NULL ; ppd++)
			*ppd = *(ppd + 1);
		if (ppc != list) ppc--;
		continue;
	}

	/* Do we have access rights to the file's directory ? */
	writeable = TRUE;
	for (pc = pd = name ; *pc != '\0' ; pc++)
		if (*pc == '/') pd = pc;
	if (pd != name) *pd = '\0';
	if (access(name, W_OK) != 0) {
		writeable = FALSE;
		msg(MSG_ERROR, "[s]Cannot write to directory $1", name);
	}
	*pd = '/';

	/* See if the entry is already on the docslist, and move or remove it */
	if (o_docscount != 0) {
		pn = gg->docshead;
		j = find_node(&pn, name, TRUE, TRUE);

		/* If it's on the list, but unwriteable, or not at the front,
		** remove it.
		*/
		if (j >= 0 && (!writeable || pn != gg->docshead)) {
			rmv_docsmenu(pn);
			del_node(pn, NULL, NULL);
			j = -1;
		}

		/* If it's writeable, add it to the front */
		if (j < 0 && writeable) {
			if ((status = add_node(&gg->docshead, name, NULL,
							BEFORE)) == OK) {
				init_enode(gg->docshead);
				add_docsmenu(gg->docshead);
			}

			if (gg->docscount > o_docscount) {
				rmv_docsmenu(gg->docshead->prev);
				del_node(gg->docshead->prev, NULL, NULL);
			}
		}
	}
	/* Failures are simply ignored */
  }

  /* Update the arglist menu */
  update_edit_files(list, next);

  g_free(pb);

  return 0;
}


/*
** n e w c m d
**
** Update the GUI's current command.
**
** Returns:	0		Always
**
** Command is one of the enums in EXCMD.  This function is called
** from the (patched) main Elvis code, and only those functions
** which require window changes use it at present.
*/
int newcmd(int cmd, int flg)
{
  GNGUI *gg = &gn_gui;
  char *pc;

  DPRINT_F(("newcmd()\n"));

  /* newcmd() isn't yet part of the GUI interface, so it has
  ** to (not) deal with being called by a non-Gnome Elvis.
  */
  if (!Exists) return 0;

  pc = get_exname(cmd);
  gg->cur_cmd = (pc == NULL) ? -1 : cmd;

  /* See the note on the #defines EX_WINUP and EX_WINDN */
  if (cmd == EX_MAIN) {
	gg->cur_cmd = EX_MAIN;
	pc = "EX_MAIN";
  }
  else if (cmd == EX_QUIT) {
	gg->cur_cmd = EX_QUIT;
	pc = "EX_QUIT";
  }
  else if (cmd == EX_WINDOW) {
	if (flg == '+') {
		gg->cur_cmd = EX_WINUP;
		pc = "EX_WINUP";
	}
	else if (flg == '-') {
		gg->cur_cmd = EX_WINDN;
		pc = "EX_WINDN";
	}
  }

  DPRINT_A(("newcmd(): command is %s\n", pc == NULL ? "invalid" : pc));

  return 0;
}


/*
** n e w u n d o
**
** Update the GUI's current undo level.
**
** Returns:	0		Always
**
** This function is called from the (patched) main Elvis code,
** so that the GUI can track the undo/redo level and display
** the relevant menu items appropriately.
*/
int newundo(int cmd, long int level, long int maxlevel)
{
  GNGUI *gg = &gn_gui;
  static int flg = -1;
  static long int redo;
  static long int undo;

  DPRINT_F(("newundo()\n"));

  /* newundo() isn't yet part of the GUI interface, so it has
  ** to (not) deal with being called by a non-Gnome Elvis.
  */
  if (!Exists) return 0;

  /* On the first call, initialise everything in sight */
  if (flg == -1) {
	flg = 0;
	redo = 0;
	undo = 0;
	if (!(gg->menu_state & MENU_UNDO))
		gg->menu_state |= MENU_UNDO;
	if ((gg->menu_state & MENU_REDO))
		gg->menu_state &= ~MENU_REDO;
	config_menus(gg);
  }

  /* Elvis treats this as a special case, with an maxlevel of 1 */
  if (maxlevel == 0)
	maxlevel = 1;

  /* newundo() is called by bufundo() and bufdo().
  ** bufundo has arguments cmd = 0, level = back where back is the number
  ** of versions back to undo (negative to redo).
  ** bufdo has arguments cmd = 1, level = 1 to wipe all redo versions.
  */
  if (cmd == 0) {
	undo += level;
	if (level < 0)
		redo += -level;
	if (undo > maxlevel)
		undo = maxlevel;
	else if (undo == 0)
		redo = 0;
  }

  /* Update the main and mouse menu states */
  if (undo == maxlevel) {	/* Undo isn't available */
	if (gg->menu_state & MENU_UNDO || !(gg->menu_state & MENU_REDO)) {
		gg->menu_state &= ~MENU_UNDO;
		gg->menu_state |= MENU_REDO;
		config_menus(gg);
	}
  }
  else if (undo == 0) {		/* Redo isn't available */
	if (!(gg->menu_state & MENU_UNDO) || (gg->menu_state |= MENU_REDO)) {
		gg->menu_state |= MENU_UNDO;
		gg->menu_state &= ~MENU_REDO;
		config_menus(gg);
	}
  }
  else {
	if (!(gg->menu_state & MENU_UNDO) || !(gg->menu_state & MENU_REDO)) {
		gg->menu_state |= MENU_UNDO;
		gg->menu_state |= MENU_REDO;
		config_menus(gg);
	}
  }

  /* Update the menu item label; display the _remaining_ levels */
  update_undolevel_menu((int) (maxlevel - undo), (int) undo);

#ifdef DPRINT_JJ
  if (cmd == 0 || cmd == 1)
	g_print("newundo() cmd %d back %2ld maxlevel %ld undo %ld redo %ld\n",
					cmd, level, maxlevel, undo, redo);
  else
	g_print("newundo(): unknown command %d\n", cmd);
#endif

   return 0;
}


/*
** m e m p r o f i l e
**
** Set up GLib memory profiling routines.
**
** Returns:	Nothing		Always
**
** Called once, directly from Elvis' main() routine.
*/
void memprofile(void)
{
#ifdef G_MEM_PROFILE
  g_mem_set_vtable(glib_mem_profiler_table);
#endif
  return;
}


/*
** EOF
*/

